-- GenGeneric Scene Controller version 2.11d
-- Supports the following devices: 
-- 		Evolve LCD1 5 button controller with LCD
--		Cooper RFWC5 5 button controller
--		Nexia One Touch 5 button controller with LCD
--		HomeSeer HS-WS100+ switch with 1 LED
--      HomeSeer HS-WD100+ dimmer with 7 LEDs
--		HomeSeer HS-WS200+ switch with 1 multi-color LED
--      HomeSeer HS-WD200+ dimmer with 7 multi-color LEDs
--      HomeSeer HS-FC200+ fan controller switch with 4 multi-color LEDs
--
-- Copyright (C) 2020 Gustavo A Fernandez
-- Thanks to Ron Luna and Ctrlable for contributing for RFWC5 support
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You can access the terms of this license at
-- https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html

-- VerboseLogging == 0: important logs and errors:    ELog, log
-- VerboseLogging == 1: Includes information logs:    ELog, ILog, log
-- VerboseLogging == 2: Includes debug logs:          ELog, ILog, log, DLog, DEntry
-- VerboseLogging == 4:	Includes verbose logs:        ELog, ILog, log, DLog, DEntry, VLog, VEntry
VerboseLogging = 4

-- DisableSceneController can be set to true as a debugging aid 
-- to completely disable all special functionality.
DisableSceneController = false

--
-- forward declarations
--
local 
	SetIndicatorValue,
    CreatePeerDevices,
    ConnectPeerDevice,
    SceneActivatedMonitorCallback,
    BasicSetMonitorCallback,
    MultiLevelSwitchStartLevelChangeMonitorCallback,
    MultiLevelSwitchStopLevelChangeMonitorCallback,
    ThermostatSetPointMonitorCallback,
    ThermostatFanModeMonitorCallback,
    ScreenMDGetMonitorCallback,
    CentralSceneNotificationMonitorCallback,
    SetScreen,
    ForAllVariables,
    ForAllModes,
    RemoveDeviceActuatorConf,
    PropagateCommFailure,
    ParseModeString,
    GetModeStr,
    TriggerVeraScene,
    GetCurrentScreen,
    SetScreenTimeout,
    GetDeviceStatus,
    OptEvent,
    GetNumLinesAndScrollOffset,
    ChooseLabelFontAndAlign,
    UpdateAssociationForPhysicalButton,
    SceneButtonPressed,
    ReadDeviceActuatorConfList,
    WriteDeviceActuatorConfList,
    SetReturnRoute,
    SetPresetScreen,
    TemperatureDeviceIsSettable,
    SetTemperatureScreen,
    SceneChange,
    ExecuteCentralScene,
    SetDeviceStatus,
    SendDeviceCommand,
    WatchDevicesAssociatedWith,
    IsVeraPrimaryController


--
-- Constants
--

local posix = require "posix"
local socket = require "socket"
local zwint = require "zwint"
local bit = require "bit"
require "L_GenGenSceneControllerShared"

local DEVTYPE_EVOLVELCD1     = "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD:1"
local DEVTYPE_COOPEREFWC5    = "urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5:1"
local DEVTYPE_NEXIAONETOUCH  = "urn:schemas-gengen_mcv-org:device:SceneControllerNexiaOneTouch:1"
local DEVTYPE_HOMESEERWS100  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS100:1"
local DEVTYPE_HOMESEERWD100  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD100:1"
local DEVTYPE_HOMESEERWS200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS200:1"
local DEVTYPE_HOMESEERWD200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD200:1"
local DEVTYPE_HOMESEERFC200  = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerFC200:1"

local DEVTYPE_ZWN        	 = "urn:schemas-micasaverde-com:device:ZWaveNetwork:1"
local DEVTYPE_BINARY     	 = "urn:schemas-upnp-org:device:BinaryLight:1"
local DEVTYPE_DIMMABLE   	 = "urn:schemas-upnp-org:device:DimmableLight:1"
local DEVTYPE_THERMOSTAT 	 = "urn:schemas-upnp-org:device:HVAC_ZoneThermostat:1"
local DEVTYPE_WINDOWCOVERING = "urn:schemas-micasaverde-com:device:WindowCovering:1"

local SID_HADEVICE       	 = "urn:micasaverde-com:serviceId:HaDevice1"
local SID_HAG            	 = "urn:micasaverde-com:serviceId:HomeAutomationGateway1"
local SID_SCTRL          	 = "urn:micasaverde-com:serviceId:SceneController1"
local SID_ZWDEVICE       	 = "urn:micasaverde-com:serviceId:ZWaveDevice1"
local SID_DIMMING        	 = "urn:upnp-org:serviceId:Dimming1"
local SID_FANMODE        	 = "urn:upnp-org:serviceId:HVAC_FanOperatingMode1"
local SID_USERMODE       	 = "urn:upnp-org:serviceId:HVAC_UserOperatingMode1"
local SID_SCENECONTROLLER	 = "urn:gengen_mcv-org:serviceId:SceneController1"
local SID_SECURITYSENSOR 	 = "urn:micasaverde-com:serviceId:SecuritySensor1"
local SID_SWITCHPOWER    	 = "urn:upnp-org:serviceId:SwitchPower1"
local SID_TEMPSENSOR     	 = "urn:upnp-org:serviceId:TemperatureSensor1"
local SID_TEMPSETPOINT   	 = "urn:upnp-org:serviceId:TemperatureSetpoint1"
local SID_COOLSETPOINT   	 = "urn:upnp-org:serviceId:TemperatureSetpoint1_Cool"
local SID_HEATSETPOINT   	 = "urn:upnp-org:serviceId:TemperatureSetpoint1_Heat"
local SID_HVACOPSTATE    	 = "urn:micasaverde-com:serviceId:HVAC_OperatingState1"
local SID_ZWN       	     = "urn:micasaverde-com:serviceId:ZWaveNetwork1"

--
-- Globals
--

-- IndicatorLocks is indexed by peer_dev_num
-- 0 (or nil) no lock
-- 1 Short term lock with timeout
-- 2 long term lock without timeout. Should be disabled if unexpected event occurred. 
-- 3 Indicator change bypassed due to lock
local IndicatorLocks = {}
local ZWaveSceneIdCacheList = {}
local param = {}
local lastChangedModes = {}
local temperatureAutoChoices = {}
local wakeUpIntervals = {}

local SCObj
local Devices = {}

local HVACOPSTATE_VAR     = "ModeState"
local HVACOPSTATE_IDLE    = "Idle"
local HVACOPSTATE_HEATING = "Heating"
local HVACOPSTATE_COOLING = "Cooling"

local ASSOCIATION_SET    = "AssociationSet"
local ACTUATOR_CONF	     = "SceneActuatorConf"
local CONFIGURED_ASSOC   = "ConfiguredAssoc"
local CURRENT_INDICATOR  = "CurrentIndicator"
local CURRENT_LED_STATUS = "CurrentLedStatus"
local LAST_EVENT_CODE    = "LastEventCode"
local PREVIOUS_LED_STATUS= "PreviousLedStatus"
local CURRENT_SCREEN     = "CurrentScreen"
local FANMODE_AUTO       = "Auto"
local FANMODE_ON         = "ContinuousOn"
local FANMODE_CYCLE	     = "PeriodicOn"
local FANMODE_VAR        = "Mode"
local NUMLINES_VAR       = "NumLines"
local PRESET_LANGUAGE    = "PresetLanguage"
local RETURN_ROUTES      = "ReturnRoutes"
local SCROLLOFFSET_VAR   = "ScrollOffset"
local SETPOINT_TARG      = "SetpointTarget"
local SETPOINT_VAR       = "CurrentSetpoint"
local ALLSETPOINTS_VAR   = "AllSetpoints"
local TEMPSENSOR_VAR     = "CurrentTemperature"
local USERMODE_VAR       = "ModeStatus"
local USERMODE_AUTO      = "AutoChangeOver"
local USERMODE_COOL      = "CoolOn"
local USERMODE_HEAT      = "HeatOn"
local USERMODE_OFF       = "Off"
local ENERGYMODE_VAR     = "EnergyModeStatus"
local ENERGYMODE_NORMAL  = "Normal"
local ENERGYMODE_SAVING  = "EnergySavingsMode"

local SCREEN_MD = {
   -- Screen flags
   ClearScreen = 0x00,
   ScrollDown  = 0x08,
   ScrollUp    = 0x10,
   NoChange    = 0x38,
   ScreenMask  = 0x38,

   MoreData    = 0x80,

   -- Line Flags
   StdFont     					= 0x00,
   Highlighted 					= 0x20,
   LargeFont   					= 0x40,
   LargeFontHighlighted 		= 0x60,
   StdFontNoScroll 				= 0x80,
   HighlightedNoScroll 			= 0xA0,
   LargeFontNoScroll   			= 0xC0,
   LargeFontHighlightedNoScroll = 0xE0,   

   NoClearLine = 0x00,
   ClearLine   = 0x10,

   Line1       = 0x00,
   Line2       = 0x01,
   Line3       = 0x02,
   Line4       = 0x03,
   Line5       = 0x04
}

-- Make sure tha all of the log functions work before even the SCObj global is set.
function GetDeviceName()
  local name = "Scene Controller"
  if SCObj then
	 name = SCObj.Name
  end
  return name
end

--[==[

--
-- Start benchmark code
--
local function findBestDelay(labels, node_id, zwave_dev_num, first, last, screenFlags)
	local save_retries = param[zwave_dev_num].SCENE_CTRL_Retries
	param[zwave_dev_num].SCENE_CTRL_Retries = 1
	local low = 1
	local high = 1000
	while high > low do
	    local test = math.floor((high + low) / 2)
	    local count = 0
		local bad = 0
		for i = 1,10 do
			log("Benchmark Testing ", labels, "delay=", test, " iteration=", i)
		    if EVLCDWrapStrings(labels, {}, {}, node_id, zwave_dev_num, screenFlags, test) then
				count = count + 1
			else
			    bad = bad + 1
				if bad > 1 then
			 		break
				end
			end
		end
		if count >= 9 then
			log("Benchmark test ", labels, " delay=", test, " passed")
			high = test
		else
			log("Benchmark test ", labels, " delay=", test, " failed")
			low = test+1
		end
	end
	log ("Benchmark test ", labels, " Best value is ", high)
	param[zwave_dev_num].SCENE_CTRL_Retries = save_retries
	return high
end

local tests = {}
local function test(labels)
	tests[labels] = findBestDelay(labels, test_node_id, test_zwave_dev_num, 1, #labels, test_clear)
end

local function Benchmark(peer_dev_num, screen)
	test_node_id, test_zwave_dev_num = GetZWaveNode(peer_dev_num)
	EVLCDWrapStrings({""}, {}, {}, test_node_id, test_zwave_dev_num, true, 1000)
	if screen == 1 then
		test_clear = SCREEN_MD.ClearScreen
	else
		test_clear = SCREEN_MD.NoChange
	end
	test {"1"}
	test {"22"}
	test {"333"}
	test {"4444"}
	test {"55555"}
	test {"666666"}
	test {"7777777"}
	test {"88888888"}
	test {"999999999"}
	printTable(tests,"Benchmark: ")
	test {"1","2"}
	test {"1","2","3"}
	test {"1","2","3","4"}
	test {"1","2","3","4","5"}
	test {"1\r1"}
	test {"1\r1","2\r2"}
	test {"1\r1","2\r2","3\r3"}
	test {"1\r1","2\r2","3\r3","4\r4"}
	test {"1\r1","2\r2","3\r3","4\r4","5\r5"}
	printTable(tests,"Benchmark: ")
end
--]==]

--
-- Parameters and defaults
--
function SceneController_ParamChange(lul_device, lul_service, lul_variable, lul_value_old, lul_value_new)
	local node_id, zwave_dev_num = GetZWaveNode(lul_device)
	local peer_dev_num =  GetPeerDevNum(lul_device)
	param[zwave_dev_num][lul_variable] = tonumber(lul_value_new)
	param[peer_dev_num][lul_variable] = tonumber(lul_value_new)
end

function SceneController_VerboseLoggingChange(lul_device, lul_service, lul_variable, lul_value_old, lul_value_new)
	local oldVerboseLogging = VerboseLogging
	VerboseLogging = tonumber(lul_value_new)
	log("VerboseLogging changed: Old level was ", oldVerboseLogging, " New level is ", VerboseLogging)
	local firstInstaller = GetFirstInstaller();
	local firstNodeId, firstZWave = GetZWaveNode(firstInstaller)
	if lul_device ~= firstZWave then
		luup.variable_set(SID_SCENECONTROLLER, "SCENE_CTRL_VerboseLogging", VerboseLogging, firstZWave)
	end
end

local SceneController_ForceDefaultParameters = false
local function SetTuningParameter(name, default_value, zwave_dev_num, peer_dev_num, functionName)
	VEntry()
	local value = luup.variable_get(SID_SCENECONTROLLER, "SCENE_CTRL_"..name, zwave_dev_num)
	if value == nil then
		value = luup.variable_get(SCObj.OldServiceId, SCObj.OldParamPrefix.."_"..name, zwave_dev_num)
		if value ~= nil then
		    luup.variable_set(SID_SCENECONTROLLER, "SCENE_CTRL_"..name, value, zwave_dev_num, true)
			luup.variable_set(SCObj.OldServiceId, SCObj.OldParamPrefix.."_"..name, nil, zwave_dev_num, true)
		end 
	end
	if SceneController_ForceDefaultParameters or value == nil then
		value = default_value
	else
		value = tonumber(value)
	end
	if not param[zwave_dev_num] then
		param[zwave_dev_num] = {}
	end
	if not param[peer_dev_num] then
		param[peer_dev_num] = {}
	end
	param[zwave_dev_num]["SCENE_CTRL_"..name] = value
	param[peer_dev_num]["SCENE_CTRL_"..name] = value
	if not functionName then
		functionName = "SceneController_ParamChange"
	end
	luup.variable_watch(functionName, SID_SCENECONTROLLER, "SCENE_CTRL_"..name, zwave_dev_num)
	luup.variable_watch(functionName, SID_SCENECONTROLLER, "SCENE_CTRL_"..name, peer_dev_num)
end

--
-- Screen Metadata display
--
local function ExpandString(s)
  local out = s:len()
  for i=1,s:len() do
     out = out .. " " .. s:byte(i)
  end
  return out
end

--EVLCDLabel(screenFlags,{line1,line2,...,linen)
--linex = {lineFlags,linePosition,label}
local function EVLCDLabel(screenFlags,lineArray, node_id, zwave_dev_num, delay_override)
	VEntry()
	if SCObj.HasScreen then
		local peer_dev_num = GetPeerDevNum(zwave_dev_num)
		if param[zwave_dev_num].SCENE_CTRL_CharDelay == nil then
			SCObj.SetTuningParameters(zwave_dev_num, GetPeerDevNum(zwave_dev_num))
		end
		local data = "146 2 " .. tostring(screenFlags) -- COMMAND_CLASS_SCREEN_META_DATA version 1 REPORT
		local part = 1
		local delay = param[zwave_dev_num].SCENE_CTRL_BaseDelay
		local screenMask = bit.band(screenFlags,SCREEN_MD.ScreenMask)
		if screenMask == SCREEN_MD.ClearScreen then
			delay = delay + param[zwave_dev_num].SCENE_CTRL_ClearDelay
		elseif screenMask ~= SCREEN_MD.NoChange then
			delay = delay + param[zwave_dev_num].SCENE_CTRL_ScrollDelay
		end
		for lnum = 1,#lineArray do
	   		local line = lineArray[lnum]
	   		assert(#line == 3,"EVDLCDLabel Line " .. lnum .. ": Each line must have 3 parts: flags, position, label")
			assert(line[1] >= 0 and line[1] <= 255, "EVDLCDLabel Line " .. lnum .. ": Flags must be 1 byte")
	   		assert(line[2] >= 0                  ,  "EVDLCDLabel Line " .. lnum .. ": position should >= 0")
	   		assert(#line[3] <= SCObj.MaxLabelSize,  "EVDLCDLabel Line " .. lnum .. ": Label too long")
	   		data = data .. " " .. tostring(line[1])
	        	        .. " " .. tostring(line[2])
	             	    .. " " .. ExpandString(line[3])
	        delay = delay + #line[3] * param[zwave_dev_num].SCENE_CTRL_CharDelay
	   		if part == 1 and lnum < #lineArray and param[zwave_dev_num].SCENE_CTRL_MaxParts > 1 then
				delay = delay + param[zwave_dev_num].SCENE_CTRL_LineDelay
	   	  		part = 2
	   		else
				if delay_override then
					delay = delay_override
				end
				EnqueueZWaveMessage(SCObj, "SetLabel_"..node_id, node_id, data, peer_dev_num, delay);
				screenFlags = SCREEN_MD.NoChange
		        data = "0x92 2 " .. tostring(screenFlags)
				part = 1
				delay = param[zwave_dev_num].SCENE_CTRL_BaseDelay
			end
	   	end -- for lnum
	end -- if SCObj.HasScreen
end -- EVDLCDLabel

--
-- Screen display function. Pass an array of up to 5 strings.
-- Lines will automatically wrap at word boundaries.
--
local function  EVLCDWrapStrings(stringArray, fontArray, alignArray, node_id, zwave_dev_num, screenFlags, delay_override)
	VEntry()
	if SCObj.HasScreen then
		local lineArray = {}
		local entryNum = 1
		for index = 1, SCObj.NumButtons do
			local lnum = index - 1;
			local displayString = stringArray[index]
			local font = fontArray[index];
			local align = alignArray[index];
			if type(displayString) == "string" then
				local part = 1
				local pos = 1
				local width = 0
				local lastSpace = 0
				local label = ""
				local label1 = ""
				local width1 = ""
				local flags1 = 0
				local word = ""
				local wordWidth = 0
				local prevc
				local cWidth
				local spaceWidth = 0
				local space = ""
				local longWord = false;
				local addLine;
				local widths = SCObj.SmallFontWidths;
				local lineFlags = 0;
				if font == nil or font == "" then
					font = "Normal"
				end
				if align == nil or align == "" then
					align = "Center"
				end
				if font == "Normal" then  -- Only the normal font can get 2 lines and thus gets - or \r subsitutions
					lineFlags = SCREEN_MD.StdFont;
					displayString = displayString:gsub("\\r",      "\r")
					displayString = displayString:gsub("-([^\r])", "-\r%1")
				elseif font == "Inverted" then
					lineFlags = SCREEN_MD.Highlighted;
				elseif font == "Compressed" or font == "Large" then -- The "Large" font is taller but actually the compressed font
					widths = SCObj.LargeFontWidths
					lineFlags = SCREEN_MD.LargeFont;
				elseif font == "Compressed, Inverted" then
					lineFlags = SCREEN_MD.LargeFontHighlighted;
				elseif font == "No Scroll" then
					lineFlags = SCREEN_MD.StdFontNoScroll;
				elseif font == "Inverted, No Scroll" then
					lineFlags = SCREEN_MD.HighlightedNoScroll;
				elseif font == "Compressed, No Scroll" then
					lineFlags = SCREEN_MD.LargeFontNoScroll;
				elseif font == "Compressed, Inverted, No Scroll" then
					lineFlags = SCREEN_MD.LargeFontHighlightedNoScroll;
				end

				if align:sub(1,3) == "Raw" then
					local offset = 0
					if align:len() > 3 then
						local offsetString = align:sub(4);
						offset = tonumber(offsetString);
					end
					local flags = lineFlags;
					if screenFlags ~= SCREEN_MD.ClearScreen then
						flags = bit.bor(flags, SCREEN_MD.ClearLine)
					end
					lineArray[entryNum] = { bit.bor(flags,lnum), offset, displayString:sub(1,SCObj.MaxLabelSize) }
		    		entryNum = entryNum + 1
				else
					local eol;
					if displayString == "" then
						eol = 0;
					else
						eol = displayString:byte(#displayString)
					end
					if eol ~= 13 and eol ~= 10 then
						displayString = displayString .. '\r';
					end
					while pos <= #displayString and lnum < SCObj.NumButtons do
				  		if longWord then
				    		longWord = false
							label = word
							width = wordWidth;
							addLine = true
							word = ""
							wordWidth = 0
						else
							addLine = pos == #displayString
							longWord = false;
							local c = displayString:byte(pos)
							if c == 32 then -- space
								if wordWidth > 0 then
						  			if font ~= "Normal" or (width + spaceWidth + wordWidth <= SCObj.ScreenWidth) then
										label = label .. space .. word
										width = width + spaceWidth + wordWidth
										word = ""
										wordWidth = 0
										space = " "
										spaceWidth = widths[1]
									else
										addLine = true
						  			end
						  		else -- wordWidth > 0
						  			space = space .. " "
						  			spaceWidth = spaceWidth + widths[1]
								end -- wordWidth > 0
				      		elseif (c == 10 or c == 13) and (font == "Normal" or pos == #displayString) then -- \n or \r
					    		if c == 13 or prevc ~= 13 then -- ignore lf after cr
						  			if width > 0 or wordWidth > 0 then
						    			if wordWidth > 0 then
							  				label = label .. space .. word
							  				width = width + spaceWidth + wordWidth
							  				word = ""
							  				wordWidth = 0
										end
						    			addLine = true;
						  			end -- cr/lf processing
								end
					  		elseif (c > 32 and c <= 127) or (c >= 160 and c <= 255) then -- normal character
								if (c <= 127) then
					      			cWidth = widths[c-31]
					    		else
						  			cWidth = widths[c-159]
								end
								if font ~= "Normal" or (width + spaceWidth + wordWidth + cWidth <= SCObj.ScreenWidth) then
						  			word = word .. string.char(c)
						  			wordWidth = wordWidth + cWidth
								elseif width > 0 then -- Current word must spill to next line
						  			if wordWidth > SCObj.ScreenWidth then -- output current line followed by long word.
						    			longWord = true;
						    			addLine = true;
						  			else -- Normal word wrap
						    			word = word .. string.char(c)
						    			wordWidth = wordWidth + cWidth
						    			addLine = true;
						  			end
					    		else -- long word but nothing else on this line
						  			label = word
						  			word = string.char(c)
						  			wordWidth = cWidth
						  			addLine = true;
								end
					  		end -- regular character processing
					  		prevc = c
					  		pos = pos + 1
						end -- not longWord
				    	if addLine then
							DLog("Adding line \"", label, "\" entryNum=", entryNum, " lnum=", lnum, " part=", part, " word=", word, " wordWidth=", wordWidth, " font=", font, " align=", align)
							local flags = lineFlags;
					  		if part == 1 then
								if screenFlags ~= SCREEN_MD.ClearScreen then
									flags = bit.bor(flags, SCREEN_MD.ClearLine)
								end
								label1 = label;
								width1 = width;
								flags1 = flags;
								local offset = 0;
								if align == "Left" then -- Left justify 1 line.
								    local spaces = math.floor((SCObj.ScreenWidth - width) / widths[1]);
								    if spaces < 0 then
								    	spaces = 0
								    end
								    while spaces > 0 and #label < SCObj.MaxLabelSize do
										label = label .. " ";
								    end
								elseif align == "Right" then -- Right justify 1 line.
									offset = math.floor((SCObj.RightJustifyScreenWidth - width) / 3);
									if offset < 0 then
										offset = 0
									end
								end
								lineArray[entryNum] = { bit.bor(flags,lnum), offset, label:sub(1,SCObj.MaxLabelSize) }
					    		entryNum = entryNum + 1
								part = 2
					  		else
								local offset = 0;
								if align == "Left" then -- Left justify 2 lines;
									while width1 < SCObj.ScreenWidth and width < SCObj.ScreenWidth and #label1 + #label < 15 do
										if width1 < SCObj.ScreenWidth and #label1 + #label < 15 then
											label1 = label1 .. " ";
											width1 = width1 + widths[1];
										end
										if width < SCObj.ScreenWidth and #label1 + #label < 15 then
											label = label .. " ";
											width = width + widths[1];
										end
									end
								elseif align == "Right" then -- Right justify 2 lines
									while (width <= width1 - widths[1]) and #label1 + #label < 15 do
										label = " " .. label;
										width = width + widths[1];
									end
									while (width1 <= width - widths[1]) and #label1 + #label < 15 do
										label1 = " " .. label1;
										width1 = width1 + widths[1];
									end
									if width > width1 then
										offset = math.floor((SCObj.RightJustifyScreenWidth - width) / 3);
									else
										offset = math.floor((SCObj.RightJustifyScreenWidth - width1) / 3);
									end
								end
								lineArray[entryNum-1][2] = offset;
						  		lineArray[entryNum-1][3] = (label1 .. "\r" .. label):sub(1,SCObj.MaxLabelSize);
								break
					  		end
					  		label = ""
					  		width = 0
					  		space = ""
					  		spaceWidth = 0
						end -- if addline
					end -- pos <= #displayString and lnum < SCObj.NumButtons
				end -- else not raw
			end -- if string
		end -- for lnum
		-- printTable(lineArray)
		if screenFlags == SCREEN_MD.NoChange then
			-- Apply a work-around for a bug in the Evolve LCD1 firmware which causes the highlighting of other lines to
			-- become wide or narrow if a different line is drawn.
			-- We temporarily turn off the indicator, draw the line(s) and turn it back on again.
			-- If the indicator is 0 (no lines highlighted) the SetIndicatorValue calls are no-ops.
			local peer_dev_num = GetPeerDevNum(zwave_dev_num)
			local indicator = SetIndicatorValue(peer_dev_num, 0, false, true)
			EVLCDLabel(screenFlags, lineArray, node_id, zwave_dev_num, delay_override)
			SetIndicatorValue(peer_dev_num, indicator, false, false)
		else
			EVLCDLabel(bit.bor(SCREEN_MD.MoreData, screenFlags), lineArray, node_id, zwave_dev_num, delay_override)
		end
	end -- if SCObj.HasScreen
end

--
-- Initialization, primary controller and peer handling
--

-- GetDeviceCounter returns 1 for the lowest numbered controller, 2 for the second, etc.
-- ordered by device ID.
local function GetDeviceCounter(our_zwave_dev_num)
	local count = 0;
	for dev_num, v in pairs(luup.devices) do
		if Devices[v.device_type] and v.device_num_parent == 1 then
			if our_zwave_dev_num >= dev_num then
				count = count + 1
			end
		end
	end
	if count == 0 then
		ELog("GetDeviceCounter: Error device ", our_zwave_dev_num, "not found.")
	end
	return count
end

-- A dummy callback used by Z-Wave intercepts where we really don't need the data or the event.
local function DummyCallback(peer_dev_num, captures)
	VEntry()
end

function SceneController_Init(lul_device)
	DEntry()
	if DisableSceneController then
		return true, "ok", "Gengen Scen Controller"
    end
	if not zwint or not zwint.register then
		ELog("ZWInt module not installed")
		return false, "ZWInt not installed", "Scene Controller";
	end
	if lul_device == 1 then
		CreatePeerDevices()
	else
		local supported, newDevice = ConnectPeerDevice(lul_device)
		if not supported then
			return false, "Not supported", "GenGenSceneController"
		end
		local peer_dev_num = lul_device
		if zwint.instance then -- Only in debug version
			DLog("ZWInt instance is ", zwint.instance())
		end
		local zwave_node, zwave_dev = GetZWaveNode(lul_device)
		if not zwave_node or not zwave_dev then
			return false, "Z-Wave device removed", SCObj.Name
		end
		RegisterClientDevice("SceneController_Init2", lul_device)
	end
	return true, "ok", SCObj.Name
end

function SceneController_Init2(lul_device)
	lul_device = tonumber(lul_device)
	local peer_dev_num = lul_device
	local zwave_node, zwave_dev = GetZWaveNode(lul_device)

	local veraZWaveNode, ZWaveNetworkDeviceId = GetVeraIDs()
    local ComPort = luup.variable_get(SID_ZWN, "ComPort", ZWaveNetworkDeviceId)
    VLog("SceneController_Init2: calling zwint.register(", ComPort, ")");
	local result, errcode, errmsg = zwint.register(ComPort);
    VLog("SceneController_Init2: zwint.register(", ComPort, ") returned result=", result, " errcode=", errcode, " errmsg=", errmsg);

	-- Devices are connected. Perform normal initialization on reload.
	if SCObj.SendsSceneActivation then
--[==[
                                                 C1                   C2   C3
42      12/11/16 20:04:55.812   0x1 0xa 0x0 0x4 0x8 0x7 0x4 0x2b 0x1 0x1 0xff 0x2e (#\n#####+###.)
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |   |    |    |
                    length = 10 ------+   |   |   |   |   |    |   |   |    |    |
                        Request ----------+   |   |   |   |    |   |   |    |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |   |    |    |
           Receive Status MULTI ------------------+   |   |    |   |   |    |    |
Device 10=Cooper RFWC5 Scene Controller Z-Wave -------+   |    |   |   |    |    |
                Data length = 4 --------------------------+    |   |   |    |    |
 COMMAND_CLASS_SCENE_ACTIVATION -------------------------------+   |   |    |    |
           SCENE_ACTIVATION_SET -----------------------------------+   |    |    |
                   Scene ID = 1 ---------------------------------------+    |    |
         Dimming Duration = 255 --------------------------------------------+    |
                    Checksum OK -------------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 2B 01 (..) (..)", -- Main RegEx
		                 "06", -- ACK response,
		                 SceneActivatedMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "SceneActivate")
	end

	if SCObj.SendsBasicSet then
--[==[
                                                 C1                  C2
42      12/11/16 20:53:12.232   0x1 0x9 0x0 0x4 0x8 0x7 0x3 0x20 0x1 0x0 0xdf (####### ###)
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |   |    |
                     length = 9 ------+   |   |   |   |   |    |   |   |    |
                        Request ----------+   |   |   |   |    |   |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |   |    |
           Receive Status MULTI ------------------+   |   |    |   |   |    |
Device 10=Cooper RFWC5 Scene Controller Z-Wave -------+   |    |   |   |    |
                Data length = 3 --------------------------+    |   |   |    |
            COMMAND_CLASS_BASIC -------------------------------+   |   |    |
                      BASIC_SET -----------------------------------+   |    |
              Value = BASIC_OFF ---------------------------------------+    |
                    Checksum OK --------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 20 01 (..)", -- Main RegEx
		                 "06", -- ACK response,
		                 BasicSetMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "BasicSet")
	end

	if SCObj.SendsSwitchMultilevel then
--[==[
                                                 C1      C2          <-----C3---->
42      10/28/17 16:50:46.391   0x1 0xb 0x0 0x4 0x8 0x2 0x5 0x26 0x4 0x20 0xff 0x6 0x4 (#######&# ###) 
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |    |    |   |   |
                    length = 11 ------+   |   |   |   |   |    |   |    |    |   |   |
                        Request ----------+   |   |   |   |    |   |    |    |   |   |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |    |    |   |   |
           Receive Status MULTI ------------------+   |   |    |   |    |    |   |   |
Device 3=Cooper RFWC5 Scene Controller Z-Wave --------+   |    |   |    |    |   |   |
                Data length = 5 --------------------------+    |   |    |    |   |   |
COMMAND_CLASS_SWITCH_MULTILEVEL -------------------------------+   |    |    |   |   |
SWITCH_MULTILEVEL_START_LEVEL_CHANGE ------------------------------+    |    |   |   |
Flags = Up | Ignore StartLevel | Ignored or Increment secondary --------+    |   |   |
               startLevel = 255 ---------------------------------------------+   |   |
                   Duration = 6 -------------------------------------------------+   |
                    Checksum OK -----------------------------------------------------+
41      10/28/17 16:50:46.391   ACK: 0x6 (#) 
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " (..) 26 04 (.+) ..$", -- Main RegEx
		                 "06", -- ACK response,
		                 MultiLevelSwitchStartLevelChangeMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "MultilevelStart")

--[==[
                                                 C1
42      01/09/17 23:22:45.021   0x1 0x8 0x0 0x4 0x8 0xc 0x2 0x26 0x5 0xd6 (#######&##) 
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |    |
                     length = 8 ------+   |   |   |   |   |    |   |    |
                        Request ----------+   |   |   |   |    |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |    |
           Receive Status MULTI ------------------+   |   |    |   |    |
Device 39=Cooper RFWC5 Scene Controller Z-Wave -------+   |    |   |    |
                Data length = 2 --------------------------+    |   |    |
COMMAND_CLASS_SWITCH_MULTILEVEL -------------------------------+   |    |
SWITCH_MULTILEVEL_STOP_LEVEL_CHANGE -------------------------------+    |
                    Checksum OK ----------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 26 05", -- Main RegEx
		                 "06", -- ACK response,
		                 MultiLevelSwitchStopLevelChangeMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "MultilevelStop")
	end

	if SCObj.HasThermostatControl then
--[==[
                                                     C1               C2  C3   C4
42      01/14/17 10:23:29.612   0x1 0xb 0x0 0x4 0x8 0xb 0x5 0x43 0x1 0x1 0x9 0x3f 0x83 (#######C###?#) 
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |   |   |    |    |
                    length = 11 ------+   |   |   |   |   |    |   |   |   |    |    |
                        Request ----------+   |   |   |   |    |   |   |   |    |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |   |   |    |    |
           Receive Status MULTI ------------------+   |   |    |   |   |   |    |    |
Device 38=Evolve LCD1 Scene Controller Z-Wave --------+   |    |   |   |   |    |    |
                Data length = 5 --------------------------+    |   |   |   |    |    |
COMMAND_CLASS_THERMOSTAT_SETPOINT -----------------------------+   |   |   |    |    |
        THERMOSTAT_SETPOINT_SET -----------------------------------+   |   |    |    |
           SetPoint = Heating 1 ---------------------------------------+   |    |    |
Level2 = Size=1 | Fahrenheit | Precision=0 --------------------------------+    |    |
                     Value = 63 ------------------------------------------------+    |
                    Checksum OK -----------------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 43 01 (..) (..) (..)", -- Main RegEx
		                 "06", -- ACK response,
		                 ThermostatSetPointMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "ThermostatSetPoint")
--[==[
                                                 c1                   C2
42      01/14/17 16:32:44.900   0x1 0x9 0x0 0x4 0x8 0xb 0x3 0x44 0x1 0x0 0xb7 (#######D###) 
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |   |    |
                     length = 9 ------+   |   |   |   |   |    |   |   |    |
                        Request ----------+   |   |   |   |    |   |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |   |    |
           Receive Status MULTI ------------------+   |   |    |   |   |    |
Device 18=Evolve LCD1 Scene Controller Z-Wave --------+   |    |   |   |    |
                Data length = 3 --------------------------+    |   |   |    |
COMMAND_CLASS_THERMOSTAT_FAN_MODE -----------------------------+   |   |    |
        THERMOSTAT_FAN_MODE_SET -----------------------------------+   |    |
              Mode = Auto | Low ---------------------------------------+    |
                    Checksum OK --------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 44 01 (..)", -- Main RegEx
		                 "06", -- ACK response,
		                 ThermostatFanModeMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "ThermostatFanMode")
	end -- if SCObj.HasThermostatControl

	if SCObj.HasScreen and not SCObj.HasBattery then
--[==[
                                                     C1
42      08/29/17 20:17:08.480       0x1 0xa 0x0 0x4 0x0 0xd7 0x4 0x92 0x1 0x5 0x0 0xb4 (#\n##########) 
               SOF - Start Of Frame --+   |   |   |   |    |   |    |   |   |   |    |
                        length = 10 ------+   |   |   |    |   |    |   |   |   |    |
                            Request ----------+   |   |    |   |    |   |   |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER --------------+   |    |   |    |   |   |   |    |
              Receive Status SINGLE ------------------+    |   |    |   |   |   |    |
Device 148=EC14 Breakfast Nook-Family Romm entrance Z-Wave +   |    |   |   |   |    |
                    Data length = 4 ---------------------------+    |   |   |   |    |
            COMMAND_CLASS_SCREEN_MD --------------------------------+   |   |   |    |
                SCREEN_METADATA_GET ------------------------------------+   |   |    |
                numberOfReports = 5 ----------------------------------------+   |    |
                Node ID: 0 : *None* --------------------------------------------+    |
                        Checksum OK -------------------------------------------------+
41      08/29/17 20:17:08.480   ACK: 0x6 (#)
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 92 01", -- Main RegEx
		                 "06", -- ACK response,
		                 ScreenMDGetMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "ScreenMDGet")
	end -- if SCObj.HasScreen and not SCObj.HasBattery

	if SCObj.SendsCentralScene then
--[==[
                                                 C1                    C2  C3  C4
42      02/25/18 14:44:37.990   0x1 0xb 0x0 0x4 0x0 0xdf 0x5 0x5b 0x3 0xd 0x0 0xe 0x71 (#######[#\r##q) 
           SOF - Start Of Frame --+   |   |   |   |    |   |    |   |   |   |   |    |
                    length = 11 ------+   |   |   |    |   |    |   |   |   |   |    |
                        Request ----------+   |   |    |   |    |   |   |   |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |    |   |    |   |   |   |   |    |
          Receive Status SINGLE ------------------+    |   |    |   |   |   |   |    |
Device 900=Nexia One Touch Scene Controller Z-Wave ----+   |    |   |   |   |   |    |
                Data length = 5 ---------------------------+    |   |   |   |   |    |
    COMMAND_CLASS_CENTRAL_SCENE --------------------------------+   |   |   |   |    |
     CENTRAL_SCENE_NOTIFICATION ------------------------------------+   |   |   |    |
           Sequence Number = 13 ----------------------------------------+   |   |    |
Key Attributes = Key Pressed 1 time ----------------------------------------+   |    |
              Scene Number = 14 ------------------------------------------------+    |
                    Checksum OK -----------------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num,
		                 nil, -- No arm_regex
		                 "^01 .. 00 04 (..) " .. string.format("%02X", zwave_node) .. " .. 5B 03 (..) (..) (..)", -- Main RegEx
		                 "06", -- ACK response,
		                 CentralSceneNotificationMonitorCallback,
		                 false, -- Not OneShot
		              	 param[lul_device].SCENE_CTRL_CentralSceneTimeout, -- multiplacket timeout
						 "CentralSceneNotify")
	end

	local currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	if not currentScreen then
		currentScreen = SCObj.DefaultScreen
	end
	if IsVeraPrimaryController() then
		if newDevice then
			-- doTimeout, force, not indicatorOnly
			SetScreen(peer_dev_num, currentScreen, true, true, false);
		else
			-- doTimeout, force, indicatorOnly
			SetScreen(peer_dev_num, currentScreen, true, true, true);
		end
	end
	RunZWaveQueue("Init2", 0)
end -- lul_device != 1


local function MigrateVariableToNewServiceId(varName, dev_num)
	local value = luup.variable_get(SCObj.OldServiceId, varName, dev_num)
	if value ~= nil then
		luup.variable_set(SID_SCENECONTROLLER, varName, value, dev_num)
		if varName ~= "PeerID" then -- Keep PeerID to not confuse old versions which may still be active.
			luup.variable_set(SCObj.OldServiceId, varName, "", dev_num)
		end		
	end
	return value
end

-- Iterate through all variables using the old service ID and move them to the new (generic) service ID
local function MigrateDeviceToNewServiceId(dev_num)
	ForAllVariables(dev_num, MigrateVariableToNewServiceId)
end

-- Work around 7.31 Vera firmate % in variable value bug
local function EscapePercent(str)
	if type(str) ~= "string" then
		ELog("type of ",str," is not string")
		return str
	end
	local s1 = str:gsub("#","##")
	return s1:gsub("%%","#p")
end

local function UnescapePercent(str)
	if type(str) ~= "string" then
		ELog("type of ",str," is not string")
		return str
	end
	local s1 = str:gsub("#p","%%")
	return s1:gsub("##","#")
end

local function FixPercentEscape(varName, dev_num)
	local value = luup.variable_get(SID_SCENECONTROLLER, varName, dev_num)
	if value then
		local newval = EscapePercent(value)
		if value ~= newval then
			if varName:match("^Label.*") then
				luup.variable_set(SID_SCENECONTROLLER, varName, newval, dev_num)
			else
				ELog("FixPercentEscape: unexpected % in valurable:", varName, "=", value, "->", newval, " in device", dev_num)
			end
		end
	end
end

--
-- 	Called at initialization to possibly create the peer (Lua) device for the Z-Wave device.
--  If a device is created, then the luup engine is reloaded and the two-way linke between
--  two devices is formed.
--
-- local 
function CreatePeerDevices()
	DEntry()
	local count = 0;
	for zwave_dev_num,v in pairs(luup.devices) do
		local candidate = Devices[v.device_type]
		if candidate and v.device_num_parent == 1 then
			SCObj = candidate
			DLog("CreatePeerDevices: Found z-wave device ", zwave_dev_num, ": ", SCObj.Name, "named ", v.description)
	       	local peerID = luup.variable_get(SID_SCENECONTROLLER,"PeerID",zwave_dev_num);
			if not peerID then
	       		peerID = luup.variable_get(SCObj.OldServiceId,"PeerID",zwave_dev_num);
				if peerID then
					MigrateDeviceToNewServiceId(zwave_dev_num)
				end
			end
			local peer_dev_num = tonumber(peerID)
			if peer_dev_num ~= nil and peer_dev_num > 0 and luup.devices[peer_dev_num] == nil and not SCObj.HasLocalControl then
			   	log("Peer Device", peer_dev_num, " Deleted. Also deleting Z-Wave device ", zwave_dev_num, ", ZWave node ID ", v.id)
			   	luup.variable_set(SID_SCENECONTROLLER,"PeerID","",zwave_dev_num)
		   		luup.call_action(SID_HAG,"DeleteDevice", {DeviceNum = zwave_dev_num}, 0);
			elseif peerID == nil or peerID == "" or (peerID ~= "Pending" and luup.devices[tonumber(peerID)] == nil) then
			   	log("Creating peer device for Vera device ID ", zwave_dev_num, " ZWave node ID ", v.id)
			   	count = count + 1
			   	luup.variable_set(SID_SCENECONTROLLER,"PeerID","Pending",zwave_dev_num)
		   		luup.variable_set(SID_SCENECONTROLLER,"ShowSingleDevice","1",zwave_dev_num)
		   		if not SCObj.HasLocalControl then
					luup.attr_set("invisible","1",zwave_dev_num)
					luup.attr_set("Invisible","1",zwave_dev_num)
				end
			    luup.call_action(SID_HAG,"CreateDevice", {
				   	deviceType = SCObj.DevType;
				   	internalID = zwave_dev_num;
					Description = v.description .. " Scene Controller";
				   	UpnpDevFilename = SCObj.DeviceXml;
				   	UpnpImplFilename =  "I_GenGenSceneController.xml";
					RoomNum = v.room_num;
					StateVariables = SID_SCENECONTROLLER .. ",PeerID=" .. zwave_dev_num .. "\n" ..
					                 SID_SCENECONTROLLER .. ",ShowSingleDevice=1" },0)
			else
				-- Fix the % bug in 7.31 firmware
				local percentEscaped = luup.variable_get(SID_SCENECONTROLLER,"percentCharsEscaped",peer_dev_num);
				if percentEscaped ~= "1" then
					ForAllVariables(peer_dev_num, FixPercentEscape)
					luup.variable_set(SID_SCENECONTROLLER,"percentCharsEscaped","1",peer_dev_num);
				end

			end
		end
	end
	if count > 0 then
		log("Peer devices created: ", count, " reloading.")
		luup.call_action(SID_HAG, "Reload", {}, 0)
	end
	VLog("Finish CreatePeerDevices()")
end


local function CheckInvisible(devNumString)
	VEntry()
	if not SCObj.HasLocalControl then
		local zwave_dev_num = tonumber(devNumString)
		local retry = false
		if zwave_dev_num < 0 then
			zwave_dev_num = -zwave_dev_num
			retry = true
		end
		local ShowSingleDevice = luup.variable_get(SID_SCENECONTROLLER,"ShowSingleDevice",zwave_dev_num)
		if type(ShowSingleDevice) ~= "string" or (ShowSingleDevice ~= "1" and ShowSingleDevice ~= "0") then
			ShowSingleDevice = "1"
		DLog("CheckInvisible: Setting ShowSingleDevice for the first time.")
			luup.variable_set(SID_SCENECONTROLLER,"ShowSingleDevice",ShowSingleDevice,zwave_dev_num)
		end
		local invisible = luup.attr_get("invisible",zwave_dev_num)
		local Invisible = luup.attr_get("Invisible",zwave_dev_num)
		DLog("CheckInvisible: zwave_dev_num=", zwave_dev_num, " ShowSingleDevice=", ShowSingleDevice, " invisible=", invisible, " retry=", retry)
		if ShowSingleDevice == "0" then
	  		if luup.version_major < 7 then
				ShowSingleDevice = "" -- invisible should be 1 or empty in UI5
			end
			retry = false -- no need to retry below if we are allowing the z-wayve device to be seen.
		end
		if invisible ~= ShowSingleDevice then
			-- UI5 bug has capitalization issues with "I/invisible" so set it both ways.
		DLog("CheckInvisible: zwave_dev_num=", zwave_dev_num, ": Setting invisible and Invisible to ", ShowSingleDevice)
			luup.attr_set("invisible",ShowSingleDevice,zwave_dev_num)
			luup.attr_set("Invisible",ShowSingleDevice,zwave_dev_num)
		end
		-- Ugly hack: Due to a bug/misfeature in UI7 which keeps resetting "invisible" to 0, we need to retry again after the initial scan
		-- SetTuningParameters calls CheckInvisible with a negative number to signify a one-time retry through call_delay.
		-- if retry then
		--  luup.call_delay("CheckInvisible",5,tostring(zwave_dev_num),true)
		-- end
	end
end

--
-- Called in the second pass of initialization after the peer device is created to connect it
-- back to the ZWave device
--
-- ZWave device --> PeerID = Peer Device
-- Peer device --> id = ZWave device
--
-- local 
function ConnectPeerDevice(peer_dev_num)
	DEntry()
	local v2 = luup.devices[peer_dev_num]
	local candidate = Devices[v2.device_type]
	if candidate then
		SCObj = candidate
		DLog("ConnectPeerDevice: peer_dev_num=", peer_dev_num, " Found device: ", SCObj.Name)
	else
		ELog("ConnectPeerDevice: Device", peer_dev_num, " type ", v2.device_type, " not supported.")
		return false, false
	end
	local newDevice = false;
	if v2.device_type == SCObj.DevType and v2.device_num_parent == 0 then
		local zwave_dev_num = tonumber(v2.id);
		local v = luup.devices[zwave_dev_num]
		if v == nil then
			log("Device ", zwave_dev_num, " for peer device ", peer_dev_num, " no longer exists. Deleting")
			ForAllModes(peer_dev_num, function(modeStr, mode, screen, virtualButton, state, eventCode)
				if mode.zWaveSceneId then
					for i = 1, #mode do
						RemoveDeviceActuatorConf(mode[i].device, mode.zWaveSceneId, mode.offZWaveSceneId)
					end
				end
			end )
			ZWaveSceneIdCacheList[zwave_dev_num] = nil
		   	luup.call_action(SID_HAG,"DeleteDevice", {DeviceNum = peer_dev_num}, 0);
		else
			CheckInvisible(-zwave_dev_num)
			SCObj.SetTuningParameters(zwave_dev_num, peer_dev_num)
			SetTuningParameter("VerboseLogging", VerboseLogging, zwave_dev_num, peer_dev_num, "SceneController_VerboseLoggingChange");
			local peerID = luup.variable_get(SID_SCENECONTROLLER,"PeerID",zwave_dev_num)
		    if peerID == "Pending" then
				log("Connecting Device ", zwave_dev_num, "(ZWave node ", v.id, ") to peer device ", peer_dev_num)
		   		luup.variable_set(SID_SCENECONTROLLER,"PeerID",peer_dev_num,zwave_dev_num)
				if IsVeraPrimaryController() then
					SCObj.SetDefaultSettings(peer_dev_num);
				end
				newDevice = true;
			elseif tonumber(peerID) ~= peer_dev_num then
				ELog("ZWave device ", zwave_dev_num, " PeerID is ", tostring(peerID), ", should have been ", peer_dev_num, ". Deleting peer device ", peer_dev_num)
		   		luup.call_action(SID_HAG,"DeleteDevice", {DeviceNum = peer_dev_num}, 0);
				return;
			end
			local zpeer = luup.variable_get(SID_SCENECONTROLLER,"PeerID",peer_dev_num)
			if not zpeer then
				zpeer = luup.variable_get(SCObj.OldServiceId,"PeerID",peer_dev_num)
				if zpeer then 
					MigrateDeviceToNewServiceId(peer_dev_num)
				end
			end
			if zpeer ~= tostring(zwave_dev_num) then
				luup.variable_set(SID_SCENECONTROLLER, "PeerID", tostring(zwave_dev_num), peer_dev_num)
			end
			if IsVeraPrimaryController() then
              VariableWatch("SceneController_ConfiguredChanged",SID_HADEVICE,"Configured",zwave_dev_num,"")
			  VariableWatch("SceneController_CommFailureChanged",SID_HADEVICE,"CommFailureTime",zwave_dev_num,"")
			  PropagateCommFailure(zwave_dev_num, peer_dev_num)
			end
			if not SCObj.HasLocalControl then
				-- Rename the Z-Wave device to match the peer device if the Z-Wave deice is not visible.
				local zwave_name = luup.attr_get("name",zwave_dev_num);
				local peer_name = luup.attr_get("name",peer_dev_num);
				local desired_name = peer_name .. " Z-Wave";
				if zwave_name ~= desired_name then
					luup.attr_set("name", desired_name, zwave_dev_num);
				end
			end
		end
	else
		ELog("Incorrect peer device ", peer_dev_num)
	end
	VLog("Finish ConnectPeerDevice()")
	return true, newDevice
end

local IsPrimaryController
-- local 
function IsVeraPrimaryController()
	if IsPrimaryController == nil then
		if param[luup.device].SCENE_CTRL_UseWithSlaveController > 0 then
		   IsPrimaryController = true
		else
			IsPrimaryController = false
			for k,v in pairs(luup.devices) do
				if v.device_type == DEVTYPE_ZWN then
					local role = luup.variable_get(SID_ZWN, "Role", k)
					local masterslave, sis, pri = tostring(role):match("(%a+) SIS:(%a+) PRI:(%a+)")
					if pri ~= "NO" then
						IsPrimaryController = true
					end
					DLog("IsVeraPrimaryController: Found Z-Wave network=", k, " Role=", role, " masterslave=", masterslave, " sis=", sis, " pri=", pri, " IsPrimaryController=", IsPrimaryController)
					break
				end
			end
		end
		if not IsPrimaryController then
			log("This is not the primary controller. Most functions disabled.")
		end
	end
	return IsPrimaryController
end

local function SetConfigurationOption(name, dev_num, node_id, option, value, delay)
	if not delay then
		delay = 0
	end
	local data = "112 4 " .. tostring(option) .. " 1 " .. tostring(value)
    EnqueueZWaveMessage(SCObj, "SetConfigurationOption("..name..":"..tostring(option).."="..tostring(value)..")"..node_id, node_id, data, dev_num, delay)
end

-- SetConfigurationOptions takes an unsorted, possibly sparse table of configuration parameters and values and generates a series of one or more
-- confguration_set or configuration_bulk_set commands to optimally set the configuration.
local function SetConfigurationOptions(name, dev_num, node_id, size, max, parameters)
	DEntry()
	local param2 = {}
	local i = 1;
	for k,v in pairs(parameters) do
		param2[i] = k
		i = i + 1
	end
	table.sort(param2)
	local first = 1;
	local data, label
	for j = 1, #param2 do
		if j == #param2 or param2[j+1] ~= param2[j] + 1 or 1 + j - first >= max then
			if first == j and param2[j] <= 255 then
				data = "112 4 " .. tostring(param2[j]) .. " " .. tostring(size) ..  " " .. tostring(parameters[param2[j]])
				label = "SetConfigurationOption("..name..":"..tostring(param2[j]).."="..tostring(parameters[param2[j]])..")"..node_id
			else
				local offset1 = math.floor(param2[first] / 256)
				local offset2 = param2[first] % 256
				data = "112 7 " .. tostring(offset1) .. " " .. tostring(offset2) .. " " .. tostring(1 + j - first) .. " " .. tostring(size)
				label = "SetConfigurationOptions("..name..":"..tostring(param2[first]) .. "-" .. tostring(param2[j]) .. "="
				for k = first, j do
					for n = size, 1, -1 do
						data = data .. " " .. tostring(bit.band(255,bit.rshift(parameters[param2[k]], (n-1) * 8)))
					end
					if k ~= first then
						label = label .. ","
					end
					label = label .. tostring(parameters[param2[k]])
				end
				label = label .. ")" .. node_id
			end
    		EnqueueZWaveMessage(SCObj, label, node_id, data, dev_num, 0)
    		first = j + 1
		end
	end
end

local timeoutSequenceNumber = 0
local timeoutList = {}
function SceneController_ScreenTimeout(data)
	VEntry()
	if IsVeraPrimaryController() then
		local peerStr, seqStr, timeoutScreen, timeoutEvent, cuurentScreen = data:match("^(%d+),(%d+),([^,]+),([^,]+),([^,]+)$")
		local peer_dev_num = tonumber(peerStr)
		local sequenceNumber = tonumber(seqStr)
		if peer_dev_num and sequenceNumber and timeoutList[peer_dev_num] == sequenceNumber then

			DLog("ScreenTimeout: peer_dev_num=", peer_dev_num, " sequenceNumber=", sequenceNumber, " timeoutScreen=", timeoutScreen, " timeoutEvent=", timeoutEvent, " cuurentScreen=", cuurentScreen)
			if SCObj.HasMultipleScreens then
				-- DoTimeout, not force, not indicatorOnly
				SetScreen(peer_dev_num, timeoutScreen, true, false, false);
			end

			if SCObj.NumStatusLEDs > 0 then
				local modeStr, mode
				if timeoutEvent == "S" then
					modeStr = "-L0-" -- Set status mode, don't change the color
					mode = ParseModeString(modeStr)
					SCObj.SetLEDStatusValue(peer_dev_num, 1, modeStr, mode, cuurentScreen, false, true);
				else
					local lastButton = luup.variable_get(SID_SCENECONTROLLER,"LastSceneAction_"..timeoutEvent, peer_dev_num)
					if lastButton then
						luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction", lastButton..timeoutEvent, peer_dev_num)
						modeStr = GetModeStr(peer_dev_num, cuurentScreen, tonumber(lastButton), nil, timeoutEvent, false, false)
						mode = ParseModeString(modeStr)
						SCObj.SetLEDStatusValue(peer_dev_num, lastButton, modeStr, mode, cuurentScreen, false, true);
					end
				end
			end

			local timeout_scene = SCObj.EncodeVeraSceneNumber(cuurentScreen, 1, 1, 1, false, true)
			TriggerVeraScene(peer_dev_num, 	timeout_scene, true); -- activate

			RunZWaveQueue("ScreenTimeout", 0)
		else
	        DLog("ScreenTimeout: Ignore stale ScreenTimeout(", data, ") Current timeoutSequenceNumber=", timeoutSequenceNumber);
		end
	end
end

function SceneController_SetScreenTimeout(peer_dev_num, screen, timeoutEnable, timeoutScreen, timeoutSeconds, timeoutEvent)
	DEntry()
	if IsVeraPrimaryController() then
		local enable = timeoutEnable == 1 or timeoutEnable == "1" or timeoutEnable == "true" or timeoutEnable == "yes" or timeoutEnable == true	-- FOr UPnP boolean spec compatibility
		luup.variable_set(SID_SCENECONTROLLER, "TimeoutEnable_" .. screen, tostring(enable), peer_dev_num)
		luup.variable_set(SID_SCENECONTROLLER, "TimeoutScreen_" .. screen, timeoutScreen, peer_dev_num)
		luup.variable_set(SID_SCENECONTROLLER, "TimeoutSeconds_" .. screen, tostring(timeoutSeconds), peer_dev_num)
		luup.variable_set(SID_SCENECONTROLLER, "TimeoutEvent_" .. screen, tostring(timeoutEvent), peer_dev_num)
		if screen == GetCurrentScreen(peer_dev_num) then
			SetScreenTimeout(peer_dev_num, true)
			RunZWaveQueue("SetScreenTimeout", 0)
		end
	end
end

-- Update the timeout for the current screen to go to the next screen if enabled
-- doTimeout is true to enable and override all previous timeouts, 
-- false to disable and override all previous timeouts,
-- nil to do nothing
-- local 
function SetScreenTimeout(peer_dev_num, doTimeout)
	DEntry()
	if doTimeout ~= nil then
		-- Whatever we do below, always make any previous timer stale
		local screen = GetCurrentScreen(peer_dev_num)
		timeoutSequenceNumber = timeoutSequenceNumber + 1
		timeoutList[peer_dev_num] = timeoutSequenceNumber
		if doTimeout then
			local enable = luup.variable_get(SID_SCENECONTROLLER, "TimeoutEnable_" .. screen, peer_dev_num)
			local seconds = luup.variable_get(SID_SCENECONTROLLER, "TimeoutSeconds_" .. screen, peer_dev_num);
			local timeoutScreen = luup.variable_get(SID_SCENECONTROLLER, "TimeoutScreen_" .. screen, peer_dev_num);
			if seconds == nil then
				seconds = 0
			end
			local timeoutEvent = luup.variable_get(SID_SCENECONTROLLER, "TimeoutEvent_" .. screen, peer_dev_num);
			if timeoutEvent == nil then
				timeoutEvent = "1"
			end
			seconds = tonumber(seconds)
			if enable == "true" and seconds > 0 and timeoutScreen then
				DLog("SetScreenTimeout: Setting timeout timer for ", seconds, " seconds. peer_dev_num=", peer_dev_num, " current screen=", screen, " timeoutScreen=", timeoutScreen, " timeoutSequenceNumber=", timeoutSequenceNumber);
				luup.call_delay("SceneController_ScreenTimeout", tonumber(seconds), tostring(peer_dev_num)..","..tostring(timeoutSequenceNumber)..","..timeoutScreen..","..timeoutEvent..","..screen, true);
			end
		end
	end
end

function SceneController_SetNumLines(peer_dev_num, screen, lines)
	VEntry()
	if IsVeraPrimaryController() then
		local oldLines = luup.variable_get(SID_SCENECONTROLLER, "NumLines_" .. screen, peer_dev_num)
		if oldLines then
			oldLines = tonumber(oldLines)
		else
			oldLines = SCObj.NumButtons
		end
		lines = tonumber(lines)
		if lines < SCObj.NumButtons then
			lines = SCObj.NumButtons
		elseif lines > SCObj.NumButtons*2 then
			lines = SCObj.NumButtons*2
		end
		luup.variable_set(SID_SCENECONTROLLER, "NumLines_" .. screen, tostring(lines), peer_dev_num)
		local currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
		if currentScreen == screen and oldLines ~= lines then
			-- doTimeout, force, not indicatorOnly
			SetScreen(peer_dev_num, currentScreen, true, true, false);
			RunZWaveQueue("SetNumLines", 0);
		end
	end
end

function SceneController_SetNumEvents(peer_dev_num, screen, button, events)
	VEntry()
	if IsVeraPrimaryController() then
		luup.variable_set(SID_SCENECONTROLLER, "Events_" .. screen.."_"..tostring(button), events, peer_dev_num)
		SCObj.SetMultiTap(peer_dev_num, screen, button, false);
		RunZWaveQueue("SetNumEvents", 0);
	end	
end

function SceneController_UpdateNightLightMode(peer_dev_num, nightLightMode)
	VEntry()
	if IsVeraPrimaryController() then
		local oldNightLight = luup.variable_get(SID_SCENECONTROLLER, "NightLightMode", peer_dev_num)
		if oldNightLight then
			oldNightLight = oldNightLight == "+"
		else
			oldNightLight = nil
		end
		luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", nightLightMode, peer_dev_num)
		nightLightMode = nightLightMode == "+"
		if oldNightLight ~= nightLightMode then
			SCObj.SetNightLightMode(peer_dev_num, nightLightMode)
			RunZWaveQueue("UpdateNightLightMode", 0);
		end
	end
end

-- Perform one of the actiona A through H. This is generally used to set the Status LEDs to a particular
-- state from within a Vera scene and cancels any LED status timeout. 
function SceneController_PerformAction(peer_dev_num, action)
	VEntry()
	local curScreen = GetCurrentScreen(peer_dev_num)
	local modeStr = GetModeStr(peer_dev_num, curScreen, SCObj.NumButtons+1, nil, action, true, false)
	local mode = ParseModeString(modeStr)
	luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction", tostring(SCObj.NumButtons+1)..action, peer_dev_num)
	luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction1", action, peer_dev_num)
	timeoutSequenceNumber = timeoutSequenceNumber + 1
	timeoutList[peer_dev_num] = timeoutSequenceNumber
	SCObj.SetLEDStatusValue(peer_dev_num, SCObj.NumButtons+1, modeStr, mode, curScreen, false, true); -- notForce, SaveRestore, notNegative
	RunZWaveQueue("PerformAction", 0);
end

-- Mode strings consist of ~? Prefix {newScreen:+}? (L..)?(B###)?[AaMmVF]?Q?{(S|C{SceneId@}?{offSceneId@}?)}? {entry}+
-- Prefix is M for momentary, T for Toggle, etc and may be preceded by ~ for an "off" negative.
-- newScreen is a letter/digit(s) such as C3 for Custom 3 or P12 for Preset 12
-- The optional L flag sets the status LED mode and is followed by two characters.
--   The first character is the action: L=normal/Level 1-9=Set status LED, A=Set All LEDs, R=Restore previous mode
--   The second character is the LED color RGBCMYWK (Red ... Black) rgbcmyw (Falshing red ... flashing white), - = No change
-- The A (All) or a (Any) flag is present if all/any device matching "on" state should turn LED on.
--   If the A/a flag is not present then if most devices can match the "on" state to turn the LED on.
-- The B flag is followed by 3 dicimal digits specifying the blink period in 100ms increments
-- For devices with multiple status LEDs, the AnyOrAll flag can take the vales:
--   M (Maximum), m (minimum), V (aVerage) of F (First) to decide the LED level when more than one deives is connected.
-- The Q (Quicktap) flag is set if the hold/release events can be ignored and tap events can be acted on immediately.
-- The S or C flags indicates that all subsequent associated devices are scene capable
--   and is optinally followed by a sceneId @ which is optional followed by an offSceneId @
-- The C flag indicates to use Cooper configuration for non-secene capable devices
-- Scene-capable modes are a ; separated list of 0 or more deviceNum,level,dimmingDuration triplets
-- Cooper condiguration modes are a ; separate list of 0 or more deviceNum,level pairs
-- Non-scene capabile modes are a ; separated list of 0 or more deviceNums
-- DeviceNums can be negative if they should only be monitored but not controlled, to set the indicator/status
-- local 
function ParseModeString(str)
	VEntry()
	local mode = {};
	if not str or (str == "") then
	    str = "M"
	end
	local prefix = str:sub(1,1)
	if prefix == "~" then
		prefix = str:sub(2,2)
		mode.prefix = prefix
		str = str:sub(3)
		mode.negative = true
	else
		mode.prefix = prefix
		mode.negative = false
		str = str:sub(2)
	end
	mode.states = 1
	mode.quickTap = false
	if prefix >= "2" and prefix <= "9" then
	   mode.states = tonumber(prefix)
	end
	if mode.prefix == "S" then
		-- Legacy direct toggle. Change into T
		mode.prefix = "T"
	end
	local newScreen, remaining = str:match("^(%a%d+):+(.*)$") -- :+ to deal with old bug
	if newScreen then
		mode.newScreen = newScreen
		str = remaining
	elseif mode.prefix == "N" then
	    -- Legacy newScreen without :
		mode.prefix = "M"
		mode.newScreen = str
		return mode
	end
	if #str >= 3 and str:sub(1,1) == "L" then
		mode.LEDAction = str:sub(2,2)
		mode.LEDColor = str:sub(3,3)
		str=str:sub(4)
	end
    if #str >= 4 and str:sub(1,1) == "B" then
    	mode.flashPeriod = tonumber(str:sub(2,4))
    	str=str:sub(5)
    end
	local mostMap = {
		A = "All",
		a = "Any",
		M = "Max",
		m = "Min",
		V = "Average",
		F = "First"
	}
	mode.allDevices = mostMap[str:sub(1,1)]
	if mode.allDevices then
		str = str:sub(2)
	else
		mode.allDevices = "Most"
	end
	local char2 = str:sub(1,1)
	if char2 == "Q" then
		mode.quickTap = true;
		str = str:sub(2)
		char2 = str:sub(1,1)
	end
	if char2 == "S" or char2 == "C" then
		mode.sceneControllable = true
		local sceneIdStr, rest = string.match(str, "^.(%d+)@(.*)$")
		if sceneIdStr then
			mode.zWaveSceneId = tonumber(sceneIdStr)
			str = rest;
			local offZWaveSceneIdStr, offRest = string.match(str, "^(%d+)@(.*)$")
			if offZWaveSceneIdStr then
				mode.offZWaveSceneId = tonumber(offZWaveSceneIdStr)
				str = offRest;
			end
		else
			str = str:sub(2)
		end
	end
	if not string.match(str,";$") then
		str = str .. ";"
	end
	while true do
		local element = nil;
		local device, level, dimmingDuration, rest = string.match(str, "^(-?%d+),(%d+),(%d+),?;(.*)")
		if device then
	    	element = {device=tonumber(device), level=tonumber(level), dimmingDuration=tonumber(dimmingDuration)};
		else
			device, level, rest = string.match(str, "^(-?%d+),(%d+);(.*)")
			if device then
	    		element = {device=tonumber(device), level=tonumber(level)};
			else
				device, rest = string.match(str, "^(-?%d+);(.*)")
				if device then
					element = {device=tonumber(device)};
				else
					break;
				end
			end
		end
		if element.device < 0 then
			element.device = -element.device
			if not mode.veraSceneSet then
				mode.veraSceneSet = {}
			end
			table.insert(mode.veraSceneSet, element)
		else
			table.insert(mode, element)
		end
		str = rest
	end
	VLog("ParseModeString: mode=",mode)
	return mode;
end

local function GenerateModeString(mode, button, eventCode)
	if not mode then
		mode = {}
	end
	if not mode.prefix then
		mode.prefix = SCObj.GetDefaultModeString(button, eventCode)
	end
	local str=mode.prefix
	if mode.negative then
		str = "~" .. str
	end
	if mode.newScreen then
		str = str .. mode.newScreen .. ":"
	end
	local first = true

	local function generateArray(arr, sign)
		if arr then
			for i = 0, #arr do
				if arr[i] and arr[i].device then
					if not first then
						str = str .. ";"
					end
					first = false
					if arr[i].dimmingDuration then
						str = str .. arr[i].device*sign .. "," .. arr[i].level .. "," .. arr[i].dimmingDuration
					elseif arr[i].level then
						str = str .. arr[i].device*sign .. "," .. arr[i].level
					else
						str = str .. arr[i].device*sign
					end
				end
			end
		end
	end

	if mode.LEDAction then
		str = str .. "L" .. mode.LEDAction
		if mode.LEDColor then
			str = str .. mode.LEDColor 
		else
			str = str .. "W"
		end
	end

	if mode.flashPeriod then
		str = str .. "B" .. string.sub("000" .. tostring(mode.flashPeriod), -3)
	end

	local mapMost = {
		All = "A",
		Any = "a",
		Max = "M",
		Min = "m",
		Average = "V",
		First = "F"
	}
	local allKey = mapMost[mode.allDevices]
	if allKey then
		str = str .. allKey
	end

	if mode.quickTap then
		str = str .. "Q"
	end

	if mode.sceneControllable then
		str = str .. "S"
		if mode.zWaveSceneId then
			str = str .. mode.zWaveSceneId .. "@"
			if mode.offZWaveSceneId then
				str = str .. mode.offZWaveSceneId .. "@"
			end
		end
	end
	generateArray(mode, 1)
	generateArray(mode.veraSceneSet, -1)
	return str;
end

-- local 
function SetIndicatorValue(peer_dev_num, indicator, force, coalesce)
	DEntry()
    local previous_indicator = 0
	if SCObj.HasIndicator then
		local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
		if not node_id then
			ELog("SetIndicatorValue: bad peer_dev_num=", peer_dev_num)
			return previous_indicator 
		end
		local previous_indicator_string = luup.variable_get(SID_SCENECONTROLLER, CURRENT_INDICATOR, peer_dev_num)
		previous_indicator = tonumber(previous_indicator_string)
		DLog("previous_indicator=",previous_indicator)
		if force or previous_indicator ~= indicator then
			local delay = 0
			if param[peer_dev_num].SCENE_CTRL_IndicatorDelay > 0 then
				if force then
					delay = param[peer_dev_num].SCENE_CTRL_IndicatorDelay * SCObj.NumButtons
				else
					local delayDiff = bit.bxor(previous_indicator, indicator)
					while delayDiff ~= 0 do
						if bit.band(1, delayDiff) ~= 0 then
						   delay = delay + param[peer_dev_num].SCENE_CTRL_IndicatorDelay	
						end
						delayDiff = bit.rshift(delayDiff, 1)
					end
				end
			end
			-- The * at the start of the label indicates that this can replace an existing command with a similar node ID and label (up to the first "(")
			-- We avoid merging in force mode to resolve redraw issues on Evolve LCD1.
			local label = "SetIndicator("..indicator..")"..node_id
			if coalesce and (param[peer_dev_num].SCENE_CTRL_IndicatorDelay == 0 or not force) then
				label = "*" .. label
			end 
			EnqueueZWaveMessage(SCObj, label, node_id,  "0x87 0x01 "..indicator, zwave_dev_num, delay);
			luup.variable_set(SID_SCENECONTROLLER, CURRENT_INDICATOR, tostring(indicator), peer_dev_num)
		end
	end
	return previous_indicator
end

-- Return a score between 0 and 1 indicating how close the devices for this mode match the target levels. Higher is closer.
local function GetDeviceScoreForMode(mode, peer_dev_num, screen, virtualButton)
	VEntry()
	local score = 1
	local count = 0
	local min = 0
	local max = 0

	local function GetScoreForElement(obj)
	    VEntry("GetScoreForElement")
		local device = obj.device
		local status, lastUpdate, service, variable = GetDeviceStatus(device)
		if VerboseLogging >= 2 then
	        DLog("GetDeviceScoreForMode: monitored device=", device, " status=", status, " lastUpdate=", os.date(nil,tonumber(lastUpdate)), " service=", service, " variable=", variable)
		end
		if status ~= nil then
			local target = 100;
			if obj.level ~= nil then
				target = obj.level
			end
			local context = tostring(peer_dev_num) .. "," .. screen
			VariableWatch("SceneController_WatchedIndicatorDeviceChanged", service, variable, device, context)
			WatchDevicesAssociatedWith(device)
			if mode.allDevices == "Min" then
				if count > 0 and status > min then
					return
				end
				min = status;
				count = 0
				score = 0 
			elseif mode.allDevices == "Max" then
				if count > 0 and status < max then
					return
				end
				max = status;
				count = 0
				score = 0 
			end
			count = count + 1
			score = score + ((target - status) ^ 2)
		else
			ELog("GetScoreForElement: Device ", device, " cannot be watched by ", luup.devices[peer_dev_num].description, ", screen: ", screen, ", button: ", virtualButton)
		end
	    VLog("Running badScore=", score, " count=", count)
	end

    local foundOne = false
	for i = 1, #mode do
		GetScoreForElement(mode[i])
		if mode.allDevices == "First" then
			foundOne = true;
			break
		end
	end
	if not foundOne and mode.veraSceneSet then
		for i = 1, #mode.veraSceneSet do
			GetScoreForElement(mode.veraSceneSet[i])
			if mode.allDevices == "First" then
				break
			end
		end
	end

	if count == 0 then
		return 0
	else
		return 1 / score
	end
end


local function ChooseBestFromMultiState(peer_dev_num, screen, virtualButton, mode, eventCode, nextState, cache)
	DEntry()
	local optEvent = OptEvent(eventCode)
	local stateVarName = "State_"..screen.."_"..virtualButton..optEvent
	local oldState = luup.variable_get(SID_SCENECONTROLLER, stateVarName, peer_dev_num)
	if cache and not nextState and oldState then
		return tonumber(oldState)
	end 
	local states = tonumber(mode.prefix)
	local bestState = 1
	local bestScore	= 0
	local negative = mode.negative
	for state = 1, states do
		if state > 1 or negative then
			local modeStr = GetModeStr(peer_dev_num, screen, virtualButton, state, eventCode, false, false)
			mode = ParseModeString(modeStr)
		end
		local score = GetDeviceScoreForMode(mode, peer_dev_num, screen, virtualButton)
	    DLog("ChooseBestFromMultiState:    state=", state, " mode=", mode, " score=", score)
		if score > bestScore then
			bestState = state
		    bestScore = score
		end
	end
	if bestScore <= 0 and oldState then
		bestState = tonumber(oldState)
	end
	DLog("ChooseBestFromMultiState: bestState=", bestState, " bestScore=", bestScore)
	if nextState then
		if negative then
			bestState = ((bestState - 2) % states) + 1
		else
			bestState = (bestState % states) + 1
		end
		DLog("ChooseBestFromMultiState: advancing bestState to ", bestState)
	end
	local newState = tostring(bestState)
	if oldState ~= newState then
		luup.variable_set(SID_SCENECONTROLLER, stateVarName, newState, peer_dev_num)
	end
	return bestState
end

-- GetBinaryStateOfDevicesOnMode returns a Boolean as to whether or net an indicator LED
-- should be highlighted based on the devices on that mode and the "allDevices" status
-- mode.allDevices can be one of 
-- Any - return true if any devices are on
-- All - return true if all devices aee on
-- First - return true if the first measurable device in the mode list is on
-- Most (default) - return true if most devices are on
-- Returns bool, true if at least one device was found.
-- Returns defaultVal, false if no devices with a measurable status are attached to the mode.
local function GetBinaryStateOfDevicesOnMode(peer_dev_num, mode, screen, watch, virtualButton, defaultVal)
	DEntry()
	local num_off = 0
	local num_on = 0
	local highlighted = defaultVal or false -- nil -> false
	local threshold = 5

	local function checkElement(obj) 
		if not obj or not obj.device or not luup.devices[obj.device] then
			return false
		end
		local status, lastUpdate, service, variable = GetDeviceStatus(obj.device)
		if VerboseLogging >= 2 then
			DLog("monitored device=", obj.device, ": name=", luup.devices[obj.device].description," threshold=", threshold, " status=", status, " lastUpdate=", os.date(nil,tonumber(lastUpdate)), " service=", service, " variable=", variable)
		end
		if status ~= nil then
			if status >= threshold then
				num_on = num_on + 1
			else
				num_off = num_off + 1
			end
			if watch then
				local context = tostring(peer_dev_num) .. "," .. screen
				VariableWatch("SceneController_WatchedIndicatorDeviceChanged", service, variable, obj.device, context)
				WatchDevicesAssociatedWith(obj.device)
			end
			return true
		else
			return false
		end
	end

	local foundOne = false
	for i = 1, #mode do
		foundOne = checkElement(mode[i])
		if foundOne and mode.allDevices == "First" then
			break
		end
	end
	if mode.veraSceneSet and (not foundOne or mode.allDevices ~= "First") then
		for i = 1, #mode.veraSceneSet do
			foundOne = checkElement(mode.veraSceneSet[i])
			if foundOne and mode.allDevices == "First" then
				break
			end
		end
	end
	if num_on > 0 then 
		if mode.allDevices == "All" then
			highlighted = num_off == 0;
		elseif mode.allDevices == "Any" then
			highlighted = true;		
		else -- Most, First
			highlighted = num_on >= num_off;
		end
	elseif num_off > 0 then
		highlighted = false
	end
	DLog("GetBinaryStateOfDevicesOnMode: num_on=", num_on, " num_off=", num_off, " highlighted=", highlighted)
	return highlighted, foundOne
end

-- GetMultilevelStateOfDevicesOnMode is similar to GetBinaryStateOfDevicesOnMode except that
-- it returns a value between 0 and 100.
-- mode.allDevices can be one of 
-- Avrage(default) - return the average value of all devices
-- Min - Return the minimum value
-- Max - Return the maximum value
-- First - Return the first measurable value in the mode list
-- Always returns 0 if no devices with a measurable status are attached to the mode.
local function GetMultilevelStateOfDevicesOnMode(peer_dev_num, mode, screen, watch, virtualButton)
	VEntry()
	local min
	local max
	local sum = 0
	local count = 0

	local function checkElement(obj) 
		if not obj or not obj.device or not luup.devices[obj.device] then
			return false
		end
		local status, lastUpdate, service, variable = GetDeviceStatus(obj.device)
		if VerboseLogging >= 2 then
			DLog("monitored device=", obj.device, ": name=", luup.devices[obj.device].description," status=", status, " lastUpdate=", os.date(nil,tonumber(lastUpdate)), " service=", service, " variable=", variable)
		end
		if status ~= nil then
			if not min or status < min then
				min = status
			end
			if not max or status > max then
				max = status
			end
			sum = sum + status
			count = count + 1
			if watch then
				local context = tostring(peer_dev_num) .. "," .. screen
				VariableWatch("SceneController_WatchedIndicatorDeviceChanged", service, variable, obj.device, context)
				WatchDevicesAssociatedWith(obj.device)
			end
			return true
		else
			return false
		end
	end

	local foundOne = false
	for i = 1, #mode do
		foundOne = checkElement(mode[i])
		if foundOne and mode.allDevices == "First" then
			break
		end
	end
	if mode.veraSceneSet and (not foundOne or mode.allDevices ~= "First") then
		for i = 1, #mode.veraSceneSet do
			foundOne = checkElement(mode.veraSceneSet[i])
			if foundOne and mode.allDevices == "First" then
				break
			end
		end
	end
	if count == 0 then
		return 0
	elseif mode.allDevices == "Min" then
		return min
	elseif mode.allDevices == "Max" then
		return max
	end
	-- Average, First
	return math.floor(sum / count + .5) 
end

local function SetIndicator(peer_dev_num, screen, force, coalesce, cache)
	DEntry()
	if force then
		cache = false
	end
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local CommFailure = luup.variable_get(SID_HADEVICE, "CommFailure", zwave_dev_num)
	if CommFailure and tonumber(CommFailure) ~= 0 and not force then
		DLog("SetIndicator skipped due to CommFailure of Z-Wave dev num=",zwave_dev_num," node id=",node_id, " peer dev nun=",peer_dev_num)
		return
	end
	if IndicatorLocks[peer_dev_num] and IndicatorLocks[peer_dev_num] > 0 then
		-- Handle a nasty race condition:
		-- 1. Cooper RFWC5 does an "All off" operation which controls a Cooper dimmer
		-- 2. Cooper dimmer sends Hail to vera
		-- 3. Vera queries dimmer status, determines it is off.
		-- 4. GenGenSceneController detects that the dimmer's status has changed, calls SetIndicator with only one bit cleared
		-- 5. This messes up the "Indicator Get/Indicator Report" operation done when GenGenSceneController detects the Basic Set Off.
		-- We solve this by temporarily deferring any SetIndicator operations on the given device while a Indicator Get is in progress.
		DLog("SetIndicator. Quick return because IndicatorLocks[", peer_dev_num, "]=", IndicatorLocks[peer_dev_num])
		IndicatorLocks[peer_dev_num] = 3
		return
	end
	local lastAction = luup.variable_get(SID_SCENECONTROLLER,"LastSceneAction", peer_dev_num)
	local virtualButton = 1
	local eventCode = "1"
	if lastAction then
		VLog("SetIndicator: lastAction=", lastAction)
		virtualButton = tonumber(lastAction:sub(1,1))
		eventCode = lastAction:sub(2,2)
	end
	local optEvent = OptEvent(eventCode)
	if SCObj.HasIndicator then
	    local indicator = 0
		local bestXbutton = 0
		local bestXhidden = false
		local bestXscore = 0
		local xSceneFound = false
		local xSceneButton = 0
		local xSceneHidden = false
		local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, screen)
		local oldIndicator = luup.variable_get(SID_SCENECONTROLLER, CURRENT_INDICATOR, peer_dev_num)
		if oldIndicator then
			oldIndicator = tonumber(oldIndicator)
		else
			oldIndicator = 0
		end
		for virtualButton = 1, numLines do
			local physicalButton = virtualButton - scrollOffset
			local hidden = physicalButton < 1 or (physicalButton == 1 and scrollOffset > 0) or
			               physicalButton > SCObj.NumButtons or (physicalButton == SCObj.NumButtons and scrollOffset < numLines-SCObj.NumButtons)
			local modeStr = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, false, false)
			local mode  = ParseModeString(modeStr)
		    local highlighted = false;
			if not hidden or mode.prefix == "X" then
				if mode.prefix == "X" then
					if #mode > 0 or mode.veraSceneSet then
						local score = GetDeviceScoreForMode(mode, peer_dev_num, screen, virtualButton)
						if score > bestXscore then
							bestXscore = score
							bestXbutton = virtualButton
							bestXhidden = hidden
						end
						VLog("  Score=",score," bestXscore=",bestXscore," bestXbutton=",bestXbutton, " bestXhidden=",bestXhidden) 
					else
				   		local scene = luup.variable_get(SID_SCTRL, "sl_SceneActivated", peer_dev_num)
						if scene then
							local sceneScreen, secenButton, sceneState, sceneTaps, sceneHold, sceneTimeout = SCOBJ.DecodeVeraSceneNumber(tonumber(scene))
							highlighted = sceneScreen == screen and sceneButton == virtualButton
							VLog("  Default scene=",scene," sceneButton=",sceneButton," virtualButton=",virtualButton," highlighted=",highlighted) 
						else
							highlighted = false
						end
						if highlighted then
							xSceneFound = true
							if hidden then
								xSceneButton = virtualButton
								xSceneHidden = true
								highlighted = false
							end
						end
					end
				elseif #mode == 0 and not mode.veraSceneSet then
					highlighted = bit.band(oldIndicator,SCObj.PhysicalButtonToIndicator(physicalButton, true)) ~= 0
				elseif mode.states > 1 then
					local oldModeStr, oldState = GetModeStr(peer_dev_num, screen, virtualButton, nil, "1", true, false)
					local state
					if SCObj.HasScreen then
						local oldLabels, oldFonts, oldAligns = {}, {}, {}
						ChooseLabelFontAndAlign(peer_dev_num, screen, physicalButton, virtualButton, oldState, oldLabels, oldFonts, oldAligns, true)
						local newLabels, newFonts, newAligns = {}, {}, {}
						state = ChooseLabelFontAndAlign(peer_dev_num, screen, physicalButton, virtualButton, nil, newLabels, newFonts, newAligns, cache)
						if (oldLabels[physicalButton] ~= newLabels[physicalButton] or oldFonts[physicalButton] ~= newFonts[physicalButton] or oldAligns[physicalButton] ~= newAligns[physicalButton]) then
							EVLCDWrapStrings(newLabels, newFonts, newAligns, node_id, zwave_dev_num, SCREEN_MD.NoChange)
						end
					else
						state = ChooseBestFromMultiState(peer_dev_num, screen, virtualButton, mode, false, cache) -- current state
					end
					if state ~= oldState or force then
						local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
						local newModeStr, nextState = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, true, true)
						UpdateAssociationForPhysicalButton(zwave_dev_num, screen, force, oldModeStr, newModeStr, physicalButton, virtualButton, nextState, eventCode)
					end
				elseif mode.prefix == "T" then
					highlighted = GetBinaryStateOfDevicesOnMode(peer_dev_num, mode, screen, true, virtualButton, false)
					DLog("button=", physicalButton, " mode=", modeStr, " mode.allDevices=", mode.allDevices, " highlighted=", highlighted)
				end
			end
			if highlighted then
				indicator = bit.bor(indicator,SCObj.PhysicalButtonToIndicator(physicalButton, true))
			end
		end

		local function MakeLineVisible(virtualButton)
			local wouldBePhysicalButton = virtualButton - scrollOffset
			VEntry("MakeLineVisible")
			if wouldBePhysicalButton == 1 or wouldBePhysicalButton == SCObj.NumButtons then
				-- Scrolling by one line up or down would make the highlighted line visible
				SceneButtonPressed(peer_dev_num, true, wouldBePhysicalButton, false, "1")
			else
				-- We need to scroll by more than one line so redraw the screen but 
				-- try to put the highlighted line in the center.
				local newScrollOffset = virtualButton - math.floor((SCObj.NumButtons+1)/2)
				if newScrollOffset < 0 then
					newScrollOffset = 0
				elseif newScrollOffset > numLines - SCObj.NumButtons then
					newScrollOffset = numLines - SCObj.NumButtons
				end
				luup.variable_set(SID_SCENECONTROLLER, SCROLLOFFSET_VAR.."_"..screen, newScrollOffset, peer_dev_num)
				SetScreen(peer_dev_num, screen, true, false, false)
			end
		end

		if xSceneFound then
			if xSceneHidden then
				MakeLineVisible(xSceneButton)
				indicator = nil
			end
		elseif bestXbutton > 0 then
			if bestXhidden then
				MakeLineVisible(bestXbutton)
				indicator = nil
			else
				indicator = bit.bor(indicator,SCObj.PhysicalButtonToIndicator(bestXbutton - scrollOffset, true))
			end
		end
		if indicator then
	   		DLog("New indicator value=", indicator)
			SetIndicatorValue(peer_dev_num, indicator, force, coalesce)
		end
	end
	if SCObj.NumStatusLEDs > 0 then
		if lastAction then
			local modeStr2 = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, false, false)
			local mode2 = ParseModeString(modeStr2)
			if mode2.negative and virtualButton > 1 then
				virtualButton = virtualButton-1
				modeStr2 = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, false, false)
				mode2 = ParseModeString(modeStr2)
				mode2.negative = true
			end
			if mode2.states > 1 then
				if not cache then
					ChooseBestFromMultiState(peer_dev_num, screen, virtualButton, mode2, eventCode, false, cache) -- current state
				end
				modeStr2 = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, true, false)
				mode2 = ParseModeString(modeStr2)
				mode2.negative = false
			end
			SCObj.SetLEDStatusValue(peer_dev_num, virtualButton, modeStr2, mode2, screen, force, true)
		end
	end
end

-- This is a variable watch trigger which gets called whenever a device status changes which
-- may affect the highlighting on the screen. It is called in a call_delay since it may call SendData
function SceneController_WatchedIndicatorDeviceChanged(device, service, variable, value_old, value_new, context)
	DEntry()
	if IsVeraPrimaryController() then
		local ix1, ix2, peer_string, screen = context:find("(%d+),(%w+)")
		local peer_dev_num = tonumber(peer_string)
		local curScreen = GetCurrentScreen(peer_dev_num)
	    DLog("WatchedIndicatorDeviceChanged: context=", context, " peer_dev_num=", peer_dev_num, " screen=", screen, " curScreen=", curScreen);
		if not SCObj.HasScreen or curScreen == screen then
			SetIndicator(peer_dev_num, screen, false, true, false) -- no force, coalesce, no cache
		end
		RunZWaveQueue("WatchedIndicatorDeviceChanged", 250) -- 250ms delay to avoid a "cannot get lock" crash as job is in progress
	end
end

-- local 
function GetNumLinesAndScrollOffset(peer_dev_num, screen)
	local numLines = luup.variable_get(SID_SCENECONTROLLER, NUMLINES_VAR .. "_"  .. screen, peer_dev_num)
	local scrollOffset = 0
	if numLines then
		numLines = tonumber(numLines)
		if numLines < SCObj.NumButtons then
			numLines = SCObj.NumButtons
		end
		scrollOffset = luup.variable_get(SID_SCENECONTROLLER, SCROLLOFFSET_VAR .. "_"  .. screen, peer_dev_num)
		if scrollOffset then
			scrollOffset = tonumber(scrollOffset)
		else
			scrollOffset = 0
		end
		local maxScroll = numLines - SCObj.NumButtons
		if scrollOffset < 0 then
			scrollOffset = 0
		elseif scrollOffset > maxScroll then
			scrollOffset = maxScroll
		end
	else
		numLines = SCObj.NumButtons
	end
	return numLines, scrollOffset
end

-- Mark a scene Id as being used either by another screen or button in this controller or by another controller
-- operating on the same target device. 
-- We only create a set of unused Z-Wave scene Ids up until
-- the maximum seen so far. If the given Z-Wave scen Id is greter than the current max, then
-- we mark the next N Z-Wave scenes Ids up to the new one-1 as unused. Otherwise, we mark
-- the given Z-Wave scene Id used.
-- Note. The first SCObj.LastFixedSceneId Z-Wave Scene Iss are "default" scene IDs matching the
-- Z-Wave scene group if the selection algorithm fails.
local function MarkZWaveSceneIdUsed(zWaveSceneId, sceneSet)
	if zWaveSceneId and zWaveSceneId > SCObj.LastFixedSceneId then
		if zWaveSceneId > sceneSet.max then
			for i = sceneSet.max+1, zWaveSceneId-1 do
				sceneSet[i] = true
			end
			sceneSet.max = zWaveSceneId
		else
			sceneSet[zWaveSceneId] = nil
		end
		return true
	end
	return false
end

local function IsScenIdUsed(sceneId, sceneSet)
	if sceneId <= SCObj.LastFixedSceneId or sceneId > 255 then
		return true
	end
	if sceneId > sceneSet.max then
		return false
	end
	return sceneSet[sceneId] == nil;
end

-- Helper function for SelectZWaveSceneId to choose an available Z-Wave scene Id from the remaining sceneSet
-- Take the current Z-Wave scene ID as a hint and reuse it if it is still available. Otherwise, if the
-- Z-Wave scene Ids have not reached the max of 255, then use the next higher number, otherwise choose
-- one scene ID from the set
-- Note, some Evolve dimmers cannot accept scene ID's > 60, hence the hack below
local function pickSceneId(current, sceneSet, needSceneId, defaultScene)
	DEntry()
	if not needSceneId then
		return defaultScene
	end
	if current and not IsScenIdUsed(current, sceneSet) then 
		MarkZWaveSceneIdUsed(current, sceneSet)
		return current
	end
	for k, v in pairs(sceneSet) do
		if k ~= "max" then
			sceneSet[k] = nil
			return k
		end
	end
	if sceneSet.max < 255 then
		sceneSet.max = sceneSet.max + 1
		return sceneSet.max
	end
	return defaultScene
end

-- Iterate through all existing modes in all screens of a given device.
-- Passes the mode string, the parsed mode(object), the screen(string), the virtual button(number) and the event code (single character string) to the function.
-- The algorithm here is simialr to MigrateDeviceToNewServiceId
-- local 
function ForAllModes(peer_dev_num, func)
	VEntry()
	local count = 0
	local oldLog = SaveLogLevel(0)
	local ScreenList = SCObj.ScreenList
	for prefix, numScreens in pairs(ScreenList) do
		for screenNum = 1, numScreens do
			local screen = prefix .. screenNum
			local numLines = SCObj.NumButtons
			local lineStep = 1
			if prefix == "C" then
				local numLinesStr = luup.variable_get(SID_SCENECONTROLLER, "NumLines_" .. screen, peer_dev_num)
				if numLinesStr then
					numLines = tonumber(numLinesStr)
				end
			elseif prefix == "T" then
				lineStep = 4 -- Just do lines 1 and 5 for temperature screens
			end
			if SCObj.NumSceneEvents > 0 then
				numLines = numLines + 1
			end
			for line = 1, numLines, lineStep do
				local lastEventIndex = 1
				local eventType = "Buttons"
				if SCObj.NumSceneEvents > 0 and line == numLines then
					eventType = "Scenes"
				end
				local lastEventIndex = #SCObj.EventCodes[eventType]
				for eventIndex = 1, lastEventIndex do
					local eventCode = SCObj.EventCodes[eventType][eventIndex]
					local modeStr = GetModeStr(peer_dev_num, screen, line, nil, eventCode, false, false)
					if modeStr then
						local mode = ParseModeString(modeStr)
						RestoreLogLevel(oldLog)
						func(modeStr, mode, screen, line, 1, eventCode)
						RestoreLogLevel(0)
						count = count + 1
						if mode.states > 1 then
							local numStates = tonumber(mode.prefix)
							for state=2, numStates do
								local modeStr2 = GetModeStr(peer_dev_num, screen, line, state, eventCode, false, false)
								if modeStr2 then
									local mode2 = ParseModeString(modeStr2)
									RestoreLogLevel(oldLog)
									func(modeStr2, mode2, screen, line, state, eventCode)
									RestoreLogLevel(0)
									count = count + 1
								end
							end
						end
					end
				end
			end
		end
	end
	RestoreLogLevel(oldLog)
	VLog("ForAllModes: Function called ", count, " times.")
end

-- Iterate through all scene controller variables for the given device number
-- Passes the name of the variable (name) and the device number (integer) to the function. 
-- Not all variables may exist but the list is comprehensive.
-- The algorithm here is similar to but more inclusive than ForAllModes
-- local 
function ForAllVariables(peer_dev_num, func)
	VEntry()
	local count = 0
	local oldLog = SaveLogLevel(0)
	local ScreenList = SCObj.ScreenList
	local varList = {"PeerID", "ShowSingleDevice", "CurrentScreen", "CurrentIndicator", "CurrentLedStatus", "PreviousLedStatus"}
	for i, v in ipairs(varList) do
		count = count + 1
		func(v, peer_dev_num)
	end 
	for prefix, numScreens in pairs(ScreenList) do
		for screenNum = 1, numScreens do
			local screen = prefix .. screenNum
			local numLines = SCObj.NumButtons
			local lineStep = 1
			count = count + 3
			func("TimeoutEnable_"..screen, peer_dev_num)
			func("TimeoutScreen_"..screen, peer_dev_num)
			func("TimeoutSeconds_"..screen, peer_dev_num)
			if prefix == "C" then
				local numLinesKey = "NumLines_" .. screen
				count = count + 1
				local numLinesStr = func(numLinesKey, peer_dev_num)
				if numLinesStr then
					numLines = tonumber(numLinesStr)
				end
			elseif prefix == "T" then
				lineStep = 4 -- Just do lines 1 and 5 for temperature screens
				count = count + 1
				func("TemperatureDevice_"..screen, peer_dev_num)
			end
			for line = 1, numLines, lineStep do
				count = count + 4
				local modeStr = func("Mode_"..screen.."_"..line, peer_dev_num)
				func("Label_"..screen.."_"..line, peer_dev_num)
				func("Font_"..screen.."_"..line, peer_dev_num)
				func("Align_"..screen.."_"..line, peer_dev_num)
				if modeStr then
					local mode = ParseModeString(modeStr)
					if mode.states > 1 then
						local numStates = tonumber(mode.prefix)
						for state=2, numStates do
							local virtualButton = line + (state-1)*1000
							count = count + 4
							func("Mode_"..screen.."_"..virtualButton, peer_dev_num)
							func("Label_"..screen.."_"..virtualButton, peer_dev_num)
							func("Font_"..screen.."_"..virtualButton, peer_dev_num)
							func("Align_"..screen.."_"..virtualButton, peer_dev_num)
						end
					end
				end
			end
		end
	end
	RestoreLogLevel(oldLog)
	VLog("ForAllVariables: Function called ", count, " times.")
end

-- Remove the old Z-Wabe scene Id from the actuator configuration list and add the new Z-Wave scene Id.
-- This also actually sets the actuator configuration.
local function UpdateActuatorConfList(actuatorConfList, oldZWaveSceneId, newZWaveSceneId, controller_dev_num, target, virtualButton, level, screen, state, eventCode, activate)
	DEntry()
	if not level then
		level = 255
	end
	local deleted = false
	local created = false
	local zWaveSceneIdCache = ZWaveSceneIdCacheList[controller_dev_num]
	if not zWaveSceneIdCache then
		zWaveSceneIdCache = {}
		ZWaveSceneIdCacheList[controller_dev_num] = zWaveSceneIdCache
	end
	local oldCache = zWaveSceneIdCache[oldZWaveSceneId]
	if not oldCache then
		oldCache = {}
	end 
	local oldActuatorConf = {}
	DLog("UpdateActuatorConfList: oldZWaveSceneId=",oldZWaveSceneId," newZWaveSceneId=", newZWaveSceneId, " oldButton=", oldCache.button, " newButton=", virtualButton, " oldState=", oldCache.state, " newState=", state, " oldEventCode=", oldCache.eventCode, " newEventCode=", eventCode, " oldActivate=", oldCache.activate, " newActivate=", activate)
	if oldZWaveSceneId and actuatorConfList[oldZWaveSceneId] then 
		oldActuatorConf = actuatorConfList[oldZWaveSceneId]
		if (oldActuatorConf.controller and oldActuatorConf.controller ~= controller_dev_num) or
		   (oldCache.button and oldCache.button ~= virtualButton) or 
		   (oldCache.activate and oldCache.activate ~= activate) or 
		   (oldCache.state and oldCache.state ~= state) or
		   (SCObj.HasMultiTap > 0 and oldCache.eventCode and oldCache.eventCode ~= eventCode) then
			ELog("UpdateActuatorConfList: Actuator configuration owner mismatch for Z-Wave scene ID", oldZWaveSceneId,".",
				 " old.controller=", oldActuatorConf.controller, " new.controller=", controller_dev_num,
				 " old.button=", oldCache.button, " new.button=", virtualButton,
				 " old.state=", oldCache.state, " new.state=", state,
				 " old.eventCode=", oldCache.eventCode, " new.eventCode=", eventCode,
                 " old.activate=", oldCache.activate, " new.activate=", activate);
		end
		actuatorConfList[oldZWaveSceneId] = nil
		zWaveSceneIdCache[oldZWaveSceneId] = nil
		deleted = true
	end
	if newZWaveSceneId then
		local target_node_id = GetZWaveNode(target.device)
		if target_node_id then
			local target_dimmingDuration = target.dimmingDuration
			if not target_dimmingDuration then
				target_dimmingDuration = 255
			end
			actuatorConfList[newZWaveSceneId]={controller=controller_dev_num, screen=screen, button=virtualButton, level=level, dimmingDuration=target_dimmingDuration, state=state, eventCode=eventCode, activate=activate}
			zWaveSceneIdCache[newZWaveSceneId]={screen=screen, button=virtualButton, activate=activate, state=state, eventCode=eventCode}
			if oldZWaveSceneId ~= newZWaveSceneId or 
			   oldActuatorConf.controller ~= controller_dev_num or 
	           oldActuatorConf.level ~= level or 
			   oldActuatorConf.button ~= virtualButton or 
			   oldActuatorConf.dimmingDuration ~= target_dimmingDuration then
				EnqueueTargetZWaveMessage(SCObj, "UpdateActuatorConfList("..newZWaveSceneId..","..level..","..target_dimmingDuration..")"..target_node_id, target_node_id,  "44 1 ".. newZWaveSceneId .. " ".. target_dimmingDuration .. " 128 ".. level, target.device, 0);
				created = true
			else
				deleted = false
			end
		end
	end
	return deleted or created
end

-- Select a zWaveSceneId and possibly an off zWaveSceneId for a given button and screen which is
-- controlling only scene-capable devices. The same scene ID is multicast to all devices
-- associated with that group (button) so the scene ID must be unique for all screens
-- and buttons of this controller and all scenes used by any controllers operating on this
-- device.
local function SelectZWaveSceneId(peer_dev_num, screen, virtualButton, physicalButton, prevModeStr, modeStr, state, eventCode, force)
	DEntry("SelectZWaveSceneId")
	local mode = ParseModeString(modeStr)
	local controller_node_id, controller_dev_num = GetZWaveNode(peer_dev_num)
	local oldZWaveSceneId = mode.zWaveSceneId
	local oldOffZWaveSceneId = mode.offZWaveSceneId
	local optEvent = OptEvent(eventCode)
	if #mode == 0 then
		mode.sceneControllable = false
		mode.zWaveSceneId = nil
		mode.offZWaveSceneId = nil
		DLog("SelectZWaveSceneId: #mode == 0 mode=",mode)
	elseif mode.sceneControllable then	
		DLog("SelectZWaveSceneId: #mode == ",#mode," mode=",mode)
		local needSceneId = false;
		local defaultSceneId = physicalButton;
		local sceneSet = {max=SCObj.LastFixedSceneId}
		-- First mark off all scene IDs used for any screens of this controller
		ForAllModes(peer_dev_num, function(modeStr2, mode2, screen2, virtualButton2, state2, eventCode2)
			if screen ~= screen2 or virtualButton ~= virtualButton2 or state ~= state2 or eventCode ~= eventCode2 then
				if mode2.zWaveSceneId then
					MarkZWaveSceneIdUsed(mode2.zWaveSceneId, sceneSet)
					VLog("  SelectZWaveSceneId->ForAllModes: Marked on  scene ", mode2.zWaveSceneId," used by screen=",screen2," button=", virtualButton2," state=", state2, " eventCode=", eventCode2, ". SceneSet now: ", sceneSet)
				end
				if mode2.offZWaveSceneId then
					MarkZWaveSceneIdUsed(mode2.offZWaveSceneId, sceneSet)
					VLog("  SelectZWaveSceneId->ForAllModes: Marked off scene ", mode2.offZWaveSceneId," used by screen=",screen2," button=", virtualButton2," state=", state2, " eventCode=", eventCode2, ". SceneSet now: ", sceneSet)
				end
			end
		end )
		-- Now mark off all scene IDs used by other controllers associated with this target
		for target = 1, #mode do
			needSceneId = true;
			local actuatorConfList = ReadDeviceActuatorConfList(mode[target].device)
			DLog("SelectZWaveSceneId: actuatorConfList for device ", mode[target].device, " is ", actuatorConfList)
			local changed = false
			for zWaveSceneId, data in pairs(actuatorConfList) do
				if not luup.devices[data.controller] then
					DLog("SelectZWaveSceneId: Removing Z-Wave Scene ID ", zWaveSceneId, " actuatior configuration for non-existant controller:", data.controller)
					actuatorConfList[zWaveSceneId] = nil
					changed = true;
				elseif data.controller == controller_dev_num and data.button == virtualButton and data.screen == screen and data.state == state then
					if data.level > 0 or (mode.prefix ~= "T" and mode.prefix ~= "O") or not SCObj.HasOffSceness then
						if not mode.zWaveSceneId then
							mode.zWaveSceneId = zWaveSceneId
						elseif mode.zWaveSceneId ~= zWaveSceneId then
							DLog("SelectZWaveSceneId: Removing Z-Wave Scene ID ", zWaveSceneId, " actuator configuration for duplicate entry:", mode.zWaveSceneId)
							actuatorConfList[zWaveSceneId] = nil
							changed = true
						end
					else
						if not mode.offZWaveSceneId then
							mode.offZWaveSceneId = zWaveSceneId
						elseif mode.offZWaveSceneId ~= zWaveSceneId then
							DLog("SelectZWaveSceneId: Removing Z-Wave Off Scene ID ", zWaveSceneId, " actuator configuration for duplicate entry:", mode.offZWaveSceneId)
							actuatorConfList[zWaveSceneId] = nil
							changed = true
						end
					end
				else
					MarkZWaveSceneIdUsed(zWaveSceneId, sceneSet)
					DLog("SelectZWaveSceneId: actuatorConfList: (device=", mode[target].device, ") Marked ", zWaveSceneId, " used. SceneSet now: ", sceneSet)
				end
			end
			if changed or force then
				WriteDeviceActuatorConfList(mode[target].device, actuatorConfList)
			end
		end
		-- The set of available scene IDs is now completed. Pick a new scene ID and perhaps a new off scene ID in case of a toggle.
		DLog("SelectZWaveSceneId: oldZWaveSceneId=", oldZWaveSceneId, " oldOffZWaveSceneId=", oldOffZWaveSceneId, " sceneSet=", sceneSet, " defaultSceneId=", defaultSceneId)
		local newZWaveSceneId =  pickSceneId(mode.zWaveSceneId, sceneSet, needSceneId, defaultSceneId)
		local newOffZWaveSceneId
		if (mode.prefix == "T" and SCObj.HasOffScenes) or mode.prefix == "C" or not SCObj.HasDirectSceneControl then
			newOffZWaveSceneId = pickSceneId(mode.offZWaveSceneId, sceneSet, needSceneId, defaultSceneId+SCObj.NumButtons)
		end
		DLog("SelectZWaveSceneId: newZWaveSceneId=", newZWaveSceneId, " newOffZWaveSceneId=", newOffZWaveSceneId, " sceneSet=", sceneSet)
		-- If either of the two scene IDs changed, then update all of the target devices
		mode.zWaveSceneId = newZWaveSceneId
		mode.offZWaveSceneId = newOffZWaveSceneId
	else
		mode.zWaveSceneId = nil
		mode.offZWaveSceneId = nil
	end
	local newModeStr = GenerateModeString(mode)
	DLog("SelectZWaveSceneId: mode=",mode," prevModeStr=",prevModeStr," newModeStr=",newModeStr)
	if prevModeStr ~= newModeStr or force then
		for i = 1, #mode do
			local actuatorConfList = ReadDeviceActuatorConfList(mode[i].device)
			local modOn  = UpdateActuatorConfList(actuatorConfList, oldZWaveSceneId,    mode.zWaveSceneId,    controller_dev_num, mode[i], virtualButton, mode[i].level, screen, state, eventCode, true)
			local modOff = UpdateActuatorConfList(actuatorConfList, oldOffZWaveSceneId, mode.offZWaveSceneId, controller_dev_num, mode[i], virtualButton, 0            , screen, state, eventCode, false)
			if modOn or modOff then
				WriteDeviceActuatorConfList(mode[i].device, actuatorConfList)
			end
		end
		if mode.prefix ~= "L" and newModeStr ~= prevModeStr then
			luup.variable_set(SID_SCENECONTROLLER, "Mode_"..screen.."_"..(virtualButton+(state-1)*1000)..optEvent, newModeStr, peer_dev_num)
		end
	end
	return mode
end

-- local 
function ReadDeviceActuatorConfList(dev_num)
	local actuatorConfStr = luup.variable_get(SID_SCENECONTROLLER, ACTUATOR_CONF, dev_num)
	VEntry()
	if not actuatorConfStr then
		return {}, false
	end
	local actuatorConfList = {}
	for sc, ct, scbt, lv, dd, ev in string.gmatch(actuatorConfStr, "(-?%d+):(%d+),([^,]+),(%d+),(%d+),?([^;]*);") do
		local scr, bt = string.match(scbt,"^(.*)_(%d+)$")
		if not scr then -- legacy, did not have screen
			scr = ""
			bt = scbt
		end
		local scene = tonumber(sc)
		local activate = true
		if scene < 0 then
			scene = -scene
			activate = false
		end
		if #ev == 0 then
			ev = "1"
		end
		local stateButton = tonumber(bt)
		local button = stateButton % 1000
		local state = math.floor(stateButton / 1000) + 1
		actuatorConfList[scene]={controller=tonumber(ct), activate=activate, screen=scr, button=button, state=state, level=tonumber(lv), dimmingDuration=tonumber(dd), eventCode=ev}
	end
	VLog("ReadDeviceActuatorConfList: actuatorConfList=",actuatorConfList)
	return actuatorConfList, true
end

-- local 
function WriteDeviceActuatorConfList(dev_num, actuatorConfList)
	DEntry()
	local str = ""
	for k, v in pairs(actuatorConfList) do
		local k2 = k;
		if not v.activate then
			k2 = -k;
		end
		str = str .. k2 .. ":" .. v.controller .. "," .. v.screen.."_".. (v.button + (v.state - 1)*1000) .. "," .. v.level .. "," .. v.dimmingDuration
		if v.eventCode then
			str = str .. "," .. v.eventCode
		end
		str = str .. ";"
	end
	local oldActuatorConfStr = luup.variable_get(SID_SCENECONTROLLER, ACTUATOR_CONF, dev_num)
	if str ~= oldActuatorConfStr then
		DLog("WriteDeviceActuatorConfList: updating: dev_num=", dev_num, " old=",oldActuatorConfStr, " new=", str)
		luup.variable_set(SID_SCENECONTROLLER, ACTUATOR_CONF, str, dev_num, false)
		return true
	end
	DLog("WriteDeviceActuatorConfList: not updating: dev_num=", dev_num, " old=new=",oldActuatorConfStr)
	return false
end

-- local 
function RemoveDeviceActuatorConf(dev_num, zWaveSceneId, offZWaveSceneId)
	local actuatorConfList, exists = ReadDeviceActuatorConfList(dev_num)
	VEntry("RemoveDeviceActuatorConf")
	if not exists then
		return
	end
	actuatorConfList[zWaveSceneId]=nil
	if offZWaveSceneId then
		actuatorConfList[offZWaveSceneId]=nil
	end
	WriteDeviceActuatorConfList(dev_num,actuatorConfList)
end

local function RemoveDeviceActuatorConfForController(controller_dev_num, target_dev_num)
	local actuatorConfList, exists = ReadDeviceActuatorConfList(target_dev_num)
	VEntry("RemoveDeviceActuatorConfForController")
	if not exists then
		return
	end
	local changed = false
	for k, v in pairs(actuatorConfList) do
		if v.controller == controller_dev_num then
		  	actuatorConfList[k] = nil
			changed = true;
		end
	end
	if changed then
		WriteDeviceActuatorConfList(target_dev_num,actuatorConfList)
	end
end

local function RemoveAllDeviceActuatorConf(dev_num)
	local actuatorConfStr = luup.variable_get(SID_SCENECONTROLLER, ACTUATOR_CONF, dev_num)
	VEntry("RemoveAllDeviceActuatorConf")
	if actuatorConfStr then
		luup.variable_set(SID_SCENECONTROLLER, ACTUATOR_CONF, nil, dev_num, false)
	end
end

-- parse a string from service urn:micasaverde-com:serviceId:ZWaveDevice1, variable AssociationSet like "17,10,11;25,37,66,77"
-- which means "group 17 is associated with devices 10 and 11; group 25 is associated with devices 37, 66, and 77"
-- into an object with the groups as keys with values as objects with key=device, value=true.
-- This is used by LuaUPnP for (re)establishing return routes after neighbor nodes are updated.
local function GetVeraAssociationList(zwave_dev_num)
	local list = luup.variable_get(SID_ZWDEVICE, ASSOCIATION_SET, zwave_dev_num)
	VEntry("GetVeraAssociationList")
	local obj = {}
	if list then
		repeat
			local groupstr, rest = list:match(";?(%d+)(.*)")
			if groupstr then
				local groupnum = tonumber(groupstr)
				local group = {}
				obj[groupnum] = group
				list = rest
				while list and list:sub(1,1) == "," do
					local nodeStr, rest2 = list:match(",([%dz%.]+)(.*)")
					if nodeStr then
						local nodeNum
						if nodeStr:sub(1,1) == "z" then
							local node = GetZWaveNode(zwave_dev_num)
							nodeNum = tonumber(tostring(node) .. nodeStr:sub(2))
						else
							nodeNum = tonumber(nodeStr)
						end
						group[nodeNum] = true;
					end
					list = rest2;
				end
			else
				list = nil
			end
		until not list or list:sub(1,1) ~= ";"
	end
	VLog("GetVeraAssociationList: obj=",obj)
	return obj
end

local function SetVeraAssociationList(zwave_dev_num, obj)
	VEntry()
	local str = ""
	local first = true
	for k, v in pairs(obj) do
		local first2 = true
		for k2, v2 in pairs(v) do
			if first2 then
				if not first then
					str = str .. ";"
				end
				first = false
				str = str .. k
				first2 = false
			end
			str = str .. "," .. k2
		end
	end
	-- Keep these in sync to avoid LuaUPnP resetting the configured flag
	luup.variable_set(SID_ZWDEVICE, ASSOCIATION_SET, str, zwave_dev_num)
	luup.variable_set(SID_ZWDEVICE, CONFIGURED_ASSOC, str, zwave_dev_num)
end

local function UnassociateDevice(zwave_dev_num, target_dev_num_list, firstGroup, lastGroup, groupStep, delay)
	DEntry()
	if not delay then
		delay = param[zwave_dev_num].SCENE_CTRL_AssociationDelay
	end
	local targets = ""
	local label = ""
	local zwave_node_id = GetZWaveNode(zwave_dev_num);
	local associationSet = GetVeraAssociationList(zwave_dev_num)
	if target_dev_num_list then
		local count = 0
		for k, v in pairs(target_dev_num_list) do
			local target_node_id = GetZWaveNode(k)
			if target_node_id then
				targets = targets .. " " .. tostring(target_node_id)
				if count > 0 then
					label = label .. ","
				end
				label = label .. tostring(k).."("..tostring(target_node_id)..")"
				count = count + 1
			else
				ELog("UnassociateDevice: Unknown target device: ", k)
			end
			for group = firstGroup, lastGroup, groupStep do
				if associationSet[group] then
					associationSet[group][k] = nil
				end
			end
		end
		if count == 0 then
			return
		end
	else
		label = "All"
		for group = firstGroup, lastGroup, groupStep do
			associationSet[group] = nil
		end
	end

	-- COMMAND_CLASS_ASSOCIATION ASSOCIATION_REMOVE
	for group = firstGroup, lastGroup, groupStep do
		EnqueueZWaveMessage(SCObj, "RemoveAssociation["..group.."]"..zwave_dev_num.."("..zwave_node_id..")-x->"..label, zwave_node_id,  "0x85 0x04 "..group..targets, zwave_dev_num, delay)
	end
	SetVeraAssociationList(zwave_dev_num, associationSet)
end

local function AssociateDevice(zwave_dev_num, target_dev_num_list, firstGroup, lastGroup, groupStep, delay)
	DEntry()
	if not delay then
		delay = param[zwave_dev_num].SCENE_CTRL_AssociationDelay
	end
	local targets = ""
	local label = ""
	local count = 0
	local zwave_node_id = GetZWaveNode(zwave_dev_num);
	local associationSet = GetVeraAssociationList(zwave_dev_num)
	for k, v in pairs(target_dev_num_list) do
		local target_node_id = GetZWaveNode(k)
		if target_node_id then
			SetReturnRoute(zwave_dev_num, zwave_node_id, target_node_id)
			targets = targets .. " " .. tostring(target_node_id)
			if count > 0 then
				label = label .. ","
			end
			label = label .. tostring(k).."("..tostring(target_node_id)..")"
			count = count + 1
		else
		    -- This is actually an important error that can occur if a directly associated device is removed.
			ELog("AssociateDevice: Unknown target device: ", k, " Group=", firstGroup, " Controlling device=", zwave_dev_num, "(", luup.devices[zwave_dev_num].description, ")")
		end
		for group = firstGroup, lastGroup, groupStep do
			if not associationSet[group] then
				associationSet[group] = {}
			end
			associationSet[group][k] = true
		end
	end
	if count == 0 then
		return
	end
	-- COMMAND_CLASS_ASSOCIATION ASSOCIATION_REMOVE
	for group = firstGroup, lastGroup, groupStep do
		EnqueueZWaveMessage(SCObj, "SetAssociation["..group.."]"..zwave_dev_num.."("..zwave_node_id..")->"..label, zwave_node_id, "0x85 0x01 "..group..targets, zwave_dev_num, delay)
	end
	SetVeraAssociationList(zwave_dev_num, associationSet)
end

local SceneController_ReturnRoutes = {}
local returnRouteCallbackNum = 1
-- local 
function SetReturnRoute(zwave_dev_num, lcd_node_id, target_node_id)
	VEntry()
  	local veraZWaveNode, ZWaveNetworkDeviceId = GetVeraIDs()
	if target_node_id ~= veraZWaveNode then
		if SceneController_ReturnRoutes[lcd_node_id] == nil then
			local returnRouteList = luup.variable_get(SID_SCENECONTROLLER, RETURN_ROUTES, zwave_dev_num)
			SceneController_ReturnRoutes[lcd_node_id] = {}
			if returnRouteList then
				for d in string.gmatch(returnRouteList, "%d+") do
	       			SceneController_ReturnRoutes[lcd_node_id][tonumber(d)] = true;
	     		end
			end
		end
		if not  SceneController_ReturnRoutes[lcd_node_id][target_node_id] then
			local retries = 0;
			local ReturnRouteCallback
			local function doSetReturnRoute()
				VEntry("doSetReturnRoute")
--[==[
41      11/05/17 9:21:45.324                 0x1 0x6 0x0 0x46 0x2 0x17 0x1 0xab (###F####) 
                        SOF - Start Of Frame --+   |   |    |   |    |   |    |
                                  length = 6 ------+   |    |   |    |   |    |
                                     Request ----------+    |   |    |   |    |
              FUNC_ID_ZW_ASSIGN_RETURN_ROUTE ---------------+   |    |   |    |
Source Node ID: 2 Device 3=Cooper RFWC5 Scene Controller Z-Wave +    |   |    |
Destination Node ID: 23 Device 29=Qubino light strip ----------------+   |    |
                             Callback ID = 1 ----------------------------+    |
                                 Checksum OK ---------------------------------+
42      11/05/17 9:21:45.352    0x6 0x1 0x4 0x1 0x46 0x1 0xbd (####F##) 
              ACK - Acknowledge --+   |   |   |    |   |    |
           SOF - Start Of Frame ------+   |   |    |   |    |
                     length = 4 ----------+   |    |   |    |
                       Response --------------+    |   |    |
 FUNC_ID_ZW_ASSIGN_RETURN_ROUTE -------------------+   |    |
     RetVal = Operation_Started -----------------------+    |
                    Checksum OK ----------------------------+
41      11/05/17 9:21:45.363    ACK: 0x6 (#)
													  C1 
42      11/05/17 9:21:45.641    0x1 0x5 0x0 0x46 0x1 0x0 0xbd (###F###) 
           SOF - Start Of Frame --+   |   |    |   |   |    |
                     length = 5 ------+   |    |   |   |    |
                        Request ----------+    |   |   |    |
 FUNC_ID_ZW_ASSIGN_RETURN_ROUTE ---------------+   |   |    |
                Callback ID = 1 -------------------+   |    |
  Status = TRANSMIT_COMPLETE_OK -----------------------+    |
                    Checksum OK ----------------------------+
41      11/05/17 9:21:45.641    ACK: 0x6 (#) 
--]==]
			EnqueueZWaveMessageWithResponse(SCObj, 
				                            "AssignReturnRoute_"..returnRouteCallbackNum.."_"..lcd_node_id.."->"..target_node_id, -- label 
										    lcd_node_id, 
											"0x00 0x46 ".. lcd_node_id .. " " .. target_node_id .. " " .. returnRouteCallbackNum, -- sendData
										    0, -- delay
										    true, -- local controller
											"^01 .. 00 46 ".. string.format("%02X", returnRouteCallbackNum) .. " (..)", -- responsePattern
										    ReturnRouteCallback, -- callback
										    true, -- oneShot 
										    zwave_dev_num, -- dev_num
											20000, -- 20 second timeout. More than ever necessary
											nil, -- armPattern
											"06") -- autoResponse
			SceneController_ReturnRoutes[lcd_node_id][target_node_id] = true
			local returnRouteList = ""
			for k,v in pairs(SceneController_ReturnRoutes[lcd_node_id]) do
			   returnRouteList = returnRouteList.. tostring(k)..","
			end
	   		luup.variable_set(SID_SCENECONTROLLER, RETURN_ROUTES, returnRouteList, zwave_dev_num)
		   		returnRouteCallbackNum = (returnRouteCallbackNum % 255) + 1 -- Don't use callback ID 0.
			end
			ReturnRouteCallback = function(peer_dev_num, response)
				VEntry("ReturnRouteCallback")
				if not response then
					log(ANSI_BRIGHT_RED, "Assign return route timed out", ANSI_RESET)
				elseif response.C1 ~= "00" then
					log(ANSI_BRIGHT_RED, "Assign return route failed", ANSI_RESET)
				else
					return
				end
				retries = retries + 1
				if retries < 3 then
					DLog("Retry ", retries)
					doSetReturnRoute()
				else
					DLog("Giving up")
				end
			end
			doSetReturnRoute()
		end
	end
end

local function ClearReturnRouteCache(zwave_dev_num)
	VEntry()
	local node_id = GetZWaveNode(zwave_dev_num);
	SceneController_ReturnRoutes[node_id] = nil;
	luup.variable_set(SID_SCENECONTROLLER, RETURN_ROUTES, "", zwave_dev_num)
end

-- local 
function UpdateAssociationForPhysicalButton(zwave_dev_num, screen, force, prevModeStr, modeStr, physicalButton, virtualButton, state, eventCode)
	DEntry("UpdateAssociationForPhysicalButton")
	local mode
	if SCObj.HasCooperConfiguration then -- Cooper RFWC5 cannot update associations incrementally
		force = true
	end
	if force or prevModeStr ~= modeStr then
		local node_id = GetZWaveNode(zwave_dev_num)
		local peer_dev_num = GetPeerDevNum(zwave_dev_num)
		mode = SelectZWaveSceneId(peer_dev_num, screen, virtualButton, physicalButton, prevModeStr, modeStr, state, eventCode, force)
		DLog("UpdateAssociationForPhysicalButton: newMode=", mode)
		if force then
			prevModeStr = mode.prefix
		end
		local prevMode = ParseModeString(prevModeStr)
  		local veraZWaveNode, ZWaveNetworkDeviceId = GetVeraIDs()
		local groupId, offGroupId, SceneControllerConfGroupId, SceneControllerConfOffGroupId = SCObj.PhysicalButtonToGroupIds(physicalButton, mode)
		local zWaveSceneId = mode.zWaveSceneId or physicalButton
		local offZWaveSceneId = SCObj.HasOffScenes and (mode.offZWaveSceneId or offGroupId)
		local prevZWaveSceneId = prevMode.zWaveSceneId or physicalButton
		local prevOffZWaveSceneId = SCObj.HasOffScenes and (prevMode.offZWaveSceneId or offGroupId)
		local associateList = {}
		local unassociateList = {}
		-- If the previous mode has any missing devices, then clear everything and start from scratch.
		if not force then
			for i = 1, #prevMode do
				local node = GetZWaveNode(prevMode[i].device)
				if not node then
					force = true
					break
				end
			end
		end
		if force then
			UnassociateDevice(zwave_dev_num, nil, groupId, groupId, 1, param[zwave_dev_num].SCENE_CTRL_UnassociateDelay) -- Unassociate activate with everything
			if offGroupId then
				UnassociateDevice(zwave_dev_num, nil, offGroupId, offGroupId, 1, param[zwave_dev_num].SCENE_CTRL_UnassociateDelay) -- Unassociate deactivate with everything
			end
		else
			if prevMode.prefix ~= "C" and SCObj.HasDirectSceneControl then
				for i = 1, #prevMode do
					unassociateList[prevMode[i].device] = true;
				end
				if #prevMode < SCObj.MaxDirectAssociations and #mode >= SCObj.MaxDirectAssociations then
					unassociateList[ZWaveNetworkDeviceId] = true
				end
			end
		end
		if mode.prefix ~= "C" and SCObj.HasDirectSceneControl then
			for i = 1, #mode do
				unassociateList[mode[i].device] = nil
				if mode[i].level then
					associateList[mode[i].device] = mode[i]
				else
					associateList[mode[i].device] = {level=255}
				end
			end
		end
		if force then
			if not SCObj.HasDirectSceneControl or #mode < SCObj.MaxDirectAssociations then
				associateList[ZWaveNetworkDeviceId] = {level=255, dimmingDuration=255}
			end
		elseif  SCObj.HasDirectSceneControl then
			if prevMode.prefix ~= "C" then
				for i = 1, #prevMode do
				    -- For Scene controllable and basic set, no need to reassociate
					associateList[prevMode[i].device] = nil;
				end
			end
			if (prevMode.prefix == "C" or #prevMode >= SCObj.MaxDirectAssociations) and 
               (mode.prefix ~= "C" and #mode < SCObj.MaxDirectAssociations) then
				associateList[ZWaveNetworkDeviceId] = {level=255, dimmingDuration=255}
			end
		end
		DLog("UpdateAssociationForPhysicalButton: prevMode=", prevMode)
		DLog("UpdateAssociationForPhysicalButton: mode=", mode)
		DLog("UpdateAssociationForPhysicalButton: unassociateList=", unassociateList)
		DLog("UpdateAssociationForPhysicalButton: associateList=", associateList)
		if SceneControllerConfGroupId and (force or zWaveSceneId ~= prevZWaveSceneId) then
			-- Set on scene controller configuration, group id, sceneid
			EnqueueZWaveMessage(SCObj, "SetSceneControllerConfig_On["..tostring(physicalButton)..":"..zWaveSceneId.."]"..zwave_dev_num.."("..node_id..")", node_id,  "0x2d 0x01 "..SceneControllerConfGroupId.." "..zWaveSceneId.." 0xFF", zwave_dev_num, param[zwave_dev_num].SCENE_CTRL_SceneControllerConfDelay)
		end
		if SCObj.HasOffScenes and mode.prefix == "T" and SceneControllerConfOffGroupId and (force or offZWaveSceneId ~= prevOffZWaveSceneId or prevMode.prefix ~= "T") then
			-- Set off scene controller configuration, group id, sceneid
			EnqueueZWaveMessage(SCObj, "SetSceneControllerConfig_Off["..physicalButton..":"..offZWaveSceneId.."]"..zwave_dev_num.."("..node_id..")", node_id,  "0x2d 0x01 "..SceneControllerConfOffGroupId.." "..offZWaveSceneId.." 0xFF", zwave_dev_num, param[zwave_dev_num].SCENE_CTRL_SceneControllerConfDelay)
		end

		UnassociateDevice(zwave_dev_num, unassociateList, groupId, groupId, 1, param[zwave_dev_num].SCENE_CTRL_AssociationDelay) -- Unassociate activate with everything
		if offGroupId then
			UnassociateDevice(zwave_dev_num, unassociateList, offGroupId, offGroupId, 1, param[zwave_dev_num].SCENE_CTRL_AssociationDelay) -- Unassociate activate with everything
		end
		
		if SCObj.HasCooperConfiguration then
			-- Special Cooper RFWC5 configuration
			-- First set associations with scene-capable devices (including the controller) but wait for the 5-second timeout.
			-- Then set association for non-scene-capable devices and set the configuration for that button to the level within the 5 second time-out.
			local sceneList = {}
			local levels = {}
			for k, v in pairs(associateList) do
				if v.dimmingDuration then
					sceneList[k] = true
				else
					if not levels[v.level] then
						levels[v.level] = true
						local list = {}
						for k2, v2 in pairs(associateList) do
							if not v2.dimmingDuration then
								if v2.level == v.level then
									list[k2] = true;
								end
							end
						end
						AssociateDevice(zwave_dev_num, list, groupId, groupId, 1, 0)
						SetConfigurationOption("CooperConfiguration", peer_dev_num, node_id, physicalButton, v.level, 5000)
					end
				end
			end
			AssociateDevice(zwave_dev_num, sceneList, groupId, groupId, 1, 5000)
		else
			-- Regular association. Can associate all devices regardless of level
			AssociateDevice(zwave_dev_num, associateList, groupId, groupId, 1, param[zwave_dev_num].SCENE_CTRL_AssociationDelay)
		end
		if offGroupId then
			AssociateDevice(zwave_dev_num, associateList, offGroupId, offGroupId, 1, param[zwave_dev_num].SCENE_CTRL_AssociationDelay)
		end	-- if SCObj.HasOffScenes
	else	-- if force or prevModeStr ~= modeStr
		mode = ParseModeString(modeStr)
	end
	return mode
end

-- local 
function OptEvent(eventCode)
	if SCObj.HasMultiTap > 0 or SCObj.NumSceneEvents > 0 then
		return "_" .. eventCode
	else
		return ""
	end 
end

-- local 
function GetModeStr(peer_dev_num, screen, virtualButton, state, eventCode, getState, nextState)
	VEntry()
	local modeStr
    local negative = false
    local optEvent = OptEvent(eventCode)
    if getState then
    	virtualButton = virtualButton % 1000
    elseif state ~= nil then
    	virtualButton = virtualButton + 1000 * (state - 1)
    end
    local varname = "???"
	if screen then
		-- modeStr  = luup.variable_get(SID_SCENECONTROLLER,"Mode_"..screen.."_"..virtualButton..OptEvent(eventCode),peer_dev_num)
		varname = "Mode_"..screen.."_"..virtualButton..optEvent
		modeStr  = luup.variable_get(SID_SCENECONTROLLER,varname,peer_dev_num)
		if not modeStr then
			modeStr = SCObj.GetDefaultModeString(virtualButton, eventCode)
		end
	else
		modeStr =  SCObj.GetDefaultModeString(virtualButton, eventCode)
	end
	local modePrefix = modeStr:sub(1,1)
	if modePrefix == "~" then
		modePrefix = modeStr:sub(2,2)
		negative = true;
	end
	if getState and modePrefix >= "2" and modePrefix <= "9" then
		local states = tonumber(modePrefix)
		if negative and virtualButton > 1 then
			virtualButton = virtualButton - 1
		end
		-- state = luup.variable_get(SID_SCENECONTROLLER,"State_"..screen.."_"..virtualButton..optEvent, peer_dev_num)
		if state == nil then
			varname = "State_"..screen.."_"..virtualButton..optEvent
			state = luup.variable_get(SID_SCENECONTROLLER, varname, peer_dev_num)
		else
			varname = "explicit_state="..tostring(state)
		end
		if state == nil then
			state = 1
		else
			state = tonumber(state)
		end
		if nextState then
			if negative then
				state = ((state - 2) % states) + 1
			else
				state = (state % states) + 1
			end
		end
		-- modeStr  = luup.variable_get(SID_SCENECONTROLLER,"Mode_"..screen.."_"..(virtualButton+(state-1)*1000)..optEvent,peer_dev_num)
		varname = "Mode_"..screen.."_"..(virtualButton+(state-1)*1000)..optEvent
		modeStr  = luup.variable_get(SID_SCENECONTROLLER, varname, peer_dev_num)
		if modeStr == nil then
			modeStr = SCObj.GetDefaultModeString(virtualButton+(state-1)*1000, eventCode)
		end
	elseif state == nil then
		state = 1
	end
	VLog("GetModeStr: varname=", varname, " modeStr=", modeStr, " state=", state)
	return modeStr, state
end

local function SetButtonMode(peer_dev_num, prevModeStr, screen, force, temperatureSettable, physicalButton, eventCode)
	DEntry()
	local modeStr 
	local state = 1
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	if not node_id then
		ELog("SetButtonMode: bad peer_dev_num=", peer_dev_num, "  screen=", screen)
		return
	end
	if not prevModeStr then
		prevModeStr = "M"
		force = true
	end
	local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, screen)
	local virtualButton = physicalButton + scrollOffset
	if physicalButton >= 2 and physicalButton <= 4 and screen:sub(1,1) == "T" then
		if temperatureSettable or physicalButton == 3 then
			modeStr = "H";
		else
			modeStr = "M";
		end
		if screen then
			-- Need to do this here so that code that calls to GetModeStr will be correct later.
			if modeStr ~= prevModeStr then
				luup.variable_set(SID_SCENECONTROLLER, "Mode_"..screen.."_"..physicalButton..OptEvent(eventCode), modeStr, peer_dev_num)
			end
		end
	elseif numLines > SCObj.NumButtons and
           ((physicalButton == 1 and scrollOffset > 0) or
            (physicalButton == SCObj.NumButtons and scrollOffset < numLines - SCObj.NumButtons)) then
		modeStr = "L"
	else
		modeStr, state = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, true, true)
	end
	local oldType
	local newType = SCObj.ModeType(ParseModeString(modeStr))
	if not force then
		oldType = SCObj.ModeType(ParseModeString(prevModeStr))
	end
	DLog("SetButtonMode: peer_dev_num=", peer_dev_num, " prevModeStr=", prevModeStr, " screen=", screen, " force=", force, " temperatureSettable=", temperatureSettable, " pysicalButton=", physicalButton, " oldType=", oldType, " modeStr=", modeStr, " newType=", newType, " NumLines=", numLines, " scrollOffset=", scrollOffset, " virtualButton=", virtualButton, " state=",state, " eventCode=", eventCode)
	SCObj.SetButtonType(peer_dev_num, node_id, physicalButton, newType, force, oldType)
	UpdateAssociationForPhysicalButton(zwave_dev_num, screen, force, prevModeStr, modeStr, physicalButton, virtualButton, state, eventCode)
end

function SceneController_UpdateCustomLabel(peer_dev_num, screen, virtualButton, label, font, align)
	DEntry("SceneController_UpdateCustomLabel")
	if IsVeraPrimaryController() then
		if label == nil then
			label = ""
		end
		local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, screen)
		virtualButton = tonumber(virtualButton)
		local physicalButton = (virtualButton % 1000) - scrollOffset
		if SCObj.HasScreen then
			luup.variable_set(SID_SCENECONTROLLER, "Label_" .. screen .. "_" .. virtualButton, label, peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "Font_" .. screen .. "_" .. virtualButton, font, peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "Align_" .. screen .. "_" .. virtualButton, align, peer_dev_num)
		end
		label = UnescapePercent(label)
		local currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
		if not SCObj.HasMultipleScreens or currentScreen == screen then
			if physicalButton >= 1 and
               physicalButton <= SCObj.NumButtons and
               (numLines <= SCObj.NumButtons or
                ((physicalButton > 1 or scrollOffset == 0) and
                 (physicalButton < SCObj.NumButtons or scrollOffset == numLines-SCObj.NumButtons))) then
				if SCObj.HasScreen then
				    local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
					if not node_id then
						ELog("SceneController_UpdateCustomLabel: Bad peer_dev_num=", peer_dev_num, " label=", label)
						return
					end
					local s = label:gsub("\\r",      "\r")
					      s =     s:gsub("-([^\r])", "-\r%1")
					local labels = {}
					local fonts = {}
					local aligns = {}
					labels[physicalButton] = s
					fonts[physicalButton] = font
					aligns[physicalButton] = align
				    EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.NoChange)
				end
			end
			SetIndicator(peer_dev_num, screen, false, false, true) -- No force, no coalesce, cache 
		else
		    -- not dotimeout, no force, not indicatorOnly
			SetScreen(peer_dev_num,screen,false,false,false);
		end
		RunZWaveQueue("UpdateCustomLabel", 0);
	end
end

function SceneController_UpdateMode(peer_dev_num, screen, stateButton, eventCode, modeStr)
	DEntry("SceneController_UpdateMode")
	if IsVeraPrimaryController() then
		local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, screen)
		local virtualButton = stateButton % 1000
		local physicalButton = virtualButton - scrollOffset
		local newState = math.floor(stateButton / 1000) + 1;
		local currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
		local optEvent = OptEvent(eventCode)
		local force = false;
		if SCObj.HasMultipleScreens and currentScreen ~= screen then
			-- not dotimeout, no force, not indicatorOnly
			SetScreen(peer_dev_num,screen,false,false,false);
		end
		local prevModeStr, prevState = GetModeStr(peer_dev_num, screen, virtualButton, nil, eventCode, true, true)
		if newState == 1 then -- The number of states is only in the state 1 mode.
			local mode = ParseModeString(modeStr);
			if prevState > mode.states then
				-- If reducing the number of states and the current state would now be invalid, reset it to state 1 and redraw the screen.
				prevState = 1
				luup.variable_set(SID_SCENECONTROLLER,"State_"..screen.."_"..virtualButton..optEvent, tostring(prevState), peer_dev_num)
				SetScreen(peer_dev_num,screen,false,false,false);
			end
		end
		if physicalButton >= 1 and
           physicalButton <= SCObj.NumButtons and
           (numLines <= SCObj.NumButtons or
            ((physicalButton > 1 or scrollOffset == 0) and
             (physicalButton < SCObj.NumButtons or scrollOffset == numLines-SCObj.NumButtons))) and
           prevState == newState then
           	-- Set the current mode immediately
           	if modeStr ~= prevModeStr then
				luup.variable_set(SID_SCENECONTROLLER, "Mode_"..screen.."_"..(virtualButton+((newState-1)*1000))..optEvent, modeStr, peer_dev_num)
			end
 			SetButtonMode(peer_dev_num, prevModeStr, screen, force, true, physicalButton, eventCode)
			SetIndicator(peer_dev_num, screen, force, true, false) -- No force, coalesce, no cache 
 		else
 			-- Not changing the mode of the current state. Just select an appropriate Scene ID and save the mode.
 			local prevModeStr = GetModeStr(peer_dev_num, screen, virtualButton, newState, eventCode, false, false)
 			SelectZWaveSceneId(peer_dev_num, screen, virtualButton, physicalButton, prevModeStr, modeStr, newState, eventCode, force)
		end
		RunZWaveQueue("UpdateCustomLabel", 0);
	end
end

local UserModeMap = {
	[USERMODE_HEAT]={label="Heat", next=USERMODE_COOL},
	[USERMODE_COOL]={label="Cool", next=USERMODE_AUTO},
	[USERMODE_AUTO]={label="Auto", next=USERMODE_OFF},
	[USERMODE_OFF] ={label="Off",  next=USERMODE_HEAT}
}

local EnergyModeMap = {
	[ENERGYMODE_NORMAL]={label="Normal", next=ENERGYMODE_SAVING},
	[ENERGYMODE_SAVING]={label="Energy Saving", next=ENERGYMODE_NORMAL}
}

-- local 
function ChooseLabelFontAndAlign(peer_dev_num, screen, line, virtualButton, stateOverride, labels, fonts, aligns, cache)
	VEntry()
	if SCObj.HasScreen then
		local modeStr = GetModeStr(peer_dev_num, screen, virtualButton, nil, "1", false, false)
		local mode = ParseModeString(modeStr)
		local label, font, align
		if mode.prefix == "P" or mode.prefix == "E" then
			local temperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, peer_dev_num);
			if temperatureDevice and temperatureDevice ~= "0" then
				temperatureDevice = tonumber(temperatureDevice)
				if mode.prefix == "P" then	
					local userMode = luup.variable_get(SID_USERMODE, USERMODE_VAR, temperatureDevice)
					if userMode and UserModeMap[userMode] then
						label = UserModeMap[userMode].label
					end
				else
					local energyMode = luup.variable_get(SID_USERMODE, ENERGYMODE_VAR, temperatureDevice)
					if energyMode and EnergyModeMap[energyMode] then
						label = EnergyModeMap[energyMode].label
					end
				end
			end
		else
			if mode.states > 1 then
				local states = tonumber(mode.prefix)
				if stateOverride ~= nil then
					if stateOverride < 1 then
						stateOverride = states
					elseif stateOverride > states then
					    stateOverride = 1
					end
				else
					stateOverride = ChooseBestFromMultiState(peer_dev_num, screen, virtualButton, mode, "1", false, cache)
				end
				virtualButton = virtualButton + ((stateOverride-1) * 1000)
			end
			label = luup.variable_get(SID_SCENECONTROLLER,"Label_"..screen.."_"..virtualButton,peer_dev_num)
			label = UnescapePercent(label)
			font = luup.variable_get(SID_SCENECONTROLLER,"Font_"..screen.."_"..virtualButton,peer_dev_num)
			align = luup.variable_get(SID_SCENECONTROLLER,"Align_"..screen.."_"..virtualButton,peer_dev_num)
		end
		if label == nil then
			label = ""
		end
		if font == nil then
			font = "Normal"
		end
		if align == nil then
			align = "Center"
		end
		labels[line] = label;
		fonts[line] = font;
		aligns[line] = align;
		VLog("ChooseLabelFontAndAlign: line=", line, " label=", label, " font=", font, " align=", align, " stateOverride=", stateOverride)
	end
	return stateOverride
end

local function SetCustomScreen(peer_dev_num, screenNum, doTimeout, forceClear, indicatorOnly)
	DEntry("SetCustomScreen")
	if not SCObj.HasScreen then
		SetPresetScreen(peer_dev_num, screenNum, doTimeout, forceClear, indicatorOnly)
		return
	end
	local screen = "C"..tostring(screenNum)
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local prevScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	if SCObj.HasMultipleScreens and (not prevScreen or SCObj.ScreenPage(prevScreen) ~= SCObj.ScreenPage(screen)) then
		forceClear = true
		indicatorOnly = false
		if not prevScreen then
			prevScreen = ""
		end
	end
	if (not indicatorOnly) and (forceClear or (SCObj.ScreenPage(screen) ~= SCObj.ScreenPage(prevScreen))) then
		SCObj.SetDeviceScreen(peer_dev_num, screen)
	end
	local prevNumLines, prevScrollOffset
	if prevScreen == nil then
		prevNumLines, prevScrollOffset = SCObj.NumButtons, 0
		forceClear = true
		indicatorOnly = false
	else
		prevNumLines, prevScrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, prevScreen)
	end
	local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, screen)
	if not indicatorOnly and forceClear then
		if scrollOffset ~= 0 then
			scrollOffset = 0
			luup.variable_set(SID_SCENECONTROLLER, SCROLLOFFSET_VAR.."_"..screen, scrollOffset, peer_dev_num)
		end
	end
	DLog("SetCustomScreen: numLines=", numLines, " scrollOffset=", scrollOffset)
	local earlyButtonConfig = false;
	if not indicatorOnly then
		SetIndicatorValue(peer_dev_num, 0, forceClear, false)
		if prevScreen and prevScreen:sub(1,1) == "T" then
			earlyButtonConfig = true;
			forceClear = true;
			for i = 2,4 do
				SetButtonMode(peer_dev_num, "T", screen, forceClear, false, i, "1");
			end
		end
		local labels = {};
		local fonts = {};
		local aligns = {};
		for i = 1,SCObj.NumButtons do
			if numLines > SCObj.NumButtons and i == SCObj.NumButtons and i + scrollOffset < numLines then
				labels[i] = "\nvvvvvv";
				fonts[i] = "Inverted";
				aligns[i] = "Center";
			elseif numLines > SCObj.NumButtons and i == 1 and scrollOffset > 0 then
				labels[i] = "^^^^^^";
				fonts[i] = "Inverted";
				aligns[i] = "Center";
			else
				ChooseLabelFontAndAlign(peer_dev_num, screen, i, i+scrollOffset, nil, labels, fonts, aligns, true)
			end
		end
		if SCObj.HasMultipleScreens then
			luup.variable_set(SID_SCENECONTROLLER, CURRENT_SCREEN, screen, peer_dev_num)
		end
		local screenFlags
		if param[peer_dev_num].SCENE_CTRL_UseClearScreen > 0 or forceClear then
			screenFlags = SCREEN_MD.ClearScreen
		else
			screenFlags = SCREEN_MD.NoChange
		end
		EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, screenFlags)
		for i = 1, SCObj.NumButtons do
			if not earlyButtonConfig or i == 1 or i == SCObj.NumButtons then
				local prevModeStr = GetModeStr(peer_dev_num, prevScreen, i+prevScrollOffset, nil, "1", true, true)
				SetButtonMode(peer_dev_num, prevModeStr, screen, forceClear, false, i, "1");
			end
		end
	end
	SetIndicator(peer_dev_num, screen, forceClear, true, false) -- coalesce, no cache
    SetScreenTimeout(peer_dev_num, doTimeout)
end

-- local
function SetPresetScreen(peer_dev_num, screenNum, doTimeout, forceClear, indicatorOnly)
	DEntry("SetPresetScreen")
	local screen = "P"..tostring(screenNum)
   	local node_id = GetZWaveNode(peer_dev_num)
	local prevScreen
	if SCObj.HasMultipleScreens then
		prevScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	else
		prevScreen = screen
	end
	local prevNumLines, prevScrollOffset
	if prevScreen == nil then
		prevNumLines, prevScrollOffset = SCObj.NumButtons, 0
		forceClear = true;
		indicatorOnly = false
	else
		prevNumLines, prevScrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, prevScreen)
	end
	if SCObj.HasMultipleScreens then
		luup.variable_set(SID_SCENECONTROLLER, CURRENT_SCREEN, screen, peer_dev_num)
	end
	if not indicatorOnly and (forceClear or prevScreen ~= screen) then
		SetIndicatorValue(peer_dev_num, 0, forceClear, true)
		SCObj.SetDeviceScreen(peer_dev_num, screen)
		for i = 1, SCObj.NumButtons do
			local prevModeStr = GetModeStr(peer_dev_num, prevScreen, i+prevScrollOffset, nil, "1", true, true)
			SetButtonMode(peer_dev_num, prevModeStr, screen, forceClear, false, i, "1");
		end
	end
	SetIndicator(peer_dev_num, screen, forceClear, true, true) -- coalesce, cache
    SetScreenTimeout(peer_dev_num, doTimeout)
end

function SceneController_SetPresetLanguage(peer_dev_num, language)
	if SCObj.HasPresetLanguages and IsVeraPrimaryController() then
	   	luup.variable_set(SID_SCENECONTROLLER, PRESET_LANGUAGE, language, peer_dev_num)
		SCObj.SetLanguage(peer_dev_num, language)
		RunZWaveQueue("SetPresetLanguage", 0);
	end
end

local function GetSetPoint(temperatureDevice, isHeat)
	local sid
	if isHeat then
		sid = SID_HEATSETPOINT
	else
		sid = SID_COOLSETPOINT
	end
	local setPointString = luup.variable_get(sid, SETPOINT_VAR, temperatureDevice)
	if not setPointString then
		local allSetpointString = luup.variable_get(SID_TEMPSETPOINT, ALLSETPOINTS_VAR, temperatureDevice)
		if allSetpointString then
			local heat, cool, auto = string.match(allSetpointString,"(%d+),(%d+),(%d+)")
			if isHeat then
				if heat then
					setPointString = heat
				end
			else
				if cool then
					setPointString = cool
				end
			end
		end
	end 
	if setPointString then
		return math.floor(tonumber(setPointString) + 0.5);
	end
	return nil
end

local function SetTemperatureLCDParameters(peer_dev_num, temperatureDevice)
	DEntry()
	local lcd_node_id = GetZWaveNode(peer_dev_num);

	local temperatureString
	if temperatureDevice > 0 then
		temperatureString = luup.variable_get(SID_TEMPSENSOR, TEMPSENSOR_VAR, temperatureDevice)
	end
	if not temperatureString then
		temperatureString = "72" -- Default if no associated temperature device
	end	
	local temperature = math.floor(tonumber(temperatureString) + 0.5);
	-- COMMAND_CLASS_SENSOR_MULTILEVEL SENSOR_MULTILEVEL_REPORT Sensor Type = Temperature Level = Size=1 | Scale=1 | Precision=0
	EnqueueZWaveMessage(SCObj, "SetLCDTemperature", lcd_node_id,  "0x31 0x5 0x1 0x9 ".. temperature, peer_dev_num, 0)

	if temperatureDevice > 0 and TemperatureDeviceIsSettable(temperatureDevice) then
		local userMode = luup.variable_get(SID_USERMODE, USERMODE_VAR, temperatureDevice)
		local choice, setPoint
		if userMode == USERMODE_COOL then
			choice = USERMODE_COOL
			setPoint = GetSetPoint(temperatureDevice, false)
		elseif userMode == USERMODE_HEAT then
			choice = USERMODE_HEAT
			setPoint = GetSetPoint(temperatureDevice, true)
		else
			local coolSetPoint = GetSetPoint(temperatureDevice, false)
			local heatSetPoint = GetSetPoint(temperatureDevice, true)
			if coolSetPoint then
				if heatSetPoint then
					local coolDiff = (coolSetPoint - temperature) ^ 2
					local heatDiff = (heatSetPoint - temperature) ^ 2
					if heatDiff <= coolDiff then
						choice = USERMODE_HEAT
						setPoint = heatSetPoint
					else
						choice = USERMODE_COOL
						setPoint = coolSetPoint 
					end
				else
					choice = USERMODE_COOL
					setPoint = coolSetPoint 
				end
			else
				if heatSetPoint then
					choice = USERMODE_HEAT
					setPoint = heatSetPoint
				end
			end
		end
		if choice then
			temperatureAutoChoices[peer_dev_num] = choice
			if setPoint then
				if choice == USERMODE_COOL then
					-- COMMAND_CLASS_THERMOSTAT_SETPOINT THERMOSTAT_SETPOINT_REPORT  SetPoint = Cooling 1 Level = Size=1 | Scale=1 | Precision=0
					EnqueueZWaveMessage(SCObj, "SetLCDSetpointCool", lcd_node_id, "0x43 0x03 0x02 0x9 ".. setPoint, peer_dev_num, 0)
				else
					-- COMMAND_CLASS_THERMOSTAT_SETPOINT THERMOSTAT_SETPOINT_REPORT  SetPoint = Heating 1 Level = Size=1 | Scale=1 | Precision=0
					EnqueueZWaveMessage(SCObj, "SetLCDSetpointHeat", lcd_node_id, "0x43 0x03 0x01 0x9 ".. setPoint, peer_dev_num, 0)
				end
			end
		end

		-- COMMAND_CLASS_THERMOSTAT_FAN_MODE  THERMOSTAT_FAN_MODE_SUPPORTED_REPORT
		-- mode supportd  Z-Wave      LCD1   Vera
		--  0     1       Auto/Low    Auto   Auto
		--  1     2       On/Low      Low    Cycle
		--  2     4       Auto/High
		--  3     8       On/High     High   On
		--  4    16       Auto/Medium
		--  5    32       On/Medium   Medium   
		local supported = 11 -- upported 1 + 2 + 4 from table above
		EnqueueZWaveMessage(SCObj, "SetLCDFanModeSupported", lcd_node_id, "0x44 0x5 ".. supported, peer_dev_num, 0)

		local fanModeString = luup.variable_get(SID_FANMODE, FANMODE_VAR, temperatureDevice)
		if fanModeString ~= nil then
			local fanMode = 3 -- high
			if fanModeString == FANMODE_AUTO then
				fanMode = 0
			elseif fanModeString == FANMODE_CYCLE then
				fanMode = 1
			end
			-- COMMAND_CLASS_THERMOSTAT_FAN_MODE  THERMOSTAT_FAN_MODE_REPORT
			EnqueueZWaveMessage(SCObj, "SetLCDFanMode", lcd_node_id, "0x44 0x3 ".. fanMode, peer_dev_num, 100)
		end
	end
end

-- This is called by the GUI to change the thermostat associated with this controller.
function SceneController_UpdateTemperatureDevice(peer_dev_num, screen, temperatureDevice)
	DEntry()
	if SCObj.HasThermostatControl and IsVeraPrimaryController() then
		temperatureDevice = tonumber(temperatureDevice);
		local curTemperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, peer_dev_num);
	DLog("Previous temperature device = ", curTemperatureDevice);
		curTemperatureDevice = tonumber(curTemperatureDevice)
		if temperatureDevice ~= curTemperatureDevice then
			luup.variable_set(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, temperatureDevice,  peer_dev_num)
			local currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num);
			if currentScreen == screen then
				SetTemperatureLCDParameters(peer_dev_num, temperatureDevice);
			end
		end
		RunZWaveQueue("UpdateTemperatureDevice", 0);
	end
end

-- A work-around for the "bug" which clears rows 1 and 5 of a custom temperature screen
function SceneController_TemperatureDeviceChanged(temperatureDevice, service, variable, old_val, new_val, context)
	DEntry()
	local peer_dev_num = tonumber(context)
	local screen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	if not screen then
		return
	end
	local prefix = screen:sub(1,1)
	if prefix ~= "T" then
		return
	end
	local screenNum = tonumber(screen:sub(2))
	SetTemperatureScreen(peer_dev_num, screenNum, nil, false, false)
	RunZWaveQueue("TemperatureDeviceChanged", 0)
end

-- local 
function TemperatureDeviceIsSettable(temperatureDevice)
	if temperatureDevice <= 0 then
	DLog("TemperatureDeviceIsSettable: return true because temperatureDevice=", temperatureDevice)
		return true  -- Dummy - No attached device. Show the arrows even if they don't do anything.
	end
	local category_num = luup.attr_get("category_num", temperatureDevice);
	if category_num ~= "5" then  -- HVAC category num
	DLog("TemperatureDeviceIsSettable: return false because category_num=", category_num)
		return false  -- Temperature sensor only
	end
	local parent = luup.attr_get("id_parent", temperatureDevice);
	if parent ~= "1" then  -- Z-Wave controller
	DLog("TemperatureDeviceIsSettable: return false because parent=", parent)
		return false  -- Non-Z-Wave thermostat
	end
	local userMode = luup.variable_get(SID_USERMODE, USERMODE_VAR, temperatureDevice)
	local result = userMode ~= USERMODE_OFF;
	DLog("TemperatureDeviceIsSettable: return ", result, " because userMode=", userMode)
	return result;
end

local function ClearScreen(dev_num, node_id)
	if SCObj.HasScreen then
		EnqueueZWaveMessage(SCObj, "ClearScreen", node_id, "146 2 0 0 0 0 0 0 0", dev_num, param[dev_num].SCENE_CTRL_ClearDelay);
	end
end

-- local 
function SetTemperatureScreen(peer_dev_num, screenNum, doTimeout, forceClear, indicatorOnly)
	local screen = "T"..screenNum;
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local screenNum = tonumber(screenNum)
	local lcdPage = SCObj.ScreenPage(screen);
	local prevScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	local prevNumLines, prevScrollOffset
	if prevScreen == nil then
		prevScreen = "";
		prevNumLines, prevScrollOffset = SCObj.NumButtons, 0
		forceClear = true;
		indicatorOnly = false;
	else
		prevNumLines, prevScrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, prevScreen)
	end
	local prevLcdPage = SCObj.ScreenPage(prevScreen);
	DEntry("SetTemperatureScreen")

	local temperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, peer_dev_num)
	if temperatureDevice == nil then
		temperatureDevice = 0
	end
	temperatureDevice = tonumber(temperatureDevice);
	local settable = TemperatureDeviceIsSettable(temperatureDevice)

	SetTemperatureLCDParameters(peer_dev_num, temperatureDevice)

	if not indicatorOnly then
		if screenNum > SCObj.NumTemperatureScreens then
		    SetIndicatorValue(peer_dev_num, 0, forceClear, false)
			if prevScreen:sub(1,1) ~= "T" then
				ClearScreen(peer_dev_num, node_id)
			end
			local labels = {};
			local fonts = {};
			local aligns = {};

			if not settable then
				-- Cover up the up/down arrows that don't work if the device is not settable
				for i = 2,4,2 do
					labels[i] = " \r ";
					fonts[i] = "Normal";
					aligns[i] = "Raw";
				end
				EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.NoChange)
				labels = {};
				fonts = {};
				aligns = {};
			end

			for i = 1, SCObj.NumButtons do
				local prevModeStr = GetModeStr(peer_dev_num, prevScreen, i+prevScrollOffset, nil, "1", true, true)
				SetButtonMode(peer_dev_num, prevModeStr, screen, forceClear, settable, i, "1");
			end

			local temperatureString = luup.variable_get(SID_TEMPSENSOR, TEMPSENSOR_VAR, temperatureDevice)
			if temperatureString == nil then
				temperatureString = "72";
			end

			for i = 1, SCObj.NumButtons, SCObj.NumButtons-1 do
				ChooseLabelFontAndAlign(peer_dev_num, screen, i, i, nil, labels, fonts, aligns, true)
			end
			EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.NoChange)
		else
			for i = 1, SCObj.NumButtons do
				local prevModeStr = GetModeStr(peer_dev_num, prevScreen, i+prevScrollOffset, nil, "1", true, true)
				SetButtonMode(peer_dev_num, prevModeStr, screen, forceClear, settable, i, "1");
			end
		end

		if prevLcdPage ~= lcdPage or forceClear then
			SetConfigurationOption("SetTemperatureScreen", peer_dev_num, node_id, 17, lcdPage, 1000)
			prevLcdPage = lcdPage;
		end
	end

	SetIndicator(peer_dev_num, screen, forceClear, false, true, false) -- no force, coalesce, no cache
	luup.variable_set(SID_SCENECONTROLLER, CURRENT_SCREEN, screen, peer_dev_num)
    SetScreenTimeout(peer_dev_num, doTimeout)

	if screenNum > SCObj.NumTemperatureScreens and temperatureDevice > 0 then
		-- Work-around to redisplay the custom labels if they get cleared
	    DLog("Watching temperature variables for temperatureDevice: ", temperatureDevice);
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_TEMPSENSOR,   TEMPSENSOR_VAR,   temperatureDevice, tostring(peer_dev_num))
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_COOLSETPOINT, SETPOINT_VAR,     temperatureDevice, tostring(peer_dev_num))
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_HEATSETPOINT, SETPOINT_VAR,     temperatureDevice, tostring(peer_dev_num))
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_FANMODE,      FANMODE_VAR,      temperatureDevice, tostring(peer_dev_num))
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_USERMODE,     USERMODE_VAR,     temperatureDevice, tostring(peer_dev_num))
		VariableWatch("SceneController_TemperatureDeviceChanged", SID_USERMODE,     ENERGYMODE_VAR,   temperatureDevice, tostring(peer_dev_num))
	end
end

local setScreenTable = {
	C=SetCustomScreen,
	P=SetPresetScreen,
	T=SetTemperatureScreen,
}

-- local 
function SetScreen(peer_dev_num, screen, doTimeout, forceClear, indicatorOnly)
	DEntry()
	if IsVeraPrimaryController() then
		local func = setScreenTable[screen:sub(1,1)]
		if func ~= nil then
			func(peer_dev_num, screen:sub(2), doTimeout, forceClear, indicatorOnly)
		else
			ELog("SetScreen: Unknown screen type: ", screen)
		end
	end
end

function SceneController_SetScreen(peer_dev_num, screen, doTimeout, forceClear, indicatorOnly)
	DEntry()
	if IsVeraPrimaryController() then
		SetScreen(peer_dev_num, screen, doTimeout, forceClear, indicatorOnly)
		RunZWaveQueue("SetScreen", 0);
	end
end

local timeout_count = 0
function SceneController_SetBacklight(peer_dev_num, timeout)
	DEntry()
	if IsVeraPrimaryController() then
		if not peer_dev_num then
			ELog("SceneController_SetBacklight: bad peer_dev_num=", peer_dev_num)
		end
		timeout_count = timeout_count + 1
		if timeout_count <= 1 then
			SCObj.SetBacklight(peer_dev_num, true)
		end
		luup.call_delay("SceneController_BacklightTimeout", timeout, peer_dev_num, true);
		RunZWaveQueue("SetBacklight", 0);
	end
end

function SceneController_BacklightTimeout(peer_dev_num_string)
	DEntry()
	if IsVeraPrimaryController() then
		timeout_count = timeout_count - 1
		if timeout_count <= 0 then
			local peer_dev_num = tonumber(peer_dev_num_string)
			SCObj.SetBacklight(peer_dev_num, false)
		end
		RunZWaveQueue("BacklightTimeout", 0);
	end
end

-- commfailure flags change on the Z-Wave device, copy it to the peer device
function SceneController_CommFailureChanged(lul_device, lul_service, lul_variable, lul_value_old, lul_value_new, context)
	DEntry()
	local zwave_dev_num = tonumber(lul_device);
  	local peer_dev_num = GetPeerDevNum(luup.device)
	PropagateCommFailure(zwave_dev_num, peer_dev_num)
end

-- local 
function PropagateCommFailure(zwave_dev_num, peer_dev_num)
	DEntry()
	for index, keyword in ipairs {"CommFailure", "CommFailureTime", "CommFailureAlarm"} do
		local zValue = luup.variable_get(SID_HADEVICE, keyword, zwave_dev_num)
		local pValue = luup.variable_get(SID_HADEVICE, keyword, peer_dev_num)
		if zValue ~= pValue then
			if keyword == "CommFailure" then
				luup.set_failure(tonumber(zValue), peer_dev_num)
			end
			luup.variable_set(SID_HADEVICE, keyword, zValue, peer_dev_num)
		end
	end
end

-- This re-associates all of the scene activation buttons after Vera has reconfigured the Z-Wave device.
-- It also possibly re-associates the thermostat if the current page is a temperature screen.
-- This is called as a watch trigger whenever the Z-Wave device is reconfigured
function SceneController_ConfiguredChanged(lul_device, lul_service, lul_variable, lul_value_old, lul_value_new, context)
	DEntry()
	if IsVeraPrimaryController() then
		local dev_num = tonumber(lul_device);
		local node_id, zwave_dev_num = GetZWaveNode(dev_num)
		if tonumber(lul_value_new) == -2 then
			-- Nexia One Touch has too many association groups and LuaUPnP stops clearing them out after group 18 out of 46
			-- So we clear this here since we will set associations and build this list as part of a reconfigure operation
			luup.variable_set(SID_ZWDEVICE, ASSOCIATION_SET, "", zwave_dev_num)
		elseif tonumber(lul_value_new) == 1 or tonumber(lul_value_new) == -3 then
			-- 10 scenes - 5 activation, 5 deactivation
			ClearReturnRouteCache(zwave_dev_num)
	  		local veraZWaveNode, ZWaveNetworkDeviceId = GetVeraIDs()
			local peer_dev_num_str = luup.variable_get(SID_SCENECONTROLLER,"PeerID",zwave_dev_num)
			local peer_dev_num = tonumber(peer_dev_num_str)
			local currentScreen = GetCurrentScreen(peer_dev_num)
			local prefix = currentScreen:sub(1,1)
			-- If this is the first event which occurred since startup, and this is a battery device,
			-- then it is already awake but we have not yet set up the monitor
			if SCObj.HasBattery then
				EnqueueInitWakeupMonitorMessage("WakeUpInit_ConfiguredChanged", node_id, peer_dev_num)
			end
			-- Force all scene actuator configurations to be re-sent eventually.
			-- We gather up a set of all direct associations and then clear
			-- the actuator configuration cache for each device associated with this controller.
			local associateList = {}
			ForAllModes(peer_dev_num, function(modeStr2, mode2, screen2, virtualButton2, state2, eventCode2)
				for i = 1, #mode2 do
					associateList[mode2[i].device] = true
				end
			end )
			DLog("  ConfiguredChanged: associateList=", associateList, " zwave_dev_num=", zwave_dev_num)
			for k, v in pairs(associateList) do
				RemoveDeviceActuatorConfForController(zwave_dev_num, k)
			end
			SCObj.DoDeviceConfiguration(peer_dev_num, node_id, zwave_dev_num)
			-- DoTimeout, force, not indicatorOnly
			SetScreen(peer_dev_num, currentScreen, true, true, false)

			-- We need to move the "Set Wakeup Interval" command for the Nexia One-Touch to the end of the configuration list
			-- so that it doesn't go to sleep too soon.
			DLog("SceneController_ConfiguredChanged: wakeUpIntervals[",peer_dev_num,"] is ", wakeUpIntervals[peer_dev_num],", node_id is ",node_id)
			if wakeUpIntervals[peer_dev_num] then
				EnqueueZWaveMessage(SCObj, "wakeUpInterval_"..node_id, node_id, wakeUpIntervals[peer_dev_num], peer_dev_num, 0);
				wakeUpIntervals[peer_dev_num] = nil;
			end
		end
		-- If LuaUPnP does not know how to configure the device, we do it for it.
		-- and then we set configured to 1 ourselves but don't re-trigger this function
		if tonumber(lul_value_new) == -3 then
    		TempVariableUnwatch(lul_service, lul_variable, lul_device) 
	  		luup.variable_set(lul_service, lul_variable, 1, lul_device)
		end
		RunZWaveQueue("ConfiguredChanged", 0);
	end
end

local SceneIds = {}
local SceneTimes = {}

-- local 
function GetCurrentScreen(peer_dev_num)
	local currentScreen
	if SCObj.HasMultipleScreens then
		currentScreen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
	end
	if not currentScreen then
		currentScreen = SCObj.DefaultScreen
	end
	return currentScreen
end

-- local 
function SceneActivatedMonitorCallback(peer_dev_num, result)
	DEntry("SceneActivatedMonitorCallback")
	if CheckDups(peer_dev_num, result, "2B"..result.C2..result.C3) then
		if IndicatorLocks[peer_dev_num] == 2 then
			-- Reenable the indicator if we did not get a switch multilevel start but
			-- then unexpectedly got a scene activation before the correspoonding switch multilevel stop 
			IndicatorLocks[peer_dev_num] = 0
		end
		local zWaveSceneId = tonumber(result.C2, 16)
		local dimmingDuration = tonumber(result.C3, 16)
		if zWaveSceneId == 0 then
			ELog("Bad Z-Wave Scene ID: ", zWaveSceneId)
			return
		end
		SceneChange(peer_dev_num, true, zWaveSceneId)
	end
end

-- local 
function BasicSetMonitorCallback(peer_dev_num, result)
	DEntry("BasicSetMonitorCallback")
	if CheckDups(peer_dev_num, result, "20"..result.C2) then
		if IndicatorLocks[peer_dev_num] == 2 then
			-- Reenable the indicator if we did not get a switch multilevel start but
			-- then unexpectedly got a basic set before the correspoonding switch multilevel stop 
			IndicatorLocks[peer_dev_num] = 0
		end
		local setValue = tonumber(result.C2, 16)
		SceneChange(peer_dev_num, false, setValue)
	end
end

-- Pass through the SWITCH_MULTILEVEL_START to non-scene-capable devices 
-- for RFWC5
-- local 
function MultiLevelSwitchStartLevelChangeMonitorCallback(peer_dev_num, result)
	DEntry("MultiLevelSwitchStartLevelChangeMonitorCallback")
	local dupString = "2604" .. result.C2 .. result.C3
	local resultC3 = result.C3
	if result.C2 == "05" then
		-- Cooper RFWC5 will send the first (multi) message with the dimmind duration
		-- but the second without. Treat the second as a dup.
		dupString = "2604" .. "04" .. string.sub(result.C3,1,5)
	else
		resultC3 = result.C3 .. " 06" -- Always add the duration if it was not there.
	end
	if CheckDups(peer_dev_num, result, dupString) then
		IndicatorLocks[peer_dev_num] = 2 -- Don't change indicators while a button is being held down
	    if SCObj.HasCooperConfiguration then
			local mode = lastChangedModes[peer_dev_num]
			VLog(" MultiLevelSwitchStartLevelChangeMonitorCallback: lastChangedModes[",peer_dev_num,"]=",mode) 
			local hexC3 = string.gsub(resultC3," "," 0x")
			if mode then
				for i = 1, #mode do
					if not mode[i].dimmingDuration then
						local node_id = GetZWaveNode(mode[i].device)
						EnqueueTargetZWaveMessage(SCObj, "StartLevelChangePassThrough", node_id, "0x26 0x04 0x" .. hexC3, mode[i].device, 0)
					end
				end
			end
		end
	end
end

-- Pass through the SWITCH_MULTILEVEL_STOP commands to non-scene-capable devices 
-- for RFWC5 and then poll the target device to decide if it is on or off.
-- local 
function MultiLevelSwitchStopLevelChangeMonitorCallback(peer_dev_num, result)
	DEntry("MultiLevelSwitchStopLevelChangeMonitorCallback")
	if CheckDups(peer_dev_num, result, "2605") then
		local mode = lastChangedModes[peer_dev_num]
		if mode then
			for i = 1, #mode do
				local node_id = GetZWaveNode(mode[i].device)
				if SCObj.HasCooperConfiguration and not mode[i].dimmingDuration then
					EnqueueTargetZWaveMessage(SCObj, "StopLevelChangePassThrough", node_id, "0x26 0x05", mode[i].device, 0)
				end
				-- Check whether polling is enabled for this particulat device (i.e. black or a non-zero value)
				-- Skip polling only if set to 0 explicitly
				local pollSettings = luup.variable_get(SID_ZWDEVICE, "PollSettings", mode[i].device)
				pollSettings = tonumber(pollSettings) -- may be nil
				if pollSettings ~= 0 then
					EnqueueLuupAction(SCObj, "ZWJob_PollNode", SID_HADEVICE, "Poll", {}, mode[i].device, 0);
				end
			end
		end
		local lock = IndicatorLocks[peer_dev_num]
		IndicatorLocks[peer_dev_num] = 0
		if lock > 2 then
			SetIndicator(peer_dev_num, GetCurrentScreen(peer_dev_num), false, true, false) -- no force, coalesce, no cache
		end
	end
end

-- local 
function ThermostatSetPointMonitorCallback(peer_dev_num, result)
	DEntry("ThermostatSetPointMonitorCallback")
	if CheckDups(peer_dev_num, result, "4301" .. result.C2 .. result.C3 .. result.C4) then
		local setPointType = tonumber(result.C2, 16)
		local levelScalePrecision = tonumber(result.C3, 16)
		local setPoint = tonumber(result.C4, 16)
		local screen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
		if not screen then
			return
		end
		SetScreenTimeout(peer_dev_num, true);
		local prefix = screen:sub(1,1)
		if prefix ~= "T" then
			return
		end
		local temperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, peer_dev_num);
		if not temperatureDevice then
			temperatureDevice = "0"
		end
		temperatureDevice = tonumber(temperatureDevice)
		local choice = temperatureAutoChoices[peer_dev_num]
		if not choice then
			choice = USERMODE_HEAT
		end
		if 	temperatureDevice <= 0 then
			return
		end
		local sid
		if choice == USERMODE_COOL then
			sid = SID_COOLSETPOINT
		else
			sid = SID_HEATSETPOINT
		end
		luup.call_action(sid, "SetCurrentSetpoint", {NewCurrentSetpoint = setPoint}, temperatureDevice)
	end
end

-- local 
function ThermostatFanModeMonitorCallback(peer_dev_num, result)
	DEntry("ThermostatFanModeMonitorCallback")
	if CheckDups(peer_dev_num, result, "4401" .. result.C2) then
		local fanModeNum = tonumber(result.C2, 16)
		local screen = luup.variable_get(SID_SCENECONTROLLER, CURRENT_SCREEN, peer_dev_num)
		if not screen then
			return
		end
		SetScreenTimeout(peer_dev_num, true);
		local prefix = screen:sub(1,1)
		if prefix ~= "T" then
			return
		end
		local temperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. screen, peer_dev_num);
		if not temperatureDevice then
			temperatureDevice = "0"
		end
		temperatureDevice = tonumber(temperatureDevice)
		if 	temperatureDevice <= 0 then
			return
		end
		local oldMode = luup.variable_get(SID_FANMODE, FANMODE_VAR, temperatureDevice)
		local fanMode
		if fanModeNum == 0 then
			fanMode = FANMODE_AUTO
		elseif fanModeNum == 1 then
			fanMode = FANMODE_CYCLE
		elseif fanModeNum == 3 then
			fanMode = FANMODE_ON
		else
			ELog("Unexpected fan mode:", fanModeNum)
			return
		end 
		if oldMode ~= fanMode then
			luup.call_action(SID_FANMODE, "SetMode", {NewMode = fanMode}, temperatureDevice)
		end
	end
end

-- This monitor callback triggers when we get an Evolve LCD controller is first
-- powered on to get its screen metadata. We respond by triggering a reconfiguration.
-- local 
function ScreenMDGetMonitorCallback(peer_dev_num, result)
	DEntry("ScreenMDGetMonitorCallback")
	if CheckDups(peer_dev_num, result, "9201") then
		local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
		luup.call_action(SID_HADEVICE,"Reconfigure", {}, zwave_dev_num)
	end
end

local centralSceneStatus = {}
local centralSceneTransactionNumber = 1
-- This monitor callback triggers when we get a Central Scene Notification event
-- local 
function CentralSceneNotificationMonitorCallback(peer_dev_num, result)
	DEntry("CentralSceneNotificationMonitorCallback")
	local status = centralSceneStatus[peer_dev_num]
	if not status then
		status = {}
		centralSceneStatus[peer_dev_num] = status
	end
	local press = status.press
	local numTaps = status.numTaps
	local lastScene = status.lastScene
	local transactionNumber = status.transactionNumber
	if numTaps == nil then
		numTaps = 1
	end
	if transactionNumber == nil then
		transactionNumber = 0
	end

	if result then
		local sequenceNum = result.C2
		if CheckDups(peer_dev_num, result, "5B"..sequenceNum, sequenceNum) then
			local C3 = tonumber(result.C3, 16)
			local keyAttributes = bit.band(C3, 7)
			local slowRefresh = bit.band(C3, 128) ~= 0
			local button = tonumber(result.C4, 16)
			local attributeNames = {
						"Key Pressed 1 time",
						"Key Released",
						"Key Held Down",
						"Key Pressed 2 times",
						"Key Pressed 3 times",
						"Key Pressed 4 times",
						"Key Pressed 5 times"}
			DLog("Central Scene notification: Sequence number=", sequenceNum, ", Key attribute=", attributeNames[keyAttributes+1], ", Slow refresh=", slowRefresh, ", Scene number=", button, " status=", status)
			if button < 1 or button > SCObj.NumButtons then
				ELog("CentralSceneNotificationMonitorCallback: Invalid scene number: ", button)
				return
			end
			if lastScene ~= nil and lastScene ~= button then
				if press == "hold" then
					press = "release"
				end
				if press ~= "ignore" then
					ExecuteCentralScene(peer_dev_num, lastScene, press, numTaps, status, false)
				end
				press = nil
				numTaps = 1
				lastScene = nil
			end

			if keyAttributes == 0 then -- Key pressed 1 time
				if press ~= nil and press ~= "ignore" then
					if press == "hold" then
						press = "release"
					end
					ExecuteCentralScene(peer_dev_num, lastScene, press, numTaps, status, false)
				end
				if ExecuteCentralScene(peer_dev_num, button, "tap", 1, status, true) then
					press = "ignore"
					numTaps = 1
					lastScene = nil
				else
					press = "tap"
					numTaps = 1
					lastScene = button
				end
			elseif keyAttributes == 1 then -- key released
				centralSceneTransactionNumber = centralSceneTransactionNumber + 1
				transactionNumber = centralSceneTransactionNumber
				if press == "tap" then
					ExecuteCentralScene(peer_dev_num, lastScene, press, numTaps, status, false)
				end
				if press ~= nil and press ~= "ignore" then
					ExecuteCentralScene(peer_dev_num, button, "release", numTaps, status, false)
				end
				press = nil
				numTaps = 1
				lastScene = nil
			elseif keyAttributes == 2 then -- key held down
				if press == "tap" then
					numTaps = numTaps + 1
					ExecuteCentralScene(peer_dev_num, button, "hold", numTaps, status, false)
				elseif press == "ignore" then
					-- do nothing for hold after quickTap
				elseif press ~= "hold" then
					lastScene = button
					numTaps = 1
					ExecuteCentralScene(peer_dev_num, button, "hold", 1, status, false)
					centralSceneTransactionNumber = centralSceneTransactionNumber + 1
					transactionNumber = centralSceneTransactionNumber
					local delaySeconds = 10
					luup.call_delay("SceneController_CentralSceneTimeout", delaySeconds, tostring(peer_dev_num)..","..tostring(button)..","..tostring(transactionNumber)..","..tostring(socket.gettime()+delaySeconds), true);
				end
				press = "hold"
			else -- 2 or more taps
				if press ~= nil and press ~= "ignore" then
					if press == "hold" then
						press = "release"
					end
					ExecuteCentralScene(peer_dev_num, lastScene, press, numTaps, status, false)
				end
				if ExecuteCentralScene(peer_dev_num, button, "tap", keyAttributes - 1, status, true) then
					press = "ignore"
					numTaps = 1
					lastScene = nil
				else
					press = "tap"
					numTaps = keyAttributes - 1
					lastScene = button
				end
			end -- if keyattribute
		end -- CheckDups
	else -- ~result = timeout
		DLog("Central Scene timeout: status=", status)
		if press == "tap" then
			ExecuteCentralScene(peer_dev_num, lastScene, press, numTaps, status, false)
			press = nil
			numTaps = 1
			lastScene = nil
		end
	end
	-- Ignore timeout when in hold press. We will get the key released event or else the call_delay SceneController_CentralSceneTimeout
	status.press=press
	status.numTaps=numTaps
	status.lastScene=lastScene
	status.transactionNumber = transactionNumber
end

function SceneController_CentralSceneTimeout(data)
	local peerStr, buttonStr, transactionNumberStr, timeStr = data:match("^(%d+),(%d+),(%d+),([%d.]+)$")
	local peer_dev_num = tonumber(peerStr)
	local button = tonumber(buttonStr)
	local transactionNumber = tonumber(transactionNumberStr)
	local time = tonumber(timeStr)
	local status = centralSceneStatus[peer_dev_num]
	DEntry("SceneController_CentralSceneTimeout")
	if status and status.transactionNumber == transactionNumber then
		local t2 = socket.gettime()
		if t2 < time then
			local delay = math.floor(time-t2)
			if delay < 1 then
				delay = 1
			end
			luup.call_delay("SceneController_CentralSceneTimeout", delay, data, true);
		else
			if status.press ~= "ignore" then
				ExecuteCentralScene(peer_dev_num, button, "release", status.numTaps, status, false)
			end
			status.press=nil
			status.numTaps=1
			status.lastScene=nil
		end
	end
end

-- local 
function ExecuteCentralScene(peer_dev_num, button, press, numTaps, status, quitNoQuickTap)
	DEntry("ExecuteCentralScene")
	local currentScreen = GetCurrentScreen(peer_dev_num)
	local eventCode = tostring(numTaps)
	local optEvent = OptEvent(eventCode)
	local modeStr = GetModeStr(peer_dev_num, currentScreen, button, nil, eventCode, false, false)
	local state = 1
	local zwave_Node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	DLog("ExecuteCentralScene: modeStr=", modeStr)
	local mode = ParseModeString(modeStr)
	local onButton = button;
	if mode.negative and button > 1 then
		onButton = button - 1
		modeStr = GetModeStr(peer_dev_num, currentScreen, onButton, nil, eventCode, false, false)
		mode = ParseModeString(modeStr)
		mode.negative = true
	end
	if quitNoQuickTap and not mode.quickTap then
		return false
	end
	if optEvent ~= "" then
		local eventsStr = luup.variable_get(SID_SCENECONTROLLER,"Events_"..currentScreen.."_"..button, peer_dev_num)
		if eventsStr and not eventsStr:find(eventCode,1,true) then
			return true
		end
		-- LastSceneAction1 is used by the Vera UI for button highlighting. It recalls the last action as a single character.
		local oldLastAction1 = luup.variable_get(SID_SCENECONTROLLER,"LastSceneAction1", peer_dev_num)
		local newLastAction1 = tostring(button)
		if oldLastAction1 ~= newLastAction1 then
			luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction1", newLastAction1, peer_dev_num)
		end
		-- LastSceneAction is used by SetIndicator. It recalls the last action regardless of the event.
		local oldLastAction = luup.variable_get(SID_SCENECONTROLLER,"LastSceneAction", peer_dev_num)
		local newLastAction = tostring(button)..eventCode
		if oldLastAction ~= newLastAction then
			luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction", newLastAction, peer_dev_num)
		end
		-- LastSceneAction_{eventCode} is used by ScreenTimeout to revert LastSceneAction to the last action of one of the other events.
		local oldLastAction2 = luup.variable_get(SID_SCENECONTROLLER,"LastSceneAction_"..eventCode, peer_dev_num)
		local newLastAction2 = tostring(button)
		if oldLastAction2 ~= newLastAction2 then
			luup.variable_set(SID_SCENECONTROLLER,"LastSceneAction_"..eventCode, newLastAction2, peer_dev_num)
		end
	end
	local now =	socket.gettime()
	if press == "tap" then -- Key pressed 1 or more times
		local isOn
		if mode.states > 1 then
			ChooseBestFromMultiState(peer_dev_num, currentScreen, onButton, mode, eventCode, true, false) -- next state, no cache
			modeStr, state = GetModeStr(peer_dev_num, currentScreen, button, nil, eventCode, true, false)
			mode = ParseModeString(modeStr)
			if quitNoQuickTap and not mode.quickTap then
				return false
			end
			isOn = false
		elseif mode.prefix == "O" then
			isOn = mode.negative
		else
			isOn = GetBinaryStateOfDevicesOnMode(peer_dev_num, mode, currentScreen, false, button, mode.negative)
		end
		status.isOn = not isOn
		status.releaseTime = now
		DLog("ExecuteCentralScene: tap=", numTaps,", Calculated isOn=", isOn) 
		for pass = 1, 2 do
			for i = 1, #mode do
				local element = mode[i]
				local level = 0
				if not isOn then
					if element.level then
						level = element.level
					else
						level = 255
					end
				end
				if pass == 1 then -- First activate any Z-Wave scenes, run the Z-Wave queu and then process any indicator triggers
					if IsZWaveChild(element.device) then
						local node_id = GetZWaveNode(element.device)
						if mode.sceneControllable then
							if isOn and mode.offZWaveSceneId then
								EnqueueTargetZWaveMessage(SCObj, "CentralSceneActivationOff", node_id, "43 1 " .. mode.offZWaveSceneId .. " 255", element.device, 0)
							else
								EnqueueTargetZWaveMessage(SCObj, "CentralSceneActivationOn", node_id, "43 1 " .. mode.zWaveSceneId .. " 255", element.device, 0)
							end
						elseif element.dimmingDuration then
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelSetWithDuration", node_id, "0x26 0x01 " .. level .. " " .. element.dimmingDuration, element.device, 0)
						elseif element.level then
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelSet", node_id, "0x26 0x01 " .. level, element.device, 0)
						else
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneBasicSet", node_id, "0x20 0x01 " .. level, element.device, 0)
						end
					else
						SendDeviceCommand(element.device, level)
					end
				else
					if level > 100 then
						level = 100
					end
					SetDeviceStatus(element.device, level)
				end
			end
			if pass == 1 then
				RunZWaveQueue("CentralSceneActivation", 0);
			end
		end
		TriggerVeraScene(peer_dev_num, SCObj.EncodeVeraSceneNumber(currentScreen, button, state, numTaps, false, false), true) -- Tap
		SetIndicator(peer_dev_num, currentScreen, false, true, true) -- no force, coalesce, cache
		SetScreenTimeout(peer_dev_num, true);
	elseif press == "release" then -- Key released
		DLog("ExecuteCentralScene: Key Released") 
		status.releaseTime = now
		for i = 1, #mode do
			local element = mode[i]
			local node_id = GetZWaveNode(element.device)
			if node_id then
				if element.level then
					EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelStopLevelChange", node_id, "0x26 0x05", element.device, 0)
					local pollSettings = luup.variable_get(SID_ZWDEVICE, "PollSettings", element.device)
					pollSettings = tonumber(pollSettings) -- may be nil
					if pollSettings ~= 0 then
						EnqueueLuupAction(SCObj, "ZWJob_PollNode", SID_HADEVICE, "Poll", {}, element.device, 0);
					end
				end
			end
		end
		TriggerVeraScene(peer_dev_num, SCObj.EncodeVeraSceneNumber(currentScreen, button, state, numTaps, true, false), false) -- Release
		SetIndicator(peer_dev_num, currentScreen, false, true, true) -- no force, coalesce, cache
		SetScreenTimeout(peer_dev_num, true);
	elseif press == "hold" then -- Key held down
		local isOn

		if not status.releaseTime then
			status.releaseTime = 0
		end
		if mode.prefix == "O" or mode.prefix == "-" or mode.states > 1 then
			isOn = mode.negative
		elseif now - status.releaseTime  < 10 then
			isOn = status.isOn -- Override previously calculated isOn so that holding down and releasing the button always alternately dims up and dims down
			DLog("  ExecuteCentralScene: Key held down: stored isOn=", isOn, " time delta=", now - status.releaseTime) 
		else
			isOn = GetBinaryStateOfDevicesOnMode(peer_dev_num, mode, currentScreen, false, button, mode.negative)
			DLog("  ExecuteCentralScene: Key held down: calculated isOn=", isOn, " time delta=", now - status.releaseTime) 
		end
		for i = 1, #mode do
			local element = mode[i]
			local node_id = GetZWaveNode(element.device)
			if node_id then
				if element.level then
					if element.dimmingDuration then
						if isOn then
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelStartLevelChangeDurationDown", node_id, "0x26 0x04 0x60 0x00 " .. element.dimmingDuration, element.device, 0)
						else
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelStartLevelChangDurationeUp", node_id, "0x26 0x04 0x20 0x00 " .. element.dimmingDuration, element.device, 0)
						end
					else
						if isOn then
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelStartLevelChangeDown", node_id, "0x26 0x04 0x60 0x00 0x06", element.device, 0)
						else
							EnqueueTargetZWaveMessage(SCObj, "CentralSceneSwitchMultiLevelStartLevelChangeUp", node_id, "0x26 0x04 0x20 0x00 0x06", element.device, 0)
						end
					end
				end
			end
		end
		status.isOn = not isOn
		TriggerVeraScene(peer_dev_num, SCObj.EncodeVeraSceneNumber(currentScreen, button, state, numTaps, true, false), true) -- Hold
		-- Don't call SetIndicator here to go into Normal (loadl load) mode. The HomeSeer dimmer has a bug
		-- whereby it goes into a strange state whereby we can no longer get into status mode until the device is power-cycled.
		-- Fortunately, it goes into normal mode itself temporarily while you hold down the buttons 
		if numTaps > 1 then
			SetIndicator(peer_dev_num, currentScreen, false, true, true) -- no force, coalesce, cache
		end
	else
		ELog("ExecuteCentralScene: Invalid press:", press)
	end
	return true
end

-- local 
function SceneButtonPressed(peer_dev_num, activate, physicalButton, indicatorUpdated, eventCode)
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local currentScreen = GetCurrentScreen(peer_dev_num)
	local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, currentScreen)
	local virtualButton = physicalButton + scrollOffset
	local optEvent = OptEvent(eventCode)
	DEntry("SceneButtonPressed")

	local oldIndicator_string = luup.variable_get(SID_SCENECONTROLLER, CURRENT_INDICATOR, peer_dev_num)
	local oldIndicator = 0
	if oldIndicator_string then
		oldIndicator = tonumber(oldIndicator_string)
	end
	local newIndicator = oldIndicator

	-- Handle scroll up and scrolll down events if numLines > SCObj.NumButtons
	-- Using the ScrollUp and ScrollDown commands, labels, indicators, and button modes all scroll together but not associations.
	-- We also need to handle to top and bottom ^^^^ and vvvv scrolling indicators which always use momentary buttons.
	if numLines > SCObj.NumButtons and physicalButton == 1 and scrollOffset > 0 then
		local labels = {}
		local fonts = {}
		local aligns = {}
		scrollOffset = scrollOffset - 1;
	  	luup.variable_set(SID_SCENECONTROLLER, "ScrollOffset_" .. currentScreen, tostring(scrollOffset), peer_dev_num)
		for i = 1, 2 do
			if i == 1 and scrollOffset > 0 then
				labels[i] = "^^^^^^"
				fonts[i] = "Inverted";
				aligns[i] = "Center";
			else
				ChooseLabelFontAndAlign(peer_dev_num, currentScreen, i, i+scrollOffset, nil, labels, fonts, aligns, true)
			end
		end
		EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.ScrollDown)
		-- Due to the Evolve LCD1 narrow/wide indicator bug, we might as well clear the indicator here and set it again below.
		SetIndicatorValue(peer_dev_num, 0, false, true)
		for i = 1, SCObj.NumButtons do
			local curModeStr
			if i <= 2 then
				curModeStr = "L"
			else
				curModeStr = GetModeStr(peer_dev_num, currentScreen, scrollOffset+i, nil, eventCode, true, true)
			end
			local prevModeStr
			if i == 1 or (i == SCObj.NumButtons and scrollOffset+i+1 < numLines) then  
				prevModeStr = "L"
			else  
				prevModeStr = GetModeStr(peer_dev_num, currentScreen, scrollOffset+i+1, nil, eventCode, true, true)
			end
			prevModeStr = string.sub(curModeStr, 1, 1) .. string.sub(prevModeStr, 2)
			SetButtonMode(peer_dev_num, prevModeStr, currentScreen, false, false, i, eventCode)
		end
		EVLCDWrapStrings({[SCObj.NumButtons]="\nvvvvvv"}, {[SCObj.NumButtons]="Inverted"}, {[SCObj.NumButtons]="Center"}, node_id, zwave_dev_num, SCREEN_MD.NoChange)
		SetIndicator(peer_dev_num, currentScreen, false, false, true) -- no force, no coalesce, cache
		luup.sleep(param[peer_dev_num].SCENE_CTRL_SwitchScreenDelay) -- Avoid a CAN when LCD1 sends the second scene activation message
		SetScreenTimeout(peer_dev_num, true);
		return
	elseif numLines > SCObj.NumButtons and physicalButton == SCObj.NumButtons and scrollOffset < numLines - SCObj.NumButtons then
		local labels = {}
		local fonts = {}
		local aligns = {}
		scrollOffset = scrollOffset + 1;
	  	luup.variable_set(SID_SCENECONTROLLER, "ScrollOffset_" .. currentScreen, tostring(scrollOffset), peer_dev_num)
		for i = SCObj.NumButtons-1, SCObj.NumButtons do
			if i == SCObj.NumButtons and scrollOffset < numLines-SCObj.NumButtons then
				labels[i] = "\nvvvvvv"
				fonts[i] = "Inverted";
				aligns[i] = "Center";
			else
				ChooseLabelFontAndAlign(peer_dev_num, currentScreen, i, i+scrollOffset, nil, labels, fonts, aligns, true)
			end
		end
		EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.ScrollUp)
		-- Due to the Evolve LCD1 narrow/wid indicator bug, we might as well clear the indicator here and set it again below.
		SetIndicatorValue(peer_dev_num, 0, false, true)
		for i = SCObj.NumButtons, 1, -1 do
			local curModeStr
			if i >= SCObj.NumButtons-1 then
				curModeStr = "L"
			else
				curModeStr = GetModeStr(peer_dev_num, currentScreen, scrollOffset+i, nil, eventCode, true, true)
			end
			local prevModeStr
			if i == SCObj.NumButtons or (i == 1 and scrollOffset > 1) then
				prevModeStr = "L"
			else
				prevModeStr = GetModeStr(peer_dev_num, currentScreen, scrollOffset+i-1, nil, eventCode, true, true)
			end
			prevModeStr = string.sub(curModeStr, 1, 1) .. string.sub(prevModeStr, 2)
			SetButtonMode(peer_dev_num, prevModeStr, currentScreen, false, false, i, eventCode)
		end
		EVLCDWrapStrings({[1]="^^^^^^"}, {[1]="Inverted"}, {[1]="Center"}, node_id, zwave_dev_num, SCREEN_MD.NoChange)
		SetIndicator(peer_dev_num, currentScreen, false, false, true) -- no force, no coalesce, cache
		luup.sleep(param[peer_dev_num].SCENE_CTRL_SwitchScreenDelay) -- Avoid a CAN when LCD1 sends the second scene activation message
		SetScreenTimeout(peer_dev_num, true);
		return
	end

	local modeStr = GetModeStr(peer_dev_num, currentScreen, virtualButton, nil, eventCode, false, false)
	DLog("SceneButtonPressed: modeStr=", modeStr)
	local mode = ParseModeString(modeStr)
	local curState = 1

	if SCObj.ModeMap[mode.prefix] == SCObj.ModeMap["T"] then
		-- A toggle button was pushed. Change our local shadow of the indicator to reflect what the device already did
	  	if activate then
	    	newIndicator = bit.bor(oldIndicator,SCObj.PhysicalButtonToIndicator(physicalButton, true))
	  	else
	    	newIndicator = bit.band(oldIndicator,SCObj.PhysicalButtonToIndicator(physicalButton, false))
	  	end
	elseif SCObj.ModeMap[mode.prefix] == SCObj.ModeMap["M"] then
		newIndicator = bit.band(oldIndicator,SCObj.PhysicalButtonToIndicator(physicalButton, false))
		if not SCObj.HasButtonModes then
			-- For "Fake" Momentary buttons, force-reset the indicator
			SetIndicatorValue(peer_dev_num, newIndicator, true, true);
		end
	end
	if newIndicator ~= oldIndicator then
  		luup.variable_set(SID_SCENECONTROLLER, CURRENT_INDICATOR, tostring(newIndicator), peer_dev_num)
	end

	if mode.prefix == "P" or mode.prefix == "E" then
		activate = true;
		local temperatureDevice = luup.variable_get(SID_SCENECONTROLLER, "TemperatureDevice_"  .. currentScreen, peer_dev_num);
		if temperatureDevice then
			temperatureDevice = tonumber(temperatureDevice)
			if mode.prefix == "P" then	
				local userMode = luup.variable_get(SID_USERMODE, USERMODE_VAR, temperatureDevice)
				if userMode and UserModeMap[userMode] then
					local nextMode = UserModeMap[userMode].next
					if nextMode == USERMODE_AUTO then
						local autoMode = luup.variable_get(SID_USERMODE, "AutoMode", temperatureDevice)
						if autoMode == "0" then
							nextMode = UserModeMap[nextMode].next
						end
					end
					luup.call_action(SID_USERMODE, "SetModeTarget", {NewModeTarget=nextMode}, temperatureDevice)
				end
			else
				local energyMode = luup.variable_get(SID_USERMODE, ENERGYMODE_VAR, temperatureDevice)
				if energyMode and EnergyModeMap[energyMode] then
					luup.call_action(SID_USERMODE, "SetEnergyModeTarget", {NewModeTarget=EnergyModeMap[energyMode].next}, temperatureDevice)
				end
			end
		end
	elseif mode.states > 1 then
		activate = true;
		local states = tonumber(mode.prefix)
		local oldState = luup.variable_get(SID_SCENECONTROLLER,"State_"..currentScreen.."_"..virtualButton..optEvent, peer_dev_num)
		if oldState == nil then
			oldState = "1"
		end

		oldState = tonumber(oldState)
		local oldModeStr = GetModeStr(peer_dev_num, currentScreen, virtualButton, oldState, eventCode, false, false)
		if oldModeStr:sub(1,1) == "T" then -- Fix legacy multi-state for Cooper.
			oldModeStr = "M" .. oldModeStr:sub(2)
		end
		DLog("SceneButtonPressed: modeStr for old state ", oldState, " = ", oldModeStr)
		local oldLabels = {}
		local oldFonts = {}
		local oldAligns = {}
		if SCObj.HasScreen then
			ChooseLabelFontAndAlign(peer_dev_num, currentScreen, physicalButton, virtualButton, oldState, oldLabels, oldFonts, oldAligns, true)
		end

		curState = oldState % states + 1
		luup.variable_set(SID_SCENECONTROLLER,"State_"..currentScreen.."_"..virtualButton..optEvent, tostring(curState), peer_dev_num)
		local curModeStr = GetModeStr(peer_dev_num, currentScreen, virtualButton, curState, eventCode, false, false)
		if curModeStr:sub(1,1) == "T" then -- Fix legacy multi-state for Cooper.
			curModeStr = "M" .. curModeStr:sub(2)
		end
		DLog("SceneButtonPressed: modeStr for current state ", curState, " = ", curModeStr)
		local labels = {}
		local fonts = {}
		local aligns = {}
		if SCObj.HasScreen then
			ChooseLabelFontAndAlign(peer_dev_num, currentScreen, physicalButton, virtualButton, curState, labels, fonts, aligns, false)
		end
		mode = ParseModeString(curModeStr)

		local nextState = curState % states + 1
		local newModeStr = GetModeStr(peer_dev_num, currentScreen, virtualButton, nextState, eventCode, false, false)
		UpdateAssociationForPhysicalButton(zwave_dev_num, currentScreen, false, curModeStr, newModeStr, physicalButton, virtualButton, nextState, eventCode)
		if SCObj.HasScreen and (oldLabels[physicalButton] ~= labels[physicalButton] or oldFonts[physicalButton] ~= fonts[physicalButton] or oldAligns[physicalButton] ~= aligns[physicalButton]) then
			EVLCDWrapStrings(labels, fonts, aligns, node_id, zwave_dev_num, SCREEN_MD.NoChange)
		end
	end
	-- For direct or toggle direct, we set the status of the target device to whatever the controller
	-- has just sent.
	DLog("SceneButtonPressed: current mode for state ", curState, " = ", mode)
	lastChangedModes[peer_dev_num] = mode
	for i = 1, #mode do
		if activate then
			if mode.sceneControllable and mode[i].level then
				SetDeviceStatus(mode[i].device, mode[i].level)
			else
				SetDeviceStatus(mode[i].device, 255)
			end
		else
			SetDeviceStatus(mode[i].device, 0)
		end
	end

	-- now calculate the scene ID on for normal Vera scene handling.
	local vera_scene = nil;
	vera_scene = SCObj.EncodeVeraSceneNumber(currentScreen, virtualButton, curState, 1, false, false);

	if mode.prefix == "X" then -- Exclusive
		activate = true
	end
	if vera_scene ~= nil then
		if mode.prefix == "X" or mode.newScreen or (#ExternalZWaveQueue > 0) then
			luup.sleep(param[peer_dev_num].SCENE_CTRL_SwitchScreenDelay) -- Avoid a CAN when LCD1 sends the second scene activation message
		end
		TriggerVeraScene(peer_dev_num, vera_scene, activate)
    	if mode.newScreen then -- Switch Screen
		    -- doTimeout, not force, not indicatorOnly
		    SetScreen(peer_dev_num,mode.newScreen,true,false,false);
		else
			SetScreenTimeout(peer_dev_num, true);
	    end
	else
		ELog("Invalid current screen: \"", currentScreen, "\"")
		-- dotimeout, not force, not indicatorOnly
		SetCustomScreen(peer_dev_num, 1, true, false, false)
	end
end

-- local 
function TriggerVeraScene(peer_dev_num, veraSceneNumber, activate)
	VEntry()
	local act_deact
	if activate then
		act_deact = "sl_SceneActivated"
	else
		act_deact = "sl_SceneDeactivated"
	end
	luup.variable_set(SID_SCTRL, act_deact, tostring(veraSceneNumber), peer_dev_num)
	local vera_scene_time_str = luup.variable_get(SID_SCTRL, "LastSceneTime", peer_dev_num)
	if not vera_scene_time_str then
		vera_scene_time_str = "0"
	end
	local vera_scene_time = tonumber(vera_scene_time_str);
	local cur_scene_time = os.time()
	if cur_scene_time <= vera_scene_time then
	    cur_scene_time = vera_scene_time + 1
	end
	luup.variable_set(SID_SCTRL, "LastSceneTime", tostring(cur_scene_time), peer_dev_num)
end

-- local
function SceneChange(peer_dev_num, isSceneId, zwave_scene)
	DEntry()
	local act_deact
	local physicalButton
	local activate
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local currentScreen = GetCurrentScreen(peer_dev_num)
	local numLines, scrollOffset = GetNumLinesAndScrollOffset(peer_dev_num, currentScreen)
	if not isSceneId then
		-- We just got a BASIC_SET command and we are not sure which button was changed.
		if SCObj.HasIndicator then
		  	local indicator_string = luup.variable_get(SID_SCENECONTROLLER, CURRENT_INDICATOR, peer_dev_num)
		  	local indicator = tonumber(indicator_string)
			activate = zwave_scene > 0
			for i = 0, SCObj.NumButtons do
				if indicator == SCObj.PhysicalButtonToIndicator(i, not activate) then
					physicalButton = i
					break
				end
			end
			if not physicalButton then
				local retry = 0
				local IndicatorReportCallback
				IndicatorReportCallback = function(peer_dev_num, response)
					DEntry("IndicatorReportCallback")
					if response then
						if CheckDups(peer_dev_num, response, "87" .. response.C2) then
							local newIndicator = tonumber(response.C2, 16)
							local oldindicator_string = luup.variable_get(SID_SCENECONTROLLER, CURRENT_INDICATOR, peer_dev_num)
							local oldIndicator = tonumber(oldindicator_string)
							DEntry("  IndicatorReportCallback")
							luup.variable_set(SID_SCENECONTROLLER, CURRENT_INDICATOR, tostring(newIndicator), peer_dev_num)
							local anyChange = false
							for physicalButton = 1, SCObj.NumButtons do
								local b = SCObj.PhysicalButtonToIndicator(physicalButton, true);
								if bit.band(b, oldIndicator) ~= bit.band(b, newIndicator) then
									local activate = bit.band(b, newIndicator) ~= 0
									anyChange = true
									SceneButtonPressed(peer_dev_num, activate, physicalButton, true, "1")
								end
							end
							-- Set the indicator here in case any of the other buttons got affected by the changes that we made
							-- This will typically be a no-op but is important for Exclusive modes
							local lock = IndicatorLocks[peer_dev_num]
							IndicatorLocks[peer_dev_num] = 0
							if anyChange or lock > 2 then
								SetIndicator(peer_dev_num, GetCurrentScreen(peer_dev_num), false, true, false) -- no force, coalesce, cache
							end
						end
					else
						retry = retry + 1
						if retry < 3 then
							-- Same as below except the label
							EnqueueZWaveMessageWithResponse(SCObj, 
															"IndicatorReportRetry"..retry, -- label
															node_id, 
															"0x87 0x02", -- sendData
															0, -- delay
															false, -- remote device
															"^01 .. 00 04 (..) "..string.format("%02x",node_id).." .. 87 03 (..)", -- responsePattern
															IndicatorReportCallback, -- callback
															true, -- oneShot
															zwave_dev_num, -- dev_num
															3000, -- timeout. 2000 is not long enough for the RFWC5 "All off" operation.
															nil, -- armPattern
															"06") -- autoResponse
						else
							log("IndicatorReportCallback: timeout")
							local lock = IndicatorLocks[peer_dev_num]
							IndicatorLocks[peer_dev_num] = 0
							if lock > 2 then
								SetIndicator(peer_dev_num, GetCurrentScreen(peer_dev_num), false, true, false) -- no force, coalesce, no cache
							end
						end
					end
				end	-- IndicatorReportCallback
--[[
                                                 C1                   C2
42      12/18/16 13:37:28.574   0x1 0x9 0x0 0x4 0x0 0xc 0x3 0x87 0x3 0x1 0x78 (##########x)
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |   |    |
                     length = 9 ------+   |   |   |   |   |    |   |   |    |
                        Request ----------+   |   |   |   |    |   |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |   |    |
          Receive Status SINGLE ------------------+   |   |    |   |   |    |
Device 20=Cooper RFWC5 Scene Controller Z-Wave -------+   |    |   |   |    |
                Data length = 3 --------------------------+    |   |   |    |
        COMMAND_CLASS_INDICATOR -------------------------------+   |   |    |
               INDICATOR_REPORT -----------------------------------+   |    |
                      Value = 1 ---------------------------------------+    |
                    Checksum OK --------------------------------------------+
41      12/18/16 13:37:28.574   ACK: 0x6 (#)
--]]
				IndicatorLocks[peer_dev_num] = 1
				EnqueueZWaveMessageWithResponse(SCObj, 
												"IndicatorReport", -- label
												node_id, 
												"0x87 0x02", -- sendData COMMAND_CLASS_INDICATOR INDICATOR_GET
												0, -- delay
												false, -- remote device
												"^01 .. 00 04 (..) "..string.format("%02x",node_id).." .. 87 03 (..)", -- responsePattern
												IndicatorReportCallback, -- callback
												true, -- oneShot
												zwave_dev_num, -- dev_num
												3000, -- timeout. 2000 is not long enough for the RFWC5 "All off" operation.
												nil, -- armPattern
												"06") -- autoResponse
				return
			end
		else
			-- Without indicator feedback, we cannot determine which button was pressed from a Basic Set command.
			DLog("Cannot trigger vera scene on generic Basic Set ",zwave_scene," because we don't know the button pressed")
			return
		end
	elseif zwave_scene > SCObj.LastFixedSceneId then
		local zWaveSceneIdCache = ZWaveSceneIdCacheList[zwave_dev_num]
		if not zWaveSceneIdCache or not zWaveSceneIdCache[zwave_scene] then
			zWaveSceneIdCache = {}
			ForAllModes(peer_dev_num, function(modeStr, mode, screen, virtualButton, state, eventCode)
				VLog("SceneChange->ForAllModes: mode=", mode, " screen=", screen, " state=", state, " virtualButton=", virtualButton)
				if mode.zWaveSceneId then
					zWaveSceneIdCache[mode.zWaveSceneId] = {screen=screen, button=virtualButton, eventCode=eventCode, state=state, activate=true}
					if mode.offZWaveSceneId then
						zWaveSceneIdCache[mode.offZWaveSceneId] = {screen=screen, button=virtualButton, eventCode=eventCode, state=state, activate=false}
					end
				end
			end )
			ZWaveSceneIdCacheList[zwave_dev_num] = zWaveSceneIdCache
		end
		local cacheEntry = zWaveSceneIdCache[zwave_scene]
		if not cacheEntry then
			ELog("SceneChange: Received unknown scene ID: ", zwave_scene, " peer_dev_num=", peer_dev_num)
			printTable(ZWaveSceneIdCacheList)
			return
		end
		activate = cacheEntry.activate;
		physicalButton = (cacheEntry.button % 1000) - scrollOffset
	elseif zwave_scene > SCObj.NumButtons then
		activate = false;
		physicalButton = zwave_scene - SCObj.NumButtons
	else
		activate = true;
		physicalButton = zwave_scene
	end
	SceneButtonPressed(peer_dev_num, activate, physicalButton, false, "1")
	-- See comment above regarding SetIndicator here.
	SetIndicator(peer_dev_num, GetCurrentScreen(peer_dev_num), false, true, false) -- no force, coalesce, no cache
end

---
--- External device interface
---

-- Returns on%, lastUpdate, service, variable if the given device is on or off. (In the case of a binary device, retun 0 or 100%
-- Returns nil if the device does not exist or is not settable
-- local 
function GetDeviceStatus(device_num)
	DEntry()
	local device = luup.devices[device_num]
  	if not device then
		ELog("GetDeviceStatus: device ",device_num," not found") 
    	return nil
  	end

	local category_num=luup.attr_get("category_num", device_num)
	local result, lastUpdate, result2, lastUpdate2
  	for i, tuple in ipairs({ 
  		{SID_SECURITYSENSOR, "Armed",           nil,               4,   100},
 		{SID_DIMMING,        "LoadLevelStatus", "LoadLevelTarget", nil, 1  },
  		{SID_SWITCHPOWER,    "Status",          "Target",          nil, 100}
	}) do
		local service, status, target, category, multiplier = unpack(tuple)
		if not category or category == category_num then
			result,  lastUpdate  = luup.variable_get(service, status, device_num)
			if target then
				result2, lastUpdate2 = luup.variable_get(service, target, device_num)
				if result then
					VLog("GetDeviceStatus: Device=",device_num,":",device.description," ",status,"=", result, ",lastUpdate=",lastUpdate," ",target,"=",result2,",lastUpdate=",lastUpdate2)
				end
				if result and result2 and lastUpdate and lastUpdate2 and lastUpdate2 > lastUpdate then
					-- Take the most recently changed of the target or status variable but always monitor the status
					result = result2
					lastUpdate = lastUpdate2
				end
			else
				if result then 
					VLog("GetDeviceStatus: Device=",device_num,":",device.description," ",status,"=", result, ",lastUpdate=",lastUpdate)
				end
			end
			if result then
				return tonumber(result)*multiplier, lastUpdate, service, status
			end
		end
	end
	ELog("GetDeviceStatus: Cannot get status of device ", device_num, ":",device.description) 
    return nil
end

-- Set the device's status without sending any Z-Wave messages when we know what the controller set it to.
-- local
function SetDeviceStatus(device_num, value)
	DEntry()
	local foundVariable = false
	local device = luup.devices[device_num]
  	if not device then
		ELog("SetDeviceStatus: device ",device_num," not found") 
    	return
  	end
  	local status = 0
	if value > 0 then
		status = 1
	end
	if value > 100 then
		value = 100
	end
  	for i, triplet in ipairs({
  		{SID_SECURITYSENSOR, "Armed"          , status},
  		{SID_DIMMING,        "LoadLevelTarget", value },
  		{SID_DIMMING,        "LoadLevelStatus", value },
  		{SID_SWITCHPOWER,    "Target"         , status},
 		{SID_SWITCHPOWER,    "Status"         , status},
  	}) do
  		local service, variable, val = unpack(triplet)
		local result, lastUpdate = luup.variable_get(service, variable, device_num)
		if result then
			local oldValue = tonumber(result)
		    DLog("SetDeviceStatus: service=", service, " variable=", variable, " oldValue=", oldValue, " value=", val)
			if oldValue ~= val then
			    -- Temporarily unwatch the variable to avoid a loop which can cause problems if more than one device is attached to a button.
			    TempVariableUnwatch(service, variable, device_num) 
			  	luup.variable_set(service, variable, tostring(val), device_num)
			end
			foundVariable = true
		end
	end
	if not foundVariable then
		ELog("SetDeviceStatus: device ", device_num, " has neither ",SID_DIMMING," LoadLevelStatus nor ",SID_SWITCHPOWER," Status")
	end
end 

-- Send a device command to LuaUPnP for a non-Z-Wave device (or a multi end point Z-Wave device)
-- local
function SendDeviceCommand(device_num, value)
	DEntry()
	local foundVariable = false
	local device = luup.devices[device_num]
  	if not device then
		ELog("SendDeviceCommand: device ",device_num," not found") 
    	return
  	end
	local service = SID_DIMMING
	local action = "SetLoadLevelTarget"
	local arguments = {newLoadlevelTarget=value}
	local statusVariable = "LoadLevelStatus"

	local result, lastUpdate = luup.variable_get(service, statusVariable, device_num)
	if not result then
	  	service = SID_SWITCHPOWER
	  	statusVariable = "Status"
	  	action = "SetTarget"
	  	if value > 0 then
	  		arguments = {newTargetValue=1}
	  	else
	  	  	arguments = {newTargetValue=0}
	  	end
	end
	EnqueueLuupAction(SCObj, action.."("..tostring(device_num)..","..tostring(value), service, action, arguments, device_num, 0)
end 

-- Watch all devices associated with a given device and trigger a poll if the target changes
local WatchedDevicesCache = {}
function WatchDevicesAssociatedWith(associated_device_num)
	DEntry()
	if WatchedDevicesCache[associated_device_num] then
		return
	end
	WatchedDevicesCache[associated_device_num] = true
	for source_device_num, v in pairs(luup.devices) do
		local peer = luup.variable_get(SID_SCENECONTROLLER,"PeerID", source_device_num)
		if not peer then -- If it has a PeerID then we are already handling it.
			local list = GetVeraAssociationList(source_device_num)
			for group, devices in pairs(list) do
				for device_endpoint, junk in pairs(devices) do
					local base_device_num = math.floor(device_endpoint) -- remove end-point.
					if base_device_num == associated_device_num then
					  	for i, obj in ipairs({
					  		{ SID_SECURITYSENSOR, "Tripped"},
					   		{ SID_DIMMING, "LoadLevelStatus"},
					   		{ SID_SWITCHPOWER, "Status"}, 		
					  	}) do
					  		local service, status = unpack(obj)
					  		local result, lastUpdate = luup.variable_get(service, status, source_device_num)
					  		if result then
					  			VariableWatch("SceneController_InvalidatePollWatchDeiveCache", SID_ZWDEVICE, ASSOCIATION_SET, source_device_num, tostring(associated_device_num))

								VariableWatch("SceneController_PollWatchedDevice", service, status, source_device_num, tostring(associated_device_num))
					  		end
					  	end
						break;
					end
				end
			end
		end
	end
end

-- This is the watch handler used by WatchDevicesAssociatedWith
function SceneController_PollWatchedDevice(source_device_num, service, status_variable, old_value, new_value, context)
	DEntry()
	local associated_device_num = tonumber(context)
	EnqueueLuupAction(SCObj, "ZWJob_PollWatchedNode", SID_HADEVICE, "Poll", {}, associated_device_num, 0)
	RunZWaveQueue("PollWatched", 0)
end

-- Invalidate the watched devices cache and rebuild the watch list in case the association list changes
-- Existing watches that may not be necessary may still be watched but this is not important.
function SceneController_InvalidatePollWatchDeiveCache(source_device_num, service, association_list, old_value, new_value, context)
	DEntry()
	local associated_device_num = tonumber(context)
	WatchedDevicesCache[associated_device_num] = nil
	WatchDevicesAssociatedWith(associated_device_num)
end

--
-- HomeSeer common functions
--
local function HomeSeer_GetDefaultModeString(button, eventCode)
	local state = math.floor(button / 1000) + 1;
	button = button % 1000;
	if button == 1 then
		local color = "W"
		local action = "N"
		if eventCode >= "2" and eventCode <= "5"  then
			local colors = "WCMYRG"
			local colorIndex = tonumber(eventCode) - 1
			color = colors:sub(colorIndex, colorIndex)
			action = eventCode
		end
		return "OL" .. action .. color
	elseif button == 2 then
		local color = "W"
		local action = "N"
		if eventCode >= "2" and eventCode <= "5"  then
			local colors = "Wcmyrg"
			local colorIndex = tonumber(eventCode) - 1
			color = colors:sub(colorIndex, colorIndex)
			action = eventCode
		end
		return "~OL" .. action .. color
	else
		-- A:   Flashing red
		-- B:   Restore
		-- C-H: solid white, red, green, blue, cyan, magenta
		local color = "rB005"
		if eventCode > "A" and eventCode <= "H" then
			local A = "A"
			local index = 1 + eventCode:byte() - A:byte()
			local colors = "r-WRGBCM"
			color = colors:sub(index,index)
		end
		if eventCode == "B" then
			return "-" .. "LR" .. color
		else
			return "-" .. "LA" .. color
		end
	end
end

-- Change the HomeSeer LED mode based on mode.LEDAction
-- LEDAction is one of:
--  N     - Set Normal mode to a color
--  L     - Level mode. Set the first n LEDs to a color and the rest to black based on value
--  #     - Set 1 status LED to a color (0 = Set status mode but don't change any LEDs)
--  A     - Set all LEDs to a color
--  C     - Set current mode. (generally used with force=true since this would otherwise be a no-op)
--  R     - Restore the previous mode 
-- LEDColor is a color in RGBCMYWK (Red ... Black) rgbcmyw (Falshing red ... flashing white), - = no change
-- CURRENT_LED_STATUS/PREVIOUS_LED_STATUS is a string which is always 4 + 2*SCObj.NumStatusLEDs long
-- encoded as MCLCSSSSSSSsssssssFFF
-- M: Current mode N (Normal) L (Level) S (Status) A(All). All but N are implemented as hardware Status
-- C: Normal color as in LEDColor above
-- L: A number between 0 and SCObj.NumStatusLEDs indicating the number of LEDs in the current level
-- C: Level color as in LEDColor above
-- S: The colors of the Status LEDs when in (virtual) status mode
-- s: The actual colors of the LEDs when in (physical) status mode
-- FFF: The flash period from 000 to 255
local function HomeSeer_SetLEDStatusValue(peer_dev_num, virtualButton, modeStr, mode, screen, force, saveRestore)
	DEntry()
	local parameters = {}
	local function CountK(s)
		local count = 0
		for k in s:gmatch("K") do
			count = count + 1
		end
		return count 
	end

	local negative = mode.negative
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	if not node_id then
		ELog("HomeSeer_SetLEDStatusValue: bad peer_dev_num=", peer_dev_num)
		return 
	end

	local LEDAction = mode.LEDAction
	if not LEDAction then
		LEDAction = "-"
	end

	local LEDColor = mode.LEDColor
	if not LEDColor then
		LEDColor = "K"
	end

	local flashPeriod = mode.flashPeriod
	if not flashPeriod then
		flashPeriod = 0
	end
	local flashPeriodStr = string.format("%03d", flashPeriod)

    local currentLEDStatus = "--0K" .. string.rep("K",SCObj.NumStatusLEDs) .. string.rep("-",SCObj.NumStatusLEDs) .. "000"
	local current = luup.variable_get(SID_SCENECONTROLLER, CURRENT_LED_STATUS, peer_dev_num)
	if current and #current == 7 + SCObj.NumStatusLEDs*2 and not force then
		currentLEDStatus = current
		if saveRestore and LEDAction ~= "R" then
			local previous = luup.variable_get(SID_SCENECONTROLLER, PREVIOUS_LED_STATUS, peer_dev_num)
			if previous ~= currentLEDStatus then
				luup.variable_set(SID_SCENECONTROLLER, PREVIOUS_LED_STATUS, currentLEDStatus, peer_dev_num)	
			end
		end
	else
		force = true
	end
	-- Normalize status
	local newLEDStatus = currentLEDStatus
	local nightlightMode = luup.variable_get(SID_SCENECONTROLLER, "NightLightMode", peer_dev_num)
	nightlightMode = nightlightMode == "+"
	local isOn

	if (LEDAction >= "1" and LEDAction <= "9") then
		ForAllModes(peer_dev_num, function(modeStr2, mode2, screen2, virtualButton2, state2, eventCode2)
			if screen == screen2 and mode2.LEDAction and mode2.LEDAction >= "1" and mode2.LEDAction <= "9" and not mode2.negative then
				VLog("modeStr2=", modeStr2, " screen2=", screen2, " virtualButton2=", virtualButton2, " state2=", state2, " eventCode2=", eventCode2)
				local LEDColor2 = mode2.LEDColor
				local defaultIsOn = mode2.negative
				if mode2.LEDAction == LEDAction then
					LEDColor2 = LEDColor
					defaultIsOn = not mode.negative
				end
				local whichLed2 = tonumber(mode2.LEDAction)
				if eventCode2 == "1" then
					-- Include the implicit local device control here
					table.insert(mode2,{device=zwave_dev_num})
				end
				local isOn, foundOne = GetBinaryStateOfDevicesOnMode(peer_dev_num, mode2, screen, true, virtualButton2, defaultIsOn)
				if not LEDColor2 then
					LEDColor2 = "K"
				elseif LEDColor2 == "-" then -- Use the current level color
					LEDColor2 = newLEDStatus:sub(4,4)
				end
				if foundOne and (isOn == nightlightMode) then
					LEDColor2 = "K"
				end
				if (whichLed2 <= SCObj.NumStatusLEDs) and ((mode2.LEDAction == LEDAction) or foundOne) then
					VLog("LEDAction2=",mode2.LEDAction, " isOn=",isOn, " LEDColo2r=",LEDColor2, " nightlightMode=", nightlightMode)
					local prevStats = newLEDStatus:sub(5,4+SCObj.NumStatusLEDs)
					local newStats = prevStats:sub(1,whichLed2-1) .. LEDColor2 .. prevStats:sub(whichLed2+1)
					newLEDStatus = "S" .. newLEDStatus:sub(2,4) .. newStats .. newStats .. flashPeriodStr
				end
			end
		end )
		-- In nightlight mode, if all LEDs would be off, ensure that at least one is on.
		if nightlightMode then
			local stats = newLEDStatus:sub(5,4+SCObj.NumStatusLEDs)
			if stats == string.rep("K", SCObj.NumStatusLEDs) then
				stats = "W" .. string.rep("K", SCObj.NumStatusLEDs-1)
				newLEDStatus = "S" .. newLEDStatus:sub(2,4) .. stats .. stats .. flashPeriodStr
			end
		end
	elseif LEDAction == "N" then
		if LEDColor == "-" then
			LEDColor = currentLEDStatus:sub(2,2)
		end
		newLEDStatus = "N" .. LEDColor .. currentLEDStatus:sub(3)
	elseif LEDAction == "L" then
		local value = GetMultilevelStateOfDevicesOnMode(peer_dev_num, mode, screen, true, virtualButton)
		if not value then
			value = 0
		end
		local onLeds = math.floor(value * (SCObj.NumStatusLEDs + 0.5) / 100)
		if onLeds > SCObj.NumStatusLEDs then
			onLeds = SCObj.NumStatusLEDs
		elseif onLeds < 0 then
			onLeds = 0
		end
		if nightlightMode and onLeds == 0 then
			onLeds = 1
		end
		if LEDColor == "-" then
			LEDColor = currentLEDStatus:sub(4,4)
		end
		newLEDStatus = "L" .. currentLEDStatus:sub(2,2) .. onLeds .. LEDColor .. currentLEDStatus:sub(5,4+SCObj.NumStatusLEDs) .. LEDColor:rep(onLeds) .. string.rep("K",SCObj.NumStatusLEDs-onLeds) .. flashPeriodStr
	elseif LEDAction == "0" then
		local prevStatus = currentLEDStatus:sub(5,4+SCObj.NumStatusLEDs)
		newLEDStatus = "S" .. currentLEDStatus:sub(2,4) .. prevStatus .. prevStatus .. flashPeriodStr
	elseif LEDAction == "A" then
		if mode.prefix == "T" then
			-- Special case for toggle status/all modes.
			isOn = GetBinaryStateOfDevicesOnMode(peer_dev_num, mode, screen, true, virtualButton, false)
			if isOn == nightlightMode then
				LEDColor = "K"
			end
			negative = false
		elseif mode.prefix == "O" then
			if negative ~= nightlightMode then
				LEDColor = "K"
			end
			negative = false
		end
		if negative then
			LEDColor = "K"
		elseif LEDColor == "-" then
			LEDColor = currentLEDStatus:sub(4,4)
		end
		newLEDStatus = "A" .. currentLEDStatus:sub(2,4)  .. currentLEDStatus:sub(5,4+SCObj.NumStatusLEDs) .. LEDColor:rep(SCObj.NumStatusLEDs) .. flashPeriodStr
	elseif LEDAction == "R" then
		local data = luup.variable_get(SID_SCENECONTROLLER, PREVIOUS_LED_STATUS, peer_dev_num)
		if data then
			newLEDStatus = data
		end
	elseif LEDAction == "C" then
		newLEDStatus = currentLEDStatus
	else
		ELog("Invalid LED action:", mode.LEDAction)
		return
	end

	local main = newLEDStatus:sub(1,1)

	if nightlightMode and main ~= "N" and SCObj.NumStatusLEDs > 1 then
		-- If in nightlight mode and all status LEDs are off then turn the bottom one on.
		-- The hardware takes care of this in normal mode.
		local LEDs = newLEDStatus:sub(5+SCObj.NumStatusLEDs, 5+(2*SCObj.NumStatusLEDs))
		local k = "K"
		local allBlack = k:rep(SCObj.NumStatusLEDs)
		if LEDs == allBlack then
			local tries = {2, 4} -- normal then level
			if main == "L" then
				tries = {4,2} -- level then normal
			end
			LEDColor = newLEDStatus:sub(tries[1],tries[1])
			if LEDColor == "K" or LEDColor == "k" or LEDColor == "-" then
				LEDColor = newLEDStatus:sub(tries[2],tries[2])
				if LEDColor == "K" or LEDColor == "k" or LEDColor == "-" then
					LEDColor = "W" -- If no other idea, set the nightlight to white
				end
			end
			LEDColor = LEDColor:upper() -- Make sure that night light is not flashing
			newLEDStatus = newLEDStatus:sub(1,4+SCObj.NumStatusLEDs) .. LEDColor .. k:rep(SCObj.NumStatusLEDs-1) .. flashPeriodStr
		end
	end

	local previous2	= luup.variable_get(SID_SCENECONTROLLER, CURRENT_LED_STATUS, peer_dev_num) 
	if previous2 ~= newLEDStatus then
		luup.variable_set(SID_SCENECONTROLLER, CURRENT_LED_STATUS, newLEDStatus, peer_dev_num)
	end	

	if VerboseLogging >= 2 then
		DLog("currentLEDStatus=", currentLEDStatus)
		DLog("newLEDStatus    =", newLEDStatus)
		change = ""
		for j = 1, #newLEDStatus do
			if currentLEDStatus:sub(j,j) ~= newLEDStatus:sub(j,j) then
				change = change .. newLEDStatus:sub(j,j)
			else
				change = change .. " "
			end
		end
		DLog("change          =", change)
	end
	
	local pmain = currentLEDStatus:sub(1,1)
	if main == "N" or force then -- Normal mode
		if main == "N" and ((pmain ~= "N") or force) then
			parameters[13] = 0 -- Set Normal LED mode
		end
		local color = newLEDStatus:sub(2,2)
		local pcolor = currentLEDStatus:sub(2,2)
		if color ~= pcolor or force then
			local index = SCObj.StatusLEDColors:find(color)
			if index then
				if index >= 8 then -- Status white = 7, Normal white = 0
					index = 1
				end
				parameters[14] = index-1 -- Set Normal LED color
			end
		end
	end
	if main ~= "N" or force then -- status mode
		if  main ~= "N" and ((pmain ~= "L" and pmain ~= "S" and pmain ~= "A") or force) then
			parameters[13] = 1 -- Set Status LED mode
		end
		local flashFlags = 0
		local prevFlashFlags = 0
		for i = 1, SCObj.NumStatusLEDs do
			local pos = 4 + i + SCObj.NumStatusLEDs
			local newColor = newLEDStatus:sub(pos, pos)
			local oldColor = currentLEDStatus:sub(pos, pos)
			local newIndex = SCObj.StatusLEDColors:find(newColor)
			local oldIndex = SCObj.StatusLEDColors:find(oldColor)
			if not oldIndex then
				force = true
			end
			if not newIndex and force then
				newIndex = 1 -- Off
			end
			if newIndex then
				if not oldIndex then
					oldIndex = 1 -- Off
				end
				local newFlash = newIndex > 8
				local oldFlash = oldIndex > 8
				if newFlash then
					newIndex = newIndex - 7
					flashFlags = bit.bor(flashFlags, bit.lshift(1, i-1))
				end
				if oldFlash then
					oldIndex = oldIndex - 7
					prevFlashFlags = bit.bor(prevFlashFlags, bit.lshift(1, i-1))
				end
				if oldIndex ~= newIndex or force then
					parameters[20+i] = newIndex-1 -- Set Status LED color
				end
			end
		end
		local prevFlashPeriod = tonumber(currentLEDStatus:sub(5 + 2 * SCObj.NumStatusLEDs))
		if not prevFlashPeriod then
			prevFlashPeriod = 0
		end
		if SCObj.NumStatusLEDs > 1 then -- Dimmers, fan controllers
			if (prevFlashPeriod ~= flashPeriod) or force then
				parameters[30] = flashPeriod
			end
			if flashFlags ~= prevFlashFlags or force then
				parameters[31] = flashFlags -- Set Status flash flags
			end
		else -- Switches
			local prevFlash = 0
			if prevFlashFlags ~= 0 then
				prevFlash = prevFlashPeriod
			end
			local flash = 0 
			if flashFlags ~= 0 then
				flash = flashPeriod
			end
			if prevFlash ~= flash or false then
				parameters[31] = flash -- Set flash period. 0 = steady
			end
		end
	end
	SetConfigurationOptions("HomeSeerLED", peer_dev_num, node_id, 1, 10, parameters)
end

local function HomeSeer_SetNightLightMode(peer_dev_num, nightlightMode)
	VEntry()
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local value = 1
	if nightlightMode then
		value = 0
	end
	SetConfigurationOption("NightLight", peer_dev_num, node_id, 3, value, 0)	
end

local function HomeSeer_SetMultiTap(peer_dev_num, screen, button, force)
	VEntry()
	local node_id, zwave_dev_num = GetZWaveNode(peer_dev_num)
	local multiap = false
	for i = 1, SCObj.NumButtons do
		local events = luup.variable_get(SID_SCENECONTROLLER, "Events_" .. screen.."_"..tostring(i), peer_dev_num)
		if events and events ~= "1" then
			multiap = true;
			break
		end
	end
	local value = 1
	if multiap then
		value = 0
	end
	SetConfigurationOption("EnableCentranScene", peer_dev_num, node_id, 6, value, 0)	
end

local function HomeSeer_EncodeVeraSceneNumber(cuurentScreen, virtualButton, state, taps, hold, timeout)
	if timeout then
		return 9000
	end
	local eventOffset = 0
	if hold then
		eventOffset = 2
	end
	return virtualButton+eventOffset+((taps-1)*4)+((state-1)*1000);
end

local function HomeSeer_DecodeVeraSceneNumber(sceneNumber)
	local virtualButton = 1
	local hold = false
	local taps = 1
	local timeout = sceneNumber >= 9000;
	local state = 1
	if not timeout then
		state = math.floor(sceneNumber / 1000) + 1;
		local t = sceneNumber % 1000 -1;
		virtualButton = t % 2 + 1;
		hold = bit.band(t, 2) ~= 0
		taps = math.floor(t / 4) + 1
	end
	return "P1", virtualButton, state, taps, hold, timeout
end


--
-- Device-specific definitions. One of these will be assigned to SCObj
--
Devices = {
	-- Evolve LCD1-specific
	[DEVTYPE_EVOLVELCD1] = {
		Id						= "EVOLVELCD1",
		Name                    = "Evolve LCD1",
		DefaultLcdVersion		= 39,
		HasScreen               = true,
		HasMultipleScreens      = true,
		ScreenWidth 		    = 65,
		RightJustifyScreenWidth = 67,
		HasPresetLanguages      = true,
		HasThermostatControl    = true,
	    NumButtons              = 5,
	    NumSceneEvents			= 0,
	    MaxScroll	        	= 10,
	    LastFixedSceneId        = 10,
		HasOffScenes            = true,
		HasIndicator			= true,
		HasMultiTap             = 0,
		HasLocalControl         = false,
	    HasDirectSceneControl	= true,
	    NumStatusLEDs			= 0,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 30,
		DefaultScreen           = "C1",
		NumTemperatureScreens   = 3, -- Preset Pages 8, 16, and 40
		StatusLEDActions		= "",
		StatusLEDTimeouts		= "",
		StatusLEDColors         = "",
		EventCodes              = {
	    							Buttons={"1"},
								    Scenes ={   }
							      },
		MaxLabelSize			= 16,
		HasBattery				= false,
		HasButtonModes			= true,
		SendsSceneActivation	= true,
		SendsBasicSet			= true,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = false,
		CentralSceneSlowRefresh = false,
		DevType                 = DEVTYPE_EVOLVELCD1,
		DeviceXml               = "D_EvolveLCD1.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:EvolveLCD1",
		OldParamPrefix          = "LCD",
		GetDefaultModeString    = function(button, eventCode)
			return "M"
		end,
		ScreenList              = {C=9,T=9,P=41},
		-- The "Large" font is actually narrower than the small font and can thus fit more characters per line.
		-- However two lines of the small font can fit in one line of the large font.
	    LargeFontWidths         = {
									3,5,7,9,9,15,11,4,6,7,8,9,4,6,4,8,
									9,9,10,9,9,9,9,9,9,9,4,4,9,8,9,7,
									11,9,9,9,9,8,8,10,10,6,6,9,9,11,10,11,
									9,11,10,9,9,10,9,14,9,10,8,6,8,6,10,11,
									6,8,8,8,8,8,6,8,8,4,5,8,4,12,8,8,
									8,8,7,7,6,8,8,12,8,8,7,8,5,9,10,11},
	    SmallFontWidths         = {
									3,4,7,11,10,16,13,3,7,7,9,11,5,6,4,10,
									10,9,10,10,11,10,10,10,10,10,5,6,11,11,11,9,
									13,11,11,11,11,10,9,11,11,8,8,11,11,13,12,12,
									10,12,12,10,10,11,11,16,11,11,10,7,9,6,10,11,
									7,9,10,9,10,9,7,10,10,4,6,9,4,14,10,10,
									10,10,8,8,7,10,9,14,9,9,8,10,6,10,12,13},

		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			local versionString
			local retry = 0
			while retry < 10 do
				versionString = luup.variable_get(SID_ZWDEVICE, "VersionInfo", zwave_dev_num)
				if 	versionString == nil then
				    log("waiting for VersionInfo for device #", zwave_dev_num, ": ", luup.devices[zwave_dev_num].description)
					luup.sleep(1000)
					retry = retry + 1;
				else
					break
				end
			end
			if not versionString then
				versionString = "1,3,20,0,39";
				log("Could not get versionInfo. Defaulting to ", versionString);
			end
			-- "VersionInfo" returns a string sunch as "1,3,20,0,37" which corresponds to firmware rev 0.37
			local version = tonumber(versionString:match("%d+,%d+,%d+,%d+,(%d+)"))
			if version == nil then
			    version = 39
			end
			if version >= 39 then
				-- This works OK for firmware version 00.00.39
				SetTuningParameter("BaseDelay"     		, 300 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ClearDelay"    		, 700 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ScrollDelay"    	, 1000 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("LineDelay"     		, 35  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("CharDelay"     		, 8   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ResponseDelay" 		, 30  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("BackoffDelay"  		, 50  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("MaxParts"      		, 2   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("UseClearScreen"		, 1   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("Retries"       		, 0   , zwave_dev_num, peer_dev_num, nil)
			elseif version >= 37 then
				-- This works OK for firmware version 00.00.37
				SetTuningParameter("BaseDelay"     		, 100 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ClearDelay"    		, 700 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ScrollDelay"    	, 1000 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("LineDelay"     		, 100 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("CharDelay"     		, 10  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ResponseDelay" 		, 30  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("BackoffDelay"  		, 50  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("MaxParts"      		, 2   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("UseClearScreen"		, 1   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("Retries"       		, 10  , zwave_dev_num, peer_dev_num, nil)
			else
				-- From Sig's feedback on forum.micasaverde.com
				-- for version 00.00.31
				-- "I tried a bunch of different settings - none of them, including these, really resulted
				--  in perfectly stable operation - the device still reboots almost every time I do a screen
				--  change.  But as I mentioned in my prior note, the controllers are now set to how I use
				--  them, and I'm not re-drawing screens during normal operation, so they're working just fine."
				SetTuningParameter("BaseDelay"     		, 3000, zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ClearDelay"    		, 5   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ScrollDelay"    	, 1000 , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("LineDelay"     		, 5   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("CharDelay"     		, 5   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("ResponseDelay" 		, 30  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("BackoffDelay"  		, 50  , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("MaxParts"      		, 1   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("UseClearScreen"		, 0   , zwave_dev_num, peer_dev_num, nil)
				SetTuningParameter("Retries"       		, 10  , zwave_dev_num, peer_dev_num, nil)
			end
			SetTuningParameter("SwitchScreenDelay"	    , 50  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("InitStaggerSeconds"	    , 15  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	    , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		, 500 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 375 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 200 , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
			-- Enable fan mode control on button 3.
			SetConfigurationOption("FanModeControl", peer_dev_num, node_id, 36, 1)
		end,

		SetDefaultSettings = function(peer_dev_num)
			local labels = {"One", "Two", "Three", "Four", "Five",
							"Six", "Seven", "Eight", "Nine", "Ten",
							"Eleven", "Twelve", "Thir-teen", "Four-teen", "Fifteen",
							"Sixteen", "Seven-teen", "Eight-teen", "Nine-teen", "Twenty",
							"Twenty one", "Twenty two", "Twenty three", "Twenty four", "Twenty five",
							"Twenty six", "Twenty seven", "Twenty eight", "Twenty nine", "Thirty"}
			for screen = 1,6 do
				for i = 1,5 do
					luup.variable_set(SID_SCENECONTROLLER, "Label_C" .. screen .. "_" .. i, labels[i + (screen - 1) * 5], peer_dev_num)
					luup.variable_set(SID_SCENECONTROLLER, "Mode_C"  .. screen .. "_" .. i, "M",  peer_dev_num)
				end
			end
			local labels2 = {"Forty-one", "Forty-five", "Forty-six", "Fifty",
							 "Fifty-one", "Fifty-five", "Fifty-six", "Sixty"}
			for screen = 1, 4 do
				luup.variable_set(SID_SCENECONTROLLER, "Label_T" .. screen+2 .. "_" .. 1, labels2[1 + (screen - 1) * 2], peer_dev_num)
				luup.variable_set(SID_SCENECONTROLLER, "Mode_T"  .. screen+2 .. "_" .. 1, "M",  peer_dev_num)
				luup.variable_set(SID_SCENECONTROLLER, "Label_T" .. screen+2 .. "_" .. 5, labels2[2 + (screen - 1) * 2], peer_dev_num)
				luup.variable_set(SID_SCENECONTROLLER, "Mode_T"  .. screen+2 .. "_" .. 5, "M",  peer_dev_num)
			end
		end,

		PhysicalButtonToIndicator = function(button, on)
			-- Bits 0 and 1 are used for "Do Not Disturb" and "Housekeeping"
			-- Bits 2-6 are buttons 1-5
			if on then
				return bit.lshift(2, button)
			else
				return bit.bxor(bit.lshift(2 ,button), 0x7C)
			end
		end,

		SetLEDStatusValue = function(peer_dev_num, virtualButton, modeStr, mode, screen, force, saveRestore)
		end,

		SetNightLightMode = function(peer_dev_num, nightlightMode)
		end,

		SetMultiTap = function(peer_dev_num, screen, button, force)
		end,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, button + 5, button, button + 5
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			local delay = 0
			if physicalButton >= 2 and physicalButton <= 4 and (newType == 3 or oldType == 3 or force) then -- Thermostat mode
				delay = 500
			end
			if force or oldType ~= newType then
				SetConfigurationOption("SetButtonType", peer_dev_num, node_id, physicalButton, newType, delay)
			end
		end,

		ScreenPage = function(screen)
			local prefix = screen:sub(1,1)
			local suffix = screen:sub(2)
			if prefix == "C" then
				return 17
			elseif prefix == "T" then
				if suffix == "1" then
					return 8
				elseif suffix == "2" then
					return 16
				elseif suffix == "3" then
					return 40
				else
					return 17
				end
			elseif prefix == "P" then
				return tonumber(suffix)
			else
				ELog("ScreenPage: Unknown screen: ", screen);
				return 17
			end
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
		   	local node_id = GetZWaveNode(peer_dev_num)
			SetConfigurationOption("SetScreen", peer_dev_num, node_id, 17, SCObj.ScreenPage(screen))
		end,

		SetLanguage = function(peer_dev_num, language)
		   	local node_id = GetZWaveNode(peer_dev_num)
		   	SetConfigurationOption("SetPresetLanguage", peer_dev_num, node_id, 16, language)
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			DEntry("SetBacklight")
			local node_id = GetZWaveNode(peer_dev_num)
			local level = 0
			if blackLightOn then
				level = 15
			end
			-- Set backlight off level (range 0-20)
			SetConfigurationOption("Backlight_on", peer_dev_num, node_id, 22, level)
			-- Set buttons off level (range 0-20)
			SetConfigurationOption("Buttons_on", peer_dev_num, node_id, 24, level)
		end,

		-- Evolve LCD1 button mode types:
		-- 0=Scene control momentary
		-- 1=Scen control toggle
		-- 2=basic set toggle
		-- 3=Temperature (buttons 2, 3, 4 only)
		-- 4=Privacy
		-- 5=Housekeeping
		-- 6=Scen Control toggle on, Basic set toggle off
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function(mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			elseif result == 1 and #mode > 0 and not mode.sceneControllable then
			    -- There are non-scene-capable direct associations, so we need to use basic set rather than scene activate/deactivate messages.
				result = 2
			end
			return result
		end,

		EncodeVeraSceneNumber = function(cuurentScreen, virtualButton, state, taps, hold, timeout)
			local sceneBases = {
				C = 1,	-- Custom      screen scenes start at base 1
				T = 31,	-- Temperature screen scenes start at base C + 5 buttons * 6 custom screens
				W = 61,	-- Welcome     screen scenes start at base T + 5 buttons * 6 temperature screens
				P = 66	-- Preset      screen scenes start at base W + 5 buttons * 1 welcome screen
				-- 271-300 - Unused	                 start at base P + 5 buttons * 41 preset screens
				-- 301-330 - Lines 6-10 for Custom screens 1-6
				-- 331-1000 - Unused
				-- 1001-1030 - State 2 for multi-state custom screens. (1031-2000 - Unused)
				-- ...
				-- 8001-8030 - State 9 for multi-state custom screens. (8031 and above - Unused)
				-- 9000, 9005, 9010, ... - Timeout scene for C1, C2, C3, etc.
			}
			local screenType, screenNum = cuurentScreen:match("(.)(%d+)")
			if not screenType then
				screenType = "C"
			end
			if not screenNum then
				screenNum = 1
			else
				screenNum = tonumber(screenNum)
			end

			local base = sceneBases[screenType]
			if not base then
				base = 1
			end
			if timeout then
				-- Timeout scene for C1, C2, C3 is 9000, 9005, 9010, etc.
				return base+((screenNum-1)*5)+8999;
			end
			local buttonGroup = math.floor((virtualButton-1)/5)
			local buttonBase = 0
			if buttonGroup > 0 then
				buttonBase = 200 + buttonGroup*100
			end
			return base+((screenNum-1)*5)+((virtualButton-1)%5)+buttonBase+((state-1)*1000);
		end,

		DecodeVeraSceneNumber = function(sceneNumber)
			local virtualButton = 1
			local hold = false
			local taps = 1
			local timeout = sceneNumber >= 9000;
			local state = 1
			if timeout then
				sceneNumber = sceneNumber - 8999
			else
				virtualButton = (sceneNumber - 1) % 5 + 1;
				state = math.floor(sceneNumber / 1000) + 1;
				sceneNumber = sceneNumber % 1000;
				if sceneNumber >= 301 and sceneNumber <= 330 then
					virtualButton = virtualButton + 5;
					sceneNumber = sceneNumber - 300
				end
			end
			local screenNum = math.floor((sceneNumber - 1) / 5) + 1
			local screenType
			if sceneNumber <= 30 then
				screenType = "C"
			elseif sceneNumber < 60 then
				screenNum = screenNum - 6
				screenType = "T"
			elseif sceneNumber <= 65 then
				screenNum = screenNum - 12
				screenType = "T"
			else
				screenNum = screenNum - 13
				screenType = "P"
			end
			return screenType .. tostring(screenNum), virtualButton, state, taps, hold, timeout
		end,

	}, -- DEVTYPE_EVOLVELCD1

	-- Cooper RFWC5-Specific
	[DEVTYPE_COOPEREFWC5] = {
		Id						= "COOPERRFWC5",
		Name                    = "Cooper RFWC5",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 5,
	    NumSceneEvents			= 0,
		MaxScroll               = 5,
	    LastFixedSceneId        = 5,
		HasOffScenes            = false,
		HasIndicator			= true,
		HasMultiTap             = 0,
		HasLocalControl         = false,
	    HasDirectSceneControl	= true,
	    NumStatusLEDs			= 0,
		HasCooperConfiguration  = true,
		MaxDirectAssociations   = 5,
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "",
		StatusLEDTimeouts		= "",
		StatusLEDColors         = "",
		EventCodes              = {
	    							Buttons={"1"},
								    Scenes ={   }
							      },
		MaxLabelSize			= 0,
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= true,
		SendsBasicSet			= true,
		SendsSwitchMultilevel   = true,
		SendsCentralScene       = false,
		CentralSceneSlowRefresh = false,
		DevType                 = DEVTYPE_COOPEREFWC5,
		DeviceXml               = "D_CooperRFWC5.xml",
		OldServiceId             = "urn:gengen_mcv-org:serviceId:CooperRFWC5",
		OldParamPrefix          = "RFWC5",

		GetDefaultModeString    = function(button, eventCode)
			return "T"
		end,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	    , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	    , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		, 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		    , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
		end,

		SetDefaultSettings = function(peer_dev_num)
			-- Empty if not HasScreen
		end,

		PhysicalButtonToIndicator = function(button, on)
			if on then
				return bit.lshift(1, button - 1)
			else
				return bit.bxor(bit.lshift(1, button - 1), 0x1F)
			end
		end,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, button, nil
		end,

		SetLEDStatusValue = function(peer_dev_num, virtualButton, modeStr, mode, screen, force, saveRestore)
		end,

		SetNightLightMode = function(peer_dev_num, nightlightMode)
		end,

		SetMultiTap = function(peer_dev_num, screen, button, force)
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = function(cuurentScreen, virtualButton, state, taps, hold, timeout)
	    	if timeout then
	    		return 9000
	    	end
			return virtualButton+((state-1)*1000);
	    end,

		DecodeVeraSceneNumber = function(sceneNumber)
			local virtualButton = 1
			local hold = false
			local taps = 1
			local timeout = sceneNumber >= 9000;
			local state = 1
			if timeout then
				sceneNumber = sceneNumber - 8999
			else
				state = math.floor(sceneNumber / 1000) + 1;
				virtualButton = sceneNumber % 1000;
			end
			return "P1", virtualButton, state, taps, hold, timeout
		end,

	}, -- DEVTYPE_COOPEREFWC5

	-- Nexia One-Touch-Specific
	[DEVTYPE_NEXIAONETOUCH] = {
		Id						= "NEXIAONETOUCH",
		Name                    = "Nexia One Touch",
	    DefaultLcdVersion       = 1,
		HasScreen               = true,
		HasMultipleScreens      = false,
		ScreenWidth 		    = 65,
		RightJustifyScreenWidth = 67,
		HasPresetLanguages      = false,
		HasThermostatControl    = true,
	    NumButtons              = 15,
	    NumSceneEvents			= 0,
	    MaxScroll	        	= 15,
	    LastFixedSceneId        = 15,
		HasOffScenes            = false,
		HasIndicator			= false,
		HasMultiTap				= 0,
		HasLocalControl         = false,
	    HasDirectSceneControl	= true,
	    NumStatusLEDs			= 0,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 2,
		DefaultScreen           = "C1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "",
		StatusLEDTimeouts		= "",
		StatusLEDColors         = "",
		EventCodes              = {
	    							Buttons={"1"},
								    Scenes ={   }
							      },
		MaxLabelSize			= 17,
		HasBattery				= true,
		HasButtonModes			= true,
		SendsSceneActivation	= true,
		SendsBasicSet			= true,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = false,
		DevType                 = DEVTYPE_NEXIAONETOUCH,
		DeviceXml               = "D_NexiaOneTouch.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:NexiaOneTouch",
		OldParamPrefix          = "NEXIA",

		GetDefaultModeString    = function(button, eventCode)
			return "M"
		end,
		ScreenList              = {C=1},
		-- The "Large" font is actually narrower than the small font and can thus fit more characters per line.
		-- However two lines of the small font can fit in one line of the large font.
	    LargeFontWidths         = {
									3,3,5,11,9,13,11,2,5,5,7,9,2,5,2,6,
									8,6,8,7,9,8,8,8,8,8,3,3,8,6,8,7,
									16,10,10,11,11,10,10,11,10,2,7,10,8,12,10,11,
									10,11,10,10,10,10,10,16,12,10,10,4,5,4,8,9,
									3,8,8,7,8,8,5,8,7,2,3,8,2,12,7,8,
									8,8,5,6,5,7,8,12,8,8,8,6,3,6,8,3},
	    SmallFontWidths         = {
									3,2,4,7,6,8,7,2,6,5,4,6,3,4,2,4,
									6,4,6,6,6,6,6,6,6,6,2,3,6,6,6,6,
									9,6,6,6,6,6,5,6,6,2,5,6,5,8,6,7,
									6,7,7,6,6,6,6,10,6,8,6,3,4,3,5,6,
									3,5,5,4,6,5,4,5,5,2,3,5,2,8,5,5,
									5,5,3,4,4,5,6,8,6,6,4,4,2,4,6,3},
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("BaseDelay"     		, 300 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ClearDelay"    		, 700 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ScrollDelay"    	, 1000 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("LineDelay"     		, 35  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CharDelay"     		, 8   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ResponseDelay" 		, 30  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("BackoffDelay"  		, 50  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MaxParts"      		, 2   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseClearScreen"		, 1   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		, 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SwitchScreenDelay"	    , 50  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("InitStaggerSeconds"	    , 15  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	    , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		, 500 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController" , 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"       , 0   , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 350 , zwave_dev_num, peer_dev_num, nil)

		   	local node_id = GetZWaveNode(zwave_dev_num)

			-- Here we set up kluges to avoid problems with UI7 with the Nexia One-Touch
			-- LuaUPnP does not handle the ASSOCIATION_GROUP_INFO_GET command in list mode
			-- correctly. The device is allowed to send all of its associations in one list
			-- or as several reports. The Nexia one-touch sends 46 such reports and this
			-- seriously confuses LuaUPnP which is trying to get other information from the
			-- device.
			-- Vera should use the results of ASSOCIATION_GROUPTINGS_GET to decide how much 
			-- data to expect from the device.
			-- Our work around is return 1 association (the lifeline) to vera (we know better)
			-- and then to fake-out the call to ASSOCIATION_GROUP_INFO_GET to return only one
			-- info packet.
			-- The nexia 1-touch for its part is sending the wrong profile for group 1
			-- It should be 0x0001 (lifeline) rather 0x0000 (N/A) but this is relatively minor
			-- and we fix that here too. 

			MonitorZWaveData( true, -- outgoing
							  peer_dev_num,
							  nil, -- No arm data
--[==[
                                              C1                       C2
41      12/29/16 18:40:07.189   0x1 0x9 0x0 0x13 0xf 0x2 0x85 0x5 0x5 0x6 0x6b (##########k) 
           SOF - Start Of Frame --+   |   |    |   |   |    |   |   |   |    |
                     length = 9 ------+   |    |   |   |    |   |   |   |    |
                        Request ----------+    |   |   |    |   |   |   |    |
           FUNC_ID_ZW_SEND_DATA ---------------+   |   |    |   |   |   |    |
Device 45=Nexia One Touch Scene Controller Z-Wave -+   |    |   |   |   |    |
                Data length = 2 -----------------------+    |   |   |   |    |
      COMMAND_CLASS_ASSOCIATION ----------------------------+   |   |   |    |
      ASSOCIATION_GROUPINGS_GET --------------------------------+   |   |    |
Xmit options = ACK | AUTO_ROUTE ------------------------------------+   |    |
                   Callback = 6 ----------------------------------------+    |
                    Checksum OK ---------------------------------------------+
--]==]
							  "^01 .. 00 (..) " .. string.format("%02X", node_id) .. " .. 85 05 .* (..) ..$",
--[==[
42      12/29/16 18:40:07.221   0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   |   |   |    |   |    |
           SOF - Start Of Frame ------+   |   |    |   |    |
                     length = 4 ----------+   |    |   |    |
                       Response --------------+    |   |    |
           FUNC_ID_ZW_SEND_DATA -------------------+   |    |
                     RetVal: OK -----------------------+    |
                    Checksum OK ----------------------------+
42      12/29/16 18:40:07.222   got expected ACK 
41      12/29/16 18:40:07.222   ACK: 0x6 (#) 
42      12/29/16 18:40:07.261   0x1 0x5 0x0 0x13 0x6 0x0 0xef (#######) 
           SOF - Start Of Frame --+   |   |    |   |   |    |
                     length = 5 ------+   |    |   |   |    |
                        Request ----------+    |   |   |    |
           FUNC_ID_ZW_SEND_DATA ---------------+   |   |    |
                   Callback = 6 -------------------+   |    |
           TRANSMIT_COMPLETE_OK -----------------------+    |
                    Checksum OK ----------------------------+
41      12/29/16 18:40:07.261   ACK: 0x6 (#) 
42      12/29/16 18:40:07.361   0x1 0x9 0x0 0x4 0x0 0xf 0x3 0x85 0x6 0x01 0x53 (#########.S) 
           SOF - Start Of Frame --+   |   |   |   |   |   |    |   |    |    |
                     length = 9 ------+   |   |   |   |   |    |   |    |    |
                        Request ----------+   |   |   |   |    |   |    |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |   |   |    |   |    |    |
          Receive Status SINGLE ------------------+   |   |    |   |    |    |
Device 45=Nexia One Touch Scene Controller Z-Wave ----+   |    |   |    |    |
                Data length = 3 --------------------------+    |   |    |    |
      COMMAND_CLASS_ASSOCIATION -------------------------------+   |    |    |
   ASSOCIATION_GROUPINGS_REPORT -----------------------------------+    |    |
        Supported Groupings = 1 ----------------------------------------+    |
                    Checksum OK ---------------------------------------------+
--]==]
							  "06 01 04 01 \\1 01 XX 01 05 00 \\1 \\2 00 XX 01 09 00 04 00 " .. string.format("%02X", node_id) .. " 03 85 06 01 XX",
							  DummyCallback, 
							  false, -- not oneShot
							  0, -- no timeout
							  "NexiaAssociationGroupings");

			MonitorZWaveData( true, -- outgoing
							  peer_dev_num,
							  nil, -- No arm data
--[==[
                                              C1                                 C2
41      12/29/16 19:20:35.302   0x1 0xb 0x0 0x13 0xf 0x4 0x59 0x3 0x40 0x0 0x5 0x2b 0xd8 (######Y#@##+#) 
           SOF - Start Of Frame --+   |   |    |   |   |    |   |    |   |   |    |    |
                    length = 11 ------+   |    |   |   |    |   |    |   |   |    |    |
                        Request ----------+    |   |   |    |   |    |   |   |    |    |
           FUNC_ID_ZW_SEND_DATA ---------------+   |   |    |   |    |   |   |    |    |
Device 45=Nexia One Touch Scene Controller Z-Wave -+   |    |   |    |   |   |    |    |
                Data length = 4 -----------------------+    |   |    |   |   |    |    |
COMMAND_CLASS_ASSOCIATION_GRP_INFO -------------------------+   |    |   |   |    |    |
     ASSOCIATION_GROUP_INFO_GET --------------------------------+    |   |   |    |    |
              Flags = List Mode -------------------------------------+   |   |    |    |
        Grouping Identifier = 0 -----------------------------------------+   |    |    |
Xmit options = ACK | AUTO_ROUTE ---------------------------------------------+    |    |
                  Callback = 43 --------------------------------------------------+    |
                    Checksum OK -------------------------------------------------------+
--]==]
							  "^01 .. 00 (..) " .. string.format("%02X", node_id) .. " .. 59 03 40 00 .* (..) ..$",
--[==[
42      12/29/16 19:20:35.321   0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   |   |   |    |   |    |
           SOF - Start Of Frame ------+   |   |    |   |    |
                     length = 4 ----------+   |    |   |    |
                       Response --------------+    |   |    |
           FUNC_ID_ZW_SEND_DATA -------------------+   |    |
                     RetVal: OK -----------------------+    |
                    Checksum OK ----------------------------+
41      12/29/16 19:20:35.321   ACK: 0x6 (#) 
42      12/29/16 19:20:35.353   0x1 0x5 0x0 0x13 0x2b 0x0 0xc2 (####+##) 
           SOF - Start Of Frame --+   |   |    |    |   |    |
                     length = 5 ------+   |    |    |   |    |
                        Request ----------+    |    |   |    |
           FUNC_ID_ZW_SEND_DATA ---------------+    |   |    |
                  Callback = 43 --------------------+   |    |
           TRANSMIT_COMPLETE_OK ------------------------+    |
                    Checksum OK -----------------------------+
41      12/29/16 19:20:35.354   ACK: 0x6 (#) 

42      12/29/16 19:20:35.387   0x1 0x10 0x0 0x4 0x0 0xf 0xa 0x59 0x4 0x81 0x1 0x0 0x0 0x0 0x0 0x0 0x0 0x33 (######\nY#########3) 
           SOF - Start Of Frame --+    |   |   |   |   |   |    |   |    |   |   | +-----+   | +-----+    |
                    length = 16 -------+   |   |   |   |   |    |   |    |   |   |    |      |    |       |
                        Request -----------+   |   |   |   |    |   |    |   |   |    |      |    |       |
FUNC_ID_APPLICATION_COMMAND_HANDLER -----------+   |   |   |    |   |    |   |   |    |      |    |       |
          Receive Status SINGLE -------------------+   |   |    |   |    |   |   |    |      |    |       |
Device 45=Nexia One Touch Scene Controller Z-Wave -----+   |    |   |    |   |   |    |      |    |       |
               Data length = 10 ---------------------------+    |   |    |   |   |    |      |    |       |
COMMAND_CLASS_ASSOCIATION_GRP_INFO -----------------------------+   |    |   |   |    |      |    |       |
  ASSOCIATION_GROUP_INFO_REPORT ------------------------------------+    |   |   |    |      |    |       |
Flags = List Mode | Group Count=1 ---------------------------------------+   |   |    |      |    |       |
        Grouping Identifier = 1 ---------------------------------------------+   |    |      |    |       |
                       Mode = 0 -------------------------------------------------+    |      |    |       |
       Profile = AGI_GENERAL_NA ------------------------------------------------------+      |    |       |
                   Reserved = 0 -------------------------------------------------------------+    |       |
                 Event Code = 0 ------------------------------------------------------------------+       |
                    Checksum OK --------------------------------------------------------------------------+
		end,
--]==]
							  "06 01 04 01 \\1 01 XX 01 05 00 \\1 \\2 00 XX 01 10 00 04 00 " .. string.format("%02X", node_id) .. " 0A 59 04 81 01 00 00 01 00 00 00 XX",
							  DummyCallback, 
							  false, -- not oneShot,
							  0, -- no timeout
							  "NexiaBypass2");

			-- The Nexisa One-Touch seems to think that getting a WAKE_UP_INTERVAL_SET command is an excuse to go back
			-- to sleep even though there is more work to do. This is not Vera's fault. This is a problem with the
			-- Nexia firmware. We do the intercept to prevent the device from receiving this command until we are done
			-- configuring it.

			local function WakeUpIntervalSetCallback(peer_dev_num, captures)
				DEntry()
				if captures and captures.C2 and captures.C3 then
					wakeUpIntervals[peer_dev_num] = "0x84 0x04 0x" .. string.gsub(captures.C2 .. " " .. captures.C3, " ", " 0x")
					DLog("WakeUpIntervalSetCallback: wakeUpIntervals[", peer_dev_num,"] set to", wakeUpIntervals[peer_dev_num]) 
				end
			end

			MonitorZWaveData( true, -- outgoing
							  peer_dev_num,
							  nil, -- No arm data
--[==[
											  C1                   ------C2-----  C3	   C4
41      11/25/17 8:48:44.793    0x1 0xd 0x0 0x13 0xdf 0x6 0x84 0x4 0x0 0x10 0x68 0x1 0x5 0x12 0xd6 (#\r########h####) 
           SOF - Start Of Frame --+   |   |    |    |   |    |   | +-----------+   |   |    |    |
                    length = 13 ------+   |    |    |   |    |   |       |         |   |    |    |
                        Request ----------+    |    |   |    |   |       |         |   |    |    |
           FUNC_ID_ZW_SEND_DATA ---------------+    |   |    |   |       |         |   |    |    |
Device 900=Nexia One Touch Scene Controller Z-Wave -+   |    |   |       |         |   |    |    |
                Data length = 6 ------------------------+    |   |       |         |   |    |    |
          COMMAND_CLASS_WAKE_UP -----------------------------+   |       |         |   |    |    |
           WAKE_UP_INTERVAL_SET ---------------------------------+       |         |   |    |    |
                 Seconds = 4200 -----------------------------------------+         |   |    |    |
nodeid: 1 Device 2=_Scene Controller ----------------------------------------------+   |    |    |
Xmit options = ACK | AUTO_ROUTE -------------------------------------------------------+    |    |
                  Callback = 18 ------------------------------------------------------------+    |
                    Checksum OK -----------------------------------------------------------------+
--]==]
							  "^01 .. 00 (..) " .. string.format("%02X", node_id) .. " 06 84 04 (.. .. ..) (..) .. (..) ..",
--[==[
42      11/25/17 8:48:44.814    0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   |   |   |    |   |    |
           SOF - Start Of Frame ------+   |   |    |   |    |
                     length = 4 ----------+   |    |   |    |
                       Response --------------+    |   |    |
           FUNC_ID_ZW_SEND_DATA -------------------+   |    |
                     RetVal: OK -----------------------+    |
                    Checksum OK ----------------------------+
42      11/25/17 8:48:44.814    got expected ACK 
41      11/25/17 8:48:44.815    ACK: 0x6 (#) 

42      11/25/17 8:48:44.892    0x1 0x5 0x0 0x13 0x12 0x0 0xfb (#######) 
           SOF - Start Of Frame --+   |   |    |    |   |    |
                     length = 5 ------+   |    |    |   |    |
                        Request ----------+    |    |   |    |
           FUNC_ID_ZW_SEND_DATA ---------------+    |   |    |
                  Callback = 18 --------------------+   |    |
           TRANSMIT_COMPLETE_OK ------------------------+    |
                    Checksum OK -----------------------------+
41      11/25/17 8:48:44.892    ACK: 0x6 (#)
--]==]
							  "06 01 04 01 \\1 01 XX 01 05 00 \\1 \\4 00 XX",
							  WakeUpIntervalSetCallback, 
							  false, -- not oneShot,
							  0, -- no timeout
							  "NexiaWakeupIntervalSet");

--[===[ Not used anymore.

		function SceneController_NexiaConfigurationReportDupTimeout(data)
			DEntry()
			local peerStr, context = data:match("^(%d+),(.*)$")
		    local peer_dev_num = tonumber(peerStr)
		    CancelZWaveMonitor(peer_dev_num, context)
		end

		function NexiaConfigurationReportDupMonitorCallback(peer_dev_num, captures)
			DEntry()
			if captures then
				DLog("Duplicate Configuration report of parameter 0x", captures.C2, " value 0x", captures.C3, " suppressed.")
			end
		end

		function NexiaConfigurationReportMonitorCallback(peer_dev_num, captures)
			DEntry();
			if captures then
				local node_id = GetZWaveNode(peer_dev_num)
				local context = MonitorZWaveData(false, -- incoming,
				 	peer_dev_num, -- peer_dev_num
                 	nil, -- No arm_regex
                 	"^01 .. 00 04 (..) " .. string.format("%02X", node_id) .. " .. 70 06 (" .. captures.C2 .. ") 01 (" .. captures.C3 ..")", -- Main RegEx
                 	"06", -- ACK response,
                 	nil, -- NexiaConfigurationReportDupMonitorCallback,
                 	false, -- Not OneShot
                 	0, -- no timeout
				 	"*NexiaConfigurationReportDup",
				 	false) -- forward
				luup.call_delay("SceneController_NexiaConfigurationReportDupTimeout", 10, tostring(peer_dev_num)..","..context)
			end
		end

--[==[
                                -------------------C1-------------------C2------C3--------------
42      10/20/18 17:18:39.009   0x1 0xd 0x0 0x4 0x0 0xdf 0x5 0x70 0x6 0x14 0x1 0xa 0xa7 0x0 0xe2 (#\r#####p###\n###) 
           SOF - Start Of Frame --+   |   |   |   |    |   |    |   |    |   |   |    |   |    |
                    length = 13 ------+   |   |   |    |   |    |   |    |   |   |    |   |    |
                        Request ----------+   |   |    |   |    |   |    |   |   |    |   |    |
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   |    |   |    |   |    |   |   |    |   |    |
          Receive Status SINGLE ------------------+    |   |    |   |    |   |   |    |   |    |
Device 900=Nexia One Touch Z-Wave ---------------------+   |    |   |    |   |   |    |   |    |
                Data length = 5 ---------------------------+    |   |    |   |   |    |   |    |
    COMMAND_CLASS_CONFIGURATION --------------------------------+   |    |   |   |    |   |    |
           CONFIGURATION_REPORT ------------------------------------+    |   |   |    |   |    |
          Parameter Number = 20 -----------------------------------------+   |   |    |   |    |
                 Level = Size=1 ---------------------------------------------+   |    |   |    |
       Configuration Value = 10 -------------------------------------------------+    |   |    |
        Received RSSI = -89 dBm ------------------------------------------------------+   |    |
            Security key = None ----------------------------------------------------------+    |
                    Checksum OK ---------------------------------------------------------------+
--]==]
		MonitorZWaveData(false, -- incoming,
						 peer_dev_num, -- peer_dev_num
		                 nil, -- No arm_regex
		                 "^(01 .. 00 04 .. " .. string.format("%02X", node_id) .. " .. 70 06 (..) 01 (..) .*)$", -- Main RegEx
		                 "\\1", -- ACK response,
		                 NexiaConfigurationReportMonitorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "NexiaConfigurationReport",
						 true) -- forward
--]===]

		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
  			local veraZWaveNode, ZWaveNetworkDeviceId = GetVeraIDs()
			local associateList	= {[ZWaveNetworkDeviceId] = {level=255, dimmingDuration=255}}
			UnassociateDevice(zwave_dev_num, nil, 1, 1, 1, 0) -- Clear central scene association
			AssociateDevice(zwave_dev_num, associateList, 1, 1, 1, 0) -- Associate central scene with Vera
		end,

		SetDefaultSettings = function(peer_dev_num)
			local labels = {"One", "Two", "Three", "Four", "Five",
							"Six", "Seven", "Eight", "Nine", "Ten",
							"Eleven", "Twelve", "Thir-teen", "Four-teen", "Fifteen",
							"Sixteen", "Seven-teen", "Eight-teen", "Nine-teen", "Twenty",
							"Twenty one", "Twenty two", "Twenty three", "Twenty four", "Twenty five",
							"Twenty six", "Twenty seven", "Twenty eight", "Twenty nine", "Thirty"}
			for i = 1,15 do
				luup.variable_set(SID_SCENECONTROLLER, "Label_C1_" .. i, labels[i], peer_dev_num)
				luup.variable_set(SID_SCENECONTROLLER, "Mode_C1_" .. i, "M",  peer_dev_num)
			end
		end,

		PhysicalButtonToIndicator = function(button, on)
			-- Nexia One Touch does not support the indicator command class
			return 0
		end,

		SetLEDStatusValue = function(peer_dev_num, virtualButton, modeStr, mode, screen, force, saveRestore)
		end,

		SetNightLightMode = function(peer_dev_num, nightlightMode)
		end,

		SetMultiTap = function(peer_dev_num, screen, button, force)
		end,

		-- Returns:
		--   Association On Group ID
		--   Association Off Group ID (or nil)
		--   Scene Controller Configuration On Group ID
		--   Scene Controller Configuration Off Group ID (or nil)
		PhysicalButtonToGroupIds = function(button, mode)
			if mode.prefix == 'H' then 
			    -- Thermostat
				return button + 31, nil, button, nil
			elseif mode.sceneControllable then -- Scene activation
				if mode.prefix == "T" then 
				    -- For scene controllable toggle we need to associate with both the scene and the basic set groups. 
					return button + 16, button + 1, button, nil
				else
				    -- Scene compatible momentary
					return button + 16, nil, button, nil
				end
			else
				if mode.prefix == "T" then 
				    -- Basic set
					return button + 1,  nil, button, nil
				else
					-- Send Scene control message back to the central controller for momentary scenes. 
					return button + 16,  nil, button, nil
				end
			end
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			DEntry("SetButtonType")
			if force or oldType ~= newType then
				SetConfigurationOption("SetButtonType", peer_dev_num, node_id, physicalButton + 1, newType)
			end
		end,

		ScreenPage = function(screen)
			-- Stub if device does not support multiple screens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- No language support
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- No backlight control
		end,

		-- Nexia One-Touch button mode types:
		-- 0=Central Scene
		-- 1=Scene Control Momentary
		-- 2=BASIC SET Toggle
		-- 3=Scene Control/BASIC SET toggle
		-- 4=Thermostat
		ModeMap = { M=1,  -- Momentary
		            D=1,  -- Direct	(Deprecated)
					T=3,  -- Toggle
					["2"]=1, -- Two-state
					["3"]=1, -- Three-state
					["4"]=1, -- Four-state
					["5"]=1, -- Five-state
					["6"]=1, -- Six-state
					["7"]=1, -- Seven-state
					["8"]=1, -- Eight-state
					["9"]=1, -- Nine-state
					S=3,  -- Direct Toggle (Deprecated)
					X=3,  -- Exclusive
					N=1,  -- Switch Screen
					H=4,  -- Temperature
					W=1,  -- Welcome
					P=1,  -- Thermostat Operating mode
					E=1,  -- Thermostat Energy mode
					C=0,  -- Central Scene
					L=1,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				return 1
			end
			if result == 3 and not mode.sceneControllable then
			    -- There are non-scene-capable direct associations, so we need to use basic set rather than scene activate/deactivate messages.
				result = 2
			end
			return result
		end,

	    EncodeVeraSceneNumber = function(cuurentScreen, virtualButton, state, taps, hold, timeout)
	    	if timeout then
	    		return 9000
	    	end
	    	local eventOffset = 0
	    	if hold then
	    		eventOffset = 330
	    	end
			return virtualButton+eventOffset+((taps-1)*660)+((state-1)*1000);
	    end,

		DecodeVeraSceneNumber = function(sceneNumber)
			local virtualButton = 1
			local hold = false
			local taps = 1
			local timeout = sceneNumber >= 9000;
			local state = 1
			if not timeout then
				state = math.floor(sceneNumber / 1000) + 1;
				local t = sceneNumber % 1000 - 1
				taps = math.floor(t / 660) + 1
				t = t % 660
				hold = t >= 330
				virtualButton = t % 330 + 1;
			end
			return "C1", virtualButton, state, taps, hold, timeout
		end,

	}, -- DEVTYPE_NEXIAONETOUCH

	-- HomeSeer HS-WS100+ switch-specific
	[DEVTYPE_HOMESEERWS100] = {
		Id						= "HOMESEERWS100",
		Name                    = "HomeSeer HD-WS100+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 0,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = false,
		HasIndicator			= false,
		HasMultiTap				= 3,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 0,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "",
		StatusLEDTimeouts 		= "",
		StatusLEDColors         = "",
		EventCodes              = {
	    							Buttons={"1", "2", "3"},
								    Scenes ={}
							      },
		MaxLabelSize			= 1,  -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERWS100,
		DeviceXml               = "D_HomeSeerWS100.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerWS100",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	    , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	    , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		, 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		    , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = function(peer_dev_num, screen, button, force)
		end,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, -- DEVTYPE_HOMESEERWS200

    -- HomeSeer HS-WD100+ dimmer-specific
	[DEVTYPE_HOMESEERWD100] = {
		Id						= "HOMESEERWD100",
		Name                    = "HomeSeer HD-WD100+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 8,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = true,
		HasIndicator			= false,
		HasMultiTap				= 3,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 0,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "",
		StatusLEDTimeouts 		= "",
		StatusLEDColors         = "",
		EventCodes              = {
	    							Buttons={"1", "2", "3", "4", "5"},
								    Scenes ={"A", "B", "C", "D", "E", "F", "G", "H" }
							      },
		MaxLabelSize			= 1, -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERWD100,
		DeviceXml               = "D_HomeSeerWD100.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerWD100",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	     , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	     , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		 , 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		     , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = function(peer_dev_num, screen, button, force)
		end,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, -- DEVTYPE_HOMESEERWD100

    -- HomeSeer HS-FC200+ fan controller-specific
	[DEVTYPE_HOMESEERFC200] = {
		Id						= "HOMESEERFC200",
		Name                    = "HomeSeer HD-FC200+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 8,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = true,
		HasIndicator			= false,
		HasMultiTap				= 5,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 4,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "-NL4321AR",
		StatusLEDTimeouts 		= "12345S",
		StatusLEDColors         = "KRGBMYCWrgbmycw",
		EventCodes              = {
	    							Buttons={"1", "2", "3", "4", "5"},
								    Scenes ={"A", "B", "C", "D", "E", "F", "G", "H" }
							      },
		MaxLabelSize			= 1, -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERFC200,
		DeviceXml               = "D_HomeSeerFC200.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerFC200",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	     , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	     , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		 , 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		     , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
			SCObj.SetMultiTap(peer_dev_num, SCObj.DefaultScreen, 1, false)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = HomeSeer_SetMultiTap,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, 

	-- HomeSeer HS-WS200+ switch-specific
	[DEVTYPE_HOMESEERWS200] = {
		Id						= "HOMESEERWS200",
		Name                    = "HomeSeer HD-WS200+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 8,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = false,
		HasIndicator			= false,
		HasMultiTap				= 5,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 1,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "-NL1R",
		StatusLEDTimeouts 		= "12345S",
		StatusLEDColors         = "KRGBMYCWrgbmycw",
		EventCodes              = {
	    							Buttons={"1", "2", "3", "4", "5"},
								    Scenes ={"A", "B", "C", "D", "E", "F", "G", "H" }
							      },
		MaxLabelSize			= 1,  -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERWS200,
		DeviceXml               = "D_HomeSeerWS200.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerWS200",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	    , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	    , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		, 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		    , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
			SCObj.SetMultiTap(peer_dev_num, SCObj.DefaultScreen, 1, false)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = HomeSeer_SetMultiTap,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, -- DEVTYPE_HOMESEERWS200

    -- HomeSeer HS-WD200+ dimmer-specific
	[DEVTYPE_HOMESEERWD200] = {
		Id						= "HOMESEERWD200",
		Name                    = "HomeSeer HD-WD200+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 8,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = true,
		HasIndicator			= false,
		HasMultiTap				= 5,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 7,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "-NL7654321AR",
		StatusLEDTimeouts 		= "12345S",
		StatusLEDColors         = "KRGBMYCWrgbmycw",
		EventCodes              = {
	    							Buttons={"1", "2", "3", "4", "5"},
								    Scenes ={"A", "B", "C", "D", "E", "F", "G", "H" }
							      },
		MaxLabelSize			= 1, -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERWD200,
		DeviceXml               = "D_HomeSeerWD200.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerWD200",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	     , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	     , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		 , 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		     , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
			SCObj.SetMultiTap(peer_dev_num, SCObj.DefaultScreen, 1, false)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = HomeSeer_SetMultiTap,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, -- DEVTYPE_HOMESEERWD200

    -- HomeSeer HS-FC200+ fan controller-specific
	[DEVTYPE_HOMESEERFC200] = {
		Id						= "HOMESEERFC200",
		Name                    = "HomeSeer HD-FC200+",
		DefaultLcdVersion       = 1,
		HasScreen               = false,
		HasMultipleScreens      = false,
		screenWidth             = 1,  -- Dummy if not HasScreen
		rightJustifyScreenWidth = 1,
		HasPresetLanguages      = false,
		HasThermostatControl    = false,
	    NumButtons              = 2,
	    NumSceneEvents			= 8,
	    MaxScroll	        	= 2,
	    LastFixedSceneId        = 2,
		HasOffScenes            = true,
		HasIndicator			= false,
		HasMultiTap				= 5,
		HasLocalControl         = true,
	    HasDirectSceneControl	= false,
	    NumStatusLEDs			= 4,
		HasCooperConfiguration  = false,
		MaxDirectAssociations   = 1, -- Actually 5 but effectively 1 since only the central controller accepts central scene.
		DefaultScreen           = "P1",
		NumTemperatureScreens   = 0,
		StatusLEDActions		= "-NL4321AR",
		StatusLEDTimeouts 		= "12345S",
		StatusLEDColors         = "KRGBMYCWrgbmycw",
		EventCodes              = {
	    							Buttons={"1", "2", "3", "4", "5"},
								    Scenes ={"A", "B", "C", "D", "E", "F", "G", "H" }
							      },
		MaxLabelSize			= 1, -- Dummy
		HasBattery				= false,
		HasButtonModes			= false,
		SendsSceneActivation	= false,
		SendsBasicSet			= false,
		SendsSwitchMultilevel   = false,
		SendsCentralScene       = true,
		CentralSceneSlowRefresh = true,
		DevType                 = DEVTYPE_HOMESEERFC200,
		DeviceXml               = "D_HomeSeerFC200.xml",
		OldServiceId            = "urn:gengen_mcv-org:serviceId:HomeSeerFC200",
		OldParamPrefix          = "HS",

		GetDefaultModeString    = HomeSeer_GetDefaultModeString,
		ScreenList              = {P=1},
	    LargeFontWidths         = {}, -- Empty if not HasScreen
	    SmallFontWidths         = {}, -- Empty if not HasScreen
		SetTuningParameters = function(zwave_dev_num, peer_dev_num)
			VEntry()
			SetTuningParameter("InitStaggerSeconds"	     , 30            , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("MinReinitSeconds"	     , 60 * 60 * 24  , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("ReturnRouteDelay"		 , 500           , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UseWithSlaveController"  , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("Retries"       		     , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("AssociationDelay"        , 5000          , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("SceneControllerConfDelay", 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("UnassociateDelay"        , 0             , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("IndicatorDelay"          , 0 			 , zwave_dev_num, peer_dev_num, nil)
			SetTuningParameter("CentralSceneTimeout"     , 100           , zwave_dev_num, peer_dev_num, nil)
		end,

		DoDeviceConfiguration = function(peer_dev_num, node_id, zwave_dev_num)
			SCObj.SetMultiTap(peer_dev_num, SCObj.DefaultScreen, 1, false)
		end,

		SetDefaultSettings = function(peer_dev_num)
			luup.variable_set(SID_SCENECONTROLLER, "NightLightMode", "-", peer_dev_num)
			SCObj.SetNightLightMode(peer_dev_num, false)
		end,

		PhysicalButtonToIndicator = function(button, on)
			return 0
		end,

		SetLEDStatusValue = HomeSeer_SetLEDStatusValue,

		SetNightLightMode = HomeSeer_SetNightLightMode,

		SetMultiTap = HomeSeer_SetMultiTap,

		PhysicalButtonToGroupIds = function(button, mode)
			return button, nil, nil, nil
		end,

		SetButtonType = function(peer_dev_num, node_id, physicalButton, newType, force, oldType)
			-- TODO: Can we switch between momentary and toggle?
		end,

		ScreenPage = function(screen)
			-- stub if not HasMultipleScreens
			return 0
		end,

	    SetDeviceScreen = function(peer_dev_num, screen)
			-- No screen-type support
		end,

		SetLanguage = function(peer_dev_num, language)
			-- empty if not HasPresetLanguages
		end,

		SetBacklight = function(peer_dev_num, blackLightOn)
			-- empty if no backlight
		end,

		-- These modes are bogus. Same as the LCD1 for now.
		ModeMap = { M=0,  -- Momentary
		            D=0,  -- Direct	(Deprecated)
					T=1,  -- Toggle
					["2"]=0, -- Two-state
					["3"]=0, -- Three-state
					["4"]=0, -- Four-state
					["5"]=0, -- Five-state
					["6"]=0, -- Six-state
					["7"]=0, -- Seven-state
					["8"]=0, -- Eight-state
					["9"]=0, -- Nine-state
					S=1,  -- Direct Toggle (Deprecated)
					X=1,  -- Exclusive
					N=0,  -- Switch Screen
					H=3,  -- Temperature
					W=0,  -- Welcome
					P=0,  -- Thermostat Operating mode
					E=0,  -- Thermostat Energy mode
					L=0,  -- scroLL
		},

		-- Convert a mode object to a mode type.
		ModeType = function (mode)
			local result = SCObj.ModeMap[mode.prefix];
			if result == nil then
				result = 0
			end
			return result
		end,

	    EncodeVeraSceneNumber = HomeSeer_EncodeVeraSceneNumber,

		DecodeVeraSceneNumber = HomeSeer_DecodeVeraSceneNumber,

	}, -- DEVTYPE_HOMESEERFC200
-- DEVTYPE_HOMESEERFC200

} -- Devices
