-- Installer for GenGeneric Scene Controller version 2.11d
--
-- Copyright (C) 2020  Gustavo A Fernandez
-- Thanks to Ron Luna and Ctrlable for contributing for RFWC5 support
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You can access the terms of this license at
-- https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html

-- Includes installation files for
--   Evolve LCD1
--   Cooper RFWC5
--   Nexia One Touch NX1000
--	 HomeSeer HS-WS100+
--   HomeSeer HS-WD100+
--	 HomeSeer HS-WS200+
--   HomeSeer HS-WD200+
--   HomeSeer HS-FC200+

-- Also installs improved support for
--   Kichler 12387 undercabinet light controller
--   Aeotec Siren
--   Shlage BE469 lock
--   Evolve T100R Thermostat (UI5 Only)
--   Aeotec Siren
--   Aeotec Trisensor (Remove S0 security)

-- This installs zwave_products_user.xml for UI5 and modifies KitDevice.json for UI7.

-- VerboseLogging == 0: important logs and errors:                      ELog, log
-- VerboseLogging == 1: Includes Info  logs:                            ELog, log, ILog,
-- VerboseLogging == 2: Includes debug logs:                            ELog, log, ILog, DLog, DEntry
-- VerboseLogging == 3: Include extended ZWave Queue                    ELog, log, ILog, DLog, DEntry, ZWaveQueue
-- VerboseLogging == 4:	Includes verbose logs:                          ELog, log, ILog, DLog, DEntry,            VLog, VEntry
-- VerboseLogging == 5:	Includes verbose logs and extended ZWave Queue: ELog, log, ILog, DLog, DEntry, ZWaveQueue VLog, VEntry

VerboseLogging = 4

-- ZWIntDebugLevel controls the messages sent by Zwint to /tmp/log.Init-LuaUPnP (not to be confused with /tmp/log/cmh/LuaUPnP.log)
-- This is the sum of a set of bit flags
--   1 - Log Z-Wavie protocal to and from ZWInt through the serial interface
--   2 - Log HTTP messages to LuaUPnP
--   4 - Log other debug messages
-- Error and initialization messages are always logged.
ZWIntDebugLevel = 3

-- Update this each time we update the installer.
local GenGenInstaller_Version = 227

local bit = require 'bit'
local nixio = require "nixio"
local socket = require "socket"
require "L_GenGenSceneControllerShared"

local SID_HAG                 = "urn:micasaverde-com:serviceId:HomeAutomationGateway1"
local SID_ZWN       	      = "urn:micasaverde-com:serviceId:ZWaveNetwork1"
local SID_SCENECONTROLLER     = "urn:gengen_mcv-org:serviceId:SceneController1"
local SID_GENGENINSTALLER     = "urn:gengen_mcv-org:serviceId:SceneControllerInstaller1"

local DEVTYPE_GENGENINSTALLER = "urn:schemas-gengen_mcv-org:device:SceneControllerInstaller:1"
local DEVTYPE_HOMESEERWS100    = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS100:1"
local DEVTYPE_HOMESEERWD100    = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD100:1"
local DEVTYPE_HOMESEERWS200    = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWS200:1"
local DEVTYPE_HOMESEERWD200    = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerWD200:1"
local DEVTYPE_HOMESEERFC200    = "urn:schemas-gengen_mcv-org:device:SceneControllerHomeSeerFC200:1"

-- Make sure tha all of the log functions work before even the SCObj global is set.
function GetDeviceName()
  return "GenGeneric Scene Controller"
end

local reload_needed = false

-- Update file with the given content if the previous version does not exist or is different.
local function UpdateFileWithContent(filename, content, permissions, version, force, decompress)
	local update = false
	local backup = false
	local oldVersion
	if not content then
		ELog("Missing content for ", filename)
		return false
	end
	local stat = nixio.fs.stat(filename)
	local oldName = filename .. ".old"
	local backupName = filename .. ".save"
	local oldversion = luup.variable_get(SID_GENGENINSTALLER, filename .. "_version", lul_device) 
	if oldversion then
		oldversion = tonumber(oldversion)
	else
		oldversion = 0
	end
	if not version then
		version = 0
	end
	if stat then
		if version > oldversion or (version == oldversion and stat.size ~= #content and not decompress) or force then
			log("Backing up ", filename, " to ", backupName, " and replacing with new version.")
			VLog("Old ", filename, " size was ", stat.size, " bytes. new size is ", #content, " bytes.")
			nixio.fs.rename(backupName, oldName)
			local result, errno, errmsg =  nixio.fs.rename(filename, backupName)
			if result then
				update = true
				backup = true
			else
				ELog("could not rename ", filename, " to", backupName, ": ", errmsg)
			end
		else
			if oldversion > version then
				VLog("Not updating ", filename, " because the old version is ", oldversion, " and the new version is ", version)
			else
				VLog("Not updating ", filename, " because the new content is ", #content, " bytes and the old is ", stat.size, " bytes.")
			end
		end
	else
		VLog("updating ", filename, " because a previous version does not exist")
		update = true
	end
	if update then
		local f2 = filename
		if decompress then
			f2 = filename .. ".lzo"
		end
		local f, errno, errmsg = nixio.open(f2, "w", permissions)
		if f then
			local result, errno, errmsg, bytesWritten = f:write(content)
			if result then
				f:close()
				if decompress then
					os.execute("pluto-lzo d " .. f2 .. " " .. filename)
					nixio.fs.remove(f2)
				end
				if backup then
					nixio.fs.remove(oldName)
				end
				VLog("Wrote ", filename, " successfully (", #content, " bytes)")
				if version > 0 then
					luup.variable_set(SID_GENGENINSTALLER, filename .. "_version", tostring(version), lul_device)
				end
				reload_needed = filename
				return true
			else
				ELog("could not write ", #content, " bytes into ", f2, ". only ", bytesWritten, " bytes written: ", errmsg)
				f:close()
				if backup then
					nixio.fs.rename(backupName, filename)
					nixio.fs.rename(oldName, backupName)
				end
			end
		else
			ELog("could not open ", f2, " for writing: ", errmsg)
			if backup then
				nixio.fs.rename(backupName, filename)
				nixio.fs.rename(oldName, backupName)
			end
		end
	end
	return false
end

-- Prepares a file to be updated.
-- Returns nixio read and write file handles if update is needed.
-- Returns nil if no update required or error occurred. (base not modified)
-- <base.old - previous base.modified (for error handling)
-- <base>.save - previous base
-- <base>.modified - new file
-- <base> - Symlink to <base>.modified
local function PrepFileForUpdate(base, force)
	VEntry()
	local base_modified = base .. ".modified"
	local base_save = base .. ".save"
	local base_old = base .. ".old"
	local base_temp = base .. ".temp"
	local base_lstat, errno, errmsg = nixio.fs.lstat(base)
	local why = nil
	if not base_lstat then
		ELog("could not lstat ", base, ": ", errmsg)
		return nil
	end
	local base_save_stat = nixio.fs.stat(base_save)
	if base_save_stat and base_lstat.type == "lnk" then
		local base_linkname, errno, errmsg = nixio.fs.readlink(base)
		if not base_linkname then
			ELog("Could not readlink ", base": ", errmsg)
			return nil
		end
		if base_linkname ~= base_save then
			why = base .. " now symlinks to " .. base_linkname .. " instead of " .. base_save .. ". Perhaps the Vera software has been updated. Removing old " .. base_save
			nixio.fs.remove(base_save)
			base_save_stat = nil
		end
	end
	local base_modified_stat = nixio.fs.stat(base_modified)
	if not why then
		if force then
			why = "Installer has been updated"
		elseif not base_modified_stat then
			why = "the file has never been modified by the installer"
		end
	end
	if why then
		if base_modified_stat then
			nixio.fs.rename(base_modified, base_old)
		end
		if not base_save_stat then
			local result, errno, errmsg = nixio.fs.rename(base, base_save)
			if not result then
				ELog("could not rename ", base, " to", base_save, ": ", errmsg)
				if base_modified_stat then
					nixio.fs.rename(base_old, base_modified)
				end
				return nil
			end
		end
		local read_file, errmsg, errno = io.open(base_save, "r")
		if not read_file then
			ELog("could not open ", base_save, " for reading: ", errmsg)
			if not base_save_stat then
		    	nixio.fs.rename(base_save, base)
			end
			if base_modified_stat then
				nixio.fs.rename(base_old, base_modified)
			end
			return nil
		end
		local write_file, errmsg, errno = io.open(base_modified, "w", 644)
		if not write_file then
			ELog("could not open ", base_modified, " for writing: ", errmsg)
			read_file:close()
			if not base_save_stat then
		    	nixio.fs.rename(base_save, base)
			end
			if base_modified_stat then
				nixio.fs.rename(base_old, base_modified)
			end
			return nil
		end
		if base_save_stat then
			local result, errno, errmsg = nixio.fs.remove(base)
			if not result then
				ELog("could not delete old symlink ", base, ": ", errmsg)
				write_file:close()
				read_file:close()
				nixio.fs.remove(base_modified)
				if base_modified_stat then
					nixio.fs.rename(base_old, base_modified)
				end
				return nil
			end
		end
		local result, errno, errmsg = nixio.fs.symlink(base_modified, base)
		if not result then
			ELog("could not symlink ", base_modified, " to", base, ": ", errmsg)
			write_file:close()
			read_file:close()
			if not base_save_stat then
		    	nixio.fs.rename(base_save, base)
			end
			nixio.fs.remove(base_modified)
			if base_modified_stat then
				nixio.fs.rename(base_old, base_modified)
			end
			return nil
		end
		nixio.fs.remove(base_old)
		return read_file, write_file, why
	end
	return nil
end

local function updateJson(filename, update_func, updated)
	VEntry()
	local read_file, write_file, why = PrepFileForUpdate(filename, updated)
	if read_file then
		log("Updating ", filename, " because ", why)
		local str = read_file:read("*a")
		read_file:close()
		local obj=json.decode(str);
		update_func(obj)
		local state = { indent = true }
		local str2 = json.encode (obj, state)
		write_file:write(str2)
		write_file:close()
		reload_needed = filename
	else
		VLog("Not updating ", filename)
	end
end

local ScannedDeviceList = {}

local function ScanForNewDevices()
	DEntry()

	local function AdoptEvolveLCD1(device_num)
		DEntry()

		local device_file = luup.attr_get("device_file");
		local impl_file = luup.attr_get("impl_file");
		local device = luup.devices[device_num];
		if device.device_type ~= 'urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD:1' and
		   device_file == "D_EvolveLCD1.xml" and
		   impl_file == "I_GenGenSceneController.xml" then
		   return false
		end

		luup.attr_set("device_type", "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD:1", device_num)
		luup.attr_set("device_file", "D_EvolveLCD1.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "Evolve Guest Controls", device_num)
		luup.attr_set("device_json", "D_EvolveLCD1.json", device_num)
		luup.attr_set("category_num", "14", device_num)
		luup.attr_set("subcategory_num", "0", device_num)
		luup.attr_set("model", "EVLCD1", device_num)
		luup.attr_set("invisible", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "5", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "20-Display Timeout (seconds),m,,"..
		                                                                                "21-Backlight ON level (1-20),m,,"..
		                                                                                "22-Backlight OFF level (0-20),m,,"..
		                                                                                "23-Button ON level (1-20),m,,"..
		                                                                                "24-Button OFF level (0-20),m,,"..
		                                                                                "25-LCD Contrast (5-20),m,,"..
		                                                                                "26-Orientation(1=rotate 180 0=normal),m,,"..
		                                                                                "27-Network Update (seconds),m,,"..
		                                                                                "29-backlight level (0-100),m,,"..
		                                                                                "32-Backlight Demo mode (0-1),m,", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "http://code.mios.com/trac/mios_evolve-lcd1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "http://code.mios.com/trac/mios_evolve-lcd1", device_num)

		return true
	end

	local function AdoptCooperRFWC5(device_num)
		DEntry()

		local device_file = luup.attr_get("device_file", device_num);
		local impl_file = luup.attr_get("impl_file", device_num);
		local device = luup.devices[device_num];
		if device.device_type == 'urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5:1' and
		   device_file == "D_CooperRFWC5.xml" and
		   impl_file == "I_GenGenSceneController.xml" then
		   return false
		end

		luup.attr_set("device_type", "urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5:1", device_num)
		luup.attr_set("device_file", "D_CooperRFWC5.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "Cooper Industries", device_num)
		luup.attr_set("device_json", "D_CooperRFWC5.json", device_num)
		luup.attr_set("category_num", "14", device_num)
		luup.attr_set("subcategory_num", "0", device_num)
		luup.attr_set("model", "RFWC5", device_num)
		luup.attr_set("invisible", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "5", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)

		return true
	end

	local function AdoptNexiaOneTouch(device_num)
		DEntry()

		local device_file = luup.attr_get("device_file", device_num);
		local impl_file = luup.attr_get("impl_file", device_num);
		local device = luup.devices[device_num];
		if device.device_type == 'urn:schemas-gengen_mcv-org:device:SceneControllerNexiaOneTouch:1' and
		   device_file == "D_NexiaOneTouch.xml" and
		   impl_file == "I_GenGenSceneController.xml" then
		   return false
		end

		luup.attr_set("device_type", "urn:schemas-gengen_mcv-org:device:SceneControllerNexiaOneTouch:1", device_num)
		luup.attr_set("device_file", "D_NexiaOneTouch.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "Ingersoll Rand", device_num)
		luup.attr_set("device_json", "D_NexiaOneTouch.json", device_num)
		luup.attr_set("category_num", "14", device_num)
		luup.attr_set("subcategory_num", "0", device_num)
		luup.attr_set("model", "NX1000", device_num)
		luup.attr_set("invisible", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "5", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "20-Touch Calibration (1-10),m,,"..
		                                                                                "21-Screen Contrast (1-10),m,,"..
		                                                                                "23-Button LED Level (1-10),m,,"..
		                                                                                "24-Backlight Level (1-10),m,,"..
		                                                                                "25-Scene Button Press Backlight Timeout (10-15),m,,"..
		                                                                                "26-Page Button Press Backlight Timeout (5-15),m,,"..
		                                                                                "28-Screen Timeout (1-240),m,,"..
		                                                                                "29-Screen Timeout Primary Page (0-3),m,,"..
		                                                                                "30-Battery Stat Shutdown Threshold % (0-20),m,,"..
		                                                                                "31-Battery Radio Cutoff Threshold % (0-40),m,,"..
																					    "32-Battery LOWBATT Indicator Threshold % (5-50),m,,"..
																					    "33-Battery Threshold Value for Midlevel % (30-80),m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "http://products.z-wavealliance.org/products/1344", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "http://products.z-wavealliance.org/products/1344", device_num)

		return true
	end

	-- Create peer devices for switches and dimmers which have well known device types already
	-- since Vera will reset our implementation back to the standard one and we will not be able
	-- to do this in CreatePeerDevices
	local function PossiblyCreatePeer(device_num, Peer_dev_type, Peer_Dev_xml)
		DEntry()
		local peerID = luup.variable_get(SID_SCENECONTROLLER,"PeerID",device_num);
		if peerID then
			if peerID == "Pending" then
				for peer_dev_num, peer_device in pairs(luup.devices) do
					local parent_id = luup.variable_get(SID_SCENECONTROLLER,"PeerID",peer_dev_num);
					if parent_id and tonumber(parent_id) == device_num then
						return false
					end
				end
			else
				peerID = tonumber(peerID)
				if luup.devices[peerID] then
					return false
				end
			end
		end
		luup.variable_set(SID_SCENECONTROLLER, "PeerID", "Pending", device_num)
		luup.variable_set(SID_SCENECONTROLLER, "ShowSingleDevice","0",device_num)
	    luup.call_action(SID_HAG,"CreateDevice", {
		   	deviceType = Peer_dev_type;
		   	internalID = device_num;
			Description = luup.devices[device_num].description .. " Controller";
		   	UpnpDevFilename = Peer_Dev_xml;
		   	UpnpImplFilename =  "I_GenGenSceneController.xml";
			RoomNum =  luup.devices[device_num].room_num;
			StateVariables = SID_SCENECONTROLLER .. ",PeerID=" .. device_num .. "\n" ..
			                 SID_SCENECONTROLLER .. ",ShowSingleDevice=0" },0)
	    return true
	end

	local function AdoptHomeSeerWS100(device_num)
		DEntry()
		if not PossiblyCreatePeer(device_num, DEVTYPE_HOMESEERWS100, "D_HomeSeerWS100.xml") then
			return false
		end
		luup.attr_set("manufacturer", "HomeSeer", device_num)
		luup.attr_set("model", "HS-WS100+", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "2", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "4-Load orientation (0=normal 1=inverted),m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "http://www.homeseer.com/guides/HS-WS100-Manual-v1_1a.pdf", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "http://www.homeseer.com/guides/HS-WS100-Manual-v1_1a.pdf", device_num)
		return true
	end

	local function AdoptHomeSeerWD100(device_num)
		DEntry()
		if not PossiblyCreatePeer(device_num, DEVTYPE_HOMESEERWD100, "D_HomeSeerWD100.xml") then
			return false
		end
		luup.attr_set("device_file", "D_HomeSeerWD100.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "HomeSeer", device_num)
		luup.attr_set("model", "HS-WD100+", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "2", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "4-Load orientation (0=normal 1=inverted),m,,"..
																						"7-Remote Ramp Step resolution (1-99),m,,"..
																						"8-Remote Ramp Step speed x10ms,m,,"..
																						"9-Local Ramp Step resolution (1-99),m,,"..
																						"10-Local Ramp Step speed x10ms,m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2018/06/HS-WD100-Manual-6a.pdf", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2018/06/HS-WD100-Manual-6a.pdf", device_num)
		return true
	end

	local function AdoptHomeSeerWS200(device_num)
		DEntry()
		if not PossiblyCreatePeer(device_num, DEVTYPE_HOMESEERWS200, "D_HomeSeerWS200.xml") then
			return false
		end
		luup.attr_set("manufacturer", "HomeSeer", device_num)
		luup.attr_set("model", "HS-WS200+", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "2", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "4-Load orientation (0=normal 1=inverted),m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "https://HomeSeer.com/wp-content/uploads/2018/02/HS-WS200-Manual-v4.pdf", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "https://HomeSeer.com/wp-content/uploads/2018/02/HS-WS200-Manual-v4.pdf", device_num)
		return true
	end

	local function AdoptHomeSeerWD200(device_num)
		DEntry()
		if not PossiblyCreatePeer(device_num, DEVTYPE_HOMESEERWD200, "D_HomeSeerWD200.xml") then
			return false
		end
		luup.attr_set("device_file", "D_HomeSeerWD200.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "HomeSeer", device_num)
		luup.attr_set("model", "HS-WD200+", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "2", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "4-Load orientation (0=normal 1=inverted),m,,"..
																						"5-Lowest dim level (1=6.5% 14=20%),m,,"..
		                                                                                "11-Remote dimmer ramp rate in seconds (0=instant),m,,"..
		                                                                                "12-Local dimmer ramp rate in seconds (0=instant),m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2018/02/HS-WD200-Manual-4.pdf", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2018/02/HS-WD200-Manual-4.pdf", device_num)
		return true
	end

	local function AdoptHomeSeerFC200(device_num)
		DEntry()
		if not PossiblyCreatePeer(device_num, DEVTYPE_HOMESEERFC200, "D_HomeSeerFC200.xml") then
			return false
		end
		luup.attr_set("device_file", "D_HomeSeerFC200.xml", device_num)
		luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num)
		luup.attr_set("manufacturer", "HomeSeer", device_num)
		luup.attr_set("model", "HS-FC200+", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "NumButtons", "2", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "FiresOffEvents", "1", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:SceneController1", "ActivationMethod", "0", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "VariablesSet", "4-Load orientation (0=normal 1=inverted),m,,"..
																						"5-Fan type (0=3 Speed 1=4 speed),m,",device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:HaDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2019/01/HS-FC200-Manual-3a.pdf", device_num)
		luup.variable_set("urn:micasaverde-com:serviceId:ZWaveDevice1", "Documentation", "https://homeseer.com/wp-content/uploads/2019/01/HS-FC200-Manual-3a.pdf", device_num)
		return true
	end


    -- This is a hack for the Aeotec Siren which includes "COMMAND_CLASS_SECURITY" in its node info but never actually responds to a Security Nonce Get command
	local function ApplyAeotecSirenHack()
		DEntry()

		local function AeotecSirenCallback(dev_num, result)
			DLog("Aeotec Siren node info intercept: peer_dev_num num=", dev_num," result=",result);
		end

		MonitorZWaveData(false, -- incoming
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
                                                         C1
42      12/02/17 21:44:16.481   0x1 0x18 0x0 0x49 0x84 0x27 0x12 0x4 0x10 0x5 0x5e 0x25 0x70 0x85 0x59 0x72 0x2b 0x2c 0x86 0x7a 0x73 0x98 0xef 0x5a 0x82 0x8c 
           SOF - Start Of Frame --+    ¦   ¦    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
                    length = 24 -------+   ¦    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
                        Request -----------+    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
  FUNC_ID_ZW_APPLICATION_UPDATE ----------------+    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Update state = Node Info received -------------------+    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Node ID: 39 Unknown node ID: 39 --------------------------+    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
          Node info length = 18 -------------------------------+   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
     Basic type = Routing slave -----------------------------------+    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
   Generic type = switch binary ----------------------------------------+   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
          Specific type = Siren --------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ---------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[2] = COMMAND_CLASS_SWITCH_BINARY ---------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[3] = COMMAND_CLASS_CONFIGURATION --------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[4] = COMMAND_CLASS_ASSOCIATION ---------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[5] = COMMAND_CLASS_ASSOCIATION_GRP_INFO -----------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[6] = COMMAND_CLASS_MANUFACTURER_SPECIFIC ---------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[7] = COMMAND_CLASS_SCENE_ACTIVATION -------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[8] = COMMAND_CLASS_SCENE_ACTUATOR_CONF ---------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[9] = COMMAND_CLASS_VERSION --------------------------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[10] = COMMAND_CLASS_FIRMWARE_UPDATE_MD -------------------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[11] = COMMAND_CLASS_POWERLEVEL --------------------------------------------------------------------------+    ¦    ¦    ¦    ¦    ¦
Can receive command class[12] = COMMAND_CLASS_SECURITY ---------------------------------------------------------------------------------+    ¦    ¦    ¦    ¦
                         Marker -------------------------------------------------------------------------------------------------------------+    ¦    ¦    ¦
Can send command class[1] = COMMAND_CLASS_DEVICE_RESET_LOCALLY -----------------------------------------------------------------------------------+    ¦    ¦
Can send command class[2] = COMMAND_CLASS_HAIL --------------------------------------------------------------------------------------------------------+    ¦
                    Checksum OK ----------------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "^01 18 00 49 84 (..) 12 04 10 05 5e 25 70 85 59 72 2b 2c 86 7a 73 98 ef 5a 82 ..", -- Main RegEx
--[==[
42      12/02/17 21:44:16.481   0x1 0x17 0x0 0x49 0x84 0x27 0x11 0x4 0x10 0x5 0x5e 0x25 0x70 0x85 0x59 0x72 0x2b 0x2c 0x86 0x7a 0x73 0xef 0x5a 0x82 0x8c
           SOF - Start Of Frame --+    ¦   ¦    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
                    length = 23 -------+   ¦    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
                        Request -----------+    ¦    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
  FUNC_ID_ZW_APPLICATION_UPDATE ----------------+    ¦    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Update state = Node Info received -------------------+    ¦    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Node ID: 39 Unknown node ID: 39 --------------------------+    ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
          Node info length = 17 -------------------------------+   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
     Basic type = Routing slave -----------------------------------+    ¦   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
   Generic type = switch binary ----------------------------------------+   ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
          Specific type = Siren --------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ---------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[2] = COMMAND_CLASS_SWITCH_BINARY ---------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[3] = COMMAND_CLASS_CONFIGURATION --------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[4] = COMMAND_CLASS_ASSOCIATION ---------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[5] = COMMAND_CLASS_ASSOCIATION_GRP_INFO -----------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[6] = COMMAND_CLASS_MANUFACTURER_SPECIFIC ---------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[7] = COMMAND_CLASS_SCENE_ACTIVATION -------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[8] = COMMAND_CLASS_SCENE_ACTUATOR_CONF ---------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[9] = COMMAND_CLASS_VERSION --------------------------------------------------------------------+    ¦    ¦    ¦    ¦    ¦    ¦
Can receive command class[10] = COMMAND_CLASS_FIRMWARE_UPDATE_MD -------------------------------------------------------------+    ¦    ¦    ¦    ¦    ¦
Can receive command class[11] = COMMAND_CLASS_POWERLEVEL --------------------------------------------------------------------------+    ¦    ¦    ¦    ¦
                         Marker --------------------------------------------------------------------------------------------------------+    ¦    ¦    ¦
Can send command class[1] = COMMAND_CLASS_DEVICE_RESET_LOCALLY ------------------------------------------------------------------------------+    ¦    ¦
Can send command class[2] = COMMAND_CLASS_HAIL ---------------------------------------------------------------------------------------------------+    ¦
                    Checksum OK -----------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "01 17 00 49 84 \\1 11 04 10 05 5e 25 70 85 59 72 2b 2c 86 7a 73 ef 5a 82 XX", -- Response
		                 AeotecSirenCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "AeotecSirenNodeInfo", -- label
						 true) -- forward
	end


    -- This is a hack for the Aeotec dry contact sensor which includes "COMMAND_CLASS_SECURITY" in its node info but never actually responds to a Security Nonce Get command
	local function ApplyAeotecDryContactSensorHack()
		DEntry()

		local function AeotecDryContactSensorCallback(dev_num, result)
			DLog("Aeotec Dry Contact Sensor node info intercept: peer_dev_num num=", dev_num," result=",result);
		end

		MonitorZWaveData(false, -- incoming
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
                                                         C1
42      03/10/18 9:02:07.757    0x1 0x19 0x0 0x49 0x84 0x2f 0x13 0x4 0x7 0x1 0x5e 0x86 0x72 0x85 0x59 0x73 0x71 0x84 0x30 0x80 0x70 0x98 0x7a 0xef 0x5a 0x82 0xf0 (###I#/####^#r#Ysq#0#p#z#Z##) 
           SOF - Start Of Frame --+    |   |    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                    length = 25 -------+   |    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                        Request -----------+    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
  FUNC_ID_ZW_APPLICATION_UPDATE ----------------+    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Update state = Node Info received -------------------+    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Node ID: 47 Device 1047=Sewage Ejector Alarm -------------+    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
          Node info length = 19 -------------------------------+   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
     Basic type = Routing slave -----------------------------------+   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Generic type = Sensor Notification ------------------------------------+   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Specific type = Notification sensor ---------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO --------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[2] = COMMAND_CLASS_VERSION --------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[3] = COMMAND_CLASS_MANUFACTURER_SPECIFIC -----------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[4] = COMMAND_CLASS_ASSOCIATION --------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[5] = COMMAND_CLASS_ASSOCIATION_GRP_INFO ----------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL -------------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[7] = COMMAND_CLASS_ALARM -----------------------------------------------------------+    |    |    |    |    |    |    |    |    |    |
Can receive command class[8] = COMMAND_CLASS_WAKE_UP --------------------------------------------------------------+    |    |    |    |    |    |    |    |    |
Can receive command class[9] = COMMAND_CLASS_SENSOR_BINARY -------------------------------------------------------------+    |    |    |    |    |    |    |    |
Can receive command class[10] = COMMAND_CLASS_BATTERY -----------------------------------------------------------------------+    |    |    |    |    |    |    |
Can receive command class[11] = COMMAND_CLASS_CONFIGURATION ----------------------------------------------------------------------+    |    |    |    |    |    |
Can receive command class[12] = COMMAND_CLASS_SECURITY --------------------------------------------------------------------------------+    |    |    |    |    |
Can receive command class[13] = COMMAND_CLASS_FIRMWARE_UPDATE_MD ---------------------------------------------------------------------------+    |    |    |    |
                         Marker -----------------------------------------------------------------------------------------------------------------+    |    |    |
Can send command class[1] = COMMAND_CLASS_DEVICE_RESET_LOCALLY ---------------------------------------------------------------------------------------+    |    |
Can send command class[2] = COMMAND_CLASS_HAIL ------------------------------------------------------------------------------------------------------------+    |
                    Checksum OK --------------------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "^01 19 00 49 84 (..) 13 04 07 01 5e 86 72 85 59 73 71 84 30 80 70 98 7a ef 5a 82 ..", -- Main RegEx
--[==[
42      03/10/18 9:02:07.757    0x1 0x19 0x0 0x49 0x84 0x2f 0x12 0x4 0x7 0x1 0x5e 0x86 0x72 0x85 0x59 0x73 0x71 0x84 0x30 0x80 0x70 0x7a 0xef 0x5a 0x82 0xf0 (###I#/####^#r#Ysq#0#p#z#Z##) 
           SOF - Start Of Frame --+    |   |    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                    length = 25 -------+   |    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                        Request -----------+    |    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
  FUNC_ID_ZW_APPLICATION_UPDATE ----------------+    |    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Update state = Node Info received -------------------+    |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Node ID: 47 Device 1047=Sewage Ejector Alarm -------------+    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
          Node info length = 18 -------------------------------+   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
     Basic type = Routing slave -----------------------------------+   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Generic type = Sensor Notification ------------------------------------+   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Specific type = Notification sensor ---------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO --------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[2] = COMMAND_CLASS_VERSION --------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[3] = COMMAND_CLASS_MANUFACTURER_SPECIFIC -----------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[4] = COMMAND_CLASS_ASSOCIATION --------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[5] = COMMAND_CLASS_ASSOCIATION_GRP_INFO ----------------------------------+    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL -------------------------------------------------+    |    |    |    |    |    |    |    |    |    |
Can receive command class[7] = COMMAND_CLASS_ALARM -----------------------------------------------------------+    |    |    |    |    |    |    |    |    |
Can receive command class[8] = COMMAND_CLASS_WAKE_UP --------------------------------------------------------------+    |    |    |    |    |    |    |    |
Can receive command class[9] = COMMAND_CLASS_SENSOR_BINARY -------------------------------------------------------------+    |    |    |    |    |    |    |
Can receive command class[10] = COMMAND_CLASS_BATTERY -----------------------------------------------------------------------+    |    |    |    |    |    |
Can receive command class[11] = COMMAND_CLASS_CONFIGURATION ----------------------------------------------------------------------+    |    |    |    |    |
Can receive command class[13] = COMMAND_CLASS_FIRMWARE_UPDATE_MD ----------------------------------------------------------------------+    |    |    |	   |
                         Marker ------------------------------------------------------------------------------------------------------------+    |    |	   |
Can send command class[1] = COMMAND_CLASS_DEVICE_RESET_LOCALLY ----------------------------------------------------------------------------------+    |	   |
Can send command class[2] = COMMAND_CLASS_HAIL -------------------------------------------------------------------------------------------------------+	   |
                    Checksum OK ---------------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "01 18 00 49 84 \\1 12 04 07 01 5e 86 72 85 59 73 71 84 30 80 70 7a ef 5a 82 XX", -- Response
		                 AeotecDryContactSensorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "AeotecDryContactSensorNodeInfo", -- label
						 true) -- forward
	end

    -- This is a hack for the Aeotec Trisensor which includes "COMMAND_CLASS_SECURITY" but cannot associate directly with non-secure devices
	local function ApplyAeotecTrisensorHack()
		DEntry()

		local function AeotecTrisensorCallback(dev_num, result)
			DLog("Aeotec Trisensor node found intercept: peer_dev_num num=", dev_num," result=",result);
		end

		MonitorZWaveData(false, -- incoming
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
                                                   C1      C2
42      02/03/20 19:58:11.701   0x1 0x1d 0x0 0x4a 0x7 0x3 0x6 0x16 0x4 0x7 0x1 0x5e 0x98 0x9f 0x55 0x86 0x73 0x85 0x8e 0x59 0x72 0x5a 0x80 0x84 0x30 0x71 0x31 0x70 0x6c 0x7a 0x2f
           SOF - Start Of Frame --+    |   |    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                    length = 29 -------+   |    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                        Request -----------+    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
 FUNC_ID_ZW_ADD_NODE_TO_NETWORK ----------------+   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                Callback ID = 7 --------------------+   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Result = ADD_NODE_STATUS_ADDING_SLAVE ------------------+   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
  Node ID: 6 Unknown node ID: 6 ----------------------------+    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                    Length = 22 ---------------------------------+   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
     Basic type = Routing slave -------------------------------------+   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Generic type = Sensor Notification --------------------------------------+   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Specific type = Notification sensor -----------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ----------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[2] = COMMAND_CLASS_SECURITY ---------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[3] = COMMAND_CLASS_SECURITY_2 ------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[4] = COMMAND_CLASS_TRANSPORT_SERVICE ----------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[5] = COMMAND_CLASS_VERSION -------------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL ---------------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[7] = COMMAND_CLASS_ASSOCIATION -------------------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[8] = COMMAND_CLASS_MULTI_INSTANCE_ASSOCIATION ---------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[9] = COMMAND_CLASS_ASSOCIATION_GRP_INFO --------------------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[10] = COMMAND_CLASS_MANUFACTURER_SPECIFIC -----------------------------------------------------------+    |    |    |    |    |    |    |    |    |    |
Can receive command class[11] = COMMAND_CLASS_DEVICE_RESET_LOCALLY -----------------------------------------------------------------+    |    |    |    |    |    |    |    |    |
Can receive command class[12] = COMMAND_CLASS_BATTERY -----------------------------------------------------------------------------------+    |    |    |    |    |    |    |    |
Can receive command class[13] = COMMAND_CLASS_WAKE_UP ----------------------------------------------------------------------------------------+    |    |    |    |    |    |    |
Can receive command class[14] = COMMAND_CLASS_SENSOR_BINARY ---------------------------------------------------------------------------------------+    |    |    |    |    |    |
Can receive command class[15] = COMMAND_CLASS_ALARM ----------------------------------------------------------------------------------------------------+    |    |    |    |    |
Can receive command class[16] = COMMAND_CLASS_SENSOR_MULTILEVEL ---------------------------------------------------------------------------------------------+    |    |    |    |
Can receive command class[17] = COMMAND_CLASS_CONFIGURATION ------------------------------------------------------------------------------------------------------+    |    |    |
Can receive command class[18] = COMMAND_CLASS_SUPERVISION -------------------------------------------------------------------------------------------------------------+    |    |
Can receive command class[19] = COMMAND_CLASS_FIRMWARE_UPDATE_MD -----------------------------------------------------------------------------------------------------------+    |
                    Checksum OK -------------------------------------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "^01 1d 00 4a (..) 03 (..) 16 04 07 01 5e 98 9f 55 86 73 85 8e 59 72 5a 80 84 30 71 31 70 6c 7a ..", -- Main RegEx
--[==[
42      02/03/20 19:58:11.701   0x1 0x1b 0x0 0x4a 0x7 0x3 0x6 0x14 0x4 0x7 0x1 0x5e 0x55 0x86 0x73 0x85 0x8e 0x59 0x72 0x5a 0x80 0x84 0x30 0x71 0x31 0x70 0x6c 0x7a 0x2f 
           SOF - Start Of Frame --+    |   |    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
                    length = 27 -------+   |    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
                        Request -----------+    |   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
 FUNC_ID_ZW_ADD_NODE_TO_NETWORK ----------------+   |   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
                Callback ID = 7 --------------------+   |   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
Result = ADD_NODE_STATUS_ADDING_SLAVE ------------------+   |    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
  Node ID: 6 Unknown node ID: 6 ----------------------------+    |   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
                    Length = 20 ---------------------------------+   |   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
     Basic type = Routing slave -------------------------------------+   |   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
Generic type = Sensor Notification --------------------------------------+   |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |  
Specific type = Notification sensor -----------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ----------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    | 
Can receive command class[4] = COMMAND_CLASS_TRANSPORT_SERVICE ------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[5] = COMMAND_CLASS_VERSION ---------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL -----------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[7] = COMMAND_CLASS_ASSOCIATION ---------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[8] = COMMAND_CLASS_MULTI_INSTANCE_ASSOCIATION -----------------------------------+    |    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[9] = COMMAND_CLASS_ASSOCIATION_GRP_INFO ----------------------------------------------+    |    |    |    |    |    |    |    |    |    |    |
Can receive command class[10] = COMMAND_CLASS_MANUFACTURER_SPECIFIC -------------------------------------------------+    |    |    |    |    |    |    |    |    |    |
Can receive command class[11] = COMMAND_CLASS_DEVICE_RESET_LOCALLY -------------------------------------------------------+    |    |    |    |    |    |    |    |    |
Can receive command class[12] = COMMAND_CLASS_BATTERY -------------------------------------------------------------------------+    |    |    |    |    |    |    |    |
Can receive command class[13] = COMMAND_CLASS_WAKE_UP ------------------------------------------------------------------------------+    |    |    |    |    |    |    |
Can receive command class[14] = COMMAND_CLASS_SENSOR_BINARY -----------------------------------------------------------------------------+    |    |    |    |    |    |
Can receive command class[15] = COMMAND_CLASS_ALARM ------------------------------------------------------------------------------------------+    |    |    |    |    |
Can receive command class[16] = COMMAND_CLASS_SENSOR_MULTILEVEL -----------------------------------------------------------------------------------+    |    |    |    |
Can receive command class[17] = COMMAND_CLASS_CONFIGURATION --------------------------------------------------------------------------------------------+    |    |    |
Can receive command class[18] = COMMAND_CLASS_SUPERVISION ---------------------------------------------------------------------------------------------------+    |    |
Can receive command class[19] = COMMAND_CLASS_FIRMWARE_UPDATE_MD -------------------------------------------------------------------------------------------------+    |
                    Checksum OK ---------------------------------------------------------------------------------------------------------------------------------------+
--]==]
		                 "01 1B 00 4a \\1 03 \\2 14 04 07 01 5e 55 86 73 85 8e 59 72 5a 80 84 30 71 31 70 6c 7a XX", -- Response
		                 AeotecTrisensorCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "AeotecTrisensornNodeFound", -- label
						 true) -- forward


		local function AeotecTrisensorCallback2(dev_num, result)
			DLog("Aeotec Trisensor node info intercept: peer_dev_num num=", dev_num," result=",result);
		end

		MonitorZWaveData(false, -- incoming
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
                                                        C1   
42      02/03/20 21:50:30.326   0x1 0x1c 0x0 0x49 0x84 0xb 0x16 0x4 0x7 0x1 0x5e 0x98 0x9f 0x55 0x86 0x73 0x85 0x8e 0x59 0x72 0x5a 0x80 0x84 0x30 0x71 0x31 0x70 0x6c 0x7a 0xa0 (###I######^##U#s##YrZ##0q1plz#) 
           SOF - Start Of Frame ──┘    │   │    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
                    length = 28 ───────┘   │    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
                        Request ───────────┘    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
  FUNC_ID_ZW_APPLICATION_UPDATE ────────────────┘    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Update state = Node Info received ───────────────────┘   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Node ID: 11 Unknown node ID: 11 ─────────────────────────┘    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
          Node info length = 22 ──────────────────────────────┘   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
     Basic type = Routing slave ──────────────────────────────────┘   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Generic type = Sensor Notification ───────────────────────────────────┘   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Specific type = Notification sensor ──────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ───────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[2] = COMMAND_CLASS_SECURITY ──────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[3] = COMMAND_CLASS_SECURITY_2 ─────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[4] = COMMAND_CLASS_TRANSPORT_SERVICE ───────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[5] = COMMAND_CLASS_VERSION ──────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL ────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[7] = COMMAND_CLASS_ASSOCIATION ────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[8] = COMMAND_CLASS_MULTI_INSTANCE_ASSOCIATION ──────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[9] = COMMAND_CLASS_ASSOCIATION_GRP_INFO ─────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[10] = COMMAND_CLASS_MANUFACTURER_SPECIFIC ────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │
Can receive command class[11] = COMMAND_CLASS_DEVICE_RESET_LOCALLY ──────────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │
Can receive command class[12] = COMMAND_CLASS_BATTERY ────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │
Can receive command class[13] = COMMAND_CLASS_WAKE_UP ─────────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │
Can receive command class[14] = COMMAND_CLASS_SENSOR_BINARY ────────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │
Can receive command class[15] = COMMAND_CLASS_ALARM ─────────────────────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │
Can receive command class[16] = COMMAND_CLASS_SENSOR_MULTILEVEL ──────────────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │
Can receive command class[17] = COMMAND_CLASS_CONFIGURATION ───────────────────────────────────────────────────────────────────────────────────────────────────┘    │    │    │
Can receive command class[18] = COMMAND_CLASS_SUPERVISION ──────────────────────────────────────────────────────────────────────────────────────────────────────────┘    │    │
Can receive command class[19] = COMMAND_CLASS_FIRMWARE_UPDATE_MD ────────────────────────────────────────────────────────────────────────────────────────────────────────┘    │
                    Checksum OK ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
--]==]
		                 "^01 1C 00 49 84 (..) 16 04 07 01 5e 98 9f 55 86 73 85 8e 59 72 5a 80 84 30 71 31 70 6c 7a ..", -- Main RegEx
--[==[
42      02/03/20 21:50:30.326   0x1 0x1A 0x0 0x49 0x84 0xb 0x14 0x4 0x7 0x1 0x5e 0x55 0x86 0x73 0x85 0x8e 0x59 0x72 0x5a 0x80 0x84 0x30 0x71 0x31 0x70 0x6c 0x7a 0xa0
           SOF - Start Of Frame ──┘    │   │    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │ 
                    length = 26 ───────┘   │    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │  
                        Request ───────────┘    │    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │  
  FUNC_ID_ZW_APPLICATION_UPDATE ────────────────┘    │   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
Update state = Node Info received ───────────────────┘   │    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │  
Node ID: 11 Unknown node ID: 11 ─────────────────────────┘    │   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
          Node info length = 20 ──────────────────────────────┘   │   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
     Basic type = Routing slave ──────────────────────────────────┘   │   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
Generic type = Sensor Notification ───────────────────────────────────┘   │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
Specific type = Notification sensor ──────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
Can receive command class[1] = COMMAND_CLASS_ZWAVEPLUS_INFO ───────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │   
Can receive command class[4] = COMMAND_CLASS_TRANSPORT_SERVICE ─────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[5] = COMMAND_CLASS_VERSION ────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[6] = COMMAND_CLASS_POWERLEVEL ──────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[7] = COMMAND_CLASS_ASSOCIATION ──────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[8] = COMMAND_CLASS_MULTI_INSTANCE_ASSOCIATION ─────────────────────-──────────┘    │    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[9] = COMMAND_CLASS_ASSOCIATION_GRP_INFO ───────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │    │
Can receive command class[10] = COMMAND_CLASS_MANUFACTURER_SPECIFIC ──────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │    │
Can receive command class[11] = COMMAND_CLASS_DEVICE_RESET_LOCALLY ────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │    │
Can receive command class[12] = COMMAND_CLASS_BATTERY ──────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │    │
Can receive command class[13] = COMMAND_CLASS_WAKE_UP ───────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │    │
Can receive command class[14] = COMMAND_CLASS_SENSOR_BINARY ──────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │    │
Can receive command class[15] = COMMAND_CLASS_ALARM ───────────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │    │
Can receive command class[16] = COMMAND_CLASS_SENSOR_MULTILEVEL ────────────────────────────────────────────────────────────────────────────────┘    │    │    │    │
Can receive command class[17] = COMMAND_CLASS_CONFIGURATION ─────────────────────────────────────────────────────────────────────────────────────────┘    │    │    │
Can receive command class[18] = COMMAND_CLASS_SUPERVISION ────────────────────────────────────────────────────────────────────────────────────────────────┘    │    │
Can receive command class[19] = COMMAND_CLASS_FIRMWARE_UPDATE_MD ──────────────────────────────────────────────────────────────────────────────────-───────────┘    │
                    Checksum OK ────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
--]==]
		                 "01 1A 00 49 84 \\1 14 04 07 01 5e 55 86 73 85 8e 59 72 5a 80 84 30 71 31 70 6c 7a XX", -- Response
		                 AeotecTrisensorCallback2,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "AeotecTrisensornNodeFound2", -- label
						 true) -- forward

	end


	-- This is a hack for UI7 1.7.2608 getting confused by inconsistent node info reports from the Kichler 12387 undercabinet light controller.
	-- Vera also does not like devices that don't support COMMAND_CLASS_VERSION so we add it to the command class list and intercept the expected
	-- version and command class version queries.
	local function ApplyKichler12387Hack(device_num, node_id)
		DEntry()

		local function Kichler12387NodeInfoCallback(peer_dev_num, result)
			DLog("Kichler 12387 node info intercept: device num=".. device_num.." node_id="..node_id.. "result=".. tableToString(result));
		end

		local function Kichler12387VersionCallback(peer_dev_num, result)
			DLog("Kichler 12387 Version intercept: device num=".. device_num.." node_id="..node_id.. "result=".. tableToString(result));
		end

		local function Kichler12387CommandClassVersionCallback(peer_dev_num, result)
			DLog("Kichler 12387 Version intercept: device num=".. device_num.." node_id="..node_id.. "result=".. tableToString(result));
		end

		MonitorZWaveData(true, -- outgoing,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
41      04/02/17 23:01:39.675       0x1 0x4 0x0 0x60 0x60 0xfb (###``#) 
               SOF - Start Of Frame --+   ¦   ¦    ¦    ¦    ¦
                         length = 4 ------+   ¦    ¦    ¦    ¦
                            Request ----------+    ¦    ¦    ¦
       FUNC_ID_ZW_REQUEST_NODE_INFO ---------------+    ¦    ¦
Node: 96 Device 204=UD17F Breakfast undercabinet lights +    ¦
                        Checksum OK -------------------------+
--]==]
		                 "^01 04 00 60 " .. string.format("%02X", node_id) .. " ..", -- Main RegEx
--[==[
42      04/02/17 23:01:39.700   0x6 0x1 0x4 0x1 0x60 0x1 0x9b (####`##) 
              ACK - Acknowledge --+   ¦   ¦   ¦    ¦   ¦    ¦
           SOF - Start Of Frame ------+   ¦   ¦    ¦   ¦    ¦
                     length = 4 ----------+   ¦    ¦   ¦    ¦
                       Response --------------+    ¦   ¦    ¦
   FUNC_ID_ZW_REQUEST_NODE_INFO -------------------+   ¦    ¦
               Result = Success -----------------------+    ¦
                    Checksum OK ----------------------------+
42      04/02/17 23:01:39.700   got expected ACK 
41      04/02/17 23:01:39.700   ACK: 0x6 (#) 

42      04/02/17 23:01:39.737     0x1 0xc 0x0 0x49 0x84 0x60 0x7 0x2 0x11 0x0 0x72 0x85 0x26 0x86 0x99 (###I#`####r#&#) 
             SOF - Start Of Frame --+   ¦   ¦    ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
                      length = 12 ------+   ¦    ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
                          Request ----------+    ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
    FUNC_ID_ZW_APPLICATION_UPDATE ---------------+    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
Update state = Node Info received --------------------+    ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
Node ID: 96 Device 204=UD17F Breakfast undercabinet lights +   ¦   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
             Node info length = 6 -----------------------------+   ¦    ¦   ¦    ¦    ¦    ¦    ¦    ¦
   Basic type = Static Controller ---------------------------------+    ¦   ¦    ¦    ¦    ¦    ¦    ¦
 Generic type = switch multilevel --------------------------------------+   ¦    ¦    ¦    ¦    ¦    ¦
         Specific type = Not used ------------------------------------------+    ¦    ¦    ¦    ¦    ¦
Can receive command class[1] = COMMAND_CLASS_MANUFACTURER_SPECIFIC --------------+    ¦    ¦    ¦    ¦
Can receive command class[2] = COMMAND_CLASS_ASSOCIATION -----------------------------+    ¦    ¦    ¦
Can receive command class[3] = COMMAND_CLASS_SWITCH_MULTILEVEL ----------------------------+    ¦    ¦
Can receive command class[3] = COMMAND_CLASS_VERSION (fake) ------------------------------------+    ¦   
                      Checksum OK -------------------------------------------------------------------+
--]==]
		                 "06 01 04 01 60 01 XX 01 0C 00 49 84 " .. string.format("%02X", node_id) .. " 07 02 11 00 72 85 26 86 XX", -- Autoresponse,
		                 Kichler12387NodeInfoCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "Kichler12387NodeInfo", -- label
						 false) -- no forward

		MonitorZWaveData(true, -- outgoing,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
41      07/17/17 7:52:58.995    0x1 0x9 0x0 0x13 0x40 0x2 0x86 0x11 0x25 0x5 0x10 (####@###%##) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦
                     length = 9 ------+   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦    ¦    ¦   ¦    ¦
Device 175=SD17-2 Breakfast workspace light --------+   ¦    ¦    ¦    ¦   ¦    ¦
                Data length = 2 ------------------------+    ¦    ¦    ¦   ¦    ¦
          COMMAND_CLASS_VERSION -----------------------------+    ¦    ¦   ¦    ¦
                    VERSION_GET ----------------------------------+    ¦   ¦    ¦
Xmit options = ACK | AUTO_ROUTE | Reserved bits : 0x20 ----------------+   ¦    ¦
                   Callback = 5 -------------------------------------------+    ¦
                    Checksum OK ------------------------------------------------+
--]==]
		                 "^01 09 00 13 " .. string.format("%02X", node_id) .. " 02 86 11 .. (..)", -- Main RegEx

--[==[
42      07/17/17 7:52:59.027    0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   ¦   ¦   ¦    ¦   ¦    ¦
           SOF - Start Of Frame ------+   ¦   ¦    ¦   ¦    ¦
                     length = 4 ----------+   ¦    ¦   ¦    ¦
                       Response --------------+    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA -------------------+   ¦    ¦
                     RetVal: OK -----------------------+    ¦
                    Checksum OK ----------------------------+
42      07/17/17 23:51:36.201   0x1 0x5 0x0 0x13 0x05 0x0 0x9d (####t##) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦
                     length = 5 ------+   ¦    ¦    ¦   ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦
                   Callback = 5 --------------------+   ¦    ¦
           TRANSMIT_COMPLETE_OK ------------------------+    ¦
                    Checksum OK -----------------------------+
41      07/17/17 23:51:36.201   ACK: 0x6 (#) 
42      07/17/17 7:52:59.060    0x1 0xd 0x0 0x4 0x0 0x40 0x7 0x86 0x12 0x6 0x3 0x2a 0x5 0x29 0x26 (#\r###@#####*#)&) 
           SOF - Start Of Frame --+   ¦   ¦   ¦   ¦    ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
                    length = 13 ------+   ¦   ¦   ¦    ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
                        Request ----------+   ¦   ¦    ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   ¦    ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
          Receive Status SINGLE ------------------+    ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
Device 175=SD17-2 Breakfast workspace light -----------+   ¦    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
                Data length = 7 ---------------------------+    ¦    ¦   ¦   ¦    ¦   ¦    ¦    ¦
          COMMAND_CLASS_VERSION --------------------------------+    ¦   ¦   ¦    ¦   ¦    ¦    ¦
                 VERSION_REPORT -------------------------------------+   ¦   ¦    ¦   ¦    ¦    ¦
Z-Wave Library Type = SLAVE_ROUTING -------------------------------------+   ¦    ¦   ¦    ¦    ¦
Z-Wave Protocol Version = 3 -------------------------------------------------+    ¦   ¦    ¦    ¦
Z-Wave Protocol Sub-Version = 42 -------------------------------------------------+   ¦    ¦    ¦
        Application Version = 5 ------------------------------------------------------+    ¦    ¦
   Application Sub-Version = 41 -----------------------------------------------------------+    ¦
                    Checksum OK ----------------------------------------------------------------+
41      07/17/17 7:52:59.061    ACK: 0x6 (#) 
--]==]
		                 "06 01 04 01 13 01 XX 01 04 00 13 \\1 00 XX 01 0D 00 04 00 " .. string.format("%02X", node_id) .. " 07 86 12 06 03 2A 99 99 XX", -- Autoresponse,
		                 Kichler12387VersionCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "Kichler12387Version", -- label
						 false) -- no forward

		MonitorZWaveData(true, -- outgoing,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
41      07/17/17 7:52:59.260    0x1 0xa 0x0 0x13 0x40 0x3 0x86 0x13 0x26 0x5 0x6 0x15 (#\n##@###&###) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦    ¦
                    length = 10 ------+   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦    ¦    ¦   ¦   ¦    ¦
Device 175=SD17-2 Breakfast workspace light --------+   ¦    ¦    ¦    ¦   ¦   ¦    ¦
                Data length = 3 ------------------------+    ¦    ¦    ¦   ¦   ¦    ¦
          COMMAND_CLASS_VERSION -----------------------------+    ¦    ¦   ¦   ¦    ¦
      VERSION_COMMAND_CLASS_GET ----------------------------------+    ¦   ¦   ¦    ¦
Requested Command Class = COMMAND_CLASS_SWITCH_MULTILEVEL -------------+   ¦   ¦    ¦
Xmit options = ACK | AUTO_ROUTE -------------------------------------------+   ¦    ¦
                   Callback = 6 -----------------------------------------------+    ¦
                    Checksum OK ----------------------------------------------------+
--]==]
		                 "^01 0A 00 13 " .. string.format("%02X", node_id) .. " 03 86 13 (..) .. (..) ..", -- Main RegEx
--[==[

42      07/17/17 7:52:59.306    0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   ¦   ¦   ¦    ¦   ¦    ¦
           SOF - Start Of Frame ------+   ¦   ¦    ¦   ¦    ¦
                     length = 4 ----------+   ¦    ¦   ¦    ¦
                       Response --------------+    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA -------------------+   ¦    ¦
                     RetVal: OK -----------------------+    ¦
                    Checksum OK ----------------------------+
41      07/17/17 7:52:59.307    ACK: 0x6 (#) 
42      07/17/17 23:51:36.660   0x1 0x5 0x0 0x13 0x06 0x0 0x9c (####u##) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦
                     length = 5 ------+   ¦    ¦    ¦   ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦
                   Callback = 6 --------------------+   ¦    ¦
           TRANSMIT_COMPLETE_OK ------------------------+    ¦
                    Checksum OK -----------------------------+
41      07/17/17 23:51:36.660   ACK: 0x6 (#) 42      
        07/17/17 7:52:59.339    0x1 0xa 0x0 0x4 0x0 0x40 0x4 0x86 0x14 0x26 0x1 0x0 (#\n###@###&##) 
           SOF - Start Of Frame --+   ¦   ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦
                    length = 10 ------+   ¦   ¦   ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦
                        Request ----------+   ¦   ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   ¦    ¦   ¦    ¦    ¦    ¦   ¦   ¦
          Receive Status SINGLE ------------------+    ¦   ¦    ¦    ¦    ¦   ¦   ¦
Device 175=SD17-2 Breakfast workspace light -----------+   ¦    ¦    ¦    ¦   ¦   ¦
                Data length = 4 ---------------------------+    ¦    ¦    ¦   ¦   ¦
          COMMAND_CLASS_VERSION --------------------------------+    ¦    ¦   ¦   ¦
   VERSION_COMMAND_CLASS_REPORT -------------------------------------+    ¦   ¦   ¦
Requested Command Class = COMMAND_CLASS_SWITCH_MULTILEVEL ----------------+   ¦   ¦
      Command CLass Version = 1 ----------------------------------------------+   ¦
                    Checksum OK --------------------------------------------------+
41      07/17/17 7:52:59.340    ACK: 0x6 (#) 
--]==]
		                 "06 01 04 01 13 01 XX 01 05 00 13 \\2 00 XX 01 0A 00 04 00 " .. string.format("%02X", node_id) .. " 04 86 14 \\1 01 XX", -- Autoresponse,
		                 Kichler12387CommandClassVersionCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "Kichler12387CommandClassVersion", -- label
						 false) -- no forward

	end

	-- This is a hack for UI7 1.7.2608 mishandling of the Shlage BE469 lock. It is incorrectly sending a Command Cleass Version, Version Command Class Get in non-secure mode.
	local function ApplySchageLockHack(device_num, node_id)
		DEntry()

		local function ShlageLockVersionCallback(peer_dev_num, result)
			DLog("Shlage lock version intercept: device num=".. device_num.." node_id="..node_id.. "result=".. tableToString(result));
		end

		MonitorZWaveData(true, -- outgoing,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
                                              C1                               C2
41      04/02/17 15:21:37.647   0x1 0xa 0x0 0x13 0xba 0x3 0x86 0x13 0x71 0x5 0xc7 0x79 (#\n######q##y) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦    ¦
                    length = 10 ------+   ¦    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦    ¦    ¦   ¦    ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦    ¦    ¦   ¦    ¦    ¦
     Device 548=Front Door Lock --------------------+   ¦    ¦    ¦    ¦   ¦    ¦    ¦
                Data length = 3 ------------------------+    ¦    ¦    ¦   ¦    ¦    ¦
          COMMAND_CLASS_VERSION -----------------------------+    ¦    ¦   ¦    ¦    ¦
      VERSION_COMMAND_CLASS_GET ----------------------------------+    ¦   ¦    ¦    ¦
Requested Command Class = COMMAND_CLASS_ALARM -------------------------+   ¦    ¦    ¦
Xmit options = ACK | AUTO_ROUTE -------------------------------------------+    ¦    ¦
                 Callback = 199 ------------------------------------------------+    ¦
                    Checksum OK -----------------------------------------------------+
--]==]
		                 "^01 .. 00 (..) " .. string.format("%02X", node_id) .. " 03 86 13 71 .. (..) ..", -- Main RegEx
--[==[
42      04/02/17 15:21:37.331   0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   ¦   ¦   ¦    ¦   ¦    ¦
           SOF - Start Of Frame ------+   ¦   ¦    ¦   ¦    ¦
                     length = 4 ----------+   ¦    ¦   ¦    ¦
                       Response --------------+    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA -------------------+   ¦    ¦
                     RetVal: OK -----------------------+    ¦
                    Checksum OK ----------------------------+
42      04/02/17 15:21:37.331   got expected ACK 
41      04/02/17 15:21:37.332   ACK: 0x6 (#) 

42      04/02/17 15:21:37.409   0x1 0x5 0x0 0x13 0xc6 0x0 0x2f (######/) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦    ¦
                     length = 5 ------+   ¦    ¦    ¦   ¦    ¦
                        Request ----------+    ¦    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦
                 Callback = 198 --------------------+   ¦    ¦
           TRANSMIT_COMPLETE_OK ------------------------+    ¦
                    Checksum OK -----------------------------+
41      04/02/17 15:21:37.410   ACK: 0x6 (#) 
42      04/02/17 15:21:37.631   0x1 0xa 0x0 0x4 0x0 0xba 0x4 0x86 0x14 0x71 0x1 0xbd 
           SOF - Start Of Frame --+   ¦   ¦   ¦   ¦    ¦   ¦    ¦    ¦   ¦    ¦    ¦
                    length = 10 ------+   ¦   ¦   ¦    ¦   ¦    ¦    ¦   ¦    ¦    ¦
                        Request ----------+   ¦   ¦    ¦   ¦    ¦    ¦   ¦    ¦    ¦
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   ¦    ¦   ¦    ¦    ¦   ¦    ¦    ¦
          Receive Status SINGLE ------------------+    ¦   ¦    ¦    ¦   ¦    ¦    ¦
     Device 548=Front Door Lock -----------------------+   ¦    ¦    ¦   ¦    ¦    ¦
                Data length = 4 ---------------------------+    ¦    ¦   ¦    ¦    ¦
          COMMAND_CLASS_VERSION --------------------------------+    ¦   ¦    ¦    ¦
   VERSION_COMMAND_CLASS_REPORT -------------------------------------+   ¦    ¦    ¦
Requested Command Class = COMMAND_CLASS_ALARM ---------------------------+    ¦    ¦
                    Version = 1 ----------------------------------------------+    ¦
                    Checksum OK ---------------------------------------------------+
--]==]
		                 "06 01 04 01 \\1 01 XX 01 05 00 \\1 \\2 00 XX 01 0A 00 04 00 " .. string.format("%02X", node_id) .. " 04 86 14 71 01 XX", -- Autoresponse,
		                 ShlageLockVersionCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "ShlageLockVersion", -- label
						 false) -- no forward
	end

--[====[ This is no longer needed with the latest version of UI7
	local function ApplyLuaUPnPAutoRouteNoACKFix()
		local function AutoRoutNoAckCallback(peer_dev_num, result)
			DLog("AutoRouteNoAckk Callback: peer_dev_num=".. peer_dev_num.." result=".. tableToString(result));
		end
		MonitorZWaveData(true, -- outgoing,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
--[==[
41      11/24/17 21:25:33.043   0x1 0xa 0x0 0x13 0xc 0x3 0x25 0x1 0x0 0x4 0x37 0xfe (#\n####%###7#) 
           SOF - Start Of Frame --+   ¦   ¦    ¦   ¦   ¦    ¦   ¦   ¦   ¦    ¦    ¦
                    length = 10 ------+   ¦    ¦   ¦   ¦    ¦   ¦   ¦   ¦    ¦    ¦
                        Request ----------+    ¦   ¦   ¦    ¦   ¦   ¦   ¦    ¦    ¦
           FUNC_ID_ZW_SEND_DATA ---------------+   ¦   ¦    ¦   ¦   ¦   ¦    ¦    ¦
  Device 5=Leviton Combo Switch -------------------+   ¦    ¦   ¦   ¦   ¦    ¦    ¦
                Data length = 3 -----------------------+    ¦   ¦   ¦   ¦    ¦    ¦
    COMMAND_CLASS_SWITCH_BINARY ----------------------------+   ¦   ¦   ¦    ¦    ¦
              SWITCH_BINARY_SET --------------------------------+   ¦   ¦    ¦    ¦
                    Value = OFF ------------------------------------+   ¦    ¦    ¦
      Xmit options = AUTO_ROUTE ----------------------------------------+    ¦    ¦
                  Callback = 55 ---------------------------------------------+    ¦
                    Checksum OK --------------------------------------------------+
--]==]
		                 "^01 .. 00 13 .+ 04 (..) ..$", -- Main RegEx
--[==[
42      11/24/17 21:25:33.086   0x6 0x1 0x4 0x1 0x13 0x1 0xe8 (#######) 
              ACK - Acknowledge --+   ¦   ¦   ¦    ¦   ¦    ¦
           SOF - Start Of Frame ------+   ¦   ¦    ¦   ¦    ¦
                     length = 4 ----------+   ¦    ¦   ¦    ¦
                       Response --------------+    ¦   ¦    ¦
           FUNC_ID_ZW_SEND_DATA -------------------+   ¦    ¦
                     RetVal: OK -----------------------+    ¦
                    Checksum OK ----------------------------+
42      11/24/17 21:25:33.087   got expected ACK 
41      11/24/17 21:25:33.087   ACK: 0x6 (#) 

42      11/24/17 21:25:33.127   0x1 0x7 0x0 0x13 0x37 0x0 0x0 0x1 0xdd (####7####) 
           SOF - Start Of Frame --+   ¦   ¦    ¦    ¦   ¦ +-----+    ¦
                     length = 7 ------+   ¦    ¦    ¦   ¦    ¦       ¦
                        Request ----------+    ¦    ¦   ¦    ¦       ¦
           FUNC_ID_ZW_SEND_DATA ---------------+    ¦   ¦    ¦       ¦
                  Callback = 55 --------------------+   ¦    ¦       ¦
           TRANSMIT_COMPLETE_OK ------------------------+    ¦       ¦
                 Tx Time = 1 ms -----------------------------+       ¦
                    Checksum OK -------------------------------------+
41      11/24/17 21:25:33.128   ACK: 0x6 (#) 

--]==]
		                 "06 01 04 01 13 01 XX 01 07 00 13 \\1 00 00 00 XX", -- Autoresponse,
		                 nil, -- AutoRoutNoAckCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "AutoRouteNoAck", -- label
						 false) -- no forward
	end
--]====]

	for device_num, device in pairs(luup.devices) do
		if device.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD:1" then
			local impl = luup.attr_get("impl_file", device_num)  
			if impl ~= "I_GenGenSceneController.xml" then
			  	reload_needed = "I_EvolveLCD1.xml"
			  	log("Updating the implementation file of the existing Evolve LCD1 peer device: ", device_num)
			  	luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num);
			end
		elseif device.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5:1" then
			local impl = luup.attr_get("impl_file", device_num) 
			if impl ~= "I_GenGenSceneController.xml" then
				reload_needed = "I_CooperRFWC5.xml"
			  	log("Updating the implementation file of the existing Cooper RFWC5 peer device: ", device_num)
			  	luup.attr_set("impl_file", "I_GenGenSceneController.xml", device_num);
			end
		elseif device.device_num_parent and
			luup.devices[device.device_num_parent] and
			luup.devices[device.device_num_parent].device_type == "urn:schemas-micasaverde-com:device:ZWaveNetwork:1" then
	  		local manufacturer_info = luup.variable_get("urn:micasaverde-com:serviceId:ZWaveDevice1", "ManufacturerInfo", device_num)
			local capabilities = luup.variable_get("urn:micasaverde-com:serviceId:ZWaveDevice1", "Capabilities", device_num)
			DLog("device_num=",device_num," name=",device.description," manufacturer_info=",manufacturer_info," capabilities=",capabilities);
		  	if manufacturer_info == "275,17750,19506" then
		  		if AdoptEvolveLCD1(device_num) then
			  		reload_needed = "Evolve LCD1 adopted"
				end
			elseif manufacturer_info == "26,22349,0" then
				if AdoptCooperRFWC5(device_num) then
		 			reload_needed = "Cooper RFWC5 adopted"
				end 
			elseif manufacturer_info == "376,21315,18229" then
				if AdoptNexiaOneTouch(device_num) then
					reload_needed = "Nexia One Touch adopted"
	  			end
	  		elseif manufacturer_info == "12,17479,12339" or capabilities == "211,156,1,4,16,1,L,R,B,RS,|37,43,44,85,89,90,91,94,108,112,114,115,122,133,134,159," then
	  			if AdoptHomeSeerWS100(device_num) then
					reload_needed = "HomeSeer HS-WS100+ switch adopted"
	  			end
	  		elseif manufacturer_info == "12,17479,12340" or capabilities == "211,156,1,4,17,1,L,R,B,RS,|38:4,43,44,85,89,90,91,94,108,112,114,115,122,133,134,159," then
	  			if AdoptHomeSeerWD100(device_num) then
					reload_needed = "HomeSeer HS-WD100+ dimmer adopted"
	  			end
	  		elseif manufacturer_info == "12,17479,12341" or capabilities == "211,156,1,4,16,1,L,R,B,RS,|37,39,43,44,85,89,90,91,94,108,112,114,115,122,133,134,159," then
	  			if AdoptHomeSeerWS200(device_num) then
					reload_needed = "HomeSeer HS-WS200+ switch adopted"
	  			end
	  		elseif manufacturer_info == "12,17479,12342" or capabilities == "211,156,1,4,17,1,L,R,B,RS,|38,39,43,44,85,89,90,91,94,108,112,114,115,122,133,134,159," then
	  			if AdoptHomeSeerWD200(device_num) then
					reload_needed = "HomeSeer HS-WD200+ dimmer adopted"
	  			end
	  		elseif manufacturer_info == "12,515,1"       or capabilities == "211,156,1,4,17,8,L,R,B,RS,|38:4,43,44,85,89,90,91,94,108,112,114,115,122,133,134,159," then
	  			if AdoptHomeSeerFC200(device_num) then
					reload_needed = "HomeSeer HS-FC200+ fan controller switch adopted"
	  			end
			elseif manufacturer_info == "59,25409,20548" or capabilities == "83,220,0,4,64,3,R,B,RS,W1,|32S,34,93S,98S,99S,112S,113S,114,122,128S,133S,134,152," then
				ApplySchageLockHack(device_num, device.id)
			elseif capabilities == "146,150,0,2,17,0,L,B,|38,114,132," or capabilities == "146,150,0,2,17,0,L,B,|38,114,133," then
				ApplyKichler12387Hack(device_num, device.id)
			end
		end
	end	-- for device_num

	local function NexiaManufacturerCallback(peer_dev_num, result)
		DEntry()
		local time = tonumber(result.time)
		local receiveStatus = tonumber(result.C1, 16)
		local node_id = tonumber(result.C2, 16)
		local device_num = NodeIdToDeviceNumber(node_id)
		DEntry("NexiaManufacturerCallback")
		if device_num and CheckDups(device_num, result, "72050178534347359e"..result.C2) then
			local device = luup.devices[device_num]
			if device and device.device_type ~= "urn:schemas-gengen_mcv-org:device:SceneControllerNexiaOneTouch:1" then
				AdoptNexiaOneTouch(device_num)
				log("Nexia One Touch adopted. Reloading LuaUPnP.")
				luup.call_action(SID_HAG, "Reload", {}, 0)
			end
		end
	end

	-- The Nexia One Touch may show up as a "Generic IO" device if it was included before the plug-in was installed
	-- because UI7 gets confused by the devices's response to the COMMAND_CLASS_ASSOCIATION_GRP_INFO/ASSOCIATION_GROUP_INFO_GET command  
--[==[
                                                 C1  C2
42      02/25/17 15:55:45.857   0x1 0xe 0x0 0x4 0x0 0xf 0x8 0x72 0x5 0x1 0x78 0x53 0x43 0x47 0x35 0x9e (#######r##xSCG5#) 
           SOF - Start Of Frame --+   ¦   ¦   ¦   ¦   ¦   ¦    ¦   ¦ +------+ +-------+ +-------+    ¦
                    length = 14 ------+   ¦   ¦   ¦   ¦   ¦    ¦   ¦    ¦         ¦         ¦        ¦
                        Request ----------+   ¦   ¦   ¦   ¦    ¦   ¦    ¦         ¦         ¦        ¦
FUNC_ID_APPLICATION_COMMAND_HANDLER ----------+   ¦   ¦   ¦    ¦   ¦    ¦         ¦         ¦        ¦
          Receive Status SINGLE ------------------+   ¦   ¦    ¦   ¦    ¦         ¦         ¦        ¦
          Device 10=_Generic IO ----------------------+   ¦    ¦   ¦    ¦         ¦         ¦        ¦
                Data length = 8 --------------------------+    ¦   ¦    ¦         ¦         ¦        ¦
COMMAND_CLASS_MANUFACTURER_SPECIFIC ---------------------------+   ¦    ¦         ¦         ¦        ¦
   MANUFACTURER_SPECIFIC_REPORT -----------------------------------+    ¦         ¦         ¦        ¦
  Manufacturer Id = Nexia (376) ----------------------------------------+         ¦         ¦        ¦
       Product Type Id = 0x5343 --------------------------------------------------+         ¦        ¦
            Procuct Id = 0x4735 ------------------------------------------------------------+        ¦
                    Checksum OK ---------------------------------------------------------------------+
--]==]
	MonitorZWaveData(false, -- incoming,
					 luup.device, -- peer_dev_num
	                 nil, -- No arm_regex
	                 "^01 .. 00 04 (..) (..) .. 72 05 01 78 53 43 47 35", -- Main RegEx
	                 nil, -- no response,
	                 NexiaManufacturerCallback,
	                 false, -- Not OneShot
	                 0, -- no timeout
					 "NexiaManufacturer", -- label
					 false ) -- not forward

	if luup.version_major < 7 then -- UI5 only.
    -- In response to a Version Get command, the Nexisa One-Touch returns a Z-Wave protocol version 4, sub-version 32
    -- which is not even in the official Z-Wave version list and certainly confuses UI5. We create a forwarding monitor
    -- which returns a (false) version which at least UI5 can handle.
--[==[								  C1		  C2  C3  C4				     C5  C6  C7
42      03/04/17 14:19:17.494   0x1 0x11 0x0 0x4 0x0 0xf 0xb 0x86 0x12 0x6 0x4 0x20 0x1 0x0 0x1 0x1 0x1 0x1a 0x42 (########### ######B) 
           SOF - Start Of Frame --+    ¦   ¦   ¦   ¦   ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦ +--------------+    ¦
                    length = 17 -------+   ¦   ¦   ¦   ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
                        Request -----------+   ¦   ¦   ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
FUNC_ID_APPLICATION_COMMAND_HANDLER -----------+   ¦   ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
          Receive Status SINGLE -------------------+   ¦   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
          Device 49=_Generic IO -----------------------+   ¦    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
               Data length = 11 ---------------------------+    ¦    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
          COMMAND_CLASS_VERSION --------------------------------+    ¦   ¦   ¦    ¦   ¦   ¦        ¦            ¦
                 VERSION_REPORT -------------------------------------+   ¦   ¦    ¦   ¦   ¦        ¦            ¦
Z-Wave Library Type = SLAVE_ROUTING -------------------------------------+   ¦    ¦   ¦   ¦        ¦            ¦
Z-Wave Protocol Version = 4 -------------------------------------------------+    ¦   ¦   ¦        ¦            ¦
Z-Wave Protocol Sub-Version = 32 -------------------------------------------------+   ¦   ¦        ¦            ¦
        Application Version = 1 ------------------------------------------------------+   ¦        ¦            ¦
    Application Sub-Version = 0 ----------------------------------------------------------+        ¦            ¦
                         ?data? -------------------------------------------------------------------+            ¦
                    Checksum OK --------------------------------------------------------------------------------+
	convert version 4, subversion 32 to version 3, subversion 28 (Z-Wave 5.03.00)
--]==]
		local function NexiaVersionCallback(peer_dev_num, result)
			DEntry("NexiaVersionCallback")
		end

		MonitorZWaveData(false, -- incoming,
						 luup.device, -- peer_dev_num
		                 nil, -- No arm_regex
		                 "^01 (..) 00 04 (..) (..) (..) 86 12 06 04 (..) (..) (..)", -- Main RegEx
		                 "01 0D 00 04 \\2 \\3 07 86 12 06 03 1C \\6 \\7 XX", -- Forwarded response,
		                 NexiaVersionCallback,
		                 false, -- Not OneShot
		                 0, -- no timeout
						 "NexiaVersion", -- label
						 true) -- forward
	end

	-- Apply the Aeotec Siren hack to anything that responds with that exact node info.
	ApplyAeotecSirenHack()

	-- Apply the Aeotec Trisensor hack to remove S0 and S2 security from added nodes.
	ApplyAeotecTrisensorHack()

	-- Apply the Aeotec Dry Contact Sensor hack to anything that responds with that exact node info.
	ApplyAeotecDryContactSensorHack()

--[====[ This is no longer needed with the latest version of UI7
	-- UI7 sometimes sends Z-Wave commands twice with xmit option AUTO_ROUTE and then again with ACK | AURO_ROUTE
	-- Only the second of these is valid.
	ApplyLuaUPnPAutoRouteNoACKFix()
--]====]

end

-- returns first, updated
-- First is true if we are the lowest device number installer
-- updated is true if the installer has been updated since the last execution.
local function IsFirstAndLatestInstallerVersion(our_dev_num, our_version)
	DEntry()
	local version = 0;
	local count = 0;
	local our_index = 0;
	local sorted = {}
	local ourVerStr = luup.variable_get(SID_GENGENINSTALLER, "Version", our_dev_num)
	local updated = not ourVerStr or tonumber(ourVerStr) ~= our_version
	if updated then
		luup.variable_set(SID_GENGENINSTALLER, "Version", our_version, our_dev_num)
	end
	for dev_num, v in pairs(luup.devices) do
		table.insert(sorted,dev_num)
	end
	table.sort(sorted)
	for i = 1, #sorted do
		local dev_num = sorted[i];
		if dev_num == our_dev_num then
			our_index = i
		end
		if luup.devices[dev_num].device_type == DEVTYPE_GENGENINSTALLER then
			local verStr = luup.variable_get(SID_GENGENINSTALLER, "Version", dev_num)
			local ver = tonumber(verStr)
			if ver then
				if our_index == 0 then
					return false, false -- We found another installer device before we found us.
				end
			end
		end
	end
	return true, updated
end

local function DeleteOldInstallers(our_dev_num)
	DEntry()
	for dev_num, v in pairs(luup.devices) do
		if dev_num ~= our_dev_num and
		   (v.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCD1Installer:1" or
		    v.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerEvolveLCDFinder:1" or
		    v.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerCooperRFWC5Installer:1") then
			log("Removing older installer: device ID ", dev_num)
			luup.call_action(SID_HAG,"DeleteDevice", {DeviceNum = dev_num}, 0);
			-- Delete only one at a time. We will reload the luup engine for more.
			return true
		end
	end
	return false
end

local nixio = require("nixio")

function SceneControllerInstaller_Init(lul_device)
  VEntry()
  -- First, delete any stale global lock. Note that there may be a race condition
  -- here if another device just took it but we will live with that.
  give_global_lock()

  -- Now, make sure that we are the latest version of this installer
  -- And if there is more than one of us, that we are the lowest numbered device.
  -- Otherwise delete ourselves.
  local first, updated = IsFirstAndLatestInstallerVersion(lul_device, GenGenInstaller_Version)
  if not first then
	log("Removing superfluous installer: device ID ", lul_device)
	luup.call_action(SID_HAG,"DeleteDevice", {DeviceNum = lul_device}, 0);
	return
  end

  -- Now look for older installers with different names and delete them one at a time
  if DeleteOldInstallers(lul_device) then
	log("Older installer deleted. Reloading LuaUPnP.")
	luup.call_action(SID_HAG, "Reload", {}, 0)
	return
  end

  luup.attr_set("invisible","1",lul_device)

  InitializeZWaveInterceptor(ZWIntDebugLevel)

  local function b642bin(str)
  	return nixio.bin.b64decode(string.gsub(str, "%s+", ""))
  end

  -- Begin zwint_so_lzo
  local zwint_so_version = 2.02
  -- lzo-compressed zwint/zwint_debug.so: compressed size = 22833 -- uncompressed size = 40040
  local zwint_so_lzo
  zwint_so_lzo = b642bin([[
	aJwAAAV/RUxGAQEBAOAAAQMACAB0AQJQEAAANFACAAEImAAABxAAdDQAIAAHACgAHAAbVAIDAHAUAQAA7QAYRAJsAAEEAAAAbAB8B0gCJwgAAdSGAABtAAWFAgGcCwSkjwAApI8BjQC8TAkC7AUAAAbdAwJdCCxMAu0A8FwCbQAHXAJ8CwFR5XRkSAIvCAD8AwFS5XRkKn0BXFwGbABsBy5sAi4QAWwEAQAAALIuXwAAEAKfFgUEAJ0AE9wA
	AiMEAAAMUAUC2A8AAA1eAKBmrCABHAIAAH4dEAq8GgqQBAAACgAAAGUEAAALXAQCEAAAAAN8AAGQAQARXAACSA8AABJdAJBdAhNeAAgAnQxwfBVdCHBsIl0jcEgDfghwCVwDXgZwWFAVXQZwdDJcBg5wDgAAAP7//28oDwAA////b3wHAvD//292YANwByAKDQBDTAV8C1gAjRRXTRM7TAACMwAAAB9dDDZgA80AFlwQfBt5AUJwAF0EI00U
	KV0CIlweThgAJkwCfBR8HsoEACpMAv0hJ/ACXDVsKvkBOF0IIUwIfCttAhxwAF0GK10ENE0RP00AF80CLE0FUUwBfVEZXABoAF4jABVgAU0DMV0ITcwBeQBATQc3XQgkTQAUTQM1TQAtTBt5JURdAD5cDFwqT18AADpgBjkNAEbNAy9dBh5NASBNCR1cAGkAD0wPbDZtATJwAM0AMF0AKE0kGlwA6QBMXQFK0QE8XRNPTQBHTQklzAL5AFRN
	Aj1NDS7AAl1DAHxGeQFJcABcCHxGaQFVUQBQTABpAFZwAF0RQV0HRcwBbT1LTQI53AHeYABI3AEgBxkAUymhAE5dD1IpPAAgFygAbGtsBwsDAAkAugMAAN1fAACEAn9lCgA0YRWQdFwDAAAQABMAnUAQtB4CEwDx/zt8A9wBAhAA8f/VUE98CWxIAhIACQAtXANsnXwLAhAACgBGXgFgkf4J8f9sOGx6vwUMAD8tfABlBRDcDwIRAPH/WF4D
	kJV0An8R8f9SXAEqPAF+MoBmdA1/FwAAMUkVcCk9AMlJA2ApPAB8UUgCiABtEWJJC1ApfQDbXRlAKT0A+l0DMCk9AKNFByApPQAuXQ0QKT0AKV0BACk9AHNeAfBlUAZMN31myF0L4Ck9ALBJA9ApPQCFXQ3AKT0Agl0DsCk9AB1cNQO1MwAAHAF/HQoAGV0LoCl9AIhdBZApPQD/XQOAKT0Akl0BcCk9ANZdBWApPQCUXAHoAF6hAIldBVAp
	fQC/XLACvSoAADLdD1RdFUApfABpATBwAN0dmF0DICk9ANFdARApPAAHfgMAAAFcAADaA70ZeV0jACl9ANhdBfDV0xJdaXFdA+ApPAAFEAEAAAkWAAB9wxJFuQhdA9ApfQCBXRfAKT0Apl4DyIRACVwGAhEADQBKXQOwKX0A6l0BoCk9ABVdBZApPQBsXQGAKT0AYl0BEF1TgEgKAhEAFgCbUSdwKX0ArV0XYCk9ALNdAVApPQBPXB8B0TUA
	AHwofRdwXQNAKXwAAkECAAAwKT0AKXwFATQAAP6cKwQKAMMBAAAgKX0Afl0BECk9AFldEQApPQDyXgPwY3ACfy0AAKlJAeApPQD4XRnQKT0Ael0PwCk9AOldA7ApPQAkXQOgKT0AQl0LkCk8AGzUSAyIAH3W6V0DgCl9AFhdB3ApPQB3XQNgKT0ANl0DUCk9AM9FCUApPQDgXQUwKT0AOV0BICk9AApdBxApPQDBXQcAKT0A9l4D8GJwAr0h
	Zl0B4Ck8AAaYAgAAcSkAAErdXQNdA9ApfQB0SQfAKT0A6EzAAh0VAADq3DMAGQBfR0xPQkFMX09GRlNFVF9UQUJMRV8AX2dwX2Rpc3AAX2ZpbmkAX19jeGFoAQAXYWxpemUAX0p2X1JlZ2lzdGVyQ2xhc3NlcwBsdWFfcHVzaGludGVnZXInPwBuaWwnLAAABHN0cmluZwBjbG9ja19nZXR0aW1lAGaUAAJsb2NhbGQBABVfcgBzdGRlcnIA
	ZnByaW50ZgBmcHV0cwBzb2NrZXQAX19lcnJub198BQl0aW9uAGNvbm5lY3RwCgZzZQB1c2xlZXBgDgRMX2NoZWNr+BEAFkx1YVVQbkNvbnRyb2xsZXJSZXNwb25zZU1vbml0b3JDYWxsYmFjawAunAAGQ29tbWFuZEluSB0BY2VwdCegAARzdHJlcnJvgB3NDWzcGwVsdWFMX2FyZyd8AAhfdG9ib29sZWFuAFQcAA1sb2MAbWVtY3B5AHN0
	cnRvbABwdGhyZWFkX211dGV4aRtrLUoAdW5oJgMAd3JpdGVwG4Qt6ApmL3Rv8A8Oc3RyY21wAG9wZW5kaXIAc27UKGAMAmxpbmsAYAFkA1Ywc2VgAWYGcHknxQFjQgR0ZdwtBnBhaXIAZHVwMoAKBAByZWdmcmWoEH03b4gtAV90eXCuAmlz/C2wEio0AARpc251bWJleDsFcmxlbgBtYWxkIQRyZWdjb21waAu8JXk9Y9wwBlBTZW5kRGF0
	YTy4BieYAAVDb25maXJtYWRDLmQHAAhfX2Zsb2F0c2lkZgBfX2RpdmRmMwBfX2FkZGABdBkEaW1lb2ZkYZkybngzCFByb3RvY29sTG9nLkACAkFib3J0MFUIQ2xOcBYAAUNBTlRvSG9zdABzYXZlX2J1ZmaEL5wcBXhlYwBwb2xsLlABBE1lc3NhZ2UwWgFaV1AICkdldFJlcG9ydFRhYmycLXA8A196d2ludCf0BqhIAWluaXR/RExfcth0
	dAGURtAwfDMABHNldGZpZWxkAGxpYmdjY19zLnNvLjF0AcBSfQEwfQFjpAEKX2Z0ZXh0AF9mZGF0YW+EAF9lqAEGX2Jzc19zdGFydAMOYnNzAF9lbmQAR0NDXzMuMACAACQcOHQALAwAkAM3EABYpUSVUJrEBLwBcAHkAbwAdAC0BGQBKHwA9AFkAjwMAHQHtAV0AGwUIuw3Itw2fAYHUCZ5CwAAAgBdBHQBKA8AuI8BI103vN0AwN0AxN0A
	yN0AzN0A0N0A1N0A2N0A3N0A4N0A5N0A6N0A7N0A8N0A9N4AXJG8DwANAgAcPCgAnCch4JkD4P+9JxAAvK8cAL+vGAC8rwEAEZQWbQQEjAQInwMkgJmPDBE5Jz7MA10FjyqeANz/bAm8BANgYjkniRSsCHwEXAoGjwgA4AMgAL0nbQqwjwWZA9BcDgUoALOvGICTj14EryxcDwAaYJFikiQAsq8gALGvHACwrxsAQBQcgYKPBQBAEByAgo8c
	gZmPCfggAwAARIxMBgAGjxiAko8ggJGPGICQj6yPUiYjiDICg4gRIjw2DBD//zEmZJECroAQAgAhECKmMVmMrgcAAFwCDo4rGFEA9/9gFAEAQiQBAAIkTQ+iTRCPTRKPTRCPTRCPTBCNFzDOF/T+fBwKmQMYgISPtI+CjAYAQIg+BECAmY8DACAiKDgGAAgAIAO0j4QkbB4o5AfEAQAc8AJqv/YMCwD0QDJp4pplXGcQ8HiacvF0mwFzE2AD
	IwJzBmAKEDDwRJrs9hBKDoQBAgz3HEoJhAECLPcUSgSEAQYM9wRKoOgAZQBfCV/2HNwJA8RkmmUE0qwKBzDwVJoS9aybAU1UAAjbQOo6ZURkIOgBaq8GP/YIzAYADppl9mQ8ZwTS8PBcmQbVB9Y6ZUDqBGcEljDwVJkGlZ5lfQGQXAEABJDwOJmQZ55lB5Y5ZUDppmd2ZCDoA2p+D9/1/Q/F3A8FUPFUmgYFAWx0DgAMBpPg8whqB5RY6waz
	RWQS6nrsASvl6BLraeLA90MzbBgBQEIPAH8In/UUJ0wCC/9kHGcK0nDwTJgkZwwEfAgJCpZw8FiYDASeZQ8FfAEDfWdaqwqWTAsACFqeZRphH/QYSlrLD5JhQjxTD9MSYcVKD9IQlAEASBDTC2HFShDSEZJnQhFLEdMEIxHSEpIBShLSkPBImBSTMPCkmICaEpLQ8BiYE5cE0mRqWusBKuXofWfRZzhlTPcQTRDqBdIRkgbSEJIH0g+SCNJa
	q0DoCdJ/ZGwXfhbf9PwfbSkcTCkJMPBEmCRnBtVi8glKfidMbGwNbiqkmHwMDAaXgJqeZThljPcATUDo0Uwp7wmf9BDNIPiOKQBqnyAQFiJuCk1sZAp0H/wJC4SYoJpQ8FCYnmWs9whMZAMLBJaeZRDxRJgCbABupGekAgsAUhLwQNmeZRJgcPBEmGwCACsw8ISYMPAEmKCarPccTCLzAUhA6DhlAWpL6mAQAGudZwjTCdMCa2zM7PcTa2vr
	bcwss4JnUPFAmBBuB9MGnC8HBJYAUp5lI2AKkm8JAUqiXTUwXx7M9wx9CUp1DARUSAVcmBLwgJmeZXwBfgsElmwRAzDxUJgYtLQChBYB8DiYnG1hnFQDCply8XCbAFIEamzqg2Eu5gMElmwfdDZkKQmeZYCaEvDAmez3DE1kGHQHBXhkoOh/AAAB/ko/8yh+BvVkfTOQVCADEPA4mAFtdBIBcvFQ2WUXSyd0BTCPAXLx0E0MHIwMAnVkIOgA
	32j/8gDcNQH2ZBKTdG8ALhxnQaMioxDweJiS8YSbBHRWYTDwaJgDbND1CEst42CjjOsCcwVgA3MFYAFzS2BHEEkiRRA5cUNgeHFEYAFxQmBlEUwnJAJsHXQovEZ0RgMt8AhN0WdtHSlEgQtkmE1sBtJi8glLQOs7ZXwWD2iYBpKeZYCbBSIw8MiYEPAQTmSGA8SYLfAATvQmAvFnTfAUjBltL3x8LgSbAGqS8UTbdHoPAWoQ8JiYBGty8ZCc
	jOvwI8wXrob/8fxzABb2ZATSmmUSkhxnKfYAayGiMPBImND1CEpJ4UCCSzQBTHjsQGsS7CBMrH0ABANrbOoCchLtBtU0YANyaWABcuDwAWFmJARqZj1s6i6MByjUAwqAmtDwWJgGl55ljfAYfB4B6jplEFyAAwRrkvFk2mwlBBDwGJgi8gGcTQcGlsD3wzXR4i4QIA93AS3xBCABdAEAFJTA94M1keJD7FhnteOp4pLxiNiS8UzYexA5cQVg
	eHEvYAFxdWBfK+0BFTXcDmyOZBsCnmWt8RA/5AEC4POIQisgDl8BTfIAP1wBART3gEJ1GW3PGWzYFU5KRJggBfwJrEWvQ973GNxgAZplCPC0bALw8FyYJMSQBjDwVJgLlZ5lkfwBAvFYmAuU/H8HkPAYmJFnomeeZXxafWsD3mue9/xrD5plBPD9ZBxnCNIQ8ViYAW0NBmx3ARHSCJZ0eQIalJ5lApR5AAgIlgDxAFoO0p5lB2Ew8MSYGpQC
	bc3yEE4kEL0EA5UEElwGfAh/BgRtDI0JD0wCrwSeZQWUBAYIlgBSENKeZQtUcmQJBwVt7fIYTpDwFJh8FAbnEFDwRJgalAaMBWROAQiWE9JkWgQEap5lbOogLe4HEpN9WghMonQ/AQTTD5OkQA0F0xCTEZYOlwbTE5MN8xRNZMYAAgfTCJaeZQ2W8PBImAFsp0YU1lJNfB0ACAiWImeeZRSWgPAUIodCUUyA2jDxSJgRlQFOdAMAI51nIPCY
	pBKTAGpA8JDBnWdA8IykYdkQkwiWQPBUwfDwWJhi2UDwk8FnQQ+UCUueZQ7TAG4LlLkHCJYAa55lJhBAc/QmBQNtbfMITokXdCkqTAAJHE5/Fw6UAUsAbkDEZAoHAUwO1BXTh2cU1+QJAAIUlxWTnmUP1wuXD5Xq7dZhQPBxwXQjCXDwSJiy8QRLg2cV03QVdIwAAr1nCJYy8lyknmUg8EjFQPBSwQFKXQHEbHwLFZNS8qCag2cCLVLyINpk
	Ag5S8qSaAiUj3QIQUvIk2pDwXIzZZAMNCJYKBTLylJrQ8FyYnmUBbnwLBAiWAXKeZQ2M3mzdbAIEoJoalCPwBZTdBx8X8PAAmBqUAW13OQFqfWT6AvACah71KHwmC0Tw+WQ8ZwbS8PBUmQkGv8UGlopMQgMDIgmSIFqEIwaZkpQBbY3zEE5sdAgYmXDwSJmy8QRIkIxdAAsQ8FiZBpZS8mianmUmI7DwQJkQ8LiZipSN01LyFIw6CQaWjZOe
	ZRIikPBcma4GBpZkCgACnmUw8MSZrfMATpDwNJlA6Tll7BF2CQFLVAkL2swRMPAEmbDwTJnN8xTMDQcGlovSnmXA8BkinAQAB2tr61LycNpjEAqiCmtu6F8oC0r8Z4zSBNJ8CgAEMPFUmTDw5J//bUoEAU3t8wROzfMUT2wkCgaWMPFAmUoEnmUKBf+EKQkGlp5lQCIIBEnkCMJnGIqVCiNMJnwCCDQq8PBYmYyUCm4IpEkGkwaWYIOeZSgr
	vMoFUvJQ2xDweJntvSR1YpkoJSMGfPlGEaSZbPki7CoGWJrt8wxNUvLQbb+ZpBIHnmUIEBDxQJmLlPwBApUqUPBMKTwAJ7QBBJsAUhJgnGd0KQQQ8JicsvEEI1wkKkwFDw30CE5VFxDwGJlQ8VCZipVS8igEBQQQ8JiZMPFcVBF0JAgAbeVnqfARTnLyFNQIdpQNIih+AY3V5AwCkpSNlfpUtGRADVSZAG5wZ3LxHEsBbAJt42dueo3TfSUS
	KewA5Ad/AnDwRIwdAqCakpTXVAikJgKQm4zqHiACxAQKcvHcmC30EE1y8RxK4SvcBLwiCRDxVJly8ZyYUvKwm6YoitJ0lCLMJQaYnHLxcJxs6h8zngGKkyhfBjDwpFUI3FUI8IwMAQTTTfTsWHVVXFUDnI5yBpYudQEYLXQBA5DwSJkGlmUKgI4JnmVvGG30HMxiCp5lipMAUzRhnGdQ8FhMZgKYnAJtUicUBm43nmW0FwFS8kzYl0ucbOog
	AZQBBFLyzJiN9BgnlAFUBwGYAFIldIScLwIGlgCaniNfNliaclcpgZowXHKsJmxK9A7sNwmSlLBnMPAkmSPwBUludC0WpC7uE0ibZAUDAUpS8kjbfAekCQfw8CCZkpSeZQFtbAYCAWpA8Hl0jyM+N57wKP0R95SPdHkJBGcy8nCaauxgnAJhRAEK2oGYgdthmICYgNtBJW1WBCW8JgYjIzDwZJlNbAgnVCgIBpYIkpDwiJmeZTJOaoCcbC4A
	BWCew56t9AxNQJvjm2OaBNNAmkOaBdLQRC28EwqeZVKACCJQ8FSZh0ARxl+eZb0BMd0BMCVEOGUYd3QXAvACat33JvwyIiw0Ilw5bxdkZ8VUFwcQmhsQQphq6hdhbIgFg5imZwbTB9Z/DQSWBkTGAQeWCipkKQSQZwFtY/cRlCkDAWoHEACYbAgG0vEASkvg3yp2ZA2fJH33BCfNIfYjXi8w8SK8OKQKCQFSnmUnYfDwTJiRZ7y2IqwrAh4i
	0PBEK0UAC04OxJhs+nwEAc30DE50wQE1EDDxLJwABcJnAhABbsvu5tkH1q3ZBs7ZuJh0GAFS8kidIow8AQFSFWB80wGDZwbVfhEGlSIsInTQCFLyqJ2RZ8LxFUjtzw1fEP90QwPdKioQ8FhMHQ2YmAbTUvKwmhDxVJhS8oycfAcktTUXdFScawGgmgaTIrUvXCMUN3QMRLQC6jplBpV125EjfjDTF7wVABBYmDLykJoUEABWQJwDYWKcau4N
	YTDwZJgG0gfWY/cRSwFtJO4wBpJmF4JndBCXLeTmKpycbBInHhUElnbmkWeMVyPMOIQybD8p/j4E8GRXAwTSUPFImSj1BQduLZkOVOUC7fQcThpsLJQBZhwI0mQLBlSZDpSeZQJtBiNEJgMElgJnnmWENAWZDpQCbQ31FCctHCSNpHCs7QFLg2cJnTAwVKgECJSwZyTwDSNEMAMCZwSWCZN/bp5lg5zxdQwglgywZ2ZXAWrtVx0p/AJ8SQAF
	EPA4mAbSWZwy8nCZ1aSKZbicbmVDZwhcDgBK+JhAmtLxAE//4h9lHmDYmvmaLmXVgg9l7Get709lHmXoZ8ln7e4OZcpn6GcFJg0nyWfb5R5lBhDiL9hnEe7rZ9/nH2XYZwFW2mChmkDcodyA3YHabuoCKjLykNl8S5WZm8WymCytHIhGPqSYj3mZgJwiDDxveS31CI55OGX0eQFA6AXShDOOZT30KHwuDD7w8WQ8ZwzSUPBcmQRnhpQiCQyW
	apRiZ55lAGoIJH0CXZQpCgyWXZOeZXJLUeMQ8UyUowUMllrSnmUQKnTVAzplCCDA6nTkABCQZ6CaTfUcTlcQwOpAmlvSABFalbDwXJkDbjhNXNWFZ2WVZDwGDJZb0p5lFSKCTPMGTJlclf9vDgYtI8QhBAyWMPBYmVolPDAP4PAAIJBnW5UvEGmSNSJal6JnbAkFGE+HZwNuXdekCQRiZ12XnmUmzAkGp2f/b13TLU8OlkYMln2UXNQJLGwB
	BF2TW9Og8BBcCwijZw4GMPAkmcLxFYSNA6QXWpNjlCI8IAUDZ2SVcEiC22z2A5BnDJZalWQiAZ5lA90kdCYAHFqUbJO9ZwyWcMSQ8WhFoKOeZbHEaZQBbQFcWpR4Z67rcsSu62qVc8QKJQFKQeBkCtwGAB6SFtoDEGqUWpOW271noPFMRUCiWpVnk51nVMV33aDxoESApVqVlcUAbABtESNmkw9GuUSZJmw7ABFnlbniwPejNEPujeNYZ4Zn
	deJaklqTmNq52m2SWttuklvbECmVJTgxtB0HWpUMloCaUIWeZSJsQALEmU31CCRsQAqZTfUUTlqSapPjmgUrdMMGLfIcTWrVaZJlInRMAAyZBNJqkgXTWpMG0m31HE1Rg2iTB9JnkgnTCNJrkgonzBhsKQJalOTyCZQpAluSMPBxZLVfgr3xHCedFghcgiIERmyMABdkZwJscvFQmkzsAiRPUwdgBGyM6oDwAyJPU4DwAGBPcxNgUHMHYE5z
	ClSVAySY7fUASSI8VGYBEEkiPFQDJJgt8hxJInRDbgIISWSsAAKDZwjTYvIJTj5lQO4G1pDw6JgEliJ+TKCfIuw5DzplB9c8EA13P2AdIlxnMPBEmt0FSm49B9ckpEV0/weQ8Eia0Wft9RhNI3QiAAGYQOo6ZQSWB5cIk55lCncRYVxnZgQw8Cb8TQKs9wBMCIxmBQSWAWqeZQ0QpAQLh2egmpDwQJgI0zplwOpUuA5qnmUIkw+U4IQBTA/U
	ui8MKrwEZwlQ8BAjPk4ATGTQlJKOd33wKP8E+GQcIrdGmAbSb4SQmxBPlIzqIXQ9KZwQIpw5I28lmICaZC0GnmXy9NiZkvLglAEGDfYATfL0XJoEbDeEGyL8KQXy9FiZoPAdInyPBZLygJug8BcsbAEI8vR8m6DwESucZ2FUBwOYnAhKCNMivigJ0iJ8TjJ8Q+QPJHQ7AgiWTfYAI7clMPBETJcKJJgJlUT2AUpi8w1JTyOEWHSwZg0GliL0
	NggS8ICbnmUJlQiWC5wpBgaWC5MK0p5lvFweDLidBGrxZ3LxkJ0jZ4zqGZceC9dNK7RMLPQBAQqWTfYmRCkHC5eeZQqTAVM0YCIcUgUS8EDZQO8/ZSfVAZm0DmQ9AQaWImcu/SoXlA06rwFt9hwlxEV8KQEBURdhrBYDAWqS8kDddh8Q8CisSiJMRyXsRQTyiNiS8mzYItxSIj1ZnCd+YATwIhRYAwjSHGcWkWQkAQRqHiUifEECbOqA8Aye
	EQrVLEQCJnVBECM8W0xVA5fDmq32BCOsTwFzEHLxIjwxI58vmArTNuQAAAOeZRSXw5pBoe32CE0E0kKhBdJDoQYnrAgFCpOBoaKhQ6FkTwABkvHEmQJ2GWEQ8NiYkvHQnoruE0wBBpiYkvGUnKrsDU0BuFwCDLidSu0EYJN0BWEDcgNhAyKdUdl0GH1HKZQYNSQCAvBYmJ5lbwyS8fBMTwZYmE33FE2S8VREV3RaA5LxWJoF0mQBdOAIkvFc
	mjhlQOgG0ngpnmNc9Sb8SDotBQCUECADLwXt9wiNKWcgDi8FLvAQzCl0Ym9MCJYKTIkCQqG8Z4FsKQydDSqS8USZr0ICXQtgEPB8JgS8na7sAiwBIB7NBI5EDyANzgQc9Cj8MQQY8PFkHGcOIh8nmAFtJDRPAQ6WnmUjJCcjHS3tJiQnlAFkGwca0g6W8PBUmJFnIiUnE4/yDpYXTXgMIv8nmJFnJiQnJBxNAWJn4BB8BQESBpFnJHdODpYY
	IvxO1AUFA20u8QhO6Bcit02RZwQjtU0bTwxQ8UiMDCSsTipUAgUFbU7xAE7PFyQML70DGUQGInwwdBJ0FAIOlgZSAEe1FdIbJZQwJBxPfggSInwRARAGkWftAhUpPQIGVwsYTqBUCyJsMa0JB40JFictAQeUFCgsAQERBpFn7QIWKSwBBAdtbvEYTnorLgEIUiIsTvUSCMQJIi5LCmGsIOwCIsQmCgcQMPDEmJFnCG2O8RAjVEMMWJgc07Lx
	BEqCZxTScPBIhLwAERuSHJMalQTSGZIXlhiXAVIF0lhnBtIVkgnTkWcH0haSCNIjTPkDAGoL0gzSZ5XE8wUj/CwBYmcOliIcMwIUlJ5lHJ+xDpYcIq02KyQUMCXuTGsQVP0BIOhDZ36X3PH8RyY8cgYw8ESaAW3l8xGNCkQlXiC88Sb8NiqdAAAvngB88f7fDvAk5GEAQSVnAGs1EA6TQIMBSw7T4mcR74hHx0waXBBhiEenTBpcDGGIR9hM
	ClwIYS13BmBfdwRgLncCYH53BmEMkxHjAUhAxADZExCcZ1xnMPFUmjDwxJwMIpRyBm2u8QROEeMX5Ww6AQSWSeAizGwABQaTAUsG0waTD5SC6wRgAJng8wVYwmF2zhj88P+LmmWOI6xQBwjSsPBYmQBtCwQiTSEGIpQhCJkIlguUrvEMSiD0bO4DSJmeZQTQbAwOIPQI0giW0PBImQyUnmUg9AyELAAICJZQ8QSZx7eeZcW2OGWCZ0Doo2cI
	lnDwHJldBZRVBJUi7XDCXAID42cIlgbSIiRPDp5lQPQIlyD0BJYH0+DzCG0NI31OClQLIuwqAUD0DJQitVCeIxRuCkD0DJaF9gFIDQQKBeIjpDgACQiWQPQYk55lTCNCmwBSBWEJayD0BNMBawMQAIQBXQ7TTAAAAZNA9BiUbDAB5DIQQJgAUilhCpNsDwkw8MSZDQUg9AiXceVsDwZ35S7yBE4g9BCUGVwAAAOTCJbgmEnjCtJA9BSTQZie
	Zfnj/+Ik/CVuEIX2JPRmIsRbdAwBCEgBS1wAhA4NIPQElGLsyGBA9ByVKCUKkCe9AQdnDRHnF00NDCPsWwJJ4AiWCiNPM0D0HCXsLwMIluJnDQQiPCdkUANA9ByWCgUqtAEpPQEDnAkC4y7yHE68PwEQ8HiZbAps5gUK0gRqjOqeZS4lThBGPQiWZAUEkPCImfL0/E9JpJnQVAIHnmWAnJLywJtO8iT0IWwbIg41CpR3QglMCCNkOwAFgPAS
	IgBqCpZA2NDwTJmHQMHYAUwNBQGcGX0KWERFArxnEvWAXAoAA2SdBGpE9gFLOyQQ8PidcvGQn4zqKHzWVUIQKSwfIj1kSC08VQKO8ghN0FwII/wnAZmS8kCk6EUWXETofBW0UQcQkztlTmxA6w0FfBUkHEsFEvVAmwDaMxAkLCMEcvGQnYzqIjHcAQGQ8EiZI/UhmSPMKQSZnmXu8gBNJc8pmdBnJ8QpIxw5dEIBTmwNBSLkQeQdAvL0GNr8
	XGULJJkQ8PifhPcBSRL1AN8idSiAJnRMIyStAYCELkF/jHv1BCVcMQACxWQG0pplQIVecgtgDpIE0g+SBdK/SwX3HY2ORd51O/UmfEoD5mQI0nqcIgSGAATlZwwjAGoE1QXSBtJbnABuA20H0uZnZRMQbKtMLwAERJoCnKOchvIZSgXXBNOQZ2/3EE4Ab20mZt0K2yd8PQAM9WQE0gDxFJ1nRX5Lb+Dt45plgVNcZwBrLmEw8KSafFADLvMU
	TcbyJHxCCQFqJRBApgDxlJ0BTlSSAAPxCJ0hRADxNN0OKAFyFGEBagDxSN0itCoACCDxjN0g8VDFAPE43QgQIPGQpQFIAPEI3Y7qdAIGAUvi69xhAGp1dBEjzjBb9Ch8h3SmJNcwBNIBJcR1AAF5IgQkSGkFJSppAxBHaQElKWkwI1xMJnyBBgSWAG+eZQ8QDlyGCaSYOmX55MCmgJtu8yIkLwkG1wSWBpeeZQFPD5IifEgBaJxC7yJUIwAH
	6WFHcQ9gSFEFYClxBmAqcRhgEhBIcQtgD3QoA5hO8whOE4QBAe31AE4iJTfETgIYTiIlN8QiJDckLDpNAhAjnHoNgJs6ZW7zCE1A6gbTBJYAaSKEUAsSEA6UKeSAokhE6EpfWiJcNQMBYSNsoJtsanQECAaTAUmeZQ+SQunrVy2EmFAiRDUBm27zDCNFNXZ0IwLwAmo78/zTIoxSJjQ6InwtBXLxWJoBUhVhJnUqMHxcKVw1In44hJgidTWe
	TBMBFEwZECL0KQNy8VSaNSLsBi4seyrMAAHO8xhMJLyGbEciPFINfWcGaljDcPBQmABspGc6ZSLeSwFvdAoDBJbQ8ByYJARaA55lBgUBbiJUOAEAaRoQpA0CAWls6hQtfAJkE70NUGQfApqeZS70JDU5USmOiTvyKPwDKjw+fBgJAXIAaoDwFGEQ8DiYxSKZTBADSwjTAVJETCo0bHgkjDwiBF8B2Jlu9CREYAN9ZxhqBJbsGwOeZaRnAW+f
	NAYIk2wbAVyYgZvcGiVNfwUzT39A9IwtjXpEVi54mCL+O3CbOOyAJmQ+A55lzvQMTDS8BTpEAmwUCZSYMPEImJ5lDpUPlncvBJYPVOQkjGoBcvF42ijuUNvwJnxNAAXO8PVkmmUI0gBUPGcPYdDwUJng8whugiRvbJYBUiJEUwSB8B1gofAFfBgDmQRqEPAYTCxvGWsiOiL8fSl0bLe5IPYQTP4BgJoFJW/tDvUcrGUFmS71EE4BcAV8MgGZ
	LfIcJGyYBplO9QRKIPYIlyKdb6SU0QQF1271HE3wI6RPZLsMggIJ4gD2DNKCAwNn4PcDNswBI+ZnBSMwzAEDcvH4mAF3MtwBKM8BrvUEI9Q/fA0FcvFYmNDwdJkinWJphA8JAWoS9UTcAGxy8ZjYdAIAShL1SNy5F4CjvWfBQyDwicXg9RjWwPIKKiD2EJcA8R8nIPYUkgDxkJoA8UyaQuwA8RZgcPBQmQFsIPAJBuRnOmUA9hTTQOoAbZ1n
	IPBJpAiWAPYUkwZynmXA8AJhTAcOldxnEPDYngDxUJ1y8XCeAUpFAd0i/TkcZGgpvAW02AEg9hSXbO60JwmeZQDx0J8A8eyf7vV+nzplbBoAICD2FJMA8VCbAPFsm2LqaWBnQiD2FJRAS0ZKvGdoM0gyEPC4nW3kSeQBm0GacvFwnUPgMr4BCJZsDXQOtAwo1AECAU4O9hApzAafDWykZ2wOB/BnRkpIMknjwZp0IGxbbA8GCJYA8VCcnmXQ
	tgPkoSI1fJlFQZRkBCIsgwbg9RiQnmXg9ggiHGGd8KBk0gmZMPCEmSLzAUou9hyEZX4bpxGk/p1sncyNL6QEZAeEml79Tva8bCJUJAIAa/L0fCJ8jgSZhPcBSn0RvC9uCZCeJvwgTDQp3CJsIAMIlv1ngJpdNZIiPIACyacA8fCEPiJMUiLnLFiZTiU0R00Hk2weAwBqAPFM3GRIRAIC0/L0XNwiBGBsDsweX0CeZb10QgGlAXI/bEEKlgFq
	4PUYl4DwQN5AxmzrAAiA8ETGQUDj6kD2HGAL4+D1FNIg9hCS4PUUl24KAVpvLLhn0FRMAUDqAWwiJS5QVioQlWQEInykAtBnQOoAXAID0PBcmSD2IiQvAuD1FJaw1G1MCAACk55lA2cg9g0QIPYQlAZyPiQg9gclbEgNkvFEnQNyeGfg9RTTAPYcK1xeAZRw8HAi/DAE5OJnO2UBbERbAtJA69BnjVqSbAO0gQM7ZeJnAGwoZACmD8JndBDs
	D1QKXJW8OAeS8WTd4PUDEENhdCUDEvWonD4ldA0GQUCD6idgC+PibRDSbB2NDtN20UDqfA5sAm4dAGx0uGy4dAJ8DKwOMLQDZAYDAWzDZzplIgWi5G8Xap5lJjwpAkjbgPUfTZ+YVJ4EtJwBdaD1AMwSDaScgPUaJQZyA2AYcoD1FWEgPXwCAqRnw2fk9zRqnWfkFHwlAxidEvVE22yJBQZyJGFS8mCYbGcNQPBQo0D1CSqBm6CbBNIF0mxN
	dKMBMPDkmSJ0KASu9hROzvYAI0xjbgOK9iSUJ35pIPX8vSJFMYwkzJs7Pg3O9iTkgnQHvGcDAmpy8VTeZAsmVGABp0JdTXw5DUPtWGdt4nLyuNxy8nzcHxAk9Hl8bSJkfyPsUimsAQcIllLyQJiQ8GiZItd+pJnAIwd+gJsOIhel6jpldgOBmqyDJiRtdBAL1CrA9AIQAXIWYZ4yByJcdAiV2GeA8MDdoPQfEHyYAAMCaoDwQN+A8ESngcdO
	7IDwhMeg9BJkAwAGlU3lgMOA8GSlbuxhpYDwhMWBQ4rqgPQfnD53FLCcrCWcg1xlL1QQbAsoZBQEnmWA8MSnLiLtrOpssUQGRNVsdgog9hSWAksBWlhnAPYQjFkB4PUc0yL0hQIBbEDquExwAAoUkwiWgPBEo55lBiIA9AwQIGyCwgNKAhAAayICfBkAHHHloKQw8ISZAUuyNi73FEyZ5sCmwMIPbsztkeXg9RyWgKTC64HCgkLjYQBqQMRk
	nQEy8lCaboZrE2xkZHwAB3CHU4eO60rrQPMcYIdnGEwBIiBMAGoE0rBE1gQKbiIFDgdAnBYDnmVA8wsqdE0iXFsCcvFwnWwkTFsppAa0omQLKvwDAcOfTvciZCx2H55ldHwDU4MlKpxnIryLAQFqU8MifF53zQDzGzN8GzA9AW4nNwWeZQFkGQeSVpoA9gjS4PEafDzMrQx/SwZLvWcA9gTTAPF03ABEr0TOXAAOfNwA8UjcIPBIxQD2ANIC
	ZyMp7BgivTgx1BBFNdMmhDEkZCNEAkw3ZIgBgJpUhiRdak4lXmrImSP0qgTQ/Wcg8EinbsDjZySMglcG0273JmQ9VAFErAAWQ2cR6ohC2EwKXARgBFACYNBLRhCIQqdMBlwEYAJQAmCpSz4Qv0oGWnQBABwOYMlLNxAgcwVhwPANIAFwDmFLEFxzA2HA8AMgCBBYcwJgeHMEYaDwHiAEcHpgfW1MKo1hmSKeO4SZ/t6u97x0fAR0FAHg9RSU
	In03ztRpZGcAD2USBliA8BdgAwoENJHigIxJ5IDqDQAXACUAMwAjAZ0AXUwcBGjCAWiGEJ1EAQdIpFAyTeMg8GjEbAkAAyD2FJUKBgFvKhBsMwoFbeVEmwBSE5QSKrYB7vcmvCUABAFrAPYA055lAGhfEANsmuoBLOXo5Zt0CAcCT1/nEu7Z5ZrvbAJuCxLvdRYmJbRPbMwDANKeZeIXI7QhB/FImwJSHGEA8Xh0BwADlv9KgKMg8bCmAW9O
	7K7sIPGQxkDDdAd+CKZnbwgg8RAl7EIBnmUDIrwQfA4ACADxjJ0A8VSdYURHTIg0keUEUwDxbN1B3ANhJ4wAA5Qg8UzcACK8OwAB3AJnAxADaAEQBGgA9giVYIUBTVwACtUFIwD2AJbf9hImvxFcGwaToPEbKwFwEGHkEHwMbBBkJfxrAaDxCSoilCBsA1VVnFRWAZyHQj+MEAqBnGPsCWADUgdggUJHIjUl5VcU3WHabIhkGlRcDJMg8QCe
	VIQA9hCVY+ACInzuZNsIAPYElgBsOmVA6vBt5pUiDCcDCJRUhZ5lbAQHDJTQ8FyZAPYElSJcIGR0Ipy8Axtg4PUUlqRcAwSZTvUQSCLccgMEmW71AEgiLH1kFr6tkGciJCZ8AXxGZJ8FAGoBaADxUN9knAc0IgDxTJ8BUjBhdH4FQ5tAgl5yKmCkf3yodb0a1G3kbayIbEUnbCUD9hCWAPHslb7u3GhsmQEBawNnItwmAQEQAGh8CwAIWpsx
	IhAEIPANAyIGoJwAVQphA2/67QEv5eh8EgAGEu2156ClARAAbQhMyuygwwFL7WEg9hCTdCcACAFuAVsiAwTTDgMF0wsDBtO4Z3uczu3g9RyXfSsHI2RQfBsBHyIfEHSRbBsJgp+jnwUqMPDImZDzJIQzCciZEPQQTg4CBNIAarbtIgd87PcjARABnBQBUoMlIiwMBCNscCACHhUv8ChEL2QeAQBqU8N8FwNRhDIi4ZzsCgEA9gDXbqps6i4l
	HZAj1EtkGCx9Fk/cCvRTI0STIyZKCJZ0V5a200psHwKUV5xGIiwUGDIWmwiWLpQBA9ef459P8LSg/QwAIuSUnAxkAiaUIAPg9RSVl50m9bq1IowrBJNYZ6niWdt0BASY24Nn5PIJlAZsMgHA8AsobiCAnFUA1HxCfAcjRJAB5Z/0DCJMNgSWDCbw8ESZb2vg9RwjdIdkBwGg8A4qZGRkZXw8fD18ZCx0LHRxbOm2OABSIkY8EWAsFAxs8H9g
	j/AMJ0Uy/F7b+J+UXAGfbOp5M5QLZlwQkiSFg0RVaxAlTDwGLvUcSgTSIPYMIsy1I3RUIswmdBAJBdMG0K/wAE1O9QRPpBIBnmVLECAVFQMPNRUD79QYdC0oJBmk4HxyAZ5lIPYibIZsEgPQ8FiZBNaHE9AP8SxcAnRuIn5GQN8iNSoFImQlDJMBSoDwQNvg9RiTAPYMlGwrDYPrgPBAnV/wGGF0KoPocmBsawEP5ONnJiQzAeD1ENMyNjMQ
	lzY1MxAoNDOUNTaeA0/xJA3SECklijItnAZVD5OmHwFzImSpIhVB5CL1Qk9louQi9UJPdCF/ItDwOIwhBRCWBdBP8RRNIi5YChAqdAEInvcZIn73BRDA8HUidE8j71NY9wwlXFgEmmWQ8PVkPF9GeJkUJcSENOYtFJZkTuf6j/EclE4BFJaeZSJ0RiWUmSWUtbYCWJki3C0LEPC4mXLxHEqA9AjSQZokhDMOF9JS8kycMvIUSzDwBJkZ0hJU
	MQ7dZyLyAUgb0kGbAWtg8GDOHUSaCGDwQs5g8GjOYPBKXQBwXQBSXQB4XAAHWs6A9AzUgPQQ1SLUSQA0FJYQ8PiZnmUy8tCf+J4fZfme2Gft7g8muGdT5YPt2Gd/59vnAFaA9ADUgPQE1gtgAG4AbwQQAW4Bb8vu6++A9ADWgPQE13QVB4D0AJCS8aScKCV0AQALkvHInJLxrJxT5oPu+Gd35f/lAFdg9AjUYPQM1waUCFUB1nQBAQBQDGFc
	AAqXwPcDNILvBmHu7A8sTAMHlAPsC2Bg9AiQCCe0AVQChAZ8CwRy8bScAnUoI4wvBHLy2Jxy8rwtrQEQTQ0UJ60BEEYHFNetDRQtrQEQzQ0QK6wBpAZvDZLyoPQaDvKonJLyjJxL5UPt+Gdv5P/jdBoCGNJg9BwnpQEYRQ0cxQ0cZg0yQnQaBuoPKmD0GJID6nUaGCukAaQGLAUSXS1lCxQjJLsJnmXgmgUgMPDEmY/xJezVApmP8RROJNRQ
	XDoF1oD0ENeS8WQj5EUDgPQU00LxJFx0TQGTvFAAAgbTYPQUk55lgPQElQ3TfGccZQrSXDsAAZRg9BCSEPB4mwXVYPQMlQTUDNJcLQOUkvJAmwlEAgIclQjUDkYCGJRsiFUMl1QNC5YR1bhnENQS0Idnr/EQI0ROfFcFAG1g9BjVQSAi/V5cZFknfKcJFJaw8EiZFwSeZQRtJAYjFJZVItIiPIwkFDoiBE00pLV2F6SZIzRYBZmeZWD0GJYv
	Jfxrbg6cZ3xoIowyIkS0AUDqOmV0AnzLC/xnUvJImgFSQ2AQ8PifIry1A3LxHE+Bn+wErCcDEvCAmgBUIu2qXCPuShSWIvRGJtU+ElSgtOh8BwIy8pSdQIQMdgGeZaQXBzplMvIUTkDqgZ5tNTxMFKT8nhHpOVSFIu7RAGoiXMQFIvIBT0DvP2V1HwhcMSIc0Lx7A5LxRJs0InwWlC8ABfFsmmLsLGGO6wYrkvFImmD0EJVD7SRhpA6V1Z4l
	lFotPCJ0JSgXMU/yECfMEbxMI0Xe36wOBHLxVJsCcjROaViZdw5y8nwn1wFy8lggGNYBb/KktmQx9A5MDdQOZ0ug8BUorQPybB0HoPAMYY7rByuS8vcdoPADIBC+A4/yKIwV5A8BkvJA32zkJvxOdRJ2LSxDI8SQKSQFZDkBSJlg9HTXvD8DgJoy8nCdfMcHWJvDm3mbS+dD700mlyQEUQZv5/hn7+MF068pZAq8SwAIMvIQmlqYDyIAawTT
	BdMG03uYAG7mZwfTkGckZNl0EQMREIKYo5gq7kcAbyJFKu+cSqRaDpxnMPBEnHGAkGdj9xFKBiMBIwSHdgcQECM0JAJtAGpY2HwcAQBredgi5SWQIxRKZASkGAsy8lCdeJpZmoNnTewMJEwXB5ZC7ghhzup/9xlElAQQl2Pvf/cURK8MGJIBUkDyCGF9Z2DwQotMLQwsMm5bFJZE1QaXgJpdZ55lwZ9sJAFg8OKKfYXvKTwEfAkLoosBaqzq
	FiLcZ4D0CJUi9FxkFlxzDZmBnVLyrJ4m9wVKAW6S8hAj1CRtFwwjHGF86CLmJk31KGUJ/V8TSo9RLXwCLjR9pDluEp5lpEJ8FARS8syfYPDqjhQP8yTcxnQNbBQBSqsBa21rvGwUVeS4VT5EVJkE+JlS8oydoYwUAQBu0vMkZCd0BiIvckSZ/UQ0A4SZYPCij6UVFCfkDXQKA1KLQPEOIqSDnnecbCVsSiicmjSvAhLwwE0V8o0VLy2tAlKM
	FQkA8RIiAGtg9AjTAWusPSSEaAMHbhLwgJ1GwNMfI6ZoFJYkpGhGApM1Tu8Ik7QFA03jFgRJ5HwHAwBrIPBkwiLUSCKkRzJcMn2LpNSxTEgEl55lT/MUTay6bDkCUGwfBUQl7YUUI/zfBLYXMSowI9wjrGSEHyAE5RKPInx1ja2e3DZvrhLwgNy3LpQWpCsEEvBAnABSNSUcXJcNnWzqLrxKMpwFdFUDEvDAn4/zJOwzfDqmeBLwIhQ2PHQY
	tA99JgAiXKYI3ARq3Gds6hDwGJ4kVIM0jgeAmqQMBPL02Jiv8xAnXYPyJhytxNIipK0BmJgXJLS2fqcRKmTFfAQGDCpAnPL0WNgERAIDeJkS9UDbItW7QEQndDEqzBO1VN31VI52VO/zKHwYfGkBWo9MIqwSBwFuFgUy8hRKgZptSUBcCVRHI+XjokUIRCd3Z+/zGCecC2wSbBQLGJtS8kCaAyJS8kSYDyp8ELQOAUDwuKZ1Di8i/HiGCp5l
	bAQL/WdA8LinQPBSoqrqDGB8Bf59T/QkFEx0MXQLBVLyAJp/8gsgdAwHeJuS8USbf/IEKvQvAlScX/IdLexTIxWzmSj8BSJsPQkUlkDw8KCAmkDwUaBElgKYBNJCmGSECAXSQPBSoG/0FE0GImSqJ/cFAWrdTI0AJFzGQPBwoEDw0aAw8UiZLiO9Z4dGYPCdxQBs50Zg8J7FE2xg8J/FgPBgxYDwwcUCT6dAgPACBE5k1wklhyRA8HFcBwAC
	lKCeZQZLJWoBLAVqFgZt5iDwRMNkEQuAa2vrSeYg8GvCDxBjRnRxDJ1nYUanQGDwfcQJTWDwHiP0lWQhDQFq/2sFEB8FUeWApAFKjutMDASU/0yC6vZhIoY53WdcAQiXkeZg8HzEOmUAbGfFHwYUIh+0mR8F9GlNBJZcmdRpIjRJNrc8z/QIJ4QHpC2calyXI1SCL2wFbDcmzjPP9Cp0CAF4mQFqpI4iHGR0KQALcvFU3Z/xAyPAmKGYBNIk
	8whqBdIAagbSB9IJ0golJWRkTE0E5JnK9xlKC4QCA872AEsI0yIsoQIM0ABsDyWEPX4iX/G1v0wrXVUUKxY2D/UohCAIX/EGEADwAmp38RAnbDcijc4EI/yvAwBrEvVk2iK8ggdS8oCbQ5xS8kDblGwBmFLyRHRsIwykfAMABASWkvFEm55lnGcFchDwOJwQ8FicLmEiFMIDBGxM7BYkIhWMBiz9YZgkRIWcJAKYnmVP9b4kBpMjNWbZTF0F
	mBDwGJiS8YhMAAesmHLymNpy8rzaI/QmBdtJEABrcvF0VgR4mHQMCnLxeJsBUwRrTOsnYRQifMcpdIko9IbvDa/1AJ05ECNd2PhFXwREGgismgFsJvcFSIvsAbzvAug4ZRUQMjQBIlTBIvzDIlzBAp5l7/UQKx+MN/AEJywFAvxkHGdkX76YmA4jnDwE0JwfkczqKyKEMwGYE9MSFykTTRMoEwAKDpYTk0CanmXCmxplHpLjmxKVAiKBogIQ
	AWyL7GQoCQTVHJUG1BPTBdWiZyJkkAKYZy/2FIRrZAcHnmUelIDwGyRBpCJEdgABQPBUoaKbwJkDIi71EGoCEIfyGC5kCQWYAWsE06vyDZVKZFUgRE4A5Jh0S2xLBAjTDNEAbI8l3m3g8CP0TBAGEQMOZEskPJACmp5lr/YkVozjEHQPvFwDAUoR6gRaIpz8AQRqKGB8wyXMey2EBqQJ5jTP9rRBZTQCZD0izHucQSbsTQLw8YRCjSK2uJ2M
	JM5vmBMtfAksRgEv9yTESwUTkwBqgpujmyJEhXwdBMSYb/cYSnqEFiiuAmhgRJxUIQWomBAEBW7Q9CQ2Uw6WJESRIjxHBAFtEAZA6gUjTJEEDpYQBVLyjCJcgAGYnmUFI233DhMgF4SKlmSgmmQ4ZEd9v68lbGMiJLMsRBEy7JEsJgOv97x3OG8E4POIJWyPCXLymNhy8nzYGhAekvQeCEiYMPDEmBDwBEoFdIlVAQROIAhOIrwtBzplivYN
	SEDqAG8ihUR8FEgWAfACalYnfp+aZSKtzCSNWgjMWgIWk8zqKo9aC9MKLU8LCJYLrFoImRpl45kKlQIjgqPkWgZImATVFJUG1AspRQtwJYRXfAYDnmUMI0KjIvTqBcSYFZOCmaOZIiR4CwTTBdJPEBeUQPBTpDgqdCBU4AKaBXIyYGwBA3LxWJotKl1PpHU5BmdPxCVgrFGWtp1sJLRZLfwJAQiWF5MiliOomGRrBp5lQPDUo7DwDJVrEFyQ
	BgBqcvFU2CAQglYQBSMSahYQ8CRkkANImFDwBEoSWBgHFZMF0vDwDE4E06w9AQSYFJd0Kq0qOhMZDngnTgX28yj9tfkjbJMIAXU8Z0VnZmcYkCRMGQ64mQRrAWhy8ZCdCtCM68DxAlyaAWSZC9KE+RNgGiKkeSLMbAWomYCbEpMLkiJcTweeZcOb8PAcTeJnIlRPAqgREPC4E5AAIr1wI2w0AplNbAzWLg5jqJkSsyaeZRYihMoBoFDxBCJU
	xQegBdJDoAbSRKAHJywgAwiWDJOeZSKMWgEVlYNnIkwkBABrCtNg8RRcMAT4mQFqkvFEI6Y5C9f8dAGC86hCIiRJJCR9BwiWkvGo3JLxbNx8EgAZgqCeZRDw2JmS8ZTdpKAQ8FiZC5eS8bze0PCgmQplQaCV5WhnoKWS8VDbQ1QDAFN4mRBoruiS8VjbAPEEIAJ14PAMYWB0gPAUYGFcHGBCdHBgQ1wMYDF0RWAyXANgKnQ2YDIQNHRHYD10
	UGAtEFJ0b2BTXANgSXRkYCYQVXRvYFh0c2AhEIZ0gPAZYIdcDWBrdHpgbEQEACxpdHNgFRBwdH9genSA8AZgDxCfdADxBmCgXARgmHQA8QFgBhCATI9MDWqD6oDwAWABaAvogBCOQiVcoPAYYCZkcAAfEOwCdHAQAXJvYARybWADcgZqoPAJYaEQj0ITXKDwBGABbEzsoPAAJAVyXmGdEI9CB0QPbAIGVyyIQv5MBVwFhAQABYDwDyxOEAly
	gPALYcgQAnJIYAVyRmABXAEFA2HCEI9CCVykCQEBQgIsfBBkAgGg8BdhRGxUEgAv6HKg8BFhURD9SgJaa2CuEARyKGAHciZgCXIkYAtyFhABcmBhpRABch1gA3IbYAZyGWAIchdgC3IVYA9yEhAFchFcAQoPYApyDWAMcgtgDnIIXBkABgdoK2ECaCkQEXICYBNyQGEBQiMQ/2gBSCLMKgEM0wvXZEsiNz5MbAgj3E4BMPComSI8WiRMLQTQ
	8QBNkvHUI4RsdEcBCJYLl2RKAkFAGyIClEcAEJLxGNsVEAFuAGy6MKg1Ee0BJA4gGebK6gdhAmyS8YTfAUpkQAYEEAFMBHQBTu0nnDMnpgzqLvRcI0x8bhEL16wRIqTaAYCaHGetERBUegMMk5LxxJ8iH9SS8VRGoaiZbmgE0k0MmySELQIF0hDyHE4WXJsSgDYBBtIKkhJIAWQeAASnFwtopRcEaKMX6GihFwFonxcDaJ0XFOEeNin/wgTw
	+yOsz26CS2wiBKhsJHQfImQiBQ6WgJqy8UCZJGxlAciYsPIkdFMDyJjQ8gRObJEDMPComNDy/PysBQTg8AMqcPBAROIBmJgAbSOfTeo6ZXfJwPATTHEBeJgBakcF2UNnInxpAdLxQNsilCYHQdoQ8PiYMvJQ2xLwLA8AbgFs8PIUS2PacPBUmAJtMvIUoBJ2CA0iFkgzBaCaFpQw8ASYIvz7bMIGyxACazDwJJgGTJwByJiw8CLcIAnomAFs
	xPMFSQnUCtQiDEcFB9II0gvTDNIi5C0F8PIcThDzHE90kQOA8AUqAmunB5DwcFwtZQcJJ80ACifUAARQ8whOcPMMlQZrJ88AUPFsOcwAApDzBE6wIkxCAek5ZVEnzQDwKZwBMHUC0F8M8PMInAwBNyoBbGQYZ7mw8JSEDQUDawbTC9QQ0/QazBoC8xhOEPSsGgYQkxsqBtMBawmfIVDxeJ4GMPRXwNIF0mQhKVYD9BCEDWQsBwMiFpSiZ1IX
	UPB98ahEOQjYmBaUcPQETbH3GCO0fwIOlhDxUEwCA55lDLcKtnxFAw6WFpQCbSJskQbImBDxCJir7XBcDHcyAWp7I5xmBSlcj8L1KABAEkJIoK0UUEUF2P+9JxwAsK8SM0cQALwTEEgNJAC/r6SPECYDAAAQ//8RJBIARwn8/xAmAAAZjvz/MRdcAxAEo0YoAL0QA5yABgAQgJmPIXjgA3wGAVYAGCQqPQBVLT0AUy09AFItPQBRLT0AUC09
	AE8tPQBOLT0ATS09AEwtPQBLLT0ASi09AEktPQBHLT0ARi09AEUtPQBELT0AQy09AEItPQBBLT0AQC09AD8tPQA+LT0APC09ADstPQA5LT0AOC09ADctPQA1LT0ANC09ADMtPQAyLT0AMC09AC8tPQAtLT0ALC09ACstPQApLT0AKC09ACctPQAmLT0AJC09ACItPQAhLT0AIC09AB8tPQAeLT0AHC09ABstPQAaLT0AGS09ABgtPQAXLT0A
	Fi09ABUtPQAULT0AEy09ABItPQAQLT0ADy09AA5dAQAtAAB/h2CpnBAYIVs8jAQUgFoDUBA5J13qEA2BWnMVBGkSdGIFdGU9aWRsZQAtTAAFcHJvY2VzcwAvXgBlbhK4ci9cAAUqdW5rbm93bhIskwIAJTJkIIAAAiUwMmQvJxAAASAlZDpwAZAAAS4lMDN0BGgdBBtbMTszMW0T1WYgFN1vIBPgawk6ICVzICVkG1swbQpDBXJlcBN4Zwlo
	dHRwX2ZkKCkKAAAtTAAFAENhbm5vdCAVCHQAAiB0byBzZXJ2ZXIuIHJldHJ5ACAgJ/QABj0lZAoAU2V0dEUaILAPCGRlYnVnIGxldmVsaAdgBJwOA3F1ZXN0ANwUDkludmFsaWQgc3BlY2lhbCBjFDx0BTogMHglMDJYtBUQAVh1bBgV4G0GOiBmdW5jX2lkygUgc9QPAWhvc3S0MXgOYDJ8HSkPAXJlcWkQIBL0cAJyY2FwdC4qATog5Bks
	LAGUAQd3YWl0X2Zvcl9jF7J3LCASZHkMb3V0ICVkIG1zIG9yIHJlFPx3AyBmcm9tICipAC4gVCwCvB0yKQIAIEVcBAExIHNlUSBkP2gEIEc9AjDAEj+xBlQSoH0LdF9ub2RlIG11c3QgYmVIAAt0d2VlbiAwIGFuZCAyNRJQlQREZWxheSBjpGZMBARuZWdhdGl2I/zhAUx1YSAS0HYLRGF0YToga2V5PSVzIHQoYAELPSVkIGRldmljZV9u
	dW2XAWF0YXEEZHQKYAIHaGFzQmF0dGVyeYRtAAQAAFRvbyBtYW55IGRhdGEgYnl0ZXMAfAsBIGlzIHB2rAIBAABCYSdsAQNwYXRoAES8DGwBASBkb2WwBQttYXRjaCBhbHJlYWR5IBZMdwdlZCBuYW1lAAAvZJgML3NlbGYvZmQvAAAlcyVzSHkHAG9yaWdpbmFsX3R2Eud4X2Zk2BMrnAF8EQJmb3VuZEluIHSOBSBmaWxlIGxpXYBDEzqA
	ZCAU/YUgEkCAAS4gZmRolZQreIYGRHVwMi4gb2xkvAsCLCBuZXfoAUNtdWx0tA4BQ2xvc36PZmRgcQNhZnRlciAScIFvbk5ld5CLeBRcC4AGA2RlbGV0ZW6KLT4tFABoBs0YIBRIe7YqbiB0hhIAikMgTm90KbEEANwFAV9udW2wBcQHAQBLZXnAAhUEh0wHAmluc2VyQoZzIDTsAXkMchJEiE0FbRQQiU9jdWxkhAkUcYMgwAMFIHN0cnVj
	dAB+TCVzJ4QJAmFybV9wkUZubEkpKABQJhN0ingCA29uZVNob2Qo9Jh8TQRwcmlvcml0dE0EIGZvcndhcsQ7AhtbMzNtTBKetzNtvLgDNW0AJXMgEj2hAHxcAUhUVFBtUTpDNXh0UrutQCVwusVhY3xicQxobc9oXzpvZmaAMlwKfAdsOG8DOiAoRGOOWykgRDIBIFdyb301ZLxdeLtoDdbSCgC50jo+mAB/JFVQbhdUhxAO8IfAdQMsIFVu
	ZXhY0wV0ZWQgZXZlbmklCiAPDwFoZXgSjIZgLmmJPbnbIBX4jxOQk+wCc2lhbmTYAVgeIAauATognNMCc3RhdGWUK3zgcg1fafAJpATCDF9jEmiV2AwrZAArTQBzeTduq5NiZXL8NBQgkycMBRASZIwgDBwFIAMcASksBfETTnimCnUgdHJhbnNtaXRSZXORgnUvnAY5fAEgTf0EUKBzlXlhYHhQXllfACYkIW2mYeCIQAMBQXJtXzHIAGDq
	EgmcAPBYLkEBRrB3mAIWRpclJSJsJQUAAEdFVCAvZF3HXyVIJAE/aWQ9e3Nvbia8xJ4qZCYi2CcKaWNlSWQ9dXJuOmdlbkgABl9tY3Ytb3JnOidxADoTGJEleiYxJqAJAT0lcyZsGgE9JWYmZN9IGQIAJkMlZBKiuSZFEoSbFVSSQQIgdHkDLzEuMQ0KEmiTBjogMTI3LjAuMGMCDQoAI/0sX6CDdQBfZQx2rF0vJBECUXVldWVMKoiQgY8g
	0BxAkQIuIGxhcyAJYBLkBwE9CgAAJ1wBAmZpcnN0iNRsCTCEATyEAipIASjsBgx0b28gbG9uZwAAABtbMza+rDM03QA1fACAhwR4IAAoJXMAInUsKUwSLSwO6EHEMRZLoTogR1zVDFNFTkRfREFUQV9BQk9SVHG7ICOUIVWbbRR+mC4gUEYDZ2V0IGFiIk4raXQi0CogB48BU3dpEuGdd2gvLawBeAggCCwTSRB3UM4xGAF5vmMn7C1wCo7R
	MW0QCXCcBTogcmVjZWl2TRt1KBAX2BwkaC5upiwgFxCdAl9sZW49JpQ0MkAB/NMCZmFrZSDDD0FDS2gxYBN4CQACIGJlY2F1c2Ugd2UgYXJlIGJ1c3l5uAByAy0+KOwCWDwSfal0FMqoLT58A2wCInAl0e4A3SAgFG2mACisCMACKjcBJXMgaDzM5lYBb2YkeCgj5C8jVCScEE31UiIAN3ArMLAA9BcoZAEBIHdhc/QYAiAgIFN3J4cHYWNr
	IkwnJ7MiIHBhIlAjAmQgb2YgIuQmBgAAICAgV3JpdGgNKWwAJgEiOiKgJgJieXRlc2ATF2Ci1B2IaCZ+IDAKbglEaSQAJNw6BCBBQ0sgKGdcWp1iWFARA3N0ZWFkKUEBIDVUAol1XylYASPNIkRYoARPawAAMDYAJz0AIGALzzxDQU6MPIgpKGoEISAusD8DZW5kaW5nehkARCN9KWQnmSYlVBxcJxPjsHN1bdjsDwoAMDEyMzQ1Njc4OUFC
	Q0RFRmxAXATcdyPhPnASFbRkVAooXAAixSppIv0ud1ZMbWVQA0wDSAAFJXMgYz0lYyBYkyUkIGAsJ7ABUAvIGkwFSAD0hQQgc3ludGF4JLBAaR0KKKwJKmgAXgdVbiPAMAplZCByZXBsYWNlbWVuEiC4aCEi5CEFbnVtUGFydHNsASsNBzGMONwaKfwCAXVuYXKEGJxFmCZgRiVoKmAgRB+MAyZgPgRNdWx0aXBhEsy2Ay4gVGltZYWVbnpJ
	bXPgBghQYXNzdGhyb3VnaMRKjAjAzojiJ4wIAWVkLiB8BSLMRIAGKGAMBHRvICVzLCCcZyLoQyRsIVwtImw1kjl1bSd8AShsBSd3AHN1bTpsAS5cAawvAVRhaWzwCikMDVYSdHJIA5gjAnV0cHV0Iy4sJXNwj8QJJvQjRxQAAENcNSKcMhLNv1MSAK0ldEYkaEgECgAlc2FsbHQJEsKwLiDNMzooLjBsbCJoSWqTOiAnbClk+gJvdXQ9JZQD
	VA8vvAADdHRwOiBhJ6gvLMAAIiQ2AmJpbmVkKZwBVBUBUG9sbCIOL3VySAPfh1RpbXQTeDtsECRgR0zeKVwAeE55d18TTrpuZ1wUMFBOYEgpzAAi4D5QkLB1hIIsTQFtJMg1AiB3aXRoIgAraEokmUcgqHFEl20fbUl6ANxPnbtfkDBcziKnMHMgPdgYKZwXNIQAnGQ17QBS8NhUP2x5BXMgKHRvdGFsdF2dASnP2XR0cCUxNDp8HewKFCq+
	CgAmfDz8AmW0RCJ6L3VlIuQlbCggEMg2QDNRUAApbDoiEEEnVBRUJW10IFvqZ3RoTFF852INIHEigCYitEIEZW1wdHkuICNgJSKoMgIAAFdhciJkVCfMRMwFA0lEIG1pc4iPK7kAACO8UQFfZG9f5UE6LThFAW5vZGVqCyVknOFz/Gd0aHiTI+gjYAGYEawEaCcoDBjc7DGMAfQNrVI9IpJEZXN8QQReMDEgLi4gVAAEMTMgKC4uKVQBSB8k
	rDwlbFYBX21haxZ0yCMQJi+CASBmXnxlZLiBAgBOZXh07CcBTm9kZSNoIwFvdmVyyD8isDdY1QFfQ0FOL/wL3GKcGjAcAXgYXCUDZmVycmVkIxw4kSQgJP1ERlgUdOwvUAEDaWRsZQoAMVwCA05vcm1hbCjwBhL6xWxlEsjENgwB/BoJQ2FwdHVyZTFQcm9jRBIqQEwi6DBwsCYBPiApEDgDcmV0dmFsJWA67i8wIOwvgzAoLiq0MNyztC0u
	DgZJX+kRMiz0BShMICqNAjqsqSwoB5A2eMQCdHJ5aW4wMBIpMAbwDSAHTwFHaXZuoXVwXC7sByMJNQAofAK891yPJfxZtTgzeKoBcmllcxNs7+0HIBOUyfTMKAwBLDwCXEMDdCBaV19TLOQwLzAHKGIDTm8o9AdQrCOoSUgUKE0NYxACHM1kE6C+Ak5PX0FDEtDwKvwAKqgAKgwBzRQyIBusCH8gYXRhyD2uNmllItsiLiBUlDcDIGFnYWlu
	L6gDzF98B2yZAk5vQWNrUBMtjDQlnCEnVCsmIDZ9yT1wWTNgSSASHAEpLAwBbm9kZSWoJyACcElQxV+ebmNlI50/dWCE3GMgBygDmE8IOiBVbmhhbmRsZWRk+xJB2yAjcEksGWcAIAZQASBMIkZBbH7VeSASSdNpE/jhA2QASW5pdIwBVG1ZMQAk3CQAAlpXSW50IHZlcnNpb24gMi4wMi4gIkQvCAAAKkR1bW15KgBeoBz4Qi79BwAwbBJU
	xNiTAS4pLirsxRrwAMwJOoBNMTwBcQkA/BAn7AAmiC8lcCcp3gAwNEjT+A68AKgILwwBiIUnKAtMDFTbVBcBMDAgMRvcBtwVKDhpJWhTKcgAKdwBsAwvzAAuoAJQL8wmBCgwWzAxXSmIFLz9TAPMMQMAaW5zdGFUWEkCACPUcUABJhBjSwEAdW4oNQBjYQVsJC1JRBKY2WAEKQwAAQEDABYS8eoQPQEAQHwDAEFAcEBAQAIQQEAUFEACABAC
	wAAQbmkAEBACQBBQEAJQUEBYQBACEBQQGOUQEEACLAACEBACFBAwQEAAbgKAYFBVUFhQYAIQAgBQAFgCKNBgAFpgYJAHAUAQEEBEAQkwagJQQEAYQEAQYBBLAgBQFRKK2EAgcAl4DyAsDQACLAAAAgADAAAAHIQJCAkAAH5+CQB+AH4JhgIJCUQCUgAJfmgEYAG8BFQBAQAACQNBAgBQBXQDAn5+fn4A8ABhAQlgAnMDCQl+YASIBisQAHgO
	YQI9YA5gAScMAHQOsAG4DXwRQBFYALQC9ABdDwlYFfgBKRwAvAYpQAAgAAAAAAAAAADeKAAB/////yLoI/wAbAEGdIQAAKERAACAXAACwRQAAIhcAALxGgAAlFwACpkgAABUbQAAWS4AAEhdADFdAKBcAggpIgAAqIQAAH0YACcAACpMASdTAAAAgCLkKhoyAbCPGvkDAEAA/AAEAACAZgAAcE0AYB3hKFBdAEBNADBNACBNABBNAABMAALw
	ZQAA4E0A0E0AwEwAAbBlAAAaMSigTQGQTQCATQBwTQBgI9eVAABQXAMCvSoAAEBdATBNACBPABBlABvBJQBMAQLwZAAA4EwAGgEl0F0AwEwAGnEksF0AoE0AkE0AgEwAGoEjcF0AYE0AUEwAGsEiQF0AMEwAAtE0AAAgXAACEGQAAABMAALwYwAA4E0A0E0AwE0AsE0AoE0AkCIAXkkEgF0AcE0AYE0AUE0AQE0AME0AIE0AEE0AAE4A8GJM
	CEwAGuEd0F0AwEwAGlAdfCcNR0NDOiAoR05VKSAzLjMuMhoQDEACD0xpbmFybyBHQ0MgNC42LTIwMSKBOilzAS4zIHgBAAEwMjAxIChwcmVyZWxlYXNlKQAAGuAyShKQ/NxBTQEAIhCPGpA5GkE+oEwuTAI3fgBREVQDMQgA/Ad8SVkDgDZ9AdWdBwMrfAAaNA/9ByEaxisAACt8ABpwO/0DaZwDLP0AeBoBIgAcQUEhG3gFLHwAKn0BbTF9
	AEAp/gDBFEEJACz8ACr8AW0yAC18ACp8AXxKOn0AJZxmOHwBvGosfAAiDF38H3xzLnwBAUgCAAD9A3FwOC18AhrwQ/0DLXA9LXwAKvwCfH4ufAEqfAB8fy58ACp9AekaxCs5fQHFGkRHLX4AiAEo/gJBJjD8BSp9AYEj5MYsfAQq/AZ8Zy58Aip8AGyDOn0A8RpESC58Ail8CHygRBQrfApoKvwrfKU6fQCBIzRWLPwCKn4FHS9UBy3+AUgE
	KH0EmRzkRyv8ASp9Cs2dAwErfAAqfgElMzB8ASr8AGyrLnwEKvwAbJ87fAAiYKwt/AIq/QYlGqRMLX4AKAYofQOxG7RILH0EqCn9A40b5EksfAAqfgLZVzB9A2Ap/gatWjB8ACp8CBpgPi58A2gS/Dsa4EgufAAagFP9A0EalE4CZ251AAEbKFkCAwAuc2gaSC4BYWIALiJESg9uZm8ALmR5bmFtaWMALmhhc2ibAXN5bbwABXRyAC5nbnUu
	JpxLKjIAX3JtCGx6BwAuIvJSAC4b8CQKLk1JUFMuc3R1YnMALhtzNS5ybyKMsAQALmVoX2ZyIkCvAi5jdG9yWQNkuwBqY3JcDQZhdGEucmVsLnJsEiJgTAQuZ290AC5zrAcazSUuIuCgBGVudAAucGTEEwRhdHRyaWJ1IvKxLm0j6E0CLmFiaTOcTmwrIAEMABqxVgZY/1iWGAQAYOUHBEABeKRosRpAUnwEASwBAABsABoQXxogWEkEAH0E
	CMw7bAJ8BBpQRG0AdFABfAP8BGwBaI9sD0oDAJBAAm4AgAUckF1sAkoCABBABmiAIkxKbAIBEAoAAGwAGtBdSAOJAAGWAAAAaGkagFx8BBpwXGwAGnBLfAhoBHwCbAB8SxpAXWwBGlBdbQAgXBZsFmwJ/AUakFYivEp8BBowX2wAGjBf/AN8AHwdfE18BmwjfPltAHhgDswAfARpAV5QAPwEGoBSbQBwGowjXAJoAGwdfQBkUAD8BCLcIW0A
	4GAgXAJoAPwJGuRg/AQinCBtAFBgBMwA/QR2XAF8BHwYAfBmAABtAOCc0kADSAD8DiIsUfwEAdCGAABsAOwIWQQA7AEagE98BGgSHMFnAB+QWiqdAI9QAfwEAqyPAQCsQAJ8JVgCiAD9CZZUAf0EtF8EtI8AJ6wBSAMnLQCbUAH8BBpRLrhcCWyBWAKIAP0JqFQB/gQAkEQBUDbsHEgDjEp9AK5QAFUDAF0YEBrAVwIQkAAATGACXAJoAGwG
	bQqzUAF4AnwEIv4nXJE0fQK6XAR8G10EABqAWAJgkQAAMGAIXAJoAHwJfQC/UAB8CXykSQEAfAQizG1sAWwAbANtAMhQAXwAWACJAKxdCWBwVJQBKA0CzUxXAfX//2/bAgAMlxqgWvYBAADtCN1IAuwBGkFbHFwESAInCADsA3wAbBjgA30E6/ABCAAAAAEAAAAAAAAAEQAA
  ]])
  -- End zwint_so_lzo

  -- These files are the compressed uninstaller. If the installer gets deleted
  -- The uninstaller remains, detects that the insatller no longer exists,
  -- cleans up the dropped files, un-adopts any devices, and cleans up itself.

  -- Begin D_GenGenSceneControllerUninstaller_json_lzo
  -- lzo-compressed D_GenGenSceneControllerUninstaller.json: compressed size = 967 -- uncompressed size = 2260
  local D_GenGenSceneControllerUninstaller_json_lzo
  D_GenGenSceneControllerUninstaller_json_lzo = b642bin([[
	1AgAAAAUew0KCSJkZWZhdWx0X2ljb24iOiAic2NlbmVfY29udHJvbGxlcl/ABAwucG5nIiwNCgkiZmxhc2j4BXwACnMvZGV2aWNlcy5zd2asBAFpbWdJQAoBQm9keXAKK1AABURpbW1hYmxlL2AAAVR1cm4zYgBNaYgTKxcBTWF4KAABAmhhbGxvYA4Bc0RpcngCAnBpY3MvlALSBm5TfBppAzCkAg9EaXNwbGF5U3RhdHVzIjoge32QGwRk
	b2NfdXJsYAICDQoJCSJ8AQRsYW5ndWFnYRYxcAS0AgJtYW51YWAFKUwAAnZlcnNpnCgpUAAJcGxhdGZvcm0iOiAwKPkAcKAKXC+IJwEiDQoJpxBUYWJ+EVt7fA8DCSJMYWJldAy6AQkicBEBX3RhZ2gZBHVuaW5zdGF4NJEaCUAWAXRleHRtA1UpbACECZQLBQkJIlBvc2l0sRQirCFUBgNUYWJUeXBkHLg72AkHIlRvcE5hdmlnYWQGA1Rh
	YiI6IKQbBAkidG9wX24nbgBfdCtwAAdDb25maWdGdW5jYAd9ETHoCoAxAkdyb3VwUCzUHAEJImlkKpAAlAssQwBsZWacGuoWCSKdPzLaCgkigU0yfCgBCQl9XYMtCSJDvFg/uAGVXmfSEiIxNBQBKrwCLWEBIvgbAAMJIkhvcml6b250YWxNdWx0aXBsaWWIUiikAswH4C4DYnV0dG9uJxwD1wNDb2SYMQFUZXN0KHABLTYICSIvuAeoF6xJ
	fEHcXEwXxUgJcDpNKzKwUW0ETK0rNShEAAJXaWR0aHA9KUwAAkhlaWdobjAyMIBKKIwBA0NvbW1hbmQ4J48BU2VyfFpcDAUidXJuOmdlbkgAB19tY3Ytb3JnOnO3A0lkOpRByx5sZXIp1Aon0wMJIkEnFAknfAD4HIYRfV14AKBcpBA72AwFc2V0dGluZ3MnyAXxKFPgA3gJ+Bs/wAwHamF2YXNjcmlwdMIKIlOXAU5hbYQ4BXNoYXJlZC5q
	/Q0iKlgMA3NpbXBsZVCslCV8DiAIGAMFYWR2YW5jZWTUDigtCEHgA3gJMxgD7SsiKdwPIBwYA+wRIBQgA7fHX29wbIX0JCg5A0SSyyBP2AM3UAPgbCAoUAPcEgJ6d2F2ZSd0AngPIAiIBgRub3RpZmljlKcwYQNOK3QAYAszsQYzwDwgL1wDLHACeQ99UNCEpwgiZXZlbnRMaXN0MkGHW6wCvDD0lnR6BXNjaGVtYXMtSHtIACd0D7wmOWkP
	OrxPzt9fdMTPIB5IAQINCn0NChEAAA==
  ]])
  -- End D_GenGenSceneControllerUninstaller_json_lzo

  -- Begin D_GenGenSceneControllerUninstaller_xml_lzo
  -- lzo-compressed D_GenGenSceneControllerUninstaller.xml: compressed size = 494 -- uncompressed size = 990
  local D_GenGenSceneControllerUninstaller_xml_lzo
  D_GenGenSceneControllerUninstaller_xml_lzo = b642bin([[
	3gMAAAA7PD94bWwgdmVyc2lvbj0iMS4wIj8+Cjxyb290IHhtbG5zPSJ1cm46c2NoZW1hcy11cG5wLW9yZzpkZXZpY2UtMS0wIj4KICA8c3BlY1a0CHwBCSAgPG1ham9yPjE8L6MBCiAgcAIDaW5vcj4wUAKAAWUHLy3pADyoC9QIsAECVHlwZT4qHwJnZW5IAAFfbWN2KTQCAAU6U2NlbmVDb250cm9sbGVyVW5pbnN0YWgBAToxPC8pKAGw
	EUwDDXRpY0pzb24+RF9HZW5HZW44BAEELmpzb248LynMALAICmZyaWVuZGx5TmFtZT6wCAJlcmljIIUJICgxAiApNgI8LyvUAKAJAAJtYW51ZmFjdHVyZXI+VmFyaW91c1wkKlQAoAUOQ2F0ZWdvcnlfTnVtPjE0PC8rQACvBFN1YitXADA8Ly5IAKMFc2VyYDkBTGlzdKQubDmsAvwBKEQAPhgG0Ak9HAbUBJAJkBCEQ9YCSWScPFA7SAAn
	fAcncAA5awExPC8nmAAp6AIGU0NQRFVSTD5TIAGEBwJ4bWw8L/wFxBMnUAGoCic4ACjABAw8aW1wbGVtZW50YXRpb26BAw30ByxsAANGaWxlPkkgBiACMegAugo8LyxoACfEAehcBz4KPC9yb290PgoRAAA=
  ]])
  -- End D_GenGenSceneControllerUninstaller_xml_lzo

  -- Begin I_GenGenSceneControllerUninstaller_xml_lzo
  -- lzo-compressed I_GenGenSceneControllerUninstaller.xml: compressed size = 276 -- uncompressed size = 436
  local I_GenGenSceneControllerUninstaller_xml_lzo
  I_GenGenSceneControllerUninstaller_xml_lzo = b642bin([[
	tAEAAAA8PD94bWwgdmVyc2lvbj0iMS4wIj8+CjxpbXBsZW1lbnRhdGlvbj4KICA8ZmlsZXM+TF9HZW5HZW5TY2VuZUNvbnRyb2xsZXJVbmluc3RhaAEDLmx1YTwvtAV8BgVzdGFydHVwPjjEAARfSW5pdDwv4QUNdgZhY2APAUxpc3RzDyAgPKACsAEAAiAgPHNlcnZpY2VJZD51cm46Z2VuSAAGX21jdi1vcmc6J3EAOjjTATE8LyeYACdU
	AQJuYW1lPicuAzwvngEKIIAABTxydW4+CgkJOEkBXyfkAAQobHVsX2RlZRQp2AcKICByZXR1cm4gdHJ1Zc4CPC+MCnoLPC8ogAPsAbcfPC9pLTAGEQAA
  ]])
  -- End I_GenGenSceneControllerUninstaller_xml_lzo

  -- Begin L_GenGenSceneControllerUninstaller_lua_lzo
  -- lzo-compressed L_GenGenSceneControllerUninstaller.lua: compressed size = 3156 -- uncompressed size = 9973
  local L_GenGenSceneControllerUninstaller_lua_lzo
  L_GenGenSceneControllerUninstaller_lua_lzo = b642bin([[
	9SYAAAAULS0gR2VuR2VuZXJpYyBTY2VuZSBDb250cm9sbGVyIFVuaW5zdGFsAQENCi0tbAAAFSBDb3B5cmlnaHQgKEMpIDIwMjAgR3VzdGF2byBBIEZlcm5hbmRleogFABNUaGFua3MgdG8gUm9uIEx1bmEgYW5kIEN0cmxhYmxlIGZvciBjdA0EaWJ1dGluZ4ACClJGV0M1IHN1cHBvcnR4CGwAAB0gVGhpcyBwcm9ncmFtIGlzIGZyZWUg
	c29mdHdhcmU7IHlvdSBjYW4gcmVkaXN0cnwJAWUgaXR7DS9vcoQIA21vZGlmeXECdUAUC3IgdGhlIHRlcm1zIG9mkAEBR05VIIQhDmFsIFB1YmxpYyBMaWNlbnNlnAcBYXMgcHQCBHNoZWQgYnmJBkZ1EFPUEAALIEZvdW5kYXRpb247IGVpdGhlciB2ZXJzaW9uIDKICMAN1AoFLCBvciAoYXRwFwFyIG9wYAcGKSBhbnkgbGF0KO0ALmgH
	bAAvDAQoqAMBZCBpboASBGhvcGUgdGhQCg9pdCB3aWxsIGJlIHVzZWZ1bCyMCAAHYnV0IFdJVEhPVVQgQU5ZIFdBUlJBTlRZO1QFBnRob3V0IGV2ZawJAAFpbXBsaWVkIHdhcnJhbnR5IG9miAgCTUVSQ0hMBgRBQklMSVRZdBgAE0ZJVE5FU1MgRk9SIEEgUEFSVElDVUxBUiBQVVJQT1NFLiAgU2VgE6gpOKQFhEAJbW9yZSBkZXRhaWxz
	KPUDWcg8A2FjY2Vzc5QdKfgGAWlzIGyyNSBhlEYAAyBodHRwczovL3d3dy5nbnUub3JnL8gEA3Mvb2xkLScwAAxncGwtMi4wLmVuLmh0bWxwHAABDQpsb2NhbCBWZXJib3NlTG9nZ3xTAj0gMA0K7AMObml4aW8gPSByZXF1aXJlICKBAiL4AwRIQUdfU0lEQAQAFiJ1cm46bWljYXNhdmVyZGUtY29tOnNlcnZpY2VJZDpIb21lQXV0b22I
	RwZHYXRld2F5MSIooAEBZnVuY3BDDSBsb2dMaXN0KC4uLikNCgmpFHNkDAEiDQoJdGkAAmkgPSAxLCBzZWxlY3QgKCIjIiwgdAUFIGRvDQoJICDPBnggPc4EKGmgBAENCgkJZAkEcyAuLiB0b2xoAW5nKHh4DAACZW5kIA0KCXJldHVybiBzDQplbmQ0cAInYAIIdXVwLmxvZygiR2UncBEpLREgYAADbG9nOiAieA0qWQMpNngBAVZMb2fc
	CwEgIGlmLmcGPiAyaUJueCI+9QF2sDo09gEgIJgbkADoOScABgxEZWxldGVGaWxlKGJhc2V0I50SIoADZ0QiLCDkA5hBBi5mcy5yZW1vdqQGex0iLm1okQFpZWQiYAg4mAB4RTeIADegBIgRjxRDYXJif2x5PbACAiwgIiBj7AWADqRLYRlf7BVUTHwBLQADeC2AZW0DX2AWK5wAYAJoUyk9AXUnSRlkK6gAKVwAiF/EVVTPAXQgPSAnkAR8
	AbgZbFK8CmwC7A8woACEA2AtKbAAKbABM3EBXymAAEAtU2NpZiB4DO1MCSo8AGURCeQCmA0nmAAoWAApRAEpdQAJJ7wCKZAIiwgpCQlYDgIJCQllbGHQCWAAKbMAbmFtpEqRBSyQEy3UAmQGmFhgASnoACcwCugMKIQB1BoqIAE+sALwCD6cAiABiAKYHcAW2AAy1AInRApsOpwEkAAvrg5SZVGtdDGsDqADLbAOsEOUWCAjxAsnXAyQRiAw
	/AopiAIniARkDYRjKoAETCrMOiIoJLREcHInVAcn+AAp/AlgKypgCi6sCXQkjATICyfAASdUASr0ACe4CSeYAIwHkTYJlACSAA0KJ5QVIxgqKBQqKeEYXyctACgnJAACZXJfZGVo6QFfbnVtaAt84ChGACwgrAEihCQEcGFpcnMobHzDsgJzKaDj0I2wAgRfdHlwZSA93QYuKVAAVCkspQBmRMFEBYQKB2F0dHJfZ2V0KCLEBnEDIvgP8BN5
	hCjABL4MPSAjfCAIc2NoZW1hcy1nZW5IAAZfbWN2LW9yZzqpBTotSAQJRXZvbHZlTENEOjEiYDtgE2zqtQVfZA8EID09ICJEXyekAAUxLnhtbCIpICI2LAkJUAUgIRQCBENvb3BlclIivC4/HAIprAAgMSACCk5leGlhT25lVG91Y2g/KAIrtAAgvjECSSb4Nj9UBCSkJi2cCSf4ANskdGhlIoQllNqsTfhjB251bV9wYXJlbnRhTzEoJRAJ
	IxwoK8wVJ+QAAXNjcmkj2DMiBCIDdG8gYSBnJSU6cyMlOmMibDgjJzoiKTtuXAkJKKkNcymoDWZpIiwsIA0uwC20Di4mDToxMIgONpQBAWZpbGV2DERfLZgErGsvqA8vHAEiRDWQc3QIIAncAQNqc29uIix0eS/cAcgDrBsnjBEw4AEDbnZpc2liuRcwMeQBKAggmDUjDCnE5K8LLmRlJ7QGKywGA2NhbGxfYSN9MiglyTMsJDosZUSYsgEi
	LCB7pwFOdW1QpbwJbBUBfSwgMMwJiMAkriNpZsSvhCV3SSJEXyKMNAZTZWVyV1MxMDCQUSLYOcgJfDIvnQAy/AQgSMwKA2JpbmFyedAXPcAKaBapOkKVBkwiQEUgJxgJIIgxB0QgBjEHRCcyB29y3CsycghGQyecACBI0AcBZGltbSMsS6xRIAjZB0TcBiC04AckfEACCQktLSCEiQNlIHZhcml4JCLcRgNhdCBtYXkiYEoCd3JpdHQitUlv
	I9hLqCgk+EoidU5jJ/kXc4QGAAtoYXZlIGJlZW4gcHJldmlvdXNseSBhc3NvY2lhdCKgSgFpaXRoeMQkMi9TSUC1DENFTkVDT05UUk9MTEVSPSOpJWcsiCUoJEYueBUi6Dqkzv0XX3TOMWMBLCAikAcJQWN0dWF0b3JDb25mX3tuaWwqVBgCLCBmYWwk2EAgCFgBBVBlZXJJRCIgKAIALCAgIlgBBlNob3dTaW5nbPW0ICAkWQFSI5BIA1Jv
	dXRlc6wWOlgBlr4NCnxEAAJVSTcgZHJvcHBlZCBmaWxlcw0KCaxHIwU3IiIYTwAEL2NtaC9za2lucy9kZWZhdWx0L2ltZ0wBIvQvbAJ9AF8iHzRlcy8o4CsHXzUweDUwLnBuZyNsQyAcVAEpGCsgK1gBK0gqICtgASvcHCA0YAFsuCA0YAEBUzIwMCA0YAF0ySAzYAGYzy1gASTcKy0ICwRraXQvS2l0JFEkLiMIJyLcOiABrgBfVtioKNAA
	Ay0tIFVJNS7oDHCFIAOfC2NvbjikDCAMEAEgHWAMnBAgHxAMkAggH8ALLBABIBZwCywQASAWIAssEAEgFtAKKxABMoAKKJQIKLRYAiIvZXRjkK0NendhdmVfcHJvZHVjdHNfdUjlI/Ate1UtLSAneDgusBbcBwUtbHVkbC9EXz1sOWRgAS5sem8weBYgDR8BeG1sIAIZAUk+OAIgBRkBTD4bAWx1YSACNQJTPhgBIAU0AiAMGAEBLS0gQyKI
	bd7sICAspB0AAXVzci9saWIvbHVhL3p3aW50LnOIMyADnAAgAsQDK0xKIAjoByg4H2BQIASYASlIEpylIBDQACAF6gVEXyvgEiAJqAEs2AAgB7ABK4ATIAgsBCzUACAI1AUn2ABgoSARsAGUBiARsAFgpiARsAGUBiARsAFgqyARsAGUBiAQsAGAsCAQsAG0BiAGsAE1dBMgBtEBSjQ3Ei5qcyAC8Aw1RRQuIAUgFDX0AANTaGFyZWQgBgwB
	NRQUIAbwA/QHCVpXYXZlTW9uaXRvciroACI8JCK4JScBc3IgE9gbKdxiIAmoEDMAHSokASAF9BA1cAggESABNbAGKiABIAXUB+wuI/A8M8hnKgQGJNA8IBDEUSlYAimoaCb0USOrbmcoIieUACMYhgIgY29tcCKwfwdkLiBSZWxvYWRpIvw4IhRaIwhaMhICICKkBSLEU5QLJyyDIANoawVJbml0KGx1bMgUIoglJBxDJwiWI2k7IC4wcSe0
	JiS0JTVoBieoJisoDQRpZiBub3QgLYQBJvBMIyhYM+gJJzQBmQdmIt2TLiiwmCLgWANhbGwgcmUiIJMlPT8uJWhMOKQHKXBvLBgEJjxTJKiFgAgqkZpjIi4jc3QisJMIcHJlc2VudC4gRG9kEwFoaW5n3A4JZW5kDQplbmQNCg0KEQAA
  ]])
  -- End L_GenGenSceneControllerUninstaller_lua_lzo

  -- Begin S_GenGenSceneControllerUninstaller_xml_lzo
  -- lzo-compressed S_GenGenSceneControllerUninstaller.xml: compressed size = 199 -- uncompressed size = 276
  local S_GenGenSceneControllerUninstaller_xml_lzo
  S_GenGenSceneControllerUninstaller_xml_lzo = b642bin([[
	FAEAAAAtPHNjcGQgeG1sbnM9InVybjpzY2hlbWFzLXVwbnAtb3JnOnNlcnZpY2UtMS0wIj4NCiAgPHNwZWNWZXJzaW9ugAIJICA8bWFqb3I+MTwvoAF4BHQCA2lub3I+MFQCgAF2AjwvLvUAPMAMB1N0YXRlVGFibGWSCjwvNVwAB2FjdGlvbkxpc3SfBCAgPKQC1AEAASAgPG5hbWU+VW5pbnN0YWxsPC+cAXAPASAgPC8psADwAcwJBjwv
	c2NwZD4NChEAAA==
  ]])
  -- End S_GenGenSceneControllerUninstaller_xml_lzo

  -- End uninstaller

	-- Drop the uninstaller files if necessary
  	UpdateFileWithContent("/usr/lib/lua/zwint.so",                                     zwint_so_lzo,                                755, zwint_so_version,        updated, true)
  	UpdateFileWithContent("/etc/cmh-ludl/D_GenGenSceneControllerUninstaller.json.lzo", D_GenGenSceneControllerUninstaller_json_lzo, 644, GenGenInstaller_Version, updated, false)
  	UpdateFileWithContent("/etc/cmh-ludl/D_GenGenSceneControllerUninstaller.xml.lzo",  D_GenGenSceneControllerUninstaller_xml_lzo,  644, GenGenInstaller_Version, updated, false)
  	UpdateFileWithContent("/etc/cmh-ludl/I_GenGenSceneControllerUninstaller.xml.lzo",  I_GenGenSceneControllerUninstaller_xml_lzo,  644, GenGenInstaller_Version, updated, false)
  	UpdateFileWithContent("/etc/cmh-ludl/L_GenGenSceneControllerUninstaller.lua.lzo",  L_GenGenSceneControllerUninstaller_lua_lzo,  644, GenGenInstaller_Version, updated, false)
  	UpdateFileWithContent("/etc/cmh-ludl/S_GenGenSceneControllerUninstaller.xml.lzo",  S_GenGenSceneControllerUninstaller_xml_lzo,  644, GenGenInstaller_Version, updated, false)

	-- Create the uninstaller device if it does not yet exist
	local uninstaller_found = false
	for dev_num, v in pairs(luup.devices) do
		if v.device_type == "urn:schemas-gengen_mcv-org:device:SceneControllerUninstaller:1" then
			uninstaller_found = true
			break
		end
	end
	if not uninstaller_found then
	    luup.call_action(SID_HAG, "CreateDevice", {
		   	deviceType = "urn:schemas-gengen_mcv-org:device:SceneControllerUninstaller:1";
			Description = "Scene Controller Uninstaller";
		   	UpnpDevFilename = "D_GenGenSceneControllerUninstaller.xml";
		   	UpnpImplFilename =  "I_GenGenSceneControllerUninstaller.xml";
			StateVariables = "invisible=0"; -- TODO Set invisible=1 when we make this into a real plug-in 
		},0)
		reload_needed = "Uninstaller created";
	end


	-- Update scene controller devices which may have been included into the Z-Wave network before the installer ran.
	ScanForNewDevices()

	if reload_needed then
		log("Files updated including ",reload_needed,". Reloading LuaUPnP.")
		luup.call_action(SID_HAG, "Reload", {}, 0)
	else
	VLog("Nothing updated. No need to reload.")
	end

	return true, "ok", "Gengeneric Scene Controller Installer"
end	-- function SceneControllerInstaller_Init
